package com.bria.common.controller.calllog

import android.content.Context
import com.bria.common.controller.IController
import com.bria.common.controller.calllog.db.*
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.core.ISettingsObserver
import com.bria.common.modules.BriaGraph
import com.bria.common.util.Log
import io.reactivex.Single
import io.reactivex.disposables.Disposable
import io.reactivex.processors.BehaviorProcessor
import io.reactivex.rxkotlin.subscribeBy
import io.reactivex.schedulers.Schedulers
import java.util.*

class CallLogApiImpl(private val context: Context, private val controller: IController): CallLogApi {
    companion object {
        private const val TAG = "CallLogApiImpl"
    }

    private var mCallLogDao: CallLogDao = getDao()
    private var mOwner = BriaGraph.settings.owner

    private val mGetAllBehaviorProcessor = BehaviorProcessor.create<List<CallLogEntity>>()
    private var mGetAllDisposable: Disposable? = null

    private val mGetCountOfUnreadMissedCallsBehaviorProcessor = BehaviorProcessor.create<Int>()
    private var mGetCountOfUnreadMissedCallsDisposable: Disposable? = null

    private val mSettingsObserver = object : ISettingsObserver {
        override fun onBeforeOwnerChange(newOwner: String?) {}

        override fun onSettingsChanged(changedSettings: MutableSet<ESetting>?) {
            if (changedSettings?.contains(ESetting.BroadWorksEnabled) == true ||
                    changedSettings?.contains(ESetting.BroadWorksEnterpriseCallLog) == true) {
                Log.d(TAG, "ISettingsObserver::onSettingsChanged: BroadWorksEnabled and/or BroadWorksEnterpriseCallLog setting changed")
                mCallLogDao = getDao()
            }
        }

        override fun onOwnerChanged() {
            val newOwner = BriaGraph.settings.owner
            Log.d(TAG, "ISettingsObserver::onOwnerChanged: old owner [$mOwner] -> new owner [$newOwner]")
            if (mOwner != newOwner) {
                mOwner = newOwner
                // Owner changed so we need to drop all existing subscriptions and subscribe again with new owner
                resubscribe()
            }
        }
    }

    init {
        Log.d(TAG, "init")
        subscribe()
    }

    override fun put(vararg entities: CallLogEntity) =
            Single.fromCallable { mCallLogDao.put(*entities) }.subscribeOn(Schedulers.io())!!

    override fun get(id: Long) = mCallLogDao.get(id, mOwner).subscribeOn(Schedulers.io())!!

    override fun getLast(callType: CallType)=
            mCallLogDao.getLast(callType, mOwner).subscribeOn(Schedulers.io())!!

    override fun getAll() =
            mGetAllBehaviorProcessor.subscribeOn(Schedulers.io()).onBackpressureLatest()!!

    override fun delete(vararg entities: CallLogEntity) =
            Single.fromCallable { mCallLogDao.delete(*entities) }.subscribeOn(Schedulers.io())!!

    override fun delete(callType: CallType) =
            Single.fromCallable { mCallLogDao.delete(callType, mOwner) }.subscribeOn(Schedulers.io())!!

    override fun deleteAll() =
            Single.fromCallable { mCallLogDao.deleteAll(mOwner) }.subscribeOn(Schedulers.io())!!

    override fun markAllRead() =
            Single.fromCallable { mCallLogDao.markAllRead(mOwner) }.subscribeOn(Schedulers.io())!!

     override fun getCountOfUnreadMissedCalls() =
             mGetCountOfUnreadMissedCallsBehaviorProcessor.subscribeOn(Schedulers.io()).onBackpressureLatest()!!

    override fun destroy() {
        Log.d(TAG, "destroy")
        unsubscribe()
        CallLogDatabase.destroy()
    }

    private fun getDao(): CallLogDao {
        val broadworksActive = isBroadworksActive()
        Log.d(TAG, "getDao: broadworksActive = $broadworksActive")
        return if (!broadworksActive)  CallLogDatabase.get(context).callLogDao() else BwCallLogDao(controller)
    }

    private fun isBroadworksActive() =
            controller.broadworksCtrl.events.isBroadworksEnabled &&
            BriaGraph.settings.getBool(ESetting.BroadWorksEnterpriseCallLog)

    private fun subscribe() {
        mGetAllDisposable = mCallLogDao.getAll(mOwner).subscribeBy(
                onNext = mGetAllBehaviorProcessor::onNext,
                onError = mGetAllBehaviorProcessor::onError,
                onComplete = mGetAllBehaviorProcessor::onComplete
        )

        mGetCountOfUnreadMissedCallsDisposable = mCallLogDao.getCountOfUnreadMissedCalls(mOwner).subscribeBy(
                onNext = mGetCountOfUnreadMissedCallsBehaviorProcessor::onNext,
                onError = mGetCountOfUnreadMissedCallsBehaviorProcessor::onError,
                onComplete = mGetCountOfUnreadMissedCallsBehaviorProcessor::onComplete
        )

        BriaGraph.settings.attachWeakObserver(mSettingsObserver,
                EnumSet.of(ESetting.BroadWorksEnabled, ESetting.BroadWorksEnterpriseCallLog), true)
    }

    private fun unsubscribe() {
        BriaGraph.settings.detachObserver(mSettingsObserver)
        mGetCountOfUnreadMissedCallsDisposable?.dispose()
        mGetAllDisposable?.dispose()
    }

    private fun resubscribe() {
        unsubscribe()
        subscribe()
    }
}
