package com.bria.common.controller.calllog.db

import com.bria.common.controller.IController
import com.bria.common.controller.settings.EAccountSetting
import com.bria.common.controller.settings.ESetting
import com.bria.common.util.Log
import com.bria.common.util.broadworks.BroadWorksConnectionManager
import com.bria.common.util.broadworks.BroadWorksException
import com.bria.common.util.broadworks.BroadWorksObjectConversion
import io.reactivex.*
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.disposables.Disposable
import io.reactivex.processors.BehaviorProcessor
import io.reactivex.rxkotlin.plusAssign
import io.reactivex.rxkotlin.subscribeBy
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.PublishSubject
import java.util.concurrent.ConcurrentHashMap

class BwCallLogDao(private val controller: IController) : CallLogDao {
    companion object {
        private const val TAG = "BwCallLogDao"
    }

    private enum class ServiceApi {
        NONE,
        BASIC,
        ENHANCED
    }

    private var mCurrentServiceApi = ServiceApi.NONE
    private val mAllDataMap = ConcurrentHashMap<Long, CallLogEntity>()

    // todo: switch to using BehaviorProcessor
    private val mGetAllSubject = PublishSubject.create<List<CallLogEntity>>()
    private val mGetAllFlowable = mGetAllSubject.toFlowable(BackpressureStrategy.LATEST)

    private var mLoadDataFromServerDisposable: Disposable? = null
    private var mDeleteDataFromServerDisposable = CompositeDisposable()

    private val mMissedCallStorage = BwMissedCallStorage(controller)

    override fun put(vararg entities: CallLogEntity): List<Long> {
        // Just trigger reload all data from the server here
        mLoadDataFromServerDisposable?.dispose()
        mLoadDataFromServerDisposable =
                loadDataFromServerObservable()
                        .subscribeOn(Schedulers.io())
                        .subscribeBy(
                                onNext = mGetAllSubject::onNext,
                                onError = { Log.e(TAG, "Load data from server error", it) },
                                onComplete = { Log.i(TAG, "Load data from server complete") }
                        )
        return listOf(0)
    }

    override fun get(id: Long, user: String?): Maybe<CallLogEntity?> {
        return Maybe.fromCallable { mAllDataMap[id] }
    }

    override fun getLast(callType: CallType, user: String?): Maybe<CallLogEntity?> {
        return Maybe.fromCallable { toSortedList(mAllDataMap).firstOrNull { it.callType == callType } }
    }

    override fun getAll(user: String?): Flowable<List<CallLogEntity>> {
        mLoadDataFromServerDisposable?.dispose()
        mLoadDataFromServerDisposable =
                loadDataFromServerObservable()
                        .subscribeOn(Schedulers.io())
                        .subscribeBy(onNext = mGetAllSubject::onNext, onError = mGetAllSubject::onError)
        return mGetAllFlowable
                // on subscription emmit current value
                .doOnSubscribe { mGetAllSubject.onNext(toSortedList(mAllDataMap)) }
                .subscribeOn(Schedulers.io())
    }

    override fun delete(vararg entities: CallLogEntity) {
        // Delete entities locally first to speed up show changes on UI
        entities.forEach { mAllDataMap.remove(key = it.id) }
        val sortedList = toSortedList(mAllDataMap)
        mMissedCallStorage.update(sortedList)
        mGetAllSubject.onNext(sortedList)

        mDeleteDataFromServerDisposable.clear()
        for (entity in entities) {
            mDeleteDataFromServerDisposable +=
                    deleteDataFromServerObservable(bwCallLogId = entity.externalId, callStatus = entity.callType.ordinal)
                            .subscribeOn(Schedulers.io())
                            .onErrorResumeNext(loadDataFromServerObservable())
                            .subscribeBy(onNext = mGetAllSubject::onNext, onError = { Log.w(TAG, "Delete data error", it) })
        }
    }

    override fun delete(callType: CallType, user: String?) {
        // Delete entities locally first to speed up show changes on UI
        val newMap = mAllDataMap.filter { it.value.callType != callType }
        mAllDataMap.clear()
        mAllDataMap.putAll(newMap)
        val sortedList = toSortedList(mAllDataMap)
        mMissedCallStorage.update(sortedList)
        mGetAllSubject.onNext(sortedList)

        mDeleteDataFromServerDisposable.clear()
        mDeleteDataFromServerDisposable +=
                deleteDataFromServerObservable(callStatus = callType.ordinal)
                        .subscribeOn(Schedulers.io())
                        .onErrorResumeNext(loadDataFromServerObservable())
                        .subscribeBy(onNext = mGetAllSubject::onNext, onError = { Log.w(TAG, "Delete data by $callType error", it) })
    }

    override fun deleteAll(user: String?) {
        // Delete entities locally first to speed up show changes on UI
        mAllDataMap.clear()
        val sortedList = toSortedList(mAllDataMap)
        mMissedCallStorage.update(sortedList)
        mGetAllSubject.onNext(sortedList)

        mDeleteDataFromServerDisposable.clear()
        mDeleteDataFromServerDisposable +=
                deleteDataFromServerObservable()
                        .subscribeOn(Schedulers.io())
                        .onErrorResumeNext(loadDataFromServerObservable())
                        .subscribeBy(onNext = mGetAllSubject::onNext, onError = { Log.w(TAG, "Delete all data error", it) })
    }

    override fun markAllRead(user: String?) = mMissedCallStorage.markAllRead()

    override fun getCountOfUnreadMissedCalls(user: String?) = mMissedCallStorage.getCountOfUnreadMissedCalls()

    override fun getCount(user: String?) = Single.just(mAllDataMap.count())!!

    private fun loadDataFromServerObservable(): Observable<List<CallLogEntity>> {
        return Observable.fromCallable {
            try {
                val bwAccount = controller.broadworksCtrl.events.broadWorksAccount
                val bwUsername = bwAccount.getStr(EAccountSetting.BwUserName)
                val bwPassword = bwAccount.getStr(EAccountSetting.BwPassword)
                val bwAccountDomain = bwAccount.getStr(EAccountSetting.Domain)
                val bwXsiServer = controller.settingsCtrl.events.getStr(ESetting.BroadWorksXsiServer)

                var freshLoadedCallLogs = ConcurrentHashMap<Long, CallLogEntity>()

                val isEnhancedCallLogsSupported = controller.bwServiceMgtCtrl.events.isEnhancedCallLogsSupported
                val isBasicCallLogsSupported = controller.bwServiceMgtCtrl.events.isBasicCallLogsSupported

                if (isEnhancedCallLogsSupported && isBasicCallLogsSupported) {
                    // if both Enhanced and Basic call logs available...
                    try {
                        // ...then try to get Enhanced call logs first...
                        val callLogs = BroadWorksConnectionManager.loadEnhancedCallLogs(bwUsername, bwAccountDomain, bwPassword, bwXsiServer)
                        freshLoadedCallLogs = BroadWorksObjectConversion.toCallLogEntityMap(callLogs, bwAccount)
                        mCurrentServiceApi = ServiceApi.ENHANCED
                    } catch (e: Exception) {
                        Log.w(TAG, "Load Enhanced call logs failed [" + e.message + "], try to switch to Basic call logs")
                        // ... if getting Enhanced was failed then show message...
                        // todo:
                        //if (!mIsServiceDegradedMessageShown) {
                        //    mIsServiceDegradedMessageShown = true
                        //    showMessage(getContext().getString(R.string.tBroadWorksServiceDegraded))
                        //}
                        // ... and try to get Basic call logs
                        val callLogs = BroadWorksConnectionManager.loadBasicCallLogs(bwUsername, bwAccountDomain, bwPassword, bwXsiServer)
                        freshLoadedCallLogs = BroadWorksObjectConversion.toCallLogEntityMap(callLogs, bwAccount)
                        mCurrentServiceApi = ServiceApi.BASIC
                    }
                } else {
                    // if only one of Enhanced or Basic available then try to get it
                    when {
                        isEnhancedCallLogsSupported -> {
                            val callLogs = BroadWorksConnectionManager.loadEnhancedCallLogs(bwUsername, bwAccountDomain, bwPassword, bwXsiServer)
                            freshLoadedCallLogs = BroadWorksObjectConversion.toCallLogEntityMap(callLogs, bwAccount)
                            mCurrentServiceApi = ServiceApi.ENHANCED
                        }
                        isBasicCallLogsSupported -> {
                            val callLogs = BroadWorksConnectionManager.loadBasicCallLogs(bwUsername, bwAccountDomain, bwPassword, bwXsiServer)
                            freshLoadedCallLogs = BroadWorksObjectConversion.toCallLogEntityMap(callLogs, bwAccount)
                            mCurrentServiceApi = ServiceApi.BASIC
                        }
                        else -> mCurrentServiceApi = ServiceApi.NONE
                    }
                }

                mAllDataMap.clear()
                mAllDataMap.putAll(freshLoadedCallLogs)

            } catch (e: BroadWorksException) {
                // todo showMessage(e.message)
                mCurrentServiceApi = ServiceApi.NONE
            }

            val sortedList = toSortedList(mAllDataMap)
            mMissedCallStorage.update(sortedList)
            sortedList
        }
    }

    private fun deleteDataFromServerObservable(bwCallLogId: String = "", callStatus: Int = -1): Observable<List<CallLogEntity>> {
        return Observable.fromCallable {
            try {
                val bwAccount = controller.broadworksCtrl.events.broadWorksAccount
                val bwUsername = bwAccount.getStr(EAccountSetting.BwUserName)
                val bwPassword = bwAccount.getStr(EAccountSetting.BwPassword)
                val bwAccountDomain = bwAccount.getStr(EAccountSetting.Domain)
                val bwXsiServer = controller.settingsCtrl.events.getStr(ESetting.BroadWorksXsiServer)

                when (mCurrentServiceApi) {
                    ServiceApi.NONE -> Log.w(TAG, "Failed to delete call log [status=$callStatus, id=$bwCallLogId] - service unavailable!")
                    ServiceApi.BASIC -> BroadWorksConnectionManager.deleteBasicCallLog(callStatus, bwCallLogId, bwUsername, bwAccountDomain, bwPassword, bwXsiServer)
                    ServiceApi.ENHANCED -> BroadWorksConnectionManager.deleteEnhancedCallLog(callStatus, bwCallLogId, bwUsername, bwAccountDomain, bwPassword, bwXsiServer)
                }
            } catch (e: BroadWorksException) {
                Log.w(TAG, "Failed to delete call log [status=$callStatus, id=$bwCallLogId] - error", e)
                // Something went wrong - show message and reload call log data
                // todo showMessage(e.message)
                throw e
            }
            toSortedList(mAllDataMap)
        }
    }

    private fun toSortedList(map: Map<Long, CallLogEntity>) = map.values.sortedByDescending { it.date }
}

private class BwMissedCallStorage(val controller: IController) {
    private val mSettings = controller.settingsCtrl.events!!

    private val mCountOfUnreadMissedCallsBehaviorProcessor = BehaviorProcessor.createDefault(calculateCountOfUnreadMissedCalls())

    fun update(list: List<CallLogEntity>) {
        val oldMissedCallsIdMap = getMissedCallsIdMap()
        val newMissedCallsIdMap = getNewMissedCallsIdMap(list)

        if (oldMissedCallsIdMap?.isNotEmpty() == true) {
            for (newEntry in newMissedCallsIdMap.entries) {
                val oldValue = oldMissedCallsIdMap[newEntry.key] ?: continue
                newEntry.setValue(oldValue)
            }
        }

        setMissedCallsIdMap(newMissedCallsIdMap)
    }

    fun getCountOfUnreadMissedCalls() =
            mCountOfUnreadMissedCallsBehaviorProcessor.subscribeOn(Schedulers.io()).onBackpressureLatest()!!

    fun calculateCountOfUnreadMissedCalls() = getMissedCallsIdMap()?.filter { it.value }?.size ?: 0

    fun markAllRead() = setMissedCallsIdMap(getMissedCallsIdMap()?.mapValues { false })

    private fun getMissedCallsIdMap(): Map<String, Boolean>? {
        return mSettings.getMap(ESetting.BroadWorksMissedCallsIdMap, String::class.java, Boolean::class.javaObjectType)
    }

    private fun setMissedCallsIdMap(map: Map<String, Boolean>?) {
        mSettings.set(ESetting.BroadWorksMissedCallsIdMap, map)
        mCountOfUnreadMissedCallsBehaviorProcessor.onNext(calculateCountOfUnreadMissedCalls())
    }

    private fun getNewMissedCallsIdMap(list: List<CallLogEntity>) =
            list.filter { it.callType == CallType.MISSED }.map { it.externalId to true }.toMap(mutableMapOf())
}
