package com.bria.common.controller.calllog.db

import android.arch.persistence.db.SupportSQLiteDatabase
import android.arch.persistence.room.Database
import android.arch.persistence.room.Room
import android.arch.persistence.room.RoomDatabase
import android.arch.persistence.room.TypeConverters
import android.arch.persistence.room.migration.Migration
import android.content.Context
import com.bria.common.controller.calllog.db.convertors.CallTypeConverter
import com.bria.common.controller.calllog.db.migration.CallLogRoomDbMigrationHelper
import com.bria.common.util.Log

@Database(entities = [CallLogEntity::class], version = 2, exportSchema = false)
@TypeConverters(CallTypeConverter::class)
abstract class CallLogDatabase : RoomDatabase() {
    abstract fun callLogDao(): CallLogDao

    companion object {
        private const val TAG = "CallLogDatabase"

        private const val DATABASE_NAME = "calllog.db"
        private const val TABLE_NAME = "calllog"

        @Volatile private var INSTANCE: CallLogDatabase? = null

        fun get(context: Context): CallLogDatabase =
                INSTANCE ?: synchronized(this) {
                    INSTANCE ?: buildDatabase(context).also { INSTANCE = it }
                }

        private fun buildDatabase(context: Context) = Room
                .databaseBuilder(context.applicationContext, CallLogDatabase::class.java, DATABASE_NAME)
                .addMigrations(MIGRATION_1_2)
                .build()
                .also { CallLogRoomDbMigrationHelper.tryToMigrateFromOldDatabaseIfNecessary(it.callLogDao()) }

        fun destroy() {
            INSTANCE?.close()
            INSTANCE = null
        }

        /**
         * Migrate from:
         * version 1 - initial
         * to
         * version 2 - The {@link CallLogEntity#recordFileName} column renamed from "recordPath" to "recordFileName"
         */
        private val MIGRATION_1_2: Migration = object : Migration(1, 2) {
            private val TABLE_NAME_TMP = "${TABLE_NAME}_tmp"

            private val TABLE_SCHEMA =
                    "id INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL, " +
                    "number TEXT NOT NULL, " +
                    "callType TEXT NOT NULL, " +
                    "date INTEGER NOT NULL, " +
                    "durationSeconds INTEGER NOT NULL, " +
                    "provisioningUser TEXT NOT NULL, " +
                    "accountId INTEGER NOT NULL, " +
                    "accountUsername TEXT NOT NULL, " +
                    "accountDomain TEXT NOT NULL, " +
                    "remoteDomain TEXT NOT NULL, " +
                    "remoteName TEXT NOT NULL, " +
                    "recordFileName TEXT NOT NULL, " +
                    "forwardedTo TEXT NOT NULL, " +
                    "hostedCollab INTEGER NOT NULL, " +
                    "externalId TEXT NOT NULL, " +
                    "unread INTEGER NOT NULL"

            private val TABLE_COLUMNS =
                    "id, " +
                    "number, " +
                    "callType, " +
                    "date, " +
                    "durationSeconds, " +
                    "provisioningUser, " +
                    "accountId, " +
                    "accountUsername, " +
                    "accountDomain, " +
                    "remoteDomain, " +
                    "remoteName, " +
                    "recordFileName, " +
                    "forwardedTo, " +
                    "hostedCollab, " +
                    "externalId, " +
                    "unread"

            private val AFFECTED_COLUMN_NAME_OLD = "recordPath"
            private val AFFECTED_COLUMN_NAME_NEW = "recordFileName"

            private val TABLE_COLUMNS_OLD = TABLE_COLUMNS.replace(AFFECTED_COLUMN_NAME_NEW, AFFECTED_COLUMN_NAME_OLD)

            override fun migrate(database: SupportSQLiteDatabase) {
                Log.d(TAG, "Migration 1 -> 2 start...")

                try {
                    // Before migration we need to be sure what current schema of table is really deprecated i.e. contains column with name "recordPath"
                    val columnNames = database.query("SELECT * FROM $TABLE_NAME WHERE 0").use { it.columnNames }
                    val shouldMigrate = columnNames.contains(AFFECTED_COLUMN_NAME_OLD)
                    if (!shouldMigrate) {
                        Log.d(TAG, "Migration 1 -> 2 current schema already in proper format, nothing to do")
                        return
                    }
                } catch (e: Exception) {
                    Log.e(TAG, "Migration 1 -> 2 something went wrong", e)
                }

                database.beginTransaction()
                try {
                    // SQLite supports a limited operations for ALTER
                    // Renaming a column is not directly supported, so this is what we need to do:
                    // 1. Rename the original table:
                    database.execSQL("ALTER TABLE $TABLE_NAME RENAME TO $TABLE_NAME_TMP")
                    // 2. Create the replacement table with the original name and corrected column name:
                    database.execSQL("CREATE TABLE $TABLE_NAME($TABLE_SCHEMA)")
                    // 3. Copy the data from the original table to the new table:
                    database.execSQL("INSERT INTO $TABLE_NAME($TABLE_COLUMNS) SELECT $TABLE_COLUMNS_OLD FROM $TABLE_NAME_TMP")
                    // 4. Drop the original table:
                    database.execSQL("DROP TABLE $TABLE_NAME_TMP")

                    Log.d(TAG, "Migration 1 -> 2 done successfully")

                    database.setTransactionSuccessful()
                } catch (e: Exception) {
                    Log.e(TAG, "Migration 1 -> 2 something went wrong", e)
                } finally {
                    database.endTransaction()
                }
            }
        }
    }
}
