package com.bria.common.controller.calllog.db.migration

import android.content.Context
import android.os.Parcel
import android.os.Parcelable
import com.bria.common.controller.calllog.db.*
import com.bria.common.controller.calllog.db.convertors.CallTypeConverter
import com.bria.common.controller.commlog.CommLogConstants
import com.bria.common.controller.commlog.db.CallLogDBHelper
import com.bria.common.controller.commlog.db.CallLogDbMigrationHelper
import com.bria.common.modules.BriaGraph
import com.bria.common.util.Log
import com.bria.common.util.SipAddressUtils
import io.reactivex.Single
import io.reactivex.schedulers.Schedulers
import java.util.*

object CallLogRoomDbMigrationHelper {
    private const val TAG = "CallLogRoomDbMigrationHelper"

    private val mCallTypeConverter = CallTypeConverter()

    @JvmStatic
    fun migrateRead(context: Context) =
            CallLogDatabase.get(context).callLogDao().getAll(null).blockingFirst().map { CallLogParcel(it) } as ArrayList<CallLogParcel>

    @JvmStatic
    fun migrateWrite(context: Context, list: List<CallLogParcel>) =
            Single.fromCallable { CallLogDatabase.get(context).callLogDao().put(*list.map { CallLogEntity(it) }.toTypedArray()) }
                    .subscribeOn(Schedulers.io()).blockingGet()!!

    @JvmStatic
    fun migrateWriteOld(context: Context, oldCallLogParcels: List<CallLogDbMigrationHelper.CallLogParcel>) =
            migrateWriteOld(oldCallLogParcels, CallLogDatabase.get(context).callLogDao())

    private fun migrateWriteOld(oldCallLogParcels: List<CallLogDbMigrationHelper.CallLogParcel>, callLogDao: CallLogDao) =
            Single.fromCallable { callLogDao.put(*oldCallLogParcels.map { convertOldParcelToEntity(it) }.toTypedArray()) }
                    .subscribeOn(Schedulers.io()).blockingGet()!!

    internal fun tryToMigrateFromOldDatabaseIfNecessary(callLogDao: CallLogDao) {
        try {
            val context = BriaGraph.application.applicationContext

            if (!isOldDatabaseExist(context)) {
                return
            }

            Log.d(TAG, "Old database exist, read old call logs...")
            val oldCallLogParcels = CallLogDbMigrationHelper.migrateCallLogRead(context)
            Log.d(TAG, "Old call logs read successfully, count ${oldCallLogParcels.size}")

            Log.d(TAG, "Try to delete old database...")
            if (tryToDeleteOldDatabase(context)) {
                Log.i(TAG, "Old database deleted successfully")
            } else {
                Log.w(TAG, "Old database deletion failed")
            }

            if (oldCallLogParcels.isEmpty()) {
                Log.d(TAG, "Old database is empty, nothing to migrate")
                return
            }

            if (callLogDao.getCount(null).subscribeOn(Schedulers.io()).blockingGet() != 0) {
                Log.w(TAG, "Room database is not empty, abort migration")
                return
            }

            Log.d(TAG, "Room database is empty and old database has records, migrate...")

            migrateWriteOld(oldCallLogParcels, callLogDao)

            Log.d(TAG, "Migration done successfully")
        } catch (e: Exception) {
            Log.e(TAG, "Something went wrong", e)
        }
    }

    private fun tryToDeleteOldDatabase(context: Context) = context.deleteDatabase(CallLogDBHelper.DATABASE_NAME)

    private fun isOldDatabaseExist(context: Context) =
            Arrays.asList(*context.databaseList()).contains(CallLogDBHelper.DATABASE_NAME)

    private fun convertOldParcelToEntity(oldCallLogParcel: CallLogDbMigrationHelper.CallLogParcel): CallLogEntity {
        var accountId = -1
        try {
            accountId = oldCallLogParcel.account?.indexOf("_")?.let { oldCallLogParcel.account.substring(it + 1).toInt() } ?: -1
        } catch (ignored: Exception) { }
        val isHostedCollab = oldCallLogParcel.number == CommLogConstants.VCCS_HOSTED_COLLAB_TAG
        val callType = when {
            oldCallLogParcel.callStatus == CommLogConstants.CALL_INCOMING -> CallType.INCOMING
            oldCallLogParcel.callStatus == CommLogConstants.CALL_OUTGOING -> CallType.OUTGOING
            oldCallLogParcel.callStatus == CommLogConstants.CALL_MISSED -> CallType.MISSED
            oldCallLogParcel.callStatus == CommLogConstants.VCCS_CALL || isHostedCollab -> CallType.CONFERENCE
            else -> throw Exception("Unknown call status ${oldCallLogParcel.callStatus}")
        }

        val number = if (isHostedCollab) oldCallLogParcel.number else SipAddressUtils.removeDomain(oldCallLogParcel.number)
        val remoteName = if (isHostedCollab) oldCallLogParcel.name else oldCallLogParcel.remoteName

        return CallLogEntity(
                number = number.orEmpty(),
                callType = callType,
                date = oldCallLogParcel.dtime,
                durationSeconds = oldCallLogParcel.duration,
                provisioningUser = oldCallLogParcel.user.orEmpty(),
                accountId = accountId,
                accountUsername = oldCallLogParcel.accUsername.orEmpty(),
                accountDomain = oldCallLogParcel.accDomain.orEmpty(),
                remoteDomain = oldCallLogParcel.remoteAccHost.orEmpty(),
                remoteName = remoteName.orEmpty(),
                recordFileName = oldCallLogParcel.recordingFile.orEmpty(),
                forwardedTo = oldCallLogParcel.forwardedTo.orEmpty(),
                hostedCollab = isHostedCollab,
                externalId = "",
                unread = oldCallLogParcel.newLog == 1
        )
    }

    data class CallLogParcel(
            override var number: String,
            override var callType: CallType,
            override var date: Long,
            override var durationSeconds: Int,
            override var provisioningUser: String,
            override var accountId: Int,
            override var accountUsername: String,
            override var accountDomain: String,
            override var remoteDomain: String,
            override var remoteName: String,
            override var recordFileName: String,
            override var forwardedTo: String,
            override var hostedCollab: Boolean,
            override var externalId: String,
            override var unread: Boolean
    ) : CallLog, Parcelable {

        constructor(callLog: CallLog) : this(
                number = callLog.number,
                callType = callLog.callType,
                date = callLog.date,
                durationSeconds = callLog.durationSeconds,
                provisioningUser = callLog.provisioningUser,
                accountId = callLog.accountId,
                accountUsername = callLog.accountUsername,
                accountDomain = callLog.accountDomain,
                remoteDomain = callLog.remoteDomain,
                remoteName = callLog.remoteName,
                recordFileName = callLog.recordFileName,
                forwardedTo = callLog.forwardedTo,
                hostedCollab = callLog.hostedCollab,
                externalId = callLog.externalId,
                unread = callLog.unread
        )

        constructor(parcel: Parcel) : this(
                number = parcel.readString().orEmpty(),
                callType = mCallTypeConverter.toCallType(parcel.readString()!!),
                date = parcel.readLong(),
                durationSeconds = parcel.readInt(),
                provisioningUser = parcel.readString().orEmpty(),
                accountId = parcel.readInt(),
                accountUsername = parcel.readString().orEmpty(),
                accountDomain = parcel.readString().orEmpty(),
                remoteDomain = parcel.readString().orEmpty(),
                remoteName = parcel.readString().orEmpty(),
                recordFileName = parcel.readString().orEmpty(),
                forwardedTo = parcel.readString().orEmpty(),
                hostedCollab = parcel.readByte() != 0.toByte(),
                externalId = parcel.readString().orEmpty(),
                unread = parcel.readByte() != 0.toByte()
        )

        override fun writeToParcel(parcel: Parcel, flags: Int) {
            parcel.writeString(number)
            parcel.writeString(mCallTypeConverter.fromCallType(callType))
            parcel.writeLong(date)
            parcel.writeInt(durationSeconds)
            parcel.writeString(provisioningUser)
            parcel.writeInt(accountId)
            parcel.writeString(accountUsername)
            parcel.writeString(accountDomain)
            parcel.writeString(remoteDomain)
            parcel.writeString(remoteName)
            parcel.writeString(recordFileName)
            parcel.writeString(forwardedTo)
            parcel.writeByte(if (hostedCollab) 1 else 0)
            parcel.writeString(externalId)
            parcel.writeByte(if (unread) 1 else 0)
        }

        override fun describeContents() = 0

        companion object CREATOR : Parcelable.Creator<CallLogParcel> {
            override fun createFromParcel(parcel: Parcel) = CallLogParcel(parcel)

            override fun newArray(size: Int) = arrayOfNulls<CallLogParcel>(size)
        }
    }
}
