package com.bria.common.controller.contacts.ldap

import android.content.Context
import android.text.TextUtils
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.core.Settings
import com.bria.common.network.NetworkModule
import com.bria.common.sdkwrapper.SipStackManager
import com.bria.common.sdkwrapper.SipStackManagerInstanceObservable
import com.bria.common.util.INotificationAction
import com.bria.common.util.Log
import com.bria.common.util.SyncObservableDelegate
import com.bria.common.util.Validator
import com.counterpath.sdk.SipLdapApi
import com.counterpath.sdk.handler.SipLdapHandler
import com.counterpath.sdk.pb.Ldap
import io.reactivex.disposables.Disposable
import java.util.*

class LdapModule(val mContext: Context, val mNetworkModule: NetworkModule) {

    private var mStackManagerDisposable: Disposable
    private val TAG: String = "LdapModule"
    private var mDirectorySearchData: MutableList<LdapContactDataObject> = ArrayList()
    private val m_observableAdapter = SyncObservableDelegate<ILdapModuleObserver>()
    private var mLdapHandle: Int = 0
    private lateinit var mApi: SipLdapApi
    private var mTotalFilters = 0
    private var mCurrentState: Ldap.LdapState = Ldap.LdapState.Disconnected
    private var mCurrentError: Ldap.LdapErrorType = Ldap.LdapErrorType.NoError
    private var isDataLoading = false


    private var mHandlerSipLdapHandler = object : SipLdapHandler.SipLdapHandlerAdapter() {
        override fun onLdapData(var1: SipLdapApi?, evt: Ldap.onLdapDataEvent) {
            Log.i(TAG, "onStateChanged state")
            val newDirectorySearchData = ArrayList<LdapContactDataObject>()
            val dataList = evt.dataEntry
            for (dataEntry in dataList) {
                Log.i(TAG, "============ LDAP DATA  =============== ")
                Log.i(TAG, "onStateChanged: DisplayName = " + dataEntry.displayName)
                Log.i(TAG, "onStateChanged: FirstName = " + dataEntry.firstName)
                Log.i(TAG, "onStateChanged: LastName = " + dataEntry.lastName)
                Log.i(TAG, "onStateChanged: Softphone = " + dataEntry.softphone)
                Log.i(TAG, "onStateChanged: JobTitle = " + dataEntry.jobTitle)
                Log.i(TAG, "onStateChanged: Department = " + dataEntry.department)
                Log.i(TAG, "onStateChanged: City = " + dataEntry.city)
                Log.i(TAG, "onStateChanged: WorkPhone = " + dataEntry.workPhone)
                Log.i(TAG, "onStateChanged: OfficePhone = " + dataEntry.officePhone)
                Log.i(TAG, "onStateChanged: HomePhone = " + dataEntry.homePhone)
                Log.i(TAG, "onStateChanged: MobilePhone = " + dataEntry.mobilePhone)
                Log.i(TAG, "onStateChanged: Email = " + dataEntry.email)
                Log.i(TAG, "onStateChanged: Jabber = " + dataEntry.jabber)

                val dataObject = LdapContactDataObject(
                        dataEntry.displayName, -1, dataEntry.firstName, dataEntry.lastName, dataEntry.mobilePhone)
                if (dataEntry.email.length != 0) {
                    dataObject.emailAddress = dataEntry.email
                }
                if (dataEntry.officePhone.length != 0) {
                    dataObject.officePhoneNumber = dataEntry.officePhone
                }
                if (dataEntry.workPhone.length != 0) {
                    dataObject.businessPhoneNumber = dataEntry.workPhone
                }
                if (dataEntry.homePhone.length != 0) {
                    dataObject.homePhoneNumber = dataEntry.homePhone
                }
                if (dataEntry.softphone.length != 0) {
                    dataObject.softphone = dataEntry.softphone
                }
                if (dataEntry.department.length != 0) {
                    dataObject.department = dataEntry.department
                }
                if (dataEntry.jabber.length != 0) {
                    dataObject.jabber = dataEntry.jabber
                }
                if (dataEntry.jobTitle.length != 0) {
                    dataObject.jobTitle = dataEntry.jobTitle
                }
                if (dataEntry.city.length != 0) {
                    dataObject.city = dataEntry.city
                }
                newDirectorySearchData.add(dataObject)
            }
            mDirectorySearchData.clear()
            mDirectorySearchData.addAll(newDirectorySearchData)
            fireOnDirectoryListUpdated()
        }

        override fun onStateChanged(var1: SipLdapApi?, evt: Ldap.onStateChangedEvent) {
            Log.i(TAG, "onStateChanged state = " + evt.ldapState)
            mCurrentState = evt.ldapState
/*
            if (evt == Ldap.LdapState.Connected && !mCurSearchString.isEmpty()) {
                setDirectorySearchString(mCurSearchString)
            }*/

        }

        override fun onError(var1: SipLdapApi?, evt: Ldap.onErrorEvent) {
            mCurrentError = evt.errorType
            //errorMessage = evt.errorText

            if (mCurrentError == Ldap.LdapErrorType.LdapConnectionError || mCurrentError == Ldap.LdapErrorType.LdapSearchError) {
                disconnect()
            }

            if (mCurrentError != Ldap.LdapErrorType.NoError) {
                mDirectorySearchData.clear()
                fireOnDirectoryListUpdated()
            }
        }
    }

    init {
        mStackManagerDisposable = SipStackManagerInstanceObservable
                .getObservable()
                .subscribe({ sipStackManager -> initLdapApi(sipStackManager) }) { error ->
                    run {

                        throw RuntimeException(error)
                    }
                }
    }


    private fun initLdapApi(sipStackManager: SipStackManager) {
        mApi = SipLdapApi.get(sipStackManager.getSipPhone())
        mLdapHandle = mApi.CreateClient()
        mApi.addHandler(mHandlerSipLdapHandler)
    }


    fun disconnect() {
        mApi.Disconnect(mLdapHandle)
    }


    fun connect() {
        isDataLoading = true
        applyDataSettings()
        mApi.Connect(mLdapHandle)
    }


    fun destroy() {
        disconnect()
        mApi.removeHandler(mHandlerSipLdapHandler)
    }

    private fun applyDataSettings() {
        val ldapSettings = Ldap.LdapClientSettings()
        val server = Settings.get(mContext).getStr(ESetting.LdapServer)

        if (TextUtils.isEmpty(server)) {
            Log.e(TAG, "Bad ldap server setting")
            mCurrentError = Ldap.LdapErrorType.LdapConnectionError
            return
        }

        var encryption: Ldap.LdapEncryption
        when (Settings.get(mContext).getStr(ESetting.LdapEncryption)) {
            "none" -> encryption = Ldap.LdapEncryption.None
            "ldaps" -> encryption = Ldap.LdapEncryption.Ldaps
            "starttls" -> encryption = Ldap.LdapEncryption.Starttls
            else -> encryption = Ldap.LdapEncryption.None
        }

        val certStrategy: Ldap.LdapCertStrategy
        when (Settings.get(mContext).getStr(ESetting.LdapCertStrategy)) {
            "never" -> certStrategy = Ldap.LdapCertStrategy.Never
            "hard" -> certStrategy = Ldap.LdapCertStrategy.Hard
            "demand" -> certStrategy = Ldap.LdapCertStrategy.Demand
            "allow" -> certStrategy = Ldap.LdapCertStrategy.Allow
            else -> certStrategy = Ldap.LdapCertStrategy.Hard
        }

        //if old "LdapUseSSL" setting is used
        if (encryption == Ldap.LdapEncryption.None && Settings.get(mContext).getBool(ESetting.LdapUseSSL)) {
            encryption = Ldap.LdapEncryption.Ldaps
        }

        val url: String
        if (encryption == Ldap.LdapEncryption.Ldaps) {
            url = "ldaps://" + server!!
        } else {
            url = "ldap://" + server!!
        }

        val username = Settings.get(mContext).getStr(ESetting.LdapUserName)
        val password = Settings.get(mContext).getStr(ESetting.LdapPassword)



        ldapSettings.userName = username
        ldapSettings.password = password
        ldapSettings.serverUrl = url
        ldapSettings.encryption = encryption
        ldapSettings.certStrategy = certStrategy
        ldapSettings.connectionTimeout = 15

        mApi.ApplySettings(mLdapHandle, ldapSettings)

        val map = Ldap.LdapDataMap()
        map.displayName = Settings.get(mContext).getStr(ESetting.DisplayName).trim({ it <= ' ' })
        map.firstName = Settings.get(mContext).getStr(ESetting.FirstName).trim({ it <= ' ' })
        map.lastName = Settings.get(mContext).getStr(ESetting.LastName).trim({ it <= ' ' })
        map.mobilePhone = Settings.get(mContext).getStr(ESetting.MobilePhone).trim({ it <= ' ' })
        map.officePhone = Settings.get(mContext).getStr(ESetting.OfficePhone).trim({ it <= ' ' })
        map.email = Settings.get(mContext).getStr(ESetting.Email).trim({ it <= ' ' })
        map.workPhone = Settings.get(mContext).getStr(ESetting.WorkPhone).trim({ it <= ' ' })
        map.homePhone = Settings.get(mContext).getStr(ESetting.HomePhone).trim({ it <= ' ' })
        map.softphone = Settings.get(mContext).getStr(ESetting.Softphone).trim({ it <= ' ' })
        map.department = Settings.get(mContext).getStr(ESetting.Department).trim({ it <= ' ' })
        map.jabber = Settings.get(mContext).getStr(ESetting.Jabber).trim({ it <= ' ' })
        map.jobTitle = Settings.get(mContext).getStr(ESetting.JobTitle).trim({ it <= ' ' })
        map.city = Settings.get(mContext).getStr(ESetting.City).trim({ it <= ' ' })

        mApi.SetDataMap(mLdapHandle, map)
    }

    fun getContact(number: String, displayName: String?): LdapContactDataObject? {
        Log.d(TAG, "getContact $number")

        val contactsMatchedByNumber = ArrayList<LdapContactDataObject>()
        for (contactObject in mDirectorySearchData) {
            if (contactObject.getSoftphone() != null && Validator.getSanitizedPhoneNumber(contactObject.getSoftphone()) == number ||
                    contactObject.getMobile() != null && Validator.getSanitizedPhoneNumber(contactObject.getMobile()) == number ||
                    contactObject.getHomePhoneNumber() != null && Validator.getSanitizedPhoneNumber(contactObject.getHomePhoneNumber()) == number ||
                    contactObject.getBusinessPhoneNumber() != null && Validator.getSanitizedPhoneNumber(contactObject.getBusinessPhoneNumber()) == number ||
                    contactObject.getOfficePhoneNumber() != null && Validator.getSanitizedPhoneNumber(contactObject.getOfficePhoneNumber()) == number) {
                contactsMatchedByNumber.add(contactObject)
            }
        }

        for (contactObject in contactsMatchedByNumber) {
            var ldapDisplayName = (contactObject.firstName + " " + contactObject.lastName).trim { it <= ' ' }
            if (TextUtils.isEmpty(ldapDisplayName)) {
                ldapDisplayName = contactObject.name
            }
            if (TextUtils.isEmpty(displayName) || displayName!!.equals(ldapDisplayName, ignoreCase = true)) {
                return contactObject
            }
        }

        return null
    }

    fun getContact(key: String?): LdapContactDataObject? {
        Log.d(TAG, "getContactByKey $key")

        for (contactDataObject in mDirectorySearchData) {
            if (contactDataObject.name != null && contactDataObject.name == key) {
                return contactDataObject
            }
        }

        return null
    }

    fun getDirectoryContactItem(position: Int): LdapContactDataObject? {
        if (position < 0 || position >= mDirectorySearchData.size) {
            Log.e(TAG, "Position [" + position + "] is out of bounds [0, " + (mDirectorySearchData.size - 1))
            return null
        } else {
            val contactItem = mDirectorySearchData[position]
            contactItem.id = position
            return contactItem
        }
    }

    fun getLdapList() = mDirectorySearchData

    fun attachWeakObserver(observer: ILdapModuleObserver) {
        m_observableAdapter.attachWeakObserver(observer)
    }

    fun detachObserver(observer: ILdapModuleObserver) {
        m_observableAdapter.detachObserver(observer)
    }


    private fun notifyObserver(notification: INotificationAction<ILdapModuleObserver>) {
        m_observableAdapter.notifyObservers(notification)
    }

    private fun fireOnDirectoryListUpdated() {
        isDataLoading = false;
        notifyObserver(INotificationAction { it.onDirectoryListUpdated() })
    }

    fun isDataLoading(): Boolean = isDataLoading

    fun setDirectorySearchString(searchString: String) {
        Log.d(TAG, "setDirectorySearchString $searchString")

        if (mCurrentState == Ldap.LdapState.Connecting) {
            //errorMessage = "Connecting to server..."
            Log.d(TAG, "setDirectorySearchString: Connecting to server...")
            return
        }

        if (mCurrentState == Ldap.LdapState.Disconnected) {
            Log.d(TAG, "setDirectorySearchString: reconnect")
            mApi.Connect(mLdapHandle)
            //   mCurSearchString = searchString
            return
        }

        if (searchString.isEmpty()) {
            //mCurSearchString = ""
            mDirectorySearchData.clear()
            fireOnDirectoryListUpdated()
            isDataLoading = false;
            return
        }

        // mCurSearchString = searchString

        val rootDn = Settings.get(mContext).getStr(ESetting.LdapRootDN)
        var baseDn = removeParenthesesIfExists(Settings.get(mContext).getStr(ESetting.LdapSearchBaseDN))

        var scopeSetting: Ldap.LdapSearchScope
        val scope = Settings.get(mContext).getInt(ESetting.LdapSearchScope)
        Log.d(TAG, "scope setting is $scope")
        when (scope) {
            0 -> scopeSetting = Ldap.LdapSearchScope.Base
            1 -> scopeSetting = Ldap.LdapSearchScope.OneLevel
            2 -> scopeSetting = Ldap.LdapSearchScope.SubTree
            3 -> scopeSetting = Ldap.LdapSearchScope.Children
            else -> throw IllegalArgumentException("LdapSearchScope invalid = $scope")
        }

        var altsearch = false
        val tempStr: String
        var altNeedle: String
        altNeedle = ""
        if (searchString.contains(" ")) {
            altsearch = true
            tempStr = searchString.substring(0, searchString.indexOf(" "))
            altNeedle = searchString.substring(searchString.indexOf(" ") + 1) + "* " + tempStr
        }

        mTotalFilters = 0
        var searchPattern = ""
        searchPattern += addLdapFilterIfEnabled(ESetting.LdapFilterDisplayName, ESetting.DisplayName, searchString)

        if (altsearch) {
            searchPattern += addLdapFilterIfEnabled(ESetting.LdapFilterDisplayName, ESetting.DisplayName, altNeedle)
        }

        searchPattern += addLdapFilterIfEnabled(ESetting.LdapFilterFirstName, ESetting.FirstName, searchString)
        searchPattern += addLdapFilterIfEnabled(ESetting.LdapFilterLastName, ESetting.LastName, searchString)
        searchPattern += addLdapFilterIfEnabled(ESetting.LdapFilterSoftphone, ESetting.Softphone, searchString)
        searchPattern += addLdapFilterIfEnabled(ESetting.LdapFilterJobTitle, ESetting.JobTitle, searchString)
        searchPattern += addLdapFilterIfEnabled(ESetting.LdapFilterDepartment, ESetting.Department, searchString)
        searchPattern += addLdapFilterIfEnabled(ESetting.LdapFilterCity, ESetting.City, searchString)    //label="Location"
        searchPattern += addLdapFilterIfEnabled(ESetting.LdapFilterWorkPhone, ESetting.WorkPhone, searchString)
        searchPattern += addLdapFilterIfEnabled(ESetting.LdapFilterOfficePhone, ESetting.OfficePhone, searchString)
        searchPattern += addLdapFilterIfEnabled(ESetting.LdapFilterHomePhone, ESetting.HomePhone, searchString)
        searchPattern += addLdapFilterIfEnabled(ESetting.LdapFilterMobilePhone, ESetting.MobilePhone, searchString)
        searchPattern += addLdapFilterIfEnabled(ESetting.LdapFilterEmail, ESetting.Email, searchString)
        searchPattern += addLdapFilterIfEnabled(ESetting.LdapFilterJabber, ESetting.Jabber, searchString)

        if (!TextUtils.isEmpty(baseDn)) {
            if (baseDn.contains("&")) { // do not add "&" if already has it
                searchPattern = "($baseDn(|$searchPattern))"
            } else {
                searchPattern = "(&($baseDn)(|$searchPattern))"
            }
        } else if (mTotalFilters > 1) {
            searchPattern = "($searchPattern)"
        }

        Log.d(TAG, "searchPattern: $searchPattern")
        mApi.Search(mLdapHandle, searchPattern, rootDn, scopeSetting, 15, 0, false)

    }

    private fun addLdapFilterIfEnabled(ldapFilter: ESetting, ldapFilterValue: ESetting, needle: String): String {
        val closingString: String
        if (needle.contains("*")) {
            closingString = ")"
        } else {
            closingString = "*)"
        }

        if (Settings.get(mContext).getBool(ldapFilter) && !TextUtils.isEmpty(Settings.get(mContext).getStr(ldapFilterValue))) {
            mTotalFilters++
            return "(" + Settings.get(mContext).getStr(ldapFilterValue) + "=" + needle + closingString
        }

        return ""
    }

    private fun removeParenthesesIfExists(input: String): String {
        return if (input.matches("\\(.+\\)".toRegex())) {
            input.substring(1, input.length - 1)
        } else input

    }

}

