package com.bria.common.controller.im.refactoring.adapter

import android.graphics.Typeface
import android.support.annotation.ColorInt
import android.text.SpannableStringBuilder
import android.text.Spanned
import android.text.TextPaint
import android.text.style.ClickableSpan
import android.text.style.ForegroundColorSpan
import android.text.style.StyleSpan
import android.view.View
import com.bria.common.controller.contacts.buddy.Buddy
import com.bria.common.controller.contacts.buddy.IBuddyCtrlEvents
import com.bria.common.controller.contacts.buddy.IBuddyCtrlObserver
import com.bria.common.controller.contacts.buddy.XmppBuddy
import com.bria.common.controller.im.ImpsUtils
import com.bria.common.controller.im.refactoring.chatroom.Mention
import com.bria.common.controller.im.refactoring.chatroom.MentionsFinder
import com.bria.common.controller.im.refactoring.chatroom.WellKnowMention
import com.bria.common.util.IObservable

class ImListAdapterMentionsHelper(
        private val buddyCtrl: IBuddyCtrlEvents,
        private val buddyObservable: IObservable<IBuddyCtrlObserver>
) {

    companion object {
        val spanTypesUsed = listOf(ForegroundColorSpan::class.java, ClickableSpan::class.java, StyleSpan::class.java)
    }

    private var mAllBuddiesWithSelfInfos: Collection<XmppBuddy>? = null

    private fun getAllBuddiesWithSelfInfos(): Collection<XmppBuddy> {
        val result = mAllBuddiesWithSelfInfos ?: buddyCtrl
            .allBuddies
            .plus(buddyCtrl.selfInfos)
            .map { it as? XmppBuddy }
            .filterNotNull()

        mAllBuddiesWithSelfInfos = result

        return result
    }

    private val mBuddyObserver: IBuddyCtrlObserver = object : IBuddyCtrlObserver {
        override fun onBuddyListUpdated() {
            // invalidate
            mAllBuddiesWithSelfInfos = null
        }

        override fun onBuddyPresenceChanged(buddy: Buddy?) {}
        override fun onBuddyFilterChanged(aBuddyFilterType: IBuddyCtrlEvents.EBuddyFilterType?) {}
    }

    init {
        buddyObservable.attachWeakObserver(mBuddyObserver)
    }


    fun splitMessageToChunks(messageText: String, mentions: List<Mention>): ChunkedString<Mention> {
        val mentionsSorted = mentions.sortedBy { it.start }
        var indexInString = 0
        var chunks = mutableListOf<ChunkedString.Chunk<Mention>>()
        for (mention in mentionsSorted) {
            if (indexInString < mention.start) {
                chunks.add(ChunkedString.Chunk(messageText.substring(indexInString, mention.start), null))
            }

            chunks.add(ChunkedString.Chunk(messageText.substring(mention.start, mention.start + mention.length), mention))

            indexInString = mention.start + mention.length
        }

        if (indexInString < messageText.length - 1) {
            chunks.add(ChunkedString.Chunk(messageText.substring(indexInString, messageText.length), null))
        }

        return ChunkedString(chunks)
    }

    fun filterKnownMentionsAndExpandUsernamesToDisplayNames(chunkedString: ChunkedString<Mention>): List<PlainTextOrMentionChunk> {
        fun isWellKnown(mentionText: String): Boolean = WellKnowMention.values().map { it.atText.toLowerCase() }.contains(mentionText.toLowerCase())

        fun findBuddy(mentionText: String): XmppBuddy? {
            val username = mentionText.trimStart('@')

            val buddy = getAllBuddiesWithSelfInfos()
                .firstOrNull { username.equals(ImpsUtils.getUsernameFromAddress(it.imAddress), true) }

            return buddy
        }

        fun mapToPlainTextOrMentionChunk(chunk: ChunkedString.Chunk<Mention>): PlainTextOrMentionChunk {
            if (chunk.tag == null) {
                return PlainTextOrMentionChunk.PlainText(chunk.text)
            }

            if (isWellKnown(chunk.text)) {
                return PlainTextOrMentionChunk.HardcodedMention(chunk.text)
            }

            val buddy = findBuddy(chunk.text)
            if (buddy == null) {
                // dont highlight
                return PlainTextOrMentionChunk.PlainText(chunk.text)
            }

            val displayName: String? = buddy.displayName
            if (!displayName.isNullOrBlank()) {
                val replacementText = "@$displayName"

                return PlainTextOrMentionChunk.BuddyMention(replacementText, buddy)
            }

            return PlainTextOrMentionChunk.BuddyMention(chunk.text, buddy)
        }

        return chunkedString.map(::mapToPlainTextOrMentionChunk)
    }

    fun processMentions(
            messageText: String?,
            @ColorInt spanForeColor: Int,
            isChatRoom: Boolean,
            onBuddyClicked: (XmppBuddy) -> Unit): CharSequence? {

        if (!isChatRoom) {
            // mentions work only in chat rooms
            return messageText
        }

        if (messageText == null || messageText.isNullOrBlank()) {
            // nothing to do when message is empty
            return messageText
        }

        val mentions = MentionsFinder.findMentions(messageText)
        if (mentions.isEmpty()) {
            // no one was mentioned
            return messageText
        }

        val chunkedString = splitMessageToChunks(messageText, mentions)

        val expandedString = filterKnownMentionsAndExpandUsernamesToDisplayNames(chunkedString)

        val spannableStringBuilder = SpannableStringBuilder()
        for (chunk in expandedString) {
            val start = spannableStringBuilder.length
            spannableStringBuilder.append(chunk.text)
            val end = spannableStringBuilder.length

            if (chunk is PlainTextOrMentionChunk.HardcodedMention ||
                    chunk is PlainTextOrMentionChunk.BuddyMention) {
                spannableStringBuilder.setSpan(ForegroundColorSpan(spanForeColor), start, end, Spanned.SPAN_EXCLUSIVE_EXCLUSIVE)
                spannableStringBuilder.setSpan(StyleSpan(Typeface.BOLD), start, end, Spanned.SPAN_EXCLUSIVE_EXCLUSIVE)
            }

            if (chunk is PlainTextOrMentionChunk.BuddyMention) {
                spannableStringBuilder.setSpan(object : ClickableSpan() {
                    override fun onClick(widget: View) {
                        onBuddyClicked(chunk.buddy)
                    }

                    override fun updateDrawState(ds: TextPaint) {
                        // do not call base
                        // base changes color, adds underline
                    }

                }, start, end, Spanned.SPAN_EXCLUSIVE_EXCLUSIVE)
            }
        }

        return spannableStringBuilder
    }
}
