package com.bria.common.controller.im.refactoring.chatroom

import android.annotation.SuppressLint
import android.content.Context
import android.text.TextUtils
import com.bria.common.controller.accounts.Account
import com.bria.common.controller.accounts.AccountsFilter
import com.bria.common.controller.accounts.IAccountsCtrlActions
import com.bria.common.controller.contacts.buddy.XmppBuddy
import com.bria.common.controller.im.ImpsUtils
import com.bria.common.controller.im.refactoring.BuddyKeyUtils
import com.bria.common.controller.im.refactoring.ChatApi
import com.bria.common.controller.im.refactoring.ChatApi.*
import com.bria.common.controller.im.refactoring.chatroom.ChatRoomCache.Companion.MEMBERS_LIST_TYPE
import com.bria.common.controller.im.refactoring.chatroom.ChatRoomCache.Companion.OWNERS_LIST_TYPE
import com.bria.common.controller.im.refactoring.chatroom.joinedroom.JoinRoomStorageField
import com.bria.common.controller.im.refactoring.chatroom.joinedroom.JoinedRoomData
import com.bria.common.controller.im.refactoring.chatroom.joinedroom.JoinedRoomSyncApi
import com.bria.common.controller.im.refactoring.chatroom.joinedroom.JoinedRoomSyncApiImpl
import com.bria.common.controller.im.refactoring.db.entities.ImConversationData
import com.bria.common.controller.im.refactoring.db.entities.ImStatusData
import com.bria.common.controller.im.refactoring.roomdb.ChatRoomRepo
import com.bria.common.controller.im.refactoring.roomdb.LocalRoomInfo
import com.bria.common.controller.im.refactoring.roomdb.entities.*
import com.bria.common.controller.im.refactoring.xmppgroup.CollabChatInfo
import com.bria.common.controller.settings.EAccountSetting
import com.bria.common.sdkwrapper.SipStackManager
import com.bria.common.uireusable.datatypes.ExtendedConversationData
import com.bria.common.util.AndroidLog
import com.bria.common.util.CrashInDebug
import com.bria.common.util.Log
import com.bria.common.util.im.Participant
import com.bria.common.util.im.ParticipantsSet
import com.counterpath.sdk.XmppAccount
import com.counterpath.sdk.XmppApiMultiUserChat
import com.counterpath.sdk.XmppMultiUserChat
import com.counterpath.sdk.handler.XmppMultiUserChatHandler
import com.counterpath.sdk.pb.Xmppmultiuserchat
import io.reactivex.Observable
import io.reactivex.Single
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.disposables.Disposable
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.PublishSubject
import io.reactivex.subjects.Subject
import java.text.SimpleDateFormat
import java.util.*
import java.util.concurrent.ScheduledExecutorService
import java.util.concurrent.TimeUnit
import kotlin.collections.HashMap
import kotlin.math.max


class ChatRoomApiImpl(val chatApi: ChatApi,
                      val chatRoomRepo: ChatRoomRepo,
                      val context: Context,
                      private val accountController: IAccountsCtrlActions,
                      val imExecutorService: ScheduledExecutorService,
                      private val chatRoomMemberCache: ChatRoomMemberCache) : ChatRoomApi {

    private val TAG = "ChatRoomApi"

    private val cache = ChatRoomCache()

    private val disposables = CompositeDisposable()

    private val tempMessages = mutableMapOf<Int, Message>()
    private val alreadySavedMessages = mutableListOf<String>()

    private val activeAccounts: MutableSet<XmppAccount> = hashSetOf()

    private val joinedRoomSyncApis: MutableMap<XmppAccount, JoinedRoomSyncApi> = mutableMapOf()

    private val syncDisposables: MutableMap<XmppAccount, Disposable> = mutableMapOf()

    private val serviceForAccount = HashMap<String, String>()

    private val waitingForInfo = mutableMapOf<XmppAccount, MutableList<String>>()

    private val roomInfoDisposables = mutableMapOf<XmppAccount, Disposable>()

    private val errorEventObservable: Subject<ChatRoomErrorEvent> = PublishSubject.create()

    private val roomStateObservable: Subject<ChatRoomStateEvent> = PublishSubject.create()

    private val roomListChangedObserver: Subject<List<String>> = PublishSubject.create()

    private val serviceAvailabilityObservable: Subject<Set<String>> = PublishSubject.create() //returns a set of accounstId with available services

    private val joinedRoomsToSync = mutableListOf<JoinedRoomData>()

    var downloadRoomInfos = HashMap<XmppAccount, Boolean>()

    private val stackManager: SipStackManager?
        get() = chatApi.stackManager


    init {
        imExecutorService.execute {
            disposables.add(chatRoomRepo.getOnRoomAddedObservable().observeOn(Schedulers.from(imExecutorService)).subscribe({ list ->
                Log.d(TAG, "rooms created")
                list.forEach { chatRoom ->
                    val time = System.currentTimeMillis()
                    val readState = activeAccounts.firstOrNull { it.getUserAtDomain() == chatRoom.accountId }?.let {
                        chatRoom.joinTime < joinedRoomSyncApis[it]?.getLastViewedMessageTimeInMs(chatRoom.chatKey) ?: 0
                    } ?: false
                    if (!chatRoom.isUserOwner() && !chatRoom.isCollab()) {
                        Log.d(TAG, "Local message added: ${chatRoom.name} public: ${chatRoom.publicRoom}")
                        val text = if (chatRoom.publicRoom) "$PARTICIPANT_LOCAL_JOINED_EVENT_ID$LOCAL_MESSAGE_SEPARATOR${chatRoom.accountId}" else
                            "$PARTICIPANT_LOCAL_INVITED_EVENT_ID$LOCAL_MESSAGE_SEPARATOR${chatRoom.accountId}"
                        val message = Message(chatId = chatRoom.id, creationTime = chatRoom.joinTime,
                                readState = if (readState) MessageReadState.READ else MessageReadState.UNREAD_NOT_SWIPED,
                                modificationTime = chatRoom.joinTime,
                                status = ImStatusData.LOCAL,
                                text = text,
                                externalId = "$time")
                        chatRoomRepo.addMessage(message)
                    }
                }
            }, {}))

            disposables.add(chatRoomRepo.getOnRoomDeletedObservable().subscribe({ chatRoomList ->
                chatRoomList.forEach { room ->
                    cache.deleteChatRoom(room.chatKey)
                }
            }, {}))
        }
    }

    private val pendingRoomSyncing = mutableMapOf<XmppAccount, List<JoinedRoomData>>()

    override val onErrorEventObservable = errorEventObservable

    override fun getOnChatRoomStateObservable() = roomStateObservable

    override fun getOnServiceAvailabilityObservable() = serviceAvailabilityObservable

    override fun accountConnected(xmppAccount: XmppAccount) {
        accountDisconnected(xmppAccount)
        clearCacheForAccount(xmppAccount)
        imExecutorService.execute {
            Log.d(TAG, "accountConnected ${xmppAccount.getUserAtDomain()}")
            activeAccounts.add(xmppAccount)
            if (joinedRoomSyncApis.containsKey(xmppAccount)) joinedRoomSyncApis.remove(xmppAccount)
            val joinedRoomSyncApi: JoinedRoomSyncApi = JoinedRoomSyncApiImpl(xmppAccount, imExecutorService, chatRoomRepo)
            joinedRoomSyncApis[xmppAccount] = joinedRoomSyncApi
            syncDisposables.put(xmppAccount, joinedRoomSyncApi.joinRoomsListChangedObservable().subscribe({ activeJoinedRoom ->
                if (serviceForAccount.containsKey(xmppAccount.getUserAtDomain())) {
                    applyServerData(activeJoinedRoom, xmppAccount)
                } else {
                    pendingRoomSyncing[xmppAccount] = activeJoinedRoom
                }
            }, { error ->
                Log.e(TAG, "on Error: ${error.message}")
            }, {}))
            joinedRoomSyncApi.fetchAndUpdate()
        }
    }

    override fun accountDisconnected(xmppAccount: XmppAccount) {
        Log.d(TAG, "accountDisconnected ${xmppAccount.getUserAtDomain()}")
        activeAccounts.remove(xmppAccount)
        joinedRoomSyncApis[xmppAccount]?.destroy()
        joinedRoomSyncApis.remove(xmppAccount)
        syncDisposables.remove(xmppAccount)?.dispose()
        serviceForAccount.remove(xmppAccount.getUserAtDomain())
        serviceAvailabilityObservable.onNext(serviceForAccount.keys)
    }


    override fun createRoom(chatRoom: ChatRoom) {
        imExecutorService.execute {
            getXmppAccount(chatRoom.accountId)?.let {
                val chatApi = XmppApiMultiUserChat.get(it)
                chatApi.createChat("")?.let { xmppChat ->
                    xmppChat.join(false, true, it.settings.username, "", getLastMessageDateString(chatRoom), arrayListOf())
                    Log.d(TAG, "requestingConfiguration")
                    xmppChat.requestConfigurations()
                    cache.add(xmppChat, chatRoom)
                    cache.setState(xmppChat, state = ChatRoomCacheState.JOINED_WAITING_FOR_SYNC)
                }
            }
        }
    }

    //just for testing
    override fun forceOffline(chatKey : String) {
        imExecutorService.execute {
            Log.d(TAG, "OnLocalUserLeft $chatKey}")
            cache.getXmppChat(chatKey)?.let { chat ->
                chatRoomRepo.getChatRoomByChatKey(chatKey)?.let { chatRoom ->
                    if (serviceForAccount.containsKey(chatRoom.accountId)) {
                        cache.remove(chat)
                        chatRoom.state = ChatRoomState.OFFLINE
                        chatRoomRepo.updateChatRoom(chatRoom)
                    }
                }
            }
        }
    }

    override fun initCollabChat(roomId: String, username: String, password: String, account: XmppAccount): Single<Long> {
        return Single.fromCallable { initCollabSync(roomId, username, password, account) }.subscribeOn(Schedulers.from(imExecutorService))
    }

    private fun initCollabSync(roomId: String, username: String, password: String, account: XmppAccount): Long {
        cache.addAsCollabChat(roomId)
        var dbChat = chatRoomRepo.getChatRoomByChatKey(roomId)
        var cachedChat = cache.getChatRoomCacheItem(roomId)?.xmppChat
        if (dbChat != null && cachedChat != null) {
            return dbChat.id
        }
        if (serviceForAccount[account.getUserAtDomain()] == null) {
            cache.addToCollabWaitingList(CollabChatInfo(roomId, username, password, account))
            return -1L
        }
        if (cachedChat == null) {
            //remove old messages
            if (dbChat != null) {
                chatRoomRepo.deleteChatRoom(dbChat)
                dbChat = null
            }
            cachedChat = XmppApiMultiUserChat.get(account).createChat(roomId)
            val tempHistorylist = ArrayList<Xmppmultiuserchat.XmppMultiUserChatHistoryItem>()
            //val deltaTimeStr = "seconds:7200" //2* 60 * 60 give me all messages in last 2 hours
            val deltaTimeStr = "seconds:0" //as decided on 2019-02-05 Chat Room metting
            cachedChat?.join(true, true, username, "", deltaTimeStr, tempHistorylist)
            cache.add(cachedChat, roomId, isCollab = true)
            setChatRoomState(roomId, ChatRoomCacheState.JOINED_NOT_SYNCABLE)
            cachedChat.requestConfigurations()
            Log.d(TAG, "initCollabChat: chat added: $roomId")
        }
        return if (dbChat == null) {
            val chatRoom = ChatRoom(
                    type = ImConversationData.GROUPCHAT_COLLAB.toLong(),
                    chatKey = roomId,
                    name = "Vccs Room",
                    accountId = account.getUserAtDomain()
            )
            chatRoomRepo.addChatRoom(chatRoom)
        } else {
            dbChat.id
        }
    }

    private fun setChatRoomState(roomId: String, state: ChatRoomCacheState) {
        val stateEvent = ChatRoomStateEvent(roomId, state)
        cache.setState(stateEvent)
        Log.d(TAG, "Room $roomId state changed: ${state.name}")
        roomStateObservable.onNext(stateEvent)
    }


    override fun isGroupChatCapable(buddyKey: String): Boolean {
        val accountKey = BuddyKeyUtils.getAccountFromNewBuddyKey(buddyKey)
        val userAtDomain = ImpsUtils.removePortFromDomain(accountKey)
        return serviceForAccount.containsKey(userAtDomain)
    }

    override fun isChatRoomActive(chatRoomKey: String) = cache.joined(chatRoomKey)

    override fun isServiceAvailable(userAtDomain: String): Boolean {
        return serviceForAccount.containsKey(userAtDomain)
    }

    override fun isServiceAvailableForAnyAccount(): Boolean {
        return serviceForAccount.isNotEmpty()
    }


    private fun getXmppAccount(accId: String): XmppAccount? {
        accountController.getAccounts(AccountsFilter.XMPP).firstOrNull { accId == it.getUserAtDomain() && it.isRegistered }?.let {
            return stackManager?.registrationManager?.getXmppAccount(it)
        }
        return null
    }

    override fun sendMessage(chatRoom: ChatRoom, message: String) {
        imExecutorService.execute {
            val id = cache.getXmppChat(chatRoom.chatKey)?.sendMessage(message, "") ?: -1
            if (id == -1) {
                Log.d(TAG, "sendMessage: Chat room not found in cache")
                return@execute
            }
            val time = Date().time
            val msg = Message(
                    chatId = chatRoom.id,
                    status = ImStatusData.OUTGOING_PENDING,
                    text = message,
                    remoteAddress = chatRoom.chatKey,
                    modificationTime = time,
                    creationTime = time)
            tempMessages[id] = msg
        }
    }

    private fun joinSyncedRooms(list: List<JoinedRoomData>, account: XmppAccount) {
        val chatApi = XmppApiMultiUserChat.get(account)
        val roomsInDb = chatRoomRepo.getAllChatRooms().blockingGet()
        val roomKeysInDb = roomsInDb.map { it.chatKey }
        list.forEach { joinedRoomData ->
            Log.d(TAG, "Cached in sync: ${joinedRoomData.roomId}")
            if (cache.getXmppChat(joinedRoomData.roomId) == null || !cache.joined(joinedRoomData.roomId)) {
                Log.d(TAG, "Joining room from sync: ${joinedRoomData.roomId} join time: ${joinedRoomData.joinTime}")
                if (!roomKeysInDb.contains(joinedRoomData.roomId)) {
                    Pair(ChatRoom(
                            type = ExtendedConversationData.CHAT_ROOM_TYPE.toLong(),
                            chatKey = joinedRoomData.roomId,
                            accountId = chatApi.account.getUserAtDomain(),
                            creationTime = joinedRoomData.joinTime * 1000,
                            joinTime = joinedRoomData.joinTime * 1000

                    ), false)
                } else {
                    Pair(roomsInDb.firstOrNull { it.hasSameChatKey(joinedRoomData.roomId) }, true)
                }.let { (chatRoom, _) ->
                    if (chatRoom != null) {
                        chatRoom.owners = if (joinedRoomData.roomCreator.isNotEmpty()) ParticipantsSet(BuddyKeyUtils.getNewBuddyKey(Participant.EParticipantType.XMPP,
                                account.getUserAtDomain(), joinedRoomData.roomCreator)) else ParticipantsSet()
                        chatRoom.joinTime = joinedRoomData.joinTime * 1000
                        chatRoom.modTime = joinedRoomData.joinTime * 1000
                        if (chatRoom.creationTime == 0L && chatRoom.isUserOwner()) {
                            chatRoom.creationTime = joinedRoomData.joinTime * 1000

                        }
                        chatRoom.lastViewedMessage = joinedRoomData.lastViewed * 1000
                        if (joinedRoomData.roomName.isNotEmpty()) {
                            chatRoom.name = joinedRoomData.roomName
                        }
                        getMultiUserChat(chatRoom.chatKey, chatApi)?.let { chat ->
                            cache.add(chat, chatRoom.chatKey)
                            setChatRoomState(chatRoom.chatKey, ChatRoomCacheState.WAITING_TO_JOIN_FROM_SYNC)
                            cache.getChatRoomCacheItem(chat)?.notSavedchatRoom = chatRoom
                            chat.getRoomInfo()
                            Log.d(TAG, "Planning to Join to room ${chatRoom.chatKey} with string: ${getLastMessageDateString(chatRoom)}")
                        }
//                        if (!cache.joined(chatRoom.chatKey)) { //if not joined
//                            //                         if (inDb) {
////                            chatRoomRepo.updateChatRoom(chatRoom)
////                        } else {
////                            chatRoomRepo.addChatRoom(chatRoom)
////                        }
//                            Log.d(TAG, "sync joining room ${chatRoom}")
//                            getMultiUserChat(chatRoom.chatKey, chatApi)?.let { chat ->
//                                chat.join(true, false, chatApi.account.settings.username, "", getLastMessageDateString(chatRoom), arrayListOf())
//                                cache.add(chat, chatRoom.chatKey)
//                                cache.setState(chat, ChatRoomCacheState.JOINED_SYNCED)
//                                Log.d(TAG, "Room joining synced room: ${chatRoom.chatKey} since string: ${getLastMessageDateString(chatRoom)}")
//                                chat.getRoomInfo()
//                            }
//                        }
                    }
                }
            } else {
                roomsInDb.firstOrNull { it.hasSameChatKey(joinedRoomData.roomId) }?.let { chatRoom ->
                    val owner = if (joinedRoomData.roomCreator.isNotEmpty()) ParticipantsSet(BuddyKeyUtils.getNewBuddyKey(Participant.EParticipantType.XMPP,
                            account.getUserAtDomain(), joinedRoomData.roomCreator)) else ParticipantsSet()
                    if (chatRoom.owners.toString() != owner.toString()
                            || chatRoom.joinTime != joinedRoomData.joinTime * 1000
                            || (joinedRoomData.roomName.isNotEmpty() && chatRoom.name != joinedRoomData.roomName)) {
                        chatRoom.owners = owner
                        chatRoom.joinTime = joinedRoomData.joinTime * 1000
                        chatRoom.modTime = max(joinedRoomData.joinTime * 1000, chatRoom.modTime)
                        if (joinedRoomData.roomName.isNotEmpty()) {
                            chatRoom.name = joinedRoomData.roomName
                        }
                        chatRoomRepo.updateChatRoom(chatRoom)
                    }
                    if (chatRoom.lastViewedMessage < joinedRoomData.lastViewed * 1000) {
                        chatRoom.lastViewedMessage = joinedRoomData.lastViewed * 1000
                    }
                }
            }
        }
    }

    private fun leaveNonSyncedRooms(list: List<JoinedRoomData>) {
        chatRoomRepo.getAllChatRooms().blockingGet()
                .forEach { dbChatRoom ->
                    if (list.firstOrNull { dbChatRoom.hasSameChatKey(it.roomId) } == null) {
                        if (cache.getState(dbChatRoom.chatKey) == ChatRoomCacheState.JOINED_SYNCED || !chatRoomRepo.roomInTheServerList(dbChatRoom.jid)) {
                            Log.d(TAG, "leaveNonSynced: ${dbChatRoom.chatKey}")
                            setChatRoomState(dbChatRoom.chatKey, state = ChatRoomCacheState.LOCALLY_LEFT)
                            cache.remove(dbChatRoom.chatKey)
                        }
                        Log.d(TAG, "deleting room from sync ${dbChatRoom.chatKey} name: ${dbChatRoom.name}")
                        chatRoomRepo.deleteChatRoom(dbChatRoom)
                    }
                }
    }

    private fun leavePublicRoomsNotOnServer() {
        chatRoomRepo.getAllChatRooms().blockingGet().forEach { dbChatRoom ->
            if (dbChatRoom.publicRoom && !chatRoomRepo.roomInTheServerList(dbChatRoom.chatKey) && dbChatRoom.state != ChatRoomState.OFFLINE) {
                dbChatRoom.state = ChatRoomState.OFFLINE
                chatRoomRepo.updateChatRoom(dbChatRoom)
            }
        }
    }

    private fun getLastMessageDateString(chatRoom: ChatRoom): String {
        val messageData = chatRoomRepo.getLastMessageForChatRoom(chatRoom.id).blockingGet()
        return "since:${(messageData?.creationTime ?: 0).div(1000)}"
        //return "message:300"
    }


    override fun inviteMembers(chatRoom: ChatRoom) {
        imExecutorService.execute {
            cache.getXmppChat(chatRoom.chatKey)?.let { chat ->
                inviteUninvited(chatRoom, chat)
            }
        }
    }

    override fun updateRoomDetails(chatRoom: ChatRoom) {
        imExecutorService.execute {
            chatRoomRepo.updateChatRoom(chatRoom)
            cache.getXmppChat(chatRoom.chatKey)?.let { chat ->
                chat.requestConfigurations()
                updateTopic(chatRoom.topic, chat)
                //inviteUninvited(chatRoom, chat)
            }
        }
    }

    private fun updateTopic(topic : String, chat: XmppMultiUserChat) {
        if (topic.isEmpty()) {
            chat.changeSubject(" ")
        } else {
            chat.changeSubject(topic)
        }
    }

    override fun removeMember(chatRoom: ChatRoom, participant: Participant) {
        imExecutorService.execute {
            cache.getXmppChat(chatRoom.chatKey)?.kick(participant.remoteAddress, "")
        }
    }

    override fun clearCacheForAccount(account: XmppAccount) {
        Log.d(TAG, "ClearCacheForAccount ${account.getUserAtDomain()}")
        imExecutorService.execute {
            cache.clearCacheForAccount(account)
        }
    }

    fun getRoomInfo(api: XmppApiMultiUserChat, chatKey: String) {
        Log.d(TAG, "getRoomInfo: $chatKey ")
        val chat = getMultiUserChat(chatKey, api)
        cache.add(chat, chatKey)
        chat.getRoomInfo()
    }

    override fun getRoomInfo(accId: String, chatKey: String) {
        getXmppAccount(accId)?.let {
            XmppApiMultiUserChat.get(it)?.let {
                getRoomInfo(it, chatKey)
            }
        }
    }

    override fun requestMembersList(chatRoom: ChatRoom) :Boolean {
        cache.getXmppChat(chatRoom.chatKey)?.let {
            requestMemberList(chatRoom = chatRoom, chat = it)

            return true
        }

        return false
    }

    @SuppressLint("CheckResult")
    override fun joinRooms(roomsJid: List<LocalRoomInfo>, account: XmppAccount) {
        imExecutorService.execute {
            val chatApi = XmppApiMultiUserChat.get(account)
            val joinTime = System.currentTimeMillis()
            roomsJid.forEach { room ->
                val chatRoom = ChatRoom(
                        type = ExtendedConversationData.CHAT_ROOM_TYPE.toLong(),
                        chatKey = room.cleanedJid,
                        name = room.name,
                        publicRoom = true, //from this point it must be public
                        accountId = account.getUserAtDomain(),
                        joinTime = joinTime,
                        modTime = joinTime
                )
                getMultiUserChat(chatRoom.chatKey, chatApi)?.let { xmppChat ->
                    cache.add(xmppChat, chatRoom.chatKey)
                    setChatRoomState(chatRoom.chatKey, ChatRoomCacheState.WAITING_TO_JOIN)
                    cache.getChatRoomCacheItem(xmppChat)?.notSavedchatRoom = chatRoom
                    xmppChat.getRoomInfo()
                    Log.d(TAG, "Planning to Join to room ${chatRoom.chatKey} with string: ${getLastMessageDateString(chatRoom)}")
                }
            }
        }
    }

    private fun getMultiUserChat(chatRoomKey: String, api: XmppApiMultiUserChat) = cache.getXmppChat(chatRoomKey) ?: api.createChat(chatRoomKey)

    override fun deleteRoom(chatRoom: ChatRoom) {
        imExecutorService.execute {
            cache.getXmppChat(chatRoom.chatKey)?.let { xmppChat ->
                if (chatRoom.isUserOwner()) {
                    xmppChat.destroy("", "", "")
                } else {
                    if (!cache.isLeft(chatRoom.chatKey)) {
                        xmppChat.changeJidAffiliation(xmppChat.api.account.getUserAtDomain(), 0, "")
                        xmppChat.leave("")
                    }
                }
                cache.remove(xmppChat)
                joinedRoomSyncApis[xmppChat.api.account]?.removeJoinedRoom(JoinedRoomData(roomId = chatRoom.chatKey))
            }
            chatRoomRepo.deleteChatRoom(chatRoom)

        }
    }

    override fun deleteCollabChatRoom(jid : String) {
        imExecutorService.execute{
            cache.getXmppChat(jid)?.let { xmppChat ->
                xmppChat.leave("")
                cache.remove(xmppChat)
            }
            chatRoomRepo.getChatRoomByChatKey(jid)?.let {
                chatRoomRepo.deleteChatRoom(it)
            }

        }
    }

    //unreachable
    override fun leaveRoom(chatRoom: ChatRoom) {
        imExecutorService.execute {
            Log.d(TAG, "leaveRoom: $chatRoom.chatKey")
            cache.getXmppChat(chatRoom.chatKey)?.let { xmppChat ->
                if (!cache.isLeft(chatRoom.chatKey)) {
                    if (!chatRoom.isUserOwner()) {
                        xmppChat.changeJidAffiliation(xmppChat.api.account.getUserAtDomain(), 0, "")
                    }
                    xmppChat.leave("")
                }
                cache.remove(xmppChat)
                joinedRoomSyncApis[xmppChat.api.account]?.removeJoinedRoom(JoinedRoomData(roomId = chatRoom.chatKey))
            }
            chatRoomRepo.deleteChatRoom(chatRoom)
        }
    }

    private fun requestMemberList(chat: XmppMultiUserChat, chatRoom: ChatRoom) {
        //nvtd add debounce for multiple listrequest
        imExecutorService.execute {
            Log.d(TAG, "requestMemberList ${chatRoom.chatKey} in cache: ${cache.getChatKeyByChat(chat)}")
            if (chatRoom.publicRoom) {
                chat.requestList(MEMBERS_LIST_TYPE)
            } else {
                chat.requestList(OWNERS_LIST_TYPE)
            }
        }
    }

    override fun getChatHandler(): XmppMultiUserChatHandler = handler

    val handler: XmppMultiUserChatHandler = object : XmppMultiUserChatHandler {

        @Synchronized
        override fun onRoomBookmarksReceived(chat: XmppMultiUserChat, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.RoomBookmarksReceivedEvent) {
            AndroidLog.d(TAG, "onRoomBookmarksReceived")
        }

        override fun onServiceAvailability(chat: XmppMultiUserChat, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.ServiceAvailabilityEvent) {
            imExecutorService.execute {
                AndroidLog.d(TAG, "onServiceAvailability")
                val manager = stackManager ?: return@execute
                com.bria.common.util.Log.d(TAG, "onServiceAvailability() - " + evt.available + " " + evt.service)
                if (!evt.available || TextUtils.isEmpty(evt.service)) {
                    return@execute
                }
                if (!(chat.api != null && chat.api.account != null)) {
                    return@execute
                }
                val xmppAccount = chat.api.account
                manager.registrationManager.getAccount(xmppAccount) ?: return@execute
                chat.api.getRoomList()
                serviceForAccount[chat.api.account.getUserAtDomain()] = evt.service
                pendingRoomSyncing[chat.api.account]?.let {
                    if (it.isNotEmpty()) {
                        applyServerData(it, xmppAccount)
                    }
                }
                downloadRoomInfos[chat.api.account] = false
                val infos = cache.getWaitingCollabList()
                infos.forEach { (roomId, username, password, account1) -> initCollabChat(roomId, username, password, account1) }
                infos.clear()
                serviceAvailabilityObservable.onNext(serviceForAccount.keys)
            }
        }


        override fun onMultiUserChatConfigurationRequested(chat: XmppMultiUserChat,
                                                           evt: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatConfigurationRequestedEvent) {
            imExecutorService.execute {
                Log.d(TAG, "onMultiUserChatConfigurationRequested")
                val key = cache.getChatKeyByChat(chat) ?: return@execute

                val chatRoom = (if (key == "INVALID") {
                    cache.getChatRoomCacheItem(chat)?.notSavedchatRoom ?: return@execute
                } else {
                    chatRoomRepo.getChatRoomByChatKey(key)
                }) ?: return@execute

                Log.d(TAG, "onMultiUserChatConfigurationRequested1: " + evt.configurations.instructions.joinToString(separator = "/n"))

                evt.xmppDataForm.fields.forEach {
                    val values = it.values.joinToString(separator = "||")
                    Log.d(TAG, "label: ${it.label} ;name: ${it.name}; required: ${it.required}; type: ${it.type}; values:$values")
                }

                val fields = mutableMapOf<String, com.counterpath.sdk.pb.nano.Xmpp.XmppDataFormField>()
                evt.xmppDataForm.fields.forEach {
                    fields[it.name] = it
                }

                //muc#roomconfig_roomname
                if (!fields.containsKey("muc#roomconfig_roomname")) {
                    val newField = com.counterpath.sdk.pb.nano.Xmpp.XmppDataFormField()
                    newField.type = 9
                    newField.name = "muc#roomconfig_roomname"
                    newField.values = arrayOf(chatRoom.name)
                    fields["muc#roomconfig_roomname"] = newField
                } else {
                    fields["muc#roomconfig_roomname"]?.values = arrayOf(chatRoom.name)
                }

                if (!fields.containsKey("muc#roomconfig_roomdesc")) {
                    val newField = com.counterpath.sdk.pb.nano.Xmpp.XmppDataFormField()
                    newField.type = 9
                    newField.name = "muc#roomconfig_roomdesc"
                    newField.values = arrayOf(chatRoom.description)
                    fields["muc#roomconfig_roomdesc"] = newField
                } else {
                    fields["muc#roomconfig_roomdesc"]?.values = arrayOf(chatRoom.description)
                }
                if (!chatRoom.publicRoom) {
                    if (!fields.containsKey("muc#roomconfig_membersonly")) {
                        val newField = com.counterpath.sdk.pb.nano.Xmpp.XmppDataFormField()
                        newField.type = 0
                        newField.name = "muc#roomconfig_membersonly"
                        newField.values = arrayOf((if (chatRoom.isCollab()) "0" else "1"))
                        fields["muc#roomconfig_membersonly"] = newField
                    } else {
                        fields["muc#roomconfig_membersonly"]?.values = arrayOf((if (chatRoom.isCollab()) "0" else "1"))
                    }
                }
                //nvtd add for nonpersistant
                if (!fields.containsKey("muc#roomconfig_persistentroom")) {
                    val newField = com.counterpath.sdk.pb.nano.Xmpp.XmppDataFormField()
                    newField.type = 0
                    newField.name = "muc#roomconfig_persistentroom"
                    newField.values = arrayOf(if (chatRoom.isCollab()) "0" else "1")
                    fields["muc#roomconfig_persistentroom"] = newField
                } else {
                    fields["muc#roomconfig_persistentroom"]?.values = arrayOf((if (chatRoom.isCollab()) "0" else "1"))
                }

                if (!fields.containsKey("muc#roomconfig_publicroom")) {
                    val newField = com.counterpath.sdk.pb.nano.Xmpp.XmppDataFormField()
                    newField.type = 0
                    newField.name = "muc#roomconfig_publicroom"
                    newField.values = arrayOf(if (!chatRoom.isCollab() && chatRoom.publicRoom) "1" else "0")
                    fields["muc#roomconfig_publicroom"] = newField
                } else {
                    fields["muc#roomconfig_publicroom"]?.values = arrayOf(if (!chatRoom.isCollab() && chatRoom.publicRoom) "1" else "0")
                }

                //max history size:
                if (!fields.containsKey("muc#maxhistoryfetch")) {
                    val newField = com.counterpath.sdk.pb.nano.Xmpp.XmppDataFormField()
                    newField.type = 9  //GLOOX_API DataFormField
                    newField.name = "muc#maxhistoryfetch"
                    newField.values = arrayOf("300")
                    fields["muc#maxhistoryfetch"] = newField
                } else {
                    fields["muc#maxhistoryfetch"]?.values = arrayOf("300")
                }


                val whois = when (chatRoom.anonymousMode) {

                    ChatRoomAnonymousMode.NON_ANONYMOUS -> "anyone"
                    ChatRoomAnonymousMode.SEMI_ANONYMOUS -> "moderators"
                    ChatRoomAnonymousMode.FULL_ANONYMOUS -> "none"
                }

                if (!fields.containsKey("muc#roomconfig_whois")) {
                    val newField = com.counterpath.sdk.pb.nano.Xmpp.XmppDataFormField()
                    newField.type = 6
                    newField.name = "muc#roomconfig_whois"
                    newField.values = arrayOf(whois)
                    fields["muc#roomconfig_whois"] = newField
                } else {
                    fields["muc#roomconfig_whois"]?.values = arrayOf(whois)
                }

                if (!fields.containsKey("muc#roomconfig_anonymity")) {
                    val newField = com.counterpath.sdk.pb.nano.Xmpp.XmppDataFormField()
                    newField.type = 0
                    newField.name = "muc#roomconfig_anonymity"
                    newField.values = arrayOf(chatRoom.anonymousMode.intValue.toString())
                    fields["muc#roomconfig_anonymity"] = newField
                } else {
                    fields["muc#roomconfig_anonymity"]?.values = arrayOf(chatRoom.anonymousMode.value)
                }

                if (!fields.containsKey("muc#roomconfig_allowinvites")) {
                    val newField = com.counterpath.sdk.pb.nano.Xmpp.XmppDataFormField()
                    newField.type = 0
                    newField.name = "muc#roomconfig_allowinvites"
                    newField.values = arrayOf("1")
                    fields["muc#roomconfig_allowinvites"] = newField
                } else {
                    fields["muc#roomconfig_allowinvites"]?.values = arrayOf("1")
                }

                if (!fields.containsKey("muc#roomconfig_changesubject")) {
                    val newField = com.counterpath.sdk.pb.nano.Xmpp.XmppDataFormField()
                    newField.type = 0
                    newField.name = "muc#roomconfig_changesubject"
                    newField.values = arrayOf("1")
                    fields["muc#roomconfig_changesubject"] = newField
                } else {
                    fields["muc#roomconfig_changesubject"]?.values = arrayOf("1")
                }

                evt.xmppDataForm.fields = fields.values.toTypedArray()
                chat.setConfigurations(evt.xmppDataForm)
            }
        }

        override fun onMultiUserChatRoomStateChanged(chat: XmppMultiUserChat, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatRoomStateChangedEvent) {
            imExecutorService.execute {
                Log.d(TAG, "onMultiUserChatRoomStateChanged: name ${msg.state.name}| is ready ${msg.state.isReady}| creation ${msg.state.creation} |description: " +
                        "${msg.state.description} |isCreatedBySelf ${msg.state.isCreatedBySelf} |isPersistent ${msg.state.isPersistent} |public ${msg.state
                                .isPublic} | anonymousMode ${msg.state.anonymousMode}| key: ${cache.getChatKeyByChat(chat)}")
                val state = msg.state
                cache.getChatKeyByChat(chat)?.let { jid ->
                    updateLocalRoomInfo(jid, state, msg)

                    var chatRoom: ChatRoom? = chatRoomRepo.getChatRoomByChatKey(jid)
                    val inDb = chatRoom != null
                    val chatRoomCachedItem: ChatRoomCacheItem? = cache.getChatRoomCacheItem(jid)
                    if (chatRoom == null) {
                        if (chatRoomCachedItem?.state in arrayListOf(ChatRoomCacheState.INVITED, ChatRoomCacheState.WAITING_TO_JOIN, ChatRoomCacheState.WAITING_TO_JOIN_FROM_SYNC)) {
                            chatRoom = chatRoomCachedItem?.notSavedchatRoom
                        }
                    }
                    chatRoom?.let { chatRoom ->
                        Log.d(TAG, "Room Info name| old ${chatRoom.name} | new ${state.name}")
                        if (state.name.isNotEmpty()) {
                            chatRoom.name = state.name
                            if (chatRoomCachedItem?.state == ChatRoomCacheState.JOINED_SYNCED) {
                                joinedRoomSyncApis[chat.api.account]?.updateField(Pair(chatRoom.chatKey, Pair(JoinRoomStorageField.ROOM_NAME, state.name)))
                            }
                        }
                        //chatRoom.creationTime = formatTZStringDate(state.creation).time
                        chatRoom.description = state.description
                        chatRoom.publicRoom = state.isPublic
                        chatRoom.anonymousMode = when (state.anonymousMode) {
                            0 -> ChatRoomAnonymousMode.NON_ANONYMOUS
                            1 -> ChatRoomAnonymousMode.SEMI_ANONYMOUS
                            2 -> ChatRoomAnonymousMode.FULL_ANONYMOUS
                            else -> ChatRoomAnonymousMode.NON_ANONYMOUS
                        }

                        chatRoom.topic = if (state.subject == " ") "" else state.subject
                        chatRoom.open = state.isOpen
                        chatRoom.moderated = state.isModerated
                        when (chatRoomCachedItem?.state) {
                            ChatRoomCacheState.NOT_IN_CHAT_ROOM -> {
                            }
                            ChatRoomCacheState.INVITED, ChatRoomCacheState.WAITING_TO_JOIN -> {
                                //do not join nonpersistant chat room if not collab
                                if (!state.isPersistent && !chatRoomCachedItem.isCollab) {
                                    chat.decline("")
                                    cache.remove(chat)
                                    return@execute
                                }
                                joinedRoomsToSync.add(JoinedRoomData(roomId = chatRoom.chatKey, roomCreator = chatRoom.roomCreator, joinTime = chatRoom
                                        .getJoinTimeInSec(), lastViewed = chatRoom.getJoinTimeInSec()))
                                if (chatRoomCachedItem.state == ChatRoomCacheState.WAITING_TO_JOIN) {
                                    Log.d(TAG, "Joining room ${chatRoom.chatKey} with string: ${getLastMessageDateString(chatRoom)}")
                                    chat.join(false, true, chat.api.account.settings.username, "", getLastMessageDateString(chatRoom), arrayListOf())
                                } else {
                                    Log.d(TAG, "Accepting join: ${chatRoom.name}")
                                    chat.accept(chat.api.account.settings.username, getLastMessageDateString(chatRoom), arrayListOf())
                                }
                                chatRoomRepo.addChatRoom(chatRoom)
                                cache.getChatRoomCacheItem(chatRoom.chatKey)?.apply {
                                    this.notSavedchatRoom = null
                                }
                                setChatRoomState(chatRoom.chatKey, ChatRoomCacheState.JOINED_SYNCED)
                                if (!cache.hasWaitingToJoinChats()) {
                                    joinedRoomSyncApis[chat.api.account]?.addJoinedRooms(joinedRoomsToSync)
                                    joinedRoomsToSync.clear()
                                }
                            }

                            ChatRoomCacheState.JOINED_SYNCED -> {
                                //BRACE-12952
                                if (!state.isPersistent && !chatRoomCachedItem.isCollab) {
                                    //do something here
                                    // mark chatroom as nonpersistent

                                    //deleteRoom(chatRoom)
                                    return@execute
                                } else {
                                    chatRoomRepo.updateChatRoom(chatRoom)
                                }
                            }
                            ChatRoomCacheState.JOINED_WAITING_FOR_SYNC -> {
                            }
                            null -> {
                            }
                            ChatRoomCacheState.JOINED_NOT_SYNCABLE -> TODO()
                            ChatRoomCacheState.WAITING_TO_JOIN_FROM_SYNC -> {
                                if (!state.isPersistent && !chatRoomCachedItem.isCollab) {
                                    cache.remove(chatRoomCachedItem.xmppChat)
                                    //joinedRoomSyncApis[chatRoomCachedItem.xmppChat.api.account]?.removeJoinedRoom(JoinedRoomData(roomId = chatRoom.chatKey))
                                    if (inDb) {
                                        chatRoom.state = ChatRoomState.OFFLINE
                                        chatRoomRepo.updateChatRoom(chatRoom)
                                    }
                                } else {
                                    chat.join(false, true, chat.api.account.settings.username, "", getLastMessageDateString(chatRoom), arrayListOf())
                                    cache.getChatRoomCacheItem(chatRoom.chatKey)?.apply {
                                        this.notSavedchatRoom?.let {
                                            chatRoom.lastViewedMessage = it.lastViewedMessage
                                        }
                                        this.notSavedchatRoom = null
                                        setChatRoomState(this.chatKey, ChatRoomCacheState.JOINED_SYNCED)
                                        chatRoom.state = ChatRoomState.ACTIVE
                                    }
                                    if (inDb) {
                                        chatRoomRepo.updateChatRoom(chatRoom)
                                    } else {
                                        chatRoomRepo.addChatRoom(chatRoom)
                                    }
                                }
                            }
                            ChatRoomCacheState.LOCALLY_LEFT -> {
                            }
                        }
                        Log.d(TAG, "room info updating chat room| name: ${chatRoom.name} jid:${chatRoom.chatKey} desc:${chatRoom.description}")
                        Log.d(TAG, "room info 2 room| time string: ${state.creation} time: ${chatRoom.creationTime}")
                        if (chatRoomCachedItem?.state == ChatRoomCacheState.JOINED_SYNCED && state.isReady) {
                            requestMemberList(chat, chatRoom)
                        }

                    }
                }

            }
        }

        override fun onRoomListRetrieved(account: XmppAccount, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.RoomListRetrievedEvent) {
            imExecutorService.execute {
                val roomDetails = StringBuilder("\n")
                roomDetails.append("RoomListRetrieved:\n ")
                val roomListInfos = mutableListOf<LocalRoomInfo>()
                val waitingForInfoInAccount = mutableListOf<String>()
                stopGettingRoomInfos(account)
                val roomsInDb = chatRoomRepo.getAllChatRooms().blockingGet()
                val roomKeysInDb = roomsInDb.map { it.chatKey }
                Log.d(TAG, "ROOM_KEYS_IN_DB ${roomKeysInDb.joinToString()}")
                val multiUserChatApi = XmppApiMultiUserChat.get(account)
                msg.roomsList.forEach { roomListItem ->
                    roomDetails.append(" Room  name").append(roomListItem.name).append(" JID: ").append(roomListItem.jid).append("\n")
                    roomListInfos.add(LocalRoomInfo(roomListItem.name, account.getUserAtDomain(), roomListItem.jid, "", ChatRoomAnonymousMode.FULL_ANONYMOUS,
                            creationTime = 0))
                    waitingForInfoInAccount.add(roomListItem.jid)
                }
                Log.d(TAG, roomDetails.toString())
                if (downloadRoomInfos[account] == true) {
                    waitingForInfo[account] = waitingForInfoInAccount
                }
                downloadRoomInfos[account] = false
                //getRoomInfo(multiUserChatApi, "h7tn4iy4xp5ujr7tmgrn3bcx8jsamznq")
                getRoomInfos(account)
                chatRoomRepo.replaceRoomInfos(roomListInfos)
                leavePublicRoomsNotOnServer()
                roomListChangedObserver.onNext(roomListInfos.map { it.name })
            }
        }

        override fun onMultiUserChatReady(chat: XmppMultiUserChat, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatReadyEvent) {
            imExecutorService.execute {
                Log.d(TAG, "onMultiUserChatReady: isNewRoom: ${msg.isNewRoom} ; room: ${msg.roomJid}")
                if (cache.getChatKeyByChat(chat) == "INVALID") {
                    cache.getChatRoomCacheItem(chat)?.let { item ->
                        if (item.state == ChatRoomCacheState.JOINED_WAITING_FOR_SYNC) {
                            item.notSavedchatRoom?.let { unsavedChatRoom ->
                                unsavedChatRoom.chatKey = msg.roomJid
                                inviteUninvited(unsavedChatRoom, chat)
                                chat.changeSubject(unsavedChatRoom.topic)
                                joinedRoomSyncApis[chat.api.account]?.addJoinedRoom(JoinedRoomData(roomId = msg.roomJid,
                                        roomCreator = unsavedChatRoom.owners.first?.remoteAddress ?: "",
                                        joinTime = unsavedChatRoom.getJoinTimeInSec(),
                                        roomName = unsavedChatRoom.name))
                                chatRoomRepo.addChatRoom(unsavedChatRoom)
                                item.chatKey = msg.roomJid
                                item.notSavedchatRoom = null
                                requestMemberList(chat, unsavedChatRoom)
                            }
                        }
                    }
                }
                cache.setReady(msg.roomJid)
            }
        }

        override fun onSendMessageSuccess(chat: XmppMultiUserChat?, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.SendMessageSuccessEvent) {
            imExecutorService.execute {
                Log.d(TAG, "onSendMessageSuccess ${evt.messageId}")
                tempMessages[evt.message]?.let {
                    it.externalId = evt.messageId
                    it.status = ImStatusData.OUTGOING_SENT
                    alreadySavedMessages.add(evt.messageId)
                    chatRoomRepo.addMessage(it)
                }
            }
        }

        override fun onSendMessageFailure(chat: XmppMultiUserChat?, evt: Xmppmultiuserchat.XmppMultiUserChatEvents.SendMessageFailureEvent) {
            imExecutorService.execute {
                AndroidLog.d(TAG, "onSendMessageFailure")
                tempMessages[evt.message]?.let {
                    it.externalId = Date().time.toString()
                    it.status = ImStatusData.OUTGOING_FLAG_FAIL_TO_SEND
                    chatRoomRepo.addMessage(it)
                }
            }
        }

        override fun onMultiUserChatNewMessage(chat: XmppMultiUserChat, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatNewMessageEvent) {
            Log.d(TAG, "onMultiUserChatNewMessage: message received: ${msg.messageId} time ${msg.timestamp * 1000 + msg.millisecond}")
            imExecutorService.execute {
                Log.d(TAG, "onMultiUserChatNewMessage")
                val time = msg.timestamp * 1000 + msg.millisecond
                cache.getChatKeyByChat(chat)?.let { chatKey ->
                    val userName = chat.api.account.settings.username
                    val domain = chat.api.account.settings.domain
                    var remoteAddress = msg.nickname
                    if (!remoteAddress.contains("@")) {
                        remoteAddress = "$remoteAddress@$domain"
                    }
                    Log.d(TAG, "onMultiUserChatNewMessage: ${chatKey}")
                    if (msg.messageId in alreadySavedMessages) {
                        alreadySavedMessages.remove(msg.messageId)
                        return@execute
                    }
                    chatRoomRepo.getRoomByRoomKey(key = chatKey)?.let { chatRoom ->
                        var isRead = false
                        joinedRoomSyncApis[chat.api.account]?.let { syncApi ->
                            if (syncApi.getMessageIdOfLastViewedMessage(chatKey) != msg.messageId) {
                                isRead = syncApi.isAlreadyRead(chatKey, time)
                            } else {
                                isRead = true
                                markMessagesReadUpToTimeForChatRoom(chatRoom.id, time)
                            }
                        }
                        val message = if (TextUtils.equals(remoteAddress, userName) || TextUtils.equals(remoteAddress, "$userName@$domain")) {
                            Message(
                                    chatId = chatRoom.id,
                                    status = ImStatusData.OUTGOING_SENT,
                                    text = msg.plain,
                                    readState = MessageReadState.READ,
                                    remoteAddress = remoteAddress,
                                    externalId = msg.messageId,
                                    modificationTime = time,
                                    creationTime = time)
                        } else {
                            val isCollab = cache.isCollab(chatKey)
                            val readStatus = when {
                                isRead -> ImStatusData.INCOMING_READ
                                !isRead && isCollab -> ImStatusData.INCOMING_COLLAB_UNREAD
                                else -> {
                                    ImStatusData.INCOMING_UNREAD
                                }
                            }
                            Message(
                                    chatId = chatRoom.id,
                                    status = readStatus,
                                    text = msg.plain,
                                    readState = if (isRead) MessageReadState.READ else MessageReadState.UNREAD_NOT_SWIPED,
                                    remoteAddress = remoteAddress,
                                    externalId = msg.messageId,
                                    modificationTime = time,
                                    creationTime = time)
                        }
                        chatRoomRepo.addMessage(message)
                    }
                }
            }
        }

        override fun onMultiUserChatError(chat: XmppMultiUserChat, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatErrorEvent) {
            Log.e(TAG, "On onMultiUserChatError ${msg.error}")
            imExecutorService.execute {
                cache.getChatRoomCacheItem(chat)?.let { cachedItem ->
                    Log.e(TAG, "On onMultiUserChatError ${msg.error} type:  ${getErrorDescFromType(msg.type)} room: ${cachedItem.chatKey}")
                    if (msg.type == 3) {//room not found
                        when (cachedItem.state) {
                            ChatRoomCacheState.INVITED -> {
                                chat.decline("")
                                cache.remove(chat)
                            }
                            ChatRoomCacheState.WAITING_TO_JOIN -> {
                                cache.remove(cachedItem.xmppChat)
                                if (!cache.hasWaitingToJoinChats()) {
                                    joinedRoomSyncApis[chat.api.account]?.addJoinedRooms(joinedRoomsToSync)
                                    joinedRoomsToSync.clear()
                                }
                            }
                            ChatRoomCacheState.LOCALLY_LEFT -> {
                                val chatRoom = chatRoomRepo.getChatRoomByChatKey(cachedItem.chatKey)
                                if (chatRoom != null && !chatRoom.isOffline()) {
                                    deleteRoom(chatRoom)
                                } else {
                                    //if not in db remove from cache and from sync
                                    //joinedRoomSyncApis[cachedItem.xmppChat.api.account]?.removeJoinedRoom(JoinedRoomData(roomId = cachedItem.chatKey))
                                    //cache.remove(chat)
                                }
                                //chatRoomRepo.getChatRoomByChatKey(cachedItem.chatKey)?.let { deleteRoom(chatRoom = it) }
                            }
//                            ChatRoomCacheState.REMOTELY_LEFT -> {
//                                chatRoomRepo.getChatRoomByChatKey(cachedItem.chatKey)?.let {
//                                    it.state = ChatRoomState.REMOTELY_DELETED
//                                    chatRoomRepo.updateChatRoom(it)
//                                    joinedRoomSyncApis[cachedItem.xmppChat.api.account]?.removeJoinedRoom(JoinedRoomData(roomId = cachedItem.chatKey))
//                                    cache.remove(chat)
//                                }
//                            }
                        }
                    } else if (msg.type == 9) {// other error
                        if (cachedItem.state == ChatRoomCacheState.LOCALLY_LEFT) {
                            joinedRoomSyncApis[chat.api.account]?.removeJoinedRoom(JoinedRoomData(roomId = cachedItem.chatKey))
                            cache.remove(chat)
                        }
                    } else if (msg.type == 6) { //not a member
                        if (!cachedItem.isCollab) {
                            joinedRoomSyncApis[chat.api.account]?.removeJoinedRoom(JoinedRoomData(roomId = cachedItem.chatKey))
                            cache.remove(chat)
                        }
                    }
                    errorEventObservable.onNext(ChatRoomErrorEvent(cachedItem.chatKey, msg.type, msg.error))
                }
            }

        }

        override fun onMultiUserChatInvitationReceived(chat: XmppMultiUserChat, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatInvitationReceivedEvent) {
            imExecutorService.execute {

                val state = cache.getChatRoomCacheItem(chat)?.state
                when (state) {
                    null,
                    ChatRoomCacheState.NOT_IN_CHAT_ROOM -> {
                        val serviceName = serviceForAccount[chat.api.account.getUserAtDomain()]
                        val roomJid = if (!msg.room.contains('@') && !serviceName.isNullOrBlank()) "${msg.room}@$serviceName" else msg.room
                        Log.d(TAG, "onMultiUserChatInvitationReceived: ${roomJid}")
                        val xmppAccount = chat.api.account
                        val chatRoom = chatRoomRepo.getChatRoomByChatKey(roomJid) ?: ChatRoom(
                                chatKey = roomJid,
                                accountId = ImpsUtils.removePortFromDomain(xmppAccount.getUserAtDomain()),
                                creationTime = msg.timeStamp + msg.millisecond,
                                joinTime = System.currentTimeMillis(),
                                modTime = System.currentTimeMillis(),
                                type = ExtendedConversationData.CHAT_ROOM_TYPE.toLong()
                        )
                        val participant = Participant("${msg.jid}@${chatRoom.accountId}", Participant.EParticipantType.XMPP, chatRoom.accountId)
                        chatRoom.owners.add(participant)
                        chatRoom.roomCreator = participant.key
                        cache.add(chat, roomJid)
                        cache.getChatRoomCacheItem(chat)?.let {
                            it.state = ChatRoomCacheState.INVITED
                            it.notSavedchatRoom = chatRoom
                        }

                        chat.getRoomInfo()
                    }
                    ChatRoomCacheState.INVITED -> {
                        // we have already requested room info. seems there was no response
                        chat.getRoomInfo()
                    }
                    else -> {
                        Log.d(TAG, "Received invite, but status was: $state")
                    }
                }
            }
        }

        override fun onParticipantAdded(chat: XmppMultiUserChat, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.ParticipantAddedEvent) {
            imExecutorService.execute {
                Log.d(TAG, "onParticipantAdded ${msg.nickname} jid: ${msg.state.jid} aff: ${msg.state.affiliation} role: ${msg.state.role}")
                if (cache.isReady(chat)) {
                    cache.getChatKeyByChat(chat)?.let { key -> chatRoomRepo.getChatRoomByChatKey(key)?.let { requestMemberList(chat, it) } }
                            ?: Log.e(TAG, "onParticipantAdded: room not found")
                } else {
                    val chatKey = cache.getChatKeyByChat(chat) ?: return@execute
                    chatRoomRepo.getChatRoomByChatKey(chatKey)?.let {
                        val participant = Participant("${msg.state.nickname}@${chat.api.account.getUserAtDomain()}",
                                Participant.EParticipantType.XMPP, chat.api.account.getUserAtDomain())
                        if (msg.state.affiliation == 3) {  //owner role (per current design no owner)
                            it.members.add(participant)
                        } else if (msg.state.affiliation == 2) { //member role
                            it.members.add(participant)
                        }
                        it
                    }
                }
            }
        }

        override fun onParticipantUpdated(chat: XmppMultiUserChat?, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.ParticipantUpdatedEvent) {
            Log.d(TAG, "onParticipantUpdated ${msg.nickname} aff: ${msg.state.affiliation} role: ${msg.state.role}")
        }


        override fun onMultiUserChatListRequested(chat: XmppMultiUserChat, args: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatListRequestedEvent) {
            imExecutorService.execute {
                Log.d(TAG, "onMultiUserChatListRequested type: ${args.xmppMultiUserChatListType}")
                val chatKey = cache.getChatKeyByChat(chat) ?: return@execute
                Log.d(TAG, "Room: ${chatKey}; type: ${args.xmppMultiUserChatListType}")
                args.configurationsListItemsList.forEach { item ->
                    Log.d(TAG, "jid: ${item.jid}; nick: ${item.nick}; reason: ${item.reason}; aff ${item.affiliation}; role ${item.role} ")
                }

                chatRoomRepo.getChatRoomByChatKey(chatKey)?.let { chatRoom ->
                    val participantsSet = ParticipantsSet()
                    args.configurationsListItemsList.forEach { item ->
                        val buddyKey = BuddyKeyUtils.getNewBuddyKey(Participant.EParticipantType.XMPP, chatRoom.accountId, item.jid)
                        participantsSet.add(Participant(buddyKey))
                    }
                    when (args.xmppMultiUserChatListType) {
                        MEMBERS_LIST_TYPE -> {
                            if (chatRoom.publicRoom) {
                                chatRoom.members.clear()
                                chatRoom.members.addAll(participantsSet)
                                chat.requestList(OWNERS_LIST_TYPE)

                                chatRoomMemberCache.putMembers(chatRoom.id, participantsSet.setOfBuddyKeys)
                            }
                        }
                        OWNERS_LIST_TYPE -> { //per current design there are no room owners only room creator and members
                            if (chatRoom.publicRoom) {
                                chatRoom.owners.clear()
                                chatRoom.owners.addAll(participantsSet)

                                chatRoomMemberCache.putOwners(chatRoom.id, participantsSet.setOfBuddyKeys)
                            } else {
                                chatRoom.members.clear()
                                chatRoom.members.addAll(participantsSet)

                                chatRoomMemberCache.putMembers(chatRoom.id, participantsSet.setOfBuddyKeys)
                            }
                        }
                        else -> {
                        }
                    }
                    chatRoomRepo.updateChatRoom(chatRoom)
                }

            }
        }

        override fun onLocalUserLeft(chat: XmppMultiUserChat, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.LocalUserLeftEvent) {
            imExecutorService.execute {
                Log.d(TAG, "OnLocalUserLeft ${msg.reason}")
                cache.getChatKeyByChat(chat)?.let {
                    chatRoomRepo.getChatRoomByChatKey(it)?.let { chatRoom ->
                        if (serviceForAccount.containsKey(chatRoom.accountId)) {
                            cache.remove(chat)
                            chatRoom.state = ChatRoomState.OFFLINE
                            chatRoomRepo.updateChatRoom(chatRoom)
                        }
                    }
                }
            }
        }

        override fun onMultiUserChatSubjectChanged(chat: XmppMultiUserChat, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatSubjectChangedEvent) {
            imExecutorService.execute {
                Log.d(TAG, "onMultiUserChatSubjectChanged type: ${msg.subject}")
                val chatKey = cache.getChatKeyByChat(chat) ?: return@execute
                chatRoomRepo.getChatRoomByChatKey(chatKey)?.let { chatRoom ->
                    chatRoom.topic = if (msg.subject == " ") "" else msg.subject
                    chatRoomRepo.updateChatRoom(chatRoom)
                }
            }
        }

        override fun onParticipantRemoved(chat: XmppMultiUserChat, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.ParticipantRemovedEvent) {
            imExecutorService.execute {
                AndroidLog.d(TAG, "onParticipantRemoved")
                if (cache.isReady(chat)) {
                    cache.getChatKeyByChat(chat)?.let { key -> chatRoomRepo.getChatRoomByChatKey(key)?.let { requestMemberList(chat, it) } }
                            ?: Log.e(TAG, "onParticipantRemoved: room not found")
                }
            }
        }


        override fun onMultiUserChatInvitationDeclined(chat: XmppMultiUserChat?, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatInvitationDeclinedEvent?) {
            Log.d(TAG, "onMultiUserChatInvitationDeclined")
        }

        override fun onParticipantSelfUpdated(chat: XmppMultiUserChat?, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.ParticipantSelfUpdatedEvent?) {
            Log.d(TAG, "onParticipantSelfUpdated")
        }

        override fun onParticipantChatStateReceived(chat: XmppMultiUserChat?, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.ParticipantChatStateEvent?) {
            Log.d(TAG, "onParticipantChatStateReceived")
        }
    }

    private fun applyServerData(activeJoinedRoom: List<JoinedRoomData>, xmppAccount: XmppAccount) {
        joinSyncedRooms(activeJoinedRoom, xmppAccount)
        leaveNonSyncedRooms(activeJoinedRoom)
        markMessagesRead(activeJoinedRoom)
    }

    private fun markMessagesRead(activeJoinedRoom: List<JoinedRoomData>) {
        activeJoinedRoom.forEach { roomData ->
            chatRoomRepo.getMessageByExternalId(roomData.lastViewedMsgId)?.let {
                markMessagesReadUpToTimeForChatRoom(it.chatId, it.creationTime)
            }
        }
    }

    private fun markMessagesReadUpToTimeForChatRoom(chatId: Long, time: Long) {
        chatRoomRepo.getUnreadMessagesForChatRoom(chatId).blockingGet().filter {
            it.creationTime <= time
        }.apply {
            Log.d(TAG, "Mark read messages: ${this.map { it.text }.joinToString()}")
            this.forEach { it.readState = MessageReadState.READ }
            chatRoomRepo.markRead(this)
        }
//        chatRoomRepo.getUnreadMessagesForChatRoomUpToTime(chatId, time).blockingGet().apply {
//            chatRoomRepo.markRead(this)
//        }
    }

    private fun updateLocalRoomInfo(jid: String, state: Xmppmultiuserchat.XmppMultiUserChatRoomState, msg: Xmppmultiuserchat.XmppMultiUserChatEvents.MultiUserChatRoomStateChangedEvent) {
        chatRoomRepo.getRoomInfoByJid(jid)?.let { localRoomInfo ->
            localRoomInfo.creationTime = formatTZStringDate(state.creation).time
            if (localRoomInfo.creationTime == 0L || !state.isPersistent) {
                chatRoomRepo.removeRoomInfoWithJid(jid)
            } else {
                localRoomInfo.description = state.description
                if (msg.state.isReady) {
                    localRoomInfo.topic = state.subject
                }
                localRoomInfo.anonymousMode = when (state.anonymousMode) {
                    0 -> ChatRoomAnonymousMode.NON_ANONYMOUS
                    1 -> ChatRoomAnonymousMode.SEMI_ANONYMOUS
                    2 -> ChatRoomAnonymousMode.FULL_ANONYMOUS
                    else -> ChatRoomAnonymousMode.NON_ANONYMOUS
                }
                localRoomInfo.hasInfo = true
                chatRoomRepo.updateRoomInfo(localRoomInfo)
            }
        }
    }

    private fun stopGettingRoomInfos(account: XmppAccount) {
        roomInfoDisposables[account]?.dispose()
    }

    @SuppressLint("RxDefaultScheduler")
    private fun getRoomInfos(account: XmppAccount) {
        roomInfoDisposables.put(account, Observable.interval(50, TimeUnit.MILLISECONDS).subscribeOn(Schedulers.from(imExecutorService)).subscribe({
            this.waitingForInfo[account]?.let {
                if (it.isNotEmpty()) {
                    getRoomInfo(XmppApiMultiUserChat.get(account), it.removeAt(0))
                }
            }

        }, {}))
    }


    private fun joinPrivateChatRoomsFromStorage() {
        Log.d(TAG, "Join local Chat Rooms")
        disposables.add(chatRoomRepo.getAllChatRooms().subscribeOn(Schedulers.io()).observeOn(Schedulers.io()).subscribe({ chatRoomList ->
            chatRoomList.forEach { chatRoom ->
                getXmppAccount(chatRoom.accountId)?.let {
                    val chatApi = XmppApiMultiUserChat.get(it)
                    getMultiUserChat(chatRoom.chatKey, chatApi)?.let { xmppChat ->
                        xmppChat.join(true, false, it.settings.username, "", getLastMessageDateString(chatRoom), arrayListOf())
                        cache.add(xmppChat, chatRoom.chatKey)
                    }
                }
            }
        }, { Log.e(TAG, "Join private Chat Rooms ERROR!!! ${it.localizedMessage}") }))
    }

    override fun updateRoomList(downloadDetails: Boolean) {
        imExecutorService.execute {
            activeAccounts.forEach {
                if (downloadDetails) {
                    downloadRoomInfos[it] = true
                }
                XmppApiMultiUserChat.get(it)?.getRoomList()
            }
        }
    }

    override fun updateJoinedRooms() {
        for (joinedRoomSyncApi in joinedRoomSyncApis.values) {
            joinedRoomSyncApi.fetchAndUpdate()
        }
    }

    private fun inviteUninvited(chatRoom: ChatRoom, chat: XmppMultiUserChat) {
        chatRoom.invited.forEach { member ->
            if (!chatRoom.members.containsRemoteAddress(member.remoteAddress)) {
                if (!chatRoom.publicRoom) {
                    Log.d(TAG, "inviting as owner ${member.remoteAddress}")
                    chat.changeJidAffiliation(member.remoteAddress, 3, "")
                } else {
                    chat.invite(member.remoteAddress, "") // nvtd add invite message
                }

                Log.d(TAG, "inviting ${member.remoteAddress}")
//                if (chatRoom.public == 1) {
//                    Log.d(TAG, "changing affiliation ${member.remoteAddress}")
//                    chat.changeAffiliation(member.remoteAddress, 3, "")
//                }
            }
        }
        chatRoom.invited.clear()
    }


    //formats "yyyy-MM-dd'T'HH:mm:ss'Z' date format to string
    private fun formatTZStringDate(stringDate: String): Date {
//        return if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
//            val instant = Instant.parse(stringDate)
//            Date.from(instant)
//        } else {
//        val format = SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'")
//        format.timeZone = TimeZone.getTimeZone("GMT")
//        return format.parse(stringDate)
//        }
        val format = SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss'Z'")
        format.timeZone = TimeZone.getTimeZone("GMT")
        return format.parse(stringDate)


    }

    fun getErrorDescFromType(type: Int): String {
        return when (type) {
            0 -> "NoError"
            1 -> "PasswordRequired"
            2 -> "UserBanned"
            3 -> "RoomNotFound"
            4 -> "CreationRestricted"
            5 -> "ReservedNicknameRequired"
            6 -> "NotAMember"
            7 -> "NicknameConflict"
            8 -> "MaximumUsersReached"
            9 -> "OtherError"
            else -> "Unknown $type"
        }
    }


    override fun reInviteToPrivateRoom(chatRoomId: Long, buddy: XmppBuddy) {
        Log.d(TAG, "$chatRoomId, ${buddy.imAddress}")
        imExecutorService.execute {
            val chatRoom = chatRoomRepo.getChatRoom(chatRoomId)
            if (chatRoom != null) {
                val chat = cache.getXmppChat(chatRoom.chatKey)
                if (chat != null) {
                    Log.d(TAG, "inviting as owner ${buddy.imAddress}")
                    chat.invite(buddy.imAddress, "")
                    if (!chatRoom.publicRoom) {
                        chat.changeJidAffiliation(buddy.imAddress, Xmppmultiuserchat.AffiliationOwner, "")
                    } else {
                        CrashInDebug.with(TAG, "Reinvite is only for private rooms.")
                    }
                }
            }
        }
    }
}

fun XmppAccount.getUserAtDomain(): String {
    return "${this.settings.username}@${ImpsUtils.removePortFromDomain(this.settings.domain)}"
}

fun Account.getUserAtDomain(): String {
    return "${this.getStr(EAccountSetting.UserName)}@${ImpsUtils.removePortFromDomain(this.getStr(EAccountSetting.Domain))}"
}


fun ChatRoom.hasSameChatKey(chatKey: String) = if (this.chatKey.contains("@") xor chatKey.contains("@")) {
    ImpsUtils.removeDomainFromAddress(this.chatKey) == ImpsUtils.removeDomainFromAddress(chatKey)
} else this.chatKey == chatKey
