package com.bria.common.controller.im.refactoring.chatroom

import com.bria.common.controller.im.refactoring.roomdb.entities.ChatRoom
import com.bria.common.controller.im.refactoring.xmppgroup.CollabChatInfo
import com.bria.common.util.Log
import com.counterpath.sdk.XmppAccount
import com.counterpath.sdk.XmppMultiUserChat

data class ChatRoomCacheItem(
        var xmppChat: XmppMultiUserChat,
        var chatKey: String,
        var notSavedchatRoom: ChatRoom? = null,
        var bookmarked: Boolean = false,
        var ready: Boolean = false,
        var state: ChatRoomCacheState = ChatRoomCacheState.NOT_IN_CHAT_ROOM,
        var isCollab: Boolean = false)

enum class ChatRoomCacheState {
    NOT_IN_CHAT_ROOM,
    INVITED,
    WAITING_TO_JOIN,
    JOINED_SYNCED,
    JOINED_WAITING_FOR_SYNC, //created but not synced
    JOINED_NOT_SYNCABLE,
    WAITING_TO_JOIN_FROM_SYNC,
    LOCALLY_LEFT,
    REMOTELY_LEFT
}

class ChatRoomCache {

    companion object {
        const val MEMBERS_LIST_TYPE = 0
        const val OWNERS_LIST_TYPE = 4
    }

    private val collabChats = mutableSetOf<String>()
    private val collabWaitingList = mutableListOf<CollabChatInfo>()

    val items = mutableListOf<ChatRoomCacheItem>()
    @Synchronized
    fun getChatKeyByChat(xmppChat: XmppMultiUserChat) = items.firstOrNull { it.xmppChat.handle() == xmppChat.handle() }?.chatKey

    @Synchronized
    fun getChatRoomCacheItem(xmppChat: XmppMultiUserChat): ChatRoomCacheItem? = items.firstOrNull { xmppChat.handle() == it.xmppChat.handle() }

    @Synchronized
    fun getChatRoomCacheItem(chatRoomKey: String): ChatRoomCacheItem? = items.firstOrNull { chatRoomKey == it.chatKey }

//    fun updateChatRoom(chatRoom: ChatRoom) = items.firstOrNull { chatRoom.chatKey == it.chatRoom?.chatKey }?.apply {
//        this.chatRoom = chatRoom.copy()
//    }

    @Synchronized
    fun getXmppChat(chatRoomKey: String): XmppMultiUserChat? = items.firstOrNull { chatRoomKey == it.chatKey }?.xmppChat

    @Synchronized
    fun add(xmppChat: XmppMultiUserChat, notSavedChatRoom: ChatRoom) {
        items.add(ChatRoomCacheItem(xmppChat = xmppChat, notSavedchatRoom = notSavedChatRoom, bookmarked = false, chatKey = "INVALID"))
    }

    @Synchronized
    fun add(xmppChat: XmppMultiUserChat, jid: String, bookmarked: Boolean = false, isCollab: Boolean = false, state: ChatRoomCacheState = ChatRoomCacheState.NOT_IN_CHAT_ROOM) {
        remove(xmppChat)
        items.add(ChatRoomCacheItem(xmppChat = xmppChat, chatKey = jid, bookmarked = bookmarked, state = state))
    }

    @Synchronized
    fun remove(xmppChat: XmppMultiUserChat) {
        items.firstOrNull { xmppChat.handle() == it.xmppChat.handle() }?.let { items.remove(it) }
    }

    @Synchronized
    fun remove(chatRoomKey: String) {
        items.firstOrNull { chatRoomKey == it.chatKey }?.let { items.remove(it) }
    }

    @Synchronized
    fun deleteChatRoom(chatKey: String) {
        val delete = mutableListOf<ChatRoomCacheItem>()
        items.firstOrNull { chatRoomCacheItem ->
            chatRoomCacheItem.chatKey == chatKey
        }?.let {
            delete.add(it)
        }
        items.removeAll(delete)
    }


    @Synchronized
    fun setReady(chatKey: String) {
        items.firstOrNull { it.chatKey == chatKey }?.ready = true
    }

    @Synchronized
    fun isReady(chat: XmppMultiUserChat) = items.firstOrNull { it.xmppChat.handle() == chat.handle() }?.ready ?: false


    @Synchronized
    fun clearCacheForAccount(account: XmppAccount) {
        Log.d("ChatRoomApi", "clearCacheForAccount old size: ${items.size}")
        val toRemove = items.filter { it.xmppChat.api.account.getUserAtDomain() == account.getUserAtDomain() }//nvtd: until Darko fix his issue.
        Log.d("ChatRoomApi", "clearCacheForAccount to remove: " + items.joinToString())
        items.removeAll(toRemove)
        Log.d("ChatRoomApi", "clearCacheForAccount new size: ${items.size}")
    }

    @Synchronized
    fun addAsCollabChat(roomId: String) {
        collabChats.add(roomId)
        getChatRoomCacheItem(roomId)?.isCollab = true
    }


    fun addToCollabWaitingList(collabChatInfo: CollabChatInfo) {
        collabWaitingList.add(collabChatInfo)
    }

    fun getWaitingCollabList(): MutableList<CollabChatInfo> {
        return collabWaitingList;
    }

    @Synchronized
    fun isCollab(chatKey: String) = collabChats.contains(chatKey)

    @Synchronized
    fun joined(chatKey: String) = items.firstOrNull { it.chatKey == chatKey }?.let {
        it.state == ChatRoomCacheState.JOINED_SYNCED || it.state ==
                ChatRoomCacheState.JOINED_WAITING_FOR_SYNC
    } ?: false

    @Synchronized
    fun isPlanningToJoin(chatKey: String) = items.firstOrNull { it.chatKey == chatKey }?.state == ChatRoomCacheState.WAITING_TO_JOIN

    @Synchronized
    fun setState(stateEvent: ChatRoomStateEvent) {
        items.firstOrNull { it.chatKey == stateEvent.key }?.apply {
            Log.d("ChatRoomApi", "set State ${state} to: ${chatKey}")
            this.state = stateEvent.state
        }
    }

    @Synchronized
    fun setState(chat: XmppMultiUserChat, state: ChatRoomCacheState) {
        getChatRoomCacheItem(chat)?.state = state
    }

    @Synchronized
    fun hasWaitingToJoinChats() = items.firstOrNull { it.state == ChatRoomCacheState.WAITING_TO_JOIN } != null

    @Synchronized
    fun getState(chatKey: String) = items.firstOrNull { it.chatKey == chatKey }?.state

    fun isLeft(chatKey: String): Boolean {
        return getState(chatKey) in listOf(ChatRoomCacheState.REMOTELY_LEFT, ChatRoomCacheState.LOCALLY_LEFT)
    }
}
