package com.bria.common.controller.im.refactoring.chatroom.joinedroom

import android.util.Xml
import com.bria.common.controller.im.refactoring.chatroom.joinedroom.XMPPDataStorageXmlHandler.Companion.ROOM_KEY
import com.bria.common.controller.im.refactoring.roomdb.ChatRoomRepo
import com.bria.common.controller.im.refactoring.roomdb.entities.Message
import com.bria.common.util.CrashInDebug
import com.bria.common.util.Log
import com.counterpath.sdk.XmppAccount
import com.counterpath.sdk.pb.Xmpp
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.PublishSubject
import io.reactivex.subjects.Subject
import org.xml.sax.InputSource
import org.xmlpull.v1.XmlSerializer
import java.io.StringReader
import java.io.StringWriter
import java.util.concurrent.ConcurrentHashMap
import java.util.concurrent.ExecutorService
import java.util.concurrent.TimeUnit
import javax.xml.parsers.SAXParserFactory


class JoinedRoomSyncApiImpl(val xmppAccount: XmppAccount, private val imExecutorService: ExecutorService, private val repo: ChatRoomRepo) : JoinedRoomSyncApi {

    companion object {
        const val KEY = "joinedRooms"
        const val TAG = "JoinedRoomSyncApiImpl"
    }

    private val disposables = CompositeDisposable()

    private val localLastReadMessageForRoom = HashMap<String, Message>()

    private val cachedMessages = HashMap<String, Message>()

    init {
        xmppAccount.addHandler(this)
        disposables.add(repo.getOnMessagesReadObservable().map { list ->
            list.maxBy { it.creationTime }?.let { message ->
                repo.getChatRoom(message.id)?.let { chatRoom ->
                    Pair(chatRoom.chatKey, message)
                }
            } ?: Pair("", null)
        }.mergeWith(
                repo.getOnRoomReadObservable().flatMapIterable {
                    it
                }.map {
                    repo.getLastMessageForChatRoom(it.id).blockingGet()?.let { message ->
                        Pair(it.chatKey, message)
                    } ?: Pair("", null)
                }
        ).filter { (chatkey, message) ->
            if (chatkey.isEmpty() || message == null) {
                return@filter false
            }
            roomsActive[chatkey]?.let { roomData ->
                getMessageByExternalId(roomData.lastViewedMsgId)?.let {oldMessage ->
                    return@filter message.creationTime > oldMessage.creationTime
                }
            } ?: true
        }
        .observeOn(Schedulers.from(imExecutorService))
        .map { (jid, message) ->
            message?.let {
                Log.d(TAG, "mark read: $jid message: ${message.externalId}")
                val existing = fieldsToBeUpdated.firstOrNull { (oldJid, nameValue) -> jid == oldJid && nameValue.first == JoinRoomStorageField.LAST_VIEWED_ID }
                if (existing == null || ((getMessageByExternalId(existing.second.second)?.creationTime ?: 0L) < message.creationTime)) {
                    if (existing != null) {
                        fieldsToBeUpdated.remove(existing)
                    }
                    fieldsToBeUpdated.add(Pair(jid, Pair(JoinRoomStorageField.LAST_VIEWED_ID, message.externalId)))
                    return@map true
                }

            } ?: return@map false
            return@map false
        }.filter { it }
        .debounce(1000, TimeUnit.MILLISECONDS, Schedulers.from(imExecutorService))
        .subscribe({
            fetchAndUpdate()
        }, { CrashInDebug.with(TAG, it) }))
    }

    private val saxParer = SAXParserFactory.newInstance().newSAXParser()

    private val joinedRoomsListChangedObservable: Subject<List<JoinedRoomData>> = PublishSubject.create()
    private val roomsToBeAdded = mutableListOf<JoinedRoomData>()
    private val roomsToBeDeleted = mutableListOf<JoinedRoomData>()
    private val fieldsToBeUpdated = mutableListOf<Pair<String, Pair<JoinRoomStorageField, String>>>()
    private val roomsActive: ConcurrentHashMap<String, JoinedRoomData> = ConcurrentHashMap()

    private fun getMessageByExternalId(externalID : String) : com.bria.common.controller.im.refactoring.roomdb.entities.Message? {
        var message = cachedMessages[externalID]
        if (message == null) {
            message = repo.getMessageByExternalId(externalID)
            message?.let {
                cachedMessages.put(it.externalId, it)
            }
        }
        return message
    }

    override fun joinRoomsListChangedObservable() = joinedRoomsListChangedObservable

    override fun fetchAndUpdate() {
        Log.d(TAG, "fetchAndUpdate")
        xmppAccount.GetPrivateStorageData()
    }

    override fun addJoinedRoom(room: JoinedRoomData) {
        imExecutorService.execute {
            Log.d(TAG, "Add joined room: ${room}")
            roomsToBeAdded.add(room)
            fetchAndUpdate()
        }
    }

    override fun addJoinedRooms(rooms: List<JoinedRoomData>) {
        Log.d(TAG, "Add joined rooms: ${rooms.map { it.roomName }.joinToString()}")
        if (rooms.isNotEmpty()) {
            roomsToBeAdded.addAll(rooms)
            fetchAndUpdate()
        }
    }

    override fun removeJoinedRoom(room: JoinedRoomData) {
        imExecutorService.execute {
            roomsToBeDeleted.add(room)
            fetchAndUpdate()
        }
    }

    override fun updateFields(newFields: List<Pair<String, Pair<JoinRoomStorageField, String>>>) {
        imExecutorService.execute {
            newFields.forEach { newField ->
                fieldsToBeUpdated.remove(fieldsToBeUpdated.firstOrNull { (jid, nameValue) -> jid == newField.first && nameValue.first == newField.second.first })
                fieldsToBeUpdated.add(newField)
                fetchAndUpdate()
            }
        }
    }

    override fun updateField(newField: Pair<String, Pair<JoinRoomStorageField, String>>) {
        //nvtd clean this please
        imExecutorService.execute {
            Log.d(TAG, "updateField: jid ${newField.first} id ${newField.second.first} value ${newField.second.second}")
            var isSame = false
            roomsActive[newField.first]?.let { existing ->
                val stringValue = when (newField.second.first) {
                    JoinRoomStorageField.ROOM_ID -> existing.roomId
                    JoinRoomStorageField.ROOM_NAME -> existing.roomName
                    JoinRoomStorageField.ROOM_CREATOR -> existing.roomCreator
                    JoinRoomStorageField.JOIN_TIME -> existing.joinTime.toString()
                    JoinRoomStorageField.LAST_VIEWED -> existing.lastViewed.toString()
                    JoinRoomStorageField.LAST_VIEWED_ID -> existing.lastViewedMsgId
                }
                isSame = stringValue == newField.second.second
                if (isSame) {
                    Log.d(TAG, "value was same for ${newField.first}")
                }
            }
            if (!isSame) {
                fieldsToBeUpdated.remove(fieldsToBeUpdated.firstOrNull { (jid, nameValue) -> jid == newField.first && nameValue.first == newField.second.first })
                fieldsToBeUpdated.add(newField)
                fetchAndUpdate()
            }
        }
    }

    override fun getLastViewedMessageTimeInMs(chatKey: String): Long {
        return (roomsActive[chatKey]?.lastViewed ?: -1L)
    }

    override fun getMessageIdOfLastViewedMessage(chatKey: String) : String {
        return (roomsActive[chatKey]?.lastViewedMsgId ?: "")
    }

    override fun isAlreadyRead(chatKey: String, time : Long) : Boolean{
        return roomsActive[chatKey]?.let {roomData ->
            (getMessageByExternalId(roomData.lastViewedMsgId)?.creationTime ?: 0) > time
        } ?: false


    }

    private fun parseXmlResponse(xmlData: String): MutableList<JoinedRoomData> {
        Log.d(TAG, "parsing xml response: ${xmlData}")
        val handler = XMPPDataStorageXmlHandler()
        try {
            saxParer.parse(InputSource(StringReader(xmlData)), handler)
        } catch (e: Exception) {
            Log.e(TAG, "XML parser exception")
        }
        return handler.joinedRooms
    }

    private fun parseToXml(list: List<JoinedRoomData>): String {
        val xmlSerializer = Xml.newSerializer()
        val writer = StringWriter()

        xmlSerializer.setOutput(writer)
        xmlSerializer.startDocument("UTF-8", true)
        xmlSerializer.startTag(null, KEY)
        list.forEach { room ->
            xmlSerializer.startTag(null, ROOM_KEY)
            addXmlElement(xmlSerializer, JoinRoomStorageField.ROOM_ID.xmlTag, room.roomId)
            addXmlElement(xmlSerializer, JoinRoomStorageField.ROOM_CREATOR.xmlTag, room.roomCreator)
            addXmlElement(xmlSerializer, JoinRoomStorageField.JOIN_TIME.xmlTag, room.joinTime.toString())
            addXmlElement(xmlSerializer, JoinRoomStorageField.ROOM_NAME.xmlTag, room.roomName)
            addXmlElement(xmlSerializer, JoinRoomStorageField.LAST_VIEWED.xmlTag, room.lastViewed.toString())
            addXmlElement(xmlSerializer, JoinRoomStorageField.LAST_VIEWED_ID.xmlTag, room.lastViewedMsgId)
            room.undefinedElements.forEach { (key, value) -> addXmlElement(xmlSerializer, key, value) }
            xmlSerializer.endTag(null, ROOM_KEY)
        }

        xmlSerializer.endTag(null, KEY)
        xmlSerializer.endDocument()
        xmlSerializer.flush()
        return writer.toString()
    }

    private fun addXmlElement(xmlSerializer : XmlSerializer, key : String, value : String) {
        xmlSerializer.startTag(null, key)
        xmlSerializer.text(value)
        xmlSerializer.endTag(null, key)
    }

    override fun onPrivateStorageDataEvent(xmppAccount: XmppAccount, evt: Xmpp.XmppAccountEvents.XmppPrivateStorageDataEvent) {
        imExecutorService.execute {
            val xmlData = evt.xmppStorageData.firstOrNull { it.name == KEY }
            val list: MutableList<JoinedRoomData> = if (xmlData != null && xmlData.value.isNotEmpty()) parseXmlResponse(xmlData.value) else mutableListOf()
            var needUpdate = false
            Log.d(TAG, "Original list: ${list.joinToString { it.roomId }}")
            val pair = removeDuplicates(list)
            val cleanList = pair.first
            roomsToBeAdded.forEach { addedRoom ->
                if (cleanList.firstOrNull { it.hasSameChatKey(addedRoom.roomId) } == null) {
                    cleanList.add(addedRoom)
                    needUpdate = true
                }
            }
            roomsToBeAdded.clear()

            //update only desired fields
            fieldsToBeUpdated.forEach { (roomKey, nameValue) ->
                Log.d(TAG, "updating ${roomKey} id ${nameValue.first} value ${nameValue.second}")
                var roomInfo = cleanList.firstOrNull { it.roomId == roomKey }
                if (roomInfo == null && nameValue.first == JoinRoomStorageField.ROOM_NAME) {
                    roomInfo = JoinedRoomData(roomId = roomKey)
                    cleanList.add(roomInfo)
                }
                when (nameValue.first) {
                    JoinRoomStorageField.ROOM_ID -> {
                    }
                    JoinRoomStorageField.ROOM_NAME -> {
                        if (roomInfo?.roomName != nameValue.second) {
                            roomInfo?.roomName = nameValue.second
                            needUpdate = true
                        }
                    }
                    JoinRoomStorageField.ROOM_CREATOR -> {
                        if (roomInfo?.roomCreator != nameValue.second) {
                            roomInfo?.roomCreator = nameValue.second
                            needUpdate = true
                        }
                    }
                    JoinRoomStorageField.JOIN_TIME -> {
                        if (roomInfo?.joinTime != nameValue.second.toLong()) {
                            roomInfo?.joinTime = nameValue.second.toLong()
                            needUpdate = true

                        }
                    }
                    JoinRoomStorageField.LAST_VIEWED -> {
                        if ((roomInfo?.lastViewed ?: 0L) < nameValue.second.toLong()) {
                            roomInfo?.lastViewed = nameValue.second.toLong()
                            needUpdate = true
                        }
                    }
                    JoinRoomStorageField.LAST_VIEWED_ID -> {
                        //nvtd if we dont have message locally do not change it
                        if (roomInfo?.lastViewedMsgId != nameValue.second) {
                            roomInfo?.lastViewedMsgId = nameValue.second
                            needUpdate = true
                        }
                    }
                }
            }
            fieldsToBeUpdated.clear()

            Log.d(TAG, "Room to delete: ${roomsToBeDeleted.joinToString { it.roomId }}")
            roomsToBeDeleted.forEach { removedRoom ->
                cleanList.firstOrNull { it.hasSameChatKey(removedRoom.roomId) }?.let {
                    cleanList.remove(it)
                    needUpdate = true
                }
            }
            roomsToBeDeleted.clear()
            fixTimesToSeconds(cleanList)
            Log.d(TAG, "sending xml: ${parseToXml(cleanList)}")
            if (needUpdate) {
                val xmlValue = parseToXml(cleanList)

                val storageData = Xmpp.XmppStorageData()
                storageData.name = KEY
                storageData.value = xmlValue
                xmppAccount.SetPrivateStorageData(listOf(storageData))
            }
            roomsActive.clear()
            cleanList.forEach { roomsActive.put(it.roomId, it) }
            Log.d(TAG, "New list: ${roomsActive.values.toList().joinToString(postfix = "/n")}")
            joinedRoomsListChangedObservable.onNext(cleanList)

        }
    }

    private fun removeDuplicates(list: MutableList<JoinedRoomData>): Pair<MutableList<JoinedRoomData>, MutableList<JoinedRoomData>> {
        val okList = mutableListOf<JoinedRoomData>()
        val toRemove = mutableListOf<JoinedRoomData>()
        list.forEach { newItem ->
            val itemIinOkList = okList.firstOrNull { newItem.hasSameChatKey(it.roomId) }
            if (itemIinOkList == null) {
                okList.add(newItem)
            } else {
                if (newItem.roomId.length > itemIinOkList.roomId.length) {
                    okList.remove(itemIinOkList)
                    okList.add(newItem)
                    if (itemIinOkList.roomName.isNotEmpty() && newItem.roomName.isEmpty()) newItem.roomName = itemIinOkList.roomName
                    if (newItem.joinTime == 0L && itemIinOkList.joinTime > 0L) newItem.joinTime = itemIinOkList.joinTime
                }
                toRemove.add(newItem)
            }
        }
        return Pair(okList, toRemove)
    }

    private fun fixTimesToSeconds(list: MutableList<JoinedRoomData>) {
        list.forEach {
            if (it.joinTime > 1000000000000) { //if old time format is used (in ms)
                Log.e(TAG, "fixtimesToSeconds old ${it.joinTime} new: ${it.joinTime/1000}")
                it.joinTime = it.joinTime / 1000
            }
        }
    }

    override fun destroy() {
        xmppAccount.removeHandler(this)
    }

    override fun onAccountStatusChangedEvent(p0: XmppAccount?, p1: Xmpp.XmppAccountEvents.XmppAccountStatusChangedEvent?) {}

    override fun onStreamManagementState(p0: XmppAccount?, p1: Xmpp.XmppAccountEvents.XmppStreamManagementStateEvent?) {}

    override fun onLicensingError(p0: XmppAccount?, p1: Xmpp.XmppAccountEvents.XmppLicensingErrorEvent?) {}

    override fun onXmppEntityTimeEvent(p0: XmppAccount?, p1: Xmpp.XmppAccountEvents.XmppEntityTimeEvent?) {}

    override fun onError(p0: XmppAccount?, p1: Xmpp.XmppAccountEvents.XmppErrorEvent?) {}


}
