package com.bria.common.controller.im.refactoring.dataprovider

import android.annotation.SuppressLint
import android.content.Context
import com.bria.common.controller.Controllers
import com.bria.common.controller.im.ImpsUtils.getLocalMidnightTime
import com.bria.common.controller.im.articles.ArticleExtractor
import com.bria.common.controller.im.refactoring.IChatApi
import com.bria.common.controller.im.refactoring.adapter.IDataProviderEvents
import com.bria.common.controller.im.refactoring.db.entities.*
import com.bria.common.controller.im.refactoring.roomdb.entities.Message
import com.bria.common.controller.im.refactoring.roomdb.entities.MessageReadState
import com.bria.common.uireusable.datatypes.ExtendedConversationData
import com.bria.common.uireusable.datatypes.MessageListItemData
import com.bria.common.uireusable.datatypes.MessageListItemData.Companion.MESSAGE_TYPE_DATE_DIVIDER
import com.bria.common.util.GlobalConstants
import com.bria.common.util.Log
import io.reactivex.BackpressureStrategy
import io.reactivex.Flowable
import io.reactivex.Observable
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.processors.PublishProcessor
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.BehaviorSubject
import java.util.concurrent.ScheduledExecutorService
import java.util.concurrent.atomic.AtomicBoolean

open class InstantMessageDataProvider(val mContext: Context, val callback: IDataProviderEvents, val imExecutorService: ScheduledExecutorService) {

    companion object {
        const val TAG = "InstantMessageDataProvider"
        private const val ARTICLE_EXTRACTOR_CACHE_SIZE = 50
        internal const val NEW_LIST_EVENT: Int = 0
        internal const val MESSAGES_ADDED_EVENT: Int = 1
        internal const val MESSAGES_REMOVED_EVENT: Int = 2
        private const val MESSAGES_UPDATED_EVENT: Int = 3
        private const val MORE_MESSAGES_EVENT: Int = 4
        private const val CLEAN_LIST: Int = 5
    }

    protected var list = ArrayList<MessageListItemData>()

    val disposables = CompositeDisposable()

    protected val removedMessagesId = HashSet<Long>()


    var articleExtractor: ArticleExtractor

    protected var conversationId: Long = ImConversationData.INVALID_CONVERSATION_ID

    var isChatRoom = false

    private val selectedIms = ArrayList<Long>()

    private val mSelectedMessageCountObservable = BehaviorSubject.createDefault(0)

    protected fun getImData() = Controllers.get().im.imData

    private val chatApi: IChatApi
        get() = Controllers.get().im.chatApi

    protected fun numberOfLoadedMessages(): Int {
        var totalMessages = list.count { it.msgType != MessageListItemData.MESSAGE_TYPE_DATE_DIVIDER }
        if (totalMessages < GlobalConstants.REMOTE_SYNC_COUNT) {
            totalMessages = GlobalConstants.REMOTE_SYNC_COUNT
        }
        return totalMessages
    }

    var loading = AtomicBoolean(false)

    init {
        articleExtractor = ArticleExtractor(ARTICLE_EXTRACTOR_CACHE_SIZE, mContext.cacheDir.absolutePath)
    }

    fun clean() {
        Log.d(TAG, "Clean")
        conversationId = ImConversationData.INVALID_CONVERSATION_ID
        accountId = ""
        clearSelection()
        cleanData()
    }

    private fun cleanData() {
        Log.d(TAG, "CleanData")
        sourceOfNewConversation.onNext(Pair(CLEAN_LIST, ArrayList()))
    }

    var accountId: String = ""

    var extendedConversationData: ExtendedConversationData? = null

    fun setImConversationData(id: Long, hardwiredSession: Boolean, isChatRoom: Boolean, newAccountId: String, extendedConversationData: ExtendedConversationData) {
        Log.d(TAG, "Set Im Conversation: oldId: $conversationId, new id: $id")
        if (conversationId != id || isChatRoom != this.isChatRoom) {
            conversationId = id
            this.extendedConversationData = extendedConversationData
            this.isChatRoom = isChatRoom
            accountId = newAccountId
            clearSelection()
            this.hardwiredSession = hardwiredSession
            reloadConversationMessages()
        }

    }

    @SuppressLint("CheckResult")
    private fun reloadConversationMessages() {
        loading.set(true)
        callback.onListLoading()
        sourceOfNewConversation.onNext(Pair(NEW_LIST_EVENT, emptyList()))
    }

    @SuppressLint("CheckResult")
    fun loadMoreMessages() {
        sourceOfLoadedMoreMessages.onNext(Pair(MORE_MESSAGES_EVENT, listOf()))
    }

    /*
     * @return -1 if message is not found, else returns message position
     */
    protected fun getMessagePosition(messageList: List<MessageListItemData>, messageId: Long): Int =
            messageList.indices.firstOrNull { messageId == messageList[it].id } ?: -1

    protected fun addMessageToTheList(messageList: MutableList<MessageListItemData>, message: MessageListItemData): Pair<MessageListItemData, Int> {
        val result = calculateMessagePosition(messageList, message)
        if (!result.first) { // if message is not there already
            if (result.second == -1) {
                messageList.add(message)
            } else {
                messageList.add(result.second, message)
            }
        }
        return Pair(message, result.second)
    }

    @Synchronized
    protected open fun addMessagesToList(inputList: List<MessageListItemData>, outputList: MutableList<MessageListItemData>): ArrayList<MessageListItemData> {
        val tempDateSeparators = HashSet<Long>()
        val addedMessages = ArrayList<MessageListItemData>()
        removedMessagesId.addAll(inputList.asSequence().filter { it.isJoinedLeavedLocalMessage() }.map { it.id }.toList())
        inputList.filter { !it.deleted && !(it.isJoinedLeavedLocalMessage()) }.forEach { uiMessage ->
            if (outputList.firstOrNull { it.id == uiMessage.id } == null) {
                addMessageToTheList(outputList, uiMessage)
                addedMessages.add(uiMessage)
                tempDateSeparators.add(getLocalMidnightTime(uiMessage.time))
            }
        }
        tempDateSeparators
                .filter { getMessagePosition(outputList, it) == -1 }
                .map { MessageListItemData.buildDateDivider(it, mContext) }
                .forEach { addMessageToTheList(outputList, it) }
        return addedMessages
    }

    @Synchronized
    private fun removeMessagesFromList(inputList: List<MessageListItemData>, outputList: MutableList<MessageListItemData>) {
        for (imData in inputList) {
            val position = getMessagePosition(outputList, imData.id)
            if (position > 0) outputList.removeAt(position)
            if (selectedIms.contains(imData.id)) {
                selectedIms.remove(imData.id)
                mSelectedMessageCountObservable.onNext(selectedIms.size)
            }
        }
    }

    @Synchronized
    fun updateMessages(inputList: List<MessageListItemData>, outputList: MutableList<MessageListItemData>) {
        inputList.map { Pair(it, getMessagePosition(outputList, it.id)) }
                .filter { it.second > 0 }.map { Pair(it.first, it.second) }.forEach { (uiMessage, position) ->
                    uiMessage.oldMessage = outputList.removeAt(position)
                    uiMessage.oldMessage?.let {
                        uiMessage.selected = it.selected
                    }
                    outputList.add(position, uiMessage)
                }
    }

    protected fun dbListToUIList(oldLIst: List<InstantMessageData>) : List<MessageListItemData> {
        var list: List<MessageListItemData> = listOf()
        try {
            list = oldLIst.map { MessageListItemData(it, mContext) }
        } catch (e : Exception) {
            Log.e(TAG, "Exception while parsing list: ${e.message}")
        }
        return list
    }


    protected fun roomDbListToUIList(oldLIst: List<Message>): List<MessageListItemData> {
        var list: List<MessageListItemData> = listOf()
        try {
            list = oldLIst.map { MessageListItemData(it, mContext) }
        } catch (e : Exception) {
            Log.e(TAG, "Exception while parsing list: ${e.message}")
        }
        return list
    }

    protected val sourceOfChatRoomAddedMessages = getImData().chatRoomRepo.getOnMessagesAddedObservable().toFlowable(BackpressureStrategy.BUFFER).observeOn(Schedulers.from(imExecutorService))
            .filter { messages -> !messages.isEmpty() && isChatRoom && messages[0].chatId == conversationId }
            .map { oldList -> roomDbListToUIList(oldList) }
            .map { filteredMessages -> Pair(MESSAGES_ADDED_EVENT, filteredMessages) }

    protected val sourceOfChatRoomRemovedMessages = getImData().chatRoomRepo.getOnMessagesRemovedObservable().toFlowable(BackpressureStrategy.BUFFER).observeOn(Schedulers.from(imExecutorService))
            .filter { messages -> !messages.isEmpty() && isChatRoom && messages[0].chatId == conversationId }
            .map { oldList -> roomDbListToUIList(oldList) }
            .map { filteredMessages -> Pair(MESSAGES_REMOVED_EVENT, filteredMessages) }

    protected val sourceOfChatRoomUpdatedMessages = getImData().chatRoomRepo.getOnMessagesUpdatedObservable().toFlowable(BackpressureStrategy.BUFFER).observeOn(Schedulers.from(imExecutorService))
            .filter { messages -> !messages.isEmpty() && isChatRoom && messages[0].chatId == conversationId }
            .map { oldList -> roomDbListToUIList(oldList) }
            .map { filteredMessages -> Pair(MESSAGES_UPDATED_EVENT, filteredMessages) }

    protected val sourceOfChatRoomReadMessages = getImData().chatRoomRepo.getOnMessagesReadObservable().toFlowable(BackpressureStrategy.BUFFER).observeOn(Schedulers.from(imExecutorService))
            .filter { messages -> !messages.isEmpty() && isChatRoom && messages[0].chatId == conversationId }
            .map { oldList -> roomDbListToUIList(oldList) }
            .map { filteredMessages -> Pair(MESSAGES_UPDATED_EVENT, filteredMessages) }

    protected val sourceOfAddedMessages = getImData().observableOnListMessagesAdded.toFlowable(BackpressureStrategy.BUFFER).observeOn(Schedulers.from(imExecutorService))
            .filter { messages -> !messages.isEmpty() && !isChatRoom && messages[0].conversationId == conversationId }
            .map { oldList -> dbListToUIList(oldList) }
            .map { filteredMessages -> Pair(MESSAGES_ADDED_EVENT, filteredMessages) }

    protected val sourceOfRemovedMessages = getImData().observableOnListMessagesRemoved.toFlowable(BackpressureStrategy.BUFFER).observeOn(Schedulers.from(imExecutorService))
            .filter { messages -> !messages.isEmpty() && !isChatRoom && messages[0].conversationId == conversationId }
            .map { oldList -> dbListToUIList(oldList) }
            .map { filteredMessages -> Pair(MESSAGES_REMOVED_EVENT, filteredMessages) }

    protected val sourceOfUpdatedMessages = getImData().observableOnListMessageUpdated.toFlowable(BackpressureStrategy.BUFFER).observeOn(Schedulers.from(imExecutorService))
            .filter { messages -> !messages.isEmpty() && !isChatRoom && messages[0].conversationId == conversationId }
            .map { oldList -> dbListToUIList(oldList) }
            .map { filteredMessages -> Pair(MESSAGES_UPDATED_EVENT, filteredMessages) }

    //assume: all messages belongs to same conversation
    protected val sourceOfReadMessages = getImData().observableOnMessagesRead.toFlowable(BackpressureStrategy.BUFFER).observeOn(Schedulers.from(imExecutorService))
            .filter { messages -> !messages.isEmpty() && !isChatRoom && messages[0].conversationId == conversationId }
            .map { oldList -> dbListToUIList(oldList) }
            .map { filteredMessages -> Pair(MESSAGES_UPDATED_EVENT, filteredMessages) }

    protected val sourceOfNewConversation = PublishProcessor.create<Pair<Int, List<MessageListItemData>>>()
    protected val sourceOfLoadedMoreMessages = PublishProcessor.create<Pair<Int, List<MessageListItemData>>>()

    protected val sourceOfManuallyChangedList = PublishProcessor.create<ArrayList<MessageListItemData>>()

    open val source: Flowable<ArrayList<MessageListItemData>> = sourceOfNewConversation
            .mergeWith(sourceOfLoadedMoreMessages)
            .mergeWith(sourceOfAddedMessages)
            .mergeWith(sourceOfRemovedMessages)
            .mergeWith(sourceOfUpdatedMessages)
            .mergeWith(sourceOfReadMessages)
            .mergeWith(sourceOfChatRoomAddedMessages)
            .mergeWith(sourceOfChatRoomRemovedMessages)
            .mergeWith(sourceOfChatRoomUpdatedMessages)
            .mergeWith(sourceOfChatRoomReadMessages)
            .observeOn(Schedulers.from(imExecutorService))
            .map {
                Log.d(TAG, "Event: ${it.first}")
                try {
                    when (it.first) {
                        NEW_LIST_EVENT -> {
                            val listOfMessages = if (isChatRoom) {
                                getChatRoomMessagesForActiveChatRoom(0, numberOfLoadedMessages())
                            } else {
                                dbListToUIList(getImData().getNMessageWithOffsetForConversation(conversationId, 0, numberOfLoadedMessages()))
                            }
                            loading.set(false)
                            callback.onListLoaded(listOfMessages.size)
                            removedMessagesId.clear()
                            list.clear()
                            addMessagesToList(listOfMessages, list)
                        }
                        MESSAGES_ADDED_EVENT -> addMessagesToList(it.second, list)
                        MESSAGES_REMOVED_EVENT -> removeMessagesFromList(it.second, list)
                        MESSAGES_UPDATED_EVENT -> updateMessages(it.second, list)
                        MORE_MESSAGES_EVENT -> {
                            val addedMessages = if (isChatRoom) {
                                getChatRoomMessagesForActiveChatRoom(calculateOffset(), GlobalConstants.REMOTE_SYNC_COUNT)
                            } else {
                                dbListToUIList(getImData().getNMessageWithOffsetForConversation(conversationId, calculateOffset(), GlobalConstants.REMOTE_SYNC_COUNT))
                            }
                            addMessagesToList(addedMessages, list)
                        }
                        CLEAN_LIST -> {
                            list.clear()
                            removedMessagesId.clear()
                        }
                        else -> {
                            Log.d(TAG, "Else event: ${it.first}")
                            updateMessages(it.second, list)
                        }
                    }
                } catch (e : Exception) {
                    Log.d(TAG, "Exception: ${e.message}")
                }
                list

            }.mergeWith(sourceOfManuallyChangedList).replay(1).autoConnect()

    protected open fun getChatRoomMessagesForActiveChatRoom(offset: Int, numberOfLoadedMessages: Int): List<MessageListItemData> {
        return roomDbListToUIList(getImData().chatRoomRepo.getNMessageWithOffsetForChat(conversationId, offset, numberOfLoadedMessages))
    }

    private val sourceOfListOfConversationsUpdated = getImData().observableOnListConversationsUpdated
            .subscribeOn(Schedulers.computation())
            .subscribe(
                    { list ->
                        list.filter { imConversationData -> imConversationData.id == conversationId }.firstOrNull()?.let {
                            reloadConversationMessages()
                        }
                    },
                    { error ->
                        Log.e(TAG, "Throwable " + error.message)
                    }
            )

    /**
     * Uses binary search to find new message position based on message time
     *  @param messageList - message list in which we are searching
     *  @param message - message which position we are looking
     *  @return pair of <Is message in list already, message position>
     */
    private fun calculateMessagePosition(messageList: List<MessageListItemData>, message: MessageListItemData): Pair<Boolean, Int> {
        if (messageList.isEmpty()) return Pair(false, 0)
        if (message > messageList[messageList.size - 1]) return Pair(false, -1) // to the end
        var lo = 0
        var hi = messageList.size - 1
        while (lo <= hi) {
            val mid = (lo + hi) / 2
            val compare = message.compareTo(messageList[mid])
            when {
                compare > 0 -> lo = mid + 1
                compare < 0 -> hi = mid - 1
                else -> return Pair(messageList[mid].id == message.id, mid)
            }
        }
        return Pair(false, hi + 1)
    }

    fun toggleSelection(id: Long) {
        list.firstOrNull { it.id == id }?.let { message ->
            message.selected = !message.selected
            if (message.selected) {
                selectedIms.add(id)
            } else {
                selectedIms.remove(id)
            }

            mSelectedMessageCountObservable.onNext(selectedIms.size)
            sourceOfManuallyChangedList.onNext(list)
        }
    }

    @Suppress("unused")
    fun setSelectedIms(ids: LongArray) {
        ids.forEach { imId ->
            list.firstOrNull { it.id == imId }?.apply {
                this.selected = true
                selectedIms.add(this.id)
            }
        }

        sourceOfManuallyChangedList.onNext(list)
        mSelectedMessageCountObservable.onNext(selectedIms.size)
    }

    fun isSelected(id: Long): Boolean = list.firstOrNull { it.id == id }?.selected ?: false

    fun clearSelection() {
        list.forEach { im -> im.selected = false }
        selectedIms.clear()
        sourceOfManuallyChangedList.onNext(list)
        mSelectedMessageCountObservable.onNext(selectedIms.size)
    }

    fun getSelectedMessagesCount(): Int = selectedIms.size

    val selectedMessageCountObservable: Observable<Int>
        get() = mSelectedMessageCountObservable

    /**
     * Used primary for markAsRead Section
     * If last message is visible then mark whole conversation as read
     * @param firstVisibleItemPosition
     * @param lastVisibleItemPosition
     */
    @SuppressLint("CheckResult")
    fun setVisibleMessages(firstVisibleItemPosition: Int, lastVisibleItemPosition: Int) {
        imExecutorService.execute {
            val listSize = list.size
            if (listSize > 0 && conversationId != -1L) {
                if (lastVisibleItemPosition >= listSize - 1) {
                    if (isChatRoom) {
                        getImData().chatRoomRepo.markReadChatRoom(conversationId)
                    } else {
                        getImData().markMessagesRead(conversationId)
                    }
                } else {
                    if (firstVisibleItemPosition < 0 || lastVisibleItemPosition < 0
                            || list.size <= firstVisibleItemPosition || list.size <= lastVisibleItemPosition) {
                        return@execute
                    }
                    if (isChatRoom) {
                        list.subList(firstVisibleItemPosition, lastVisibleItemPosition + 1).filter { !it.isRead }.map { it.id }.let { visibleIds ->
                            if (visibleIds.isNotEmpty()) {
                                disposables.add(getImData().chatRoomRepo.getMessagesWithIds(visibleIds)
                                        .observeOn(Schedulers.from(imExecutorService)).subscribe({ messages ->
                                            messages.filter { it.readState != MessageReadState.READ || it.isFileTransferUnread() }.map {
                                                it.readState = MessageReadState.READ
                                                it
                                            }.apply {
                                                getImData().chatRoomRepo.markRead(this)
                                            }
                                        }, {
                                            Log.e(TAG, "No messages found")
                                        }))
                            }
                        }
                    } else {
                        val visibleImList = ArrayList<InstantMessageData>()
                        for (i in firstVisibleItemPosition..lastVisibleItemPosition) {
                            val item = list[i]
                            if (!item.isRead || item.isFileTransferUnread()) {
                                getImData().getMessageById(list[i].id)?.let {
                                    it.markRead()
                                    visibleImList.add(it)
                                }
                            }
                        }
                        if (!visibleImList.isEmpty()) {
                            getImData().markMessagesRead(visibleImList)
                        }
                    }
                }
            }
        }
    }

    private fun calculateOffset(): Int = list.filter { it.msgType != MESSAGE_TYPE_DATE_DIVIDER }.size + removedMessagesId.count()

    @SuppressLint("CheckResult")
    fun accept(messageListItemData: MessageListItemData) {
        imExecutorService.execute {
            val message = getImData().getMessageById(messageListItemData.id)
            chatApi.acceptMessage(message)
        }
    }

    @SuppressLint("CheckResult")
    fun reject(messageListItemData: MessageListItemData) {
        imExecutorService.execute {
            val message = getImData().getMessageById(messageListItemData.id)
            chatApi.rejectMessage(message)
        }
    }

    fun retry(messageListItemData: MessageListItemData) {
        imExecutorService.execute {
            getImData().getMessageById(messageListItemData.id)?.let {
                chatApi.retryTransfer(it)
            }
        }
    }

    fun isEmpty(): Boolean = list.isEmpty()

    @Suppress("unused")
    @SuppressLint("CheckResult")
    fun incomingFileIsNotImage(messageListItemData: MessageListItemData) {
        if (messageListItemData.status != ImStatusData.INCOMING_FT_RECEIVED_CHECKED_READ) {
            imExecutorService.execute {
                val message = getImData().getMessageById(messageListItemData.id)
                chatApi.incomingFileIsNotImage(message)
            }
        }
    }

    @Suppress("unused")
    @SuppressLint("CheckResult")
    fun incomingFileIsImage(messageListItemData: MessageListItemData) {
        if (messageListItemData.status != ImStatusData.INCOMING_IMAGE_TRANSFER_RECEIVED_READ) {
            imExecutorService.execute {
                val message = getImData().getMessageById(messageListItemData.id)
                chatApi.incomingFileIsImage(message)
            }
        }
    }

    fun hasTransferMessagesInSelection() = getSelectedMessages().firstOrNull { it.isTransfer() } != null
    //nvtd add rxjava event
    fun deleteSelectedMessages() {
        if (isChatRoom) {
            disposables.add(Observable.fromCallable { getSelectedChatRoomMessages() }.subscribeOn(Schedulers.io()).observeOn(Schedulers.io()).subscribe({
                getImData().chatRoomRepo.deleteMessages(it)
            }, { Log.e(TAG, it.message) }))

        } else {
            getImData().removeBulkMessages(getSelectedMessages())
        }
    }

    private fun getSelectedMessages(): List<InstantMessageData> {
        val selected = ArrayList<InstantMessageData>()
        selectedIms.mapNotNullTo(selected) { getImData().getMessageById(it) }
        return selected
    }

    fun getSelectedFileTransferMessages() = getSelectedMessages().filter { it.isTransfer() }

    private fun getSelectedNonFileTransferMessages() = getSelectedMessages().filter { !it.isTransfer() }

    private fun getSelectedChatRoomMessages(): List<Message> {
        val selected = ArrayList<Message>()
        selectedIms.mapNotNullTo(selected) { getImData().chatRoomRepo.getMessageById(it).blockingGet() }
        return selected
    }

    data class GetSelectedMessagesAsStringResult(val text: String, val someWereSkipped: Boolean)

    fun getSelectedMessagesAsString(): GetSelectedMessagesAsStringResult {
        if (isChatRoom) {
            val text = getSelectedChatRoomMessages()
                .sortedBy { it.creationTime }
                .map { it.text }
                .joinToString(separator = System.lineSeparator())

            val someWereSkipped = false

            return GetSelectedMessagesAsStringResult(text, someWereSkipped)
        } else {
            val text = getSelectedNonFileTransferMessages()
                .sortedBy { it.time }
                .map { it.message }
                .joinToString(separator = System.lineSeparator())

            val someWereSkipped = getSelectedFileTransferMessages().any()

            return GetSelectedMessagesAsStringResult(text, someWereSkipped)
        }
    }

    var hardwiredSession: Boolean = false


}
