package com.bria.common.controller.im.refactoring.roomdb

import com.bria.common.controller.im.refactoring.roomdb.entities.ChatRoom
import com.bria.common.controller.im.refactoring.roomdb.entities.Message
import com.bria.common.controller.im.refactoring.roomdb.entities.MessageReadState
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.PublishSubject
import java.util.concurrent.ConcurrentHashMap
import java.util.concurrent.ScheduledExecutorService

class ChatCounter(private val chatRoomRepo: ChatRoomRepo, imExecutorService : ScheduledExecutorService) {


    private val compositeDisposable = CompositeDisposable()

    private val unreadMessages = ConcurrentHashMap<Long, MutableSet<Long>>()
    private val unreadCollabMessages = ConcurrentHashMap<Long, MutableSet<Long>>()

    fun getCounterChangedObservable() = observableCounterChanged

    private val observableCounterChanged = PublishSubject.create<Long>() //in which chat room counter changed

    init {
        reinit()
        compositeDisposable.add(chatRoomRepo.getOnMessagesAddedObservable().observeOn(Schedulers.from(imExecutorService))
                .subscribe ({ list -> list.filter { (it.readState == MessageReadState.UNREAD_NOT_SWIPED || it.readState == MessageReadState.UNREAD_SWIPED) &&
                        !it.isOutgoing()}.forEach { addMessage(it) } },{}))
        compositeDisposable.add(chatRoomRepo.getOnMessagesRemovedObservable().mergeWith(chatRoomRepo.getOnMessagesReadObservable())
                .observeOn(Schedulers.from(imExecutorService))
                .subscribe ({ list -> list.filter { !it.isOutgoing()}.forEach {
                    removeMessage(it)
                } },{}))
        compositeDisposable.add(chatRoomRepo.getOnMessagesUpdatedObservable().observeOn(Schedulers.from(imExecutorService))
                .subscribe ({ list -> list.filter { !it.isOutgoing() }.forEach{ if (it.readState == MessageReadState.READ) removeMessage(it) else addMessage(it)
                }},{}))
        compositeDisposable.add(chatRoomRepo.getOnRoomDeletedObservable().mergeWith(chatRoomRepo.getOnRoomReadObservable())
                .observeOn(Schedulers.from(imExecutorService))
                .subscribe ({ list -> list.forEach{ clearChat(it)}},{}))

    }

    private fun clearChat(chat: ChatRoom) {
        unreadMessages.remove(chat.id)
        unreadCollabMessages.remove(chat.id)
        observableCounterChanged.onNext(chat.id)
    }

    fun reinit() {
        unreadMessages.clear()
        unreadCollabMessages.clear()
        compositeDisposable.add(chatRoomRepo.getAllChatRooms().subscribeOn(Schedulers.io()).observeOn(Schedulers.io()).subscribe ({ chats ->
            chats.forEach { chatRoom -> updateMessagesForChat(chatRoom.id)}
        },{}))
    }

    private fun updateMessagesForChat(id : Long) {
        compositeDisposable.add(chatRoomRepo.getUnreadMessagesForChatRoom(id).subscribeOn(Schedulers.io()).observeOn(Schedulers.io())
                .subscribe({messages -> messages.forEach { addMessage(it) }
        },{}))
    }

    private fun addMessage(message : Message) {
        val map = if (message.isCollab()) unreadCollabMessages else unreadMessages
        if (!map.containsKey(message.chatId)) {
            map[message.chatId] = mutableSetOf(message.id)
        } else {
            map[message.chatId]?.add(message.id)
        }
        observableCounterChanged.onNext(message.chatId)
    }

    private fun removeMessage(message : Message) {
        val list = if (message.isCollab()) unreadCollabMessages else unreadMessages
        if (list.containsKey(message.chatId)) {
            list[message.chatId]?.remove(message.id)
        }
        observableCounterChanged.onNext(message.chatId)
    }

    fun getNumberOfUnreadMessagesForChat(id : Long) : Int = unreadMessages[id]?.size ?: unreadCollabMessages[id]?.size ?: 0

    fun getNumberOfUnreadMessages() : Int {
        var count = 0
        unreadMessages.mapKeys { entry -> count+= entry.value.size }
        return count
    }

    fun getNumberOfChatsWithUnreadMessages() : Int{
        var count = 0
        unreadMessages.mapKeys { entry -> if (entry.value.size > 0) count+=1 }
        return count
    }

    fun getNumberOfUnreadCollabMessages() : Int {
        var count = 0
        unreadCollabMessages.mapKeys { entry -> count+= entry.value.size }
        return count
    }

    fun destroy() {
        compositeDisposable.clear()
    }

}
