package com.bria.common.controller.im.refactoring.roomdb

import android.arch.persistence.room.Room
import android.content.Context
import com.bria.common.controller.Controllers
import com.bria.common.controller.im.refactoring.roomdb.entities.ChatRoom
import com.bria.common.controller.im.refactoring.roomdb.entities.ChatRoomAnonymousMode
import com.bria.common.controller.im.refactoring.roomdb.entities.Message
import com.bria.common.controller.im.refactoring.roomdb.entities.MessageReadState
import com.bria.common.uireusable.datatypes.IdentifiableDataItem
import com.bria.common.util.Log
import io.reactivex.Maybe
import io.reactivex.Observable
import io.reactivex.Single
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.PublishSubject
import io.reactivex.subjects.Subject

class ChatRoomRepoImpl(val context: Context, val owner: String) : ChatRoomRepo {

    companion object {
        const val TAG = "ChatRoomRepoImpl"
        const val MESSAGES_CHANGED = 0
        const val CHAT_ROOMS_CHANGED = 1
    }

    //nvtd should be done per account
    private val serverRoomsMap = mutableMapOf<String, LocalRoomInfo>()

    val localRoomJids = mutableListOf<String>()

    val imExecutorService = Controllers.get().im.getImExecutorService()

    var imDb: ImProviderDb

    var chatCounter: ChatCounter

    var observableChatRoomAdded: Subject<List<ChatRoom>> = PublishSubject.create()
    var observableChatRoomUpdated: Subject<List<ChatRoom>> = PublishSubject.create()
    var observableChatRoomMessagesRead: Subject<List<ChatRoom>> = PublishSubject.create()
    var observableChatRoomDelete: Subject<List<ChatRoom>> = PublishSubject.create()
    var observableMessagesAdded: Subject<List<Message>> = PublishSubject.create()
    var observableMessagesUpdated: Subject<List<Message>> = PublishSubject.create()
    var observableMessagesRemoved: Subject<List<Message>> = PublishSubject.create()
    var observableMessagesRead: Subject<List<Message>> = PublishSubject.create()

    val observableChatRoomChanged: Observable<List<ChatRoom>> = observableChatRoomAdded
            .mergeWith(observableChatRoomUpdated)
            .mergeWith(observableChatRoomDelete)

    val observableMessagesChanged: Observable<List<Message>> = observableMessagesAdded
            .mergeWith(observableMessagesUpdated)
            .mergeWith(observableMessagesRemoved)
            .mergeWith(observableMessagesRead)

    val imDataChangedObservable: Observable<Int> = observableMessagesChanged.map { MESSAGES_CHANGED }.mergeWith(observableChatRoomChanged.map {
        CHAT_ROOMS_CHANGED
    })

    var observableServerRoomListChanged: Subject<List<LocalRoomInfo>> = PublishSubject.create()

    init {

        imDb = Room.databaseBuilder(context.applicationContext, ImProviderDb::class.java, "im$owner").build()
        chatCounter = ChatCounter(this, imExecutorService)
        val disp = getAllChatRooms().observeOn(Schedulers.io()).subscribeOn(Schedulers.io()).subscribe({ roomList ->
            localRoomJids.clear()
            localRoomJids.addAll(roomList.map { it.chatKey })
        }, {
            Log.e(TAG, "Error in loading roomList ${it.message}")
        })

    }


    override fun getOnRoomAddedObservable(): Subject<List<ChatRoom>> = observableChatRoomAdded

    override fun getOnRoomUpdatedObservable(): Subject<List<ChatRoom>> = observableChatRoomUpdated
    override fun getOnRoomReadObservable(): Subject<List<ChatRoom>> = observableChatRoomMessagesRead
    override fun getOnRoomDeletedObservable(): Subject<List<ChatRoom>> = observableChatRoomDelete

    override fun getOnMessagesUpdatedObservable() = observableMessagesUpdated

    override fun getOnMessagesAddedObservable(): Subject<List<Message>> = observableMessagesAdded

    override fun getOnMessagesRemovedObservable(): Subject<List<Message>> = observableMessagesRemoved

    override fun getOnMessagesReadObservable(): Subject<List<Message>> = observableMessagesRead

    override fun getOnServerRoomListChangedObservable() = observableServerRoomListChanged

    override fun getAllChatRooms(): Single<List<ChatRoom>> {
        return imDb.getChatRoomDao().getAllRooms()
    }

    override fun getRoomByRoomKey(key: String): ChatRoom? {
        return imDb.getChatRoomDao().getRoomByRoomKey(key)
    }

    override fun getChatRoom(id: Long): ChatRoom? {
        return imDb.getChatRoomDao().getRoom(id)
    }

    override fun getChatRoomByChatKey(roomKey: String): ChatRoom? {
        return imDb.getChatRoomDao().getRoomByRoomKey(roomKey)
    }

    override fun getDataChangedObservable() = imDataChangedObservable

    @Synchronized
    override fun addChatRoom(chatRoom: ChatRoom): Long {
        val id = imDb.getChatRoomDao().insertRoom(chatRoom)
        localRoomJids.add(chatRoom.chatKey)
        addRoomInfo(LocalRoomInfo(name = chatRoom.name, accId = chatRoom.accountId, cleanedJid = chatRoom.chatKey, description = "", anonymousMode =
        ChatRoomAnonymousMode.NON_ANONYMOUS, creationTime = 0))
        if (id != null) {
            chatRoom.id = id
            observableChatRoomAdded.onNext(listOf(chatRoom))
        }
        return id ?: -1L

    }

    @Synchronized
    override fun updateChatRoom(chatRoom: ChatRoom) {
        imDb.getChatRoomDao().updateChatRoom(chatRoom)
        observableChatRoomUpdated.onNext(listOf(chatRoom))
    }

    override fun deleteChatRoom(chatRoom: ChatRoom) {
        Log.d(TAG, "deleteChatRoom: ${chatRoom.chatKey}")
        imDb.getChatRoomDao().deleteRoom(chatRoom)
        localRoomJids.remove(chatRoom.chatKey)
        //serverRoomsMap.remove(chatRoom.chatKey)
        //serverRoomsMap.remove(ImpsUtils.removeDomainFromAddress(chatRoom.chatKey))
        imDb.getChatRoomDao().deleteMessagesForChatRoom(chatRoom.id)
        observableChatRoomDelete.onNext(listOf(chatRoom))
    }

    override fun getLastMessageForChatRoom(id: Long) = imDb.getChatRoomDao().getLastMessageForChatRoom(id)

    override fun getLastIncomingMessageForChatRoom(id: Long) = imDb.getChatRoomDao().getLastIncomingMessageForChatRoom(id)

    override fun getNumberOfUnreadMessagesForChatRoom(id: Long) = chatCounter.getNumberOfUnreadMessagesForChat(id)

    override fun getNumberOfUnreadMessages() = chatCounter.getNumberOfUnreadMessages()

    override fun getNumberOfChatsWithUnreadMessages() = chatCounter.getNumberOfChatsWithUnreadMessages()

    override fun getNumberOfUnreadCollabMessages() = chatCounter.getNumberOfUnreadCollabMessages()

    override fun getUnreadMessagesAllChatRooms(): Single<List<Message>> {
        return imDb.getChatRoomDao().getUnreadMessagesAllChatRooms()
    }

    override fun getUnreadMessagesNotSwipedAllChatRooms(): Single<List<Message>> {
        return imDb.getChatRoomDao().getUnreadNotSwipedMessages()
    }

    override fun getUnreadMessagesForChatRoom(id: Long): Single<List<Message>> {
        return imDb.getChatRoomDao().getUnreadMessagesForChatRoom(id)
    }

    override fun getUnreadMessagesForChatRoomUpToTime(id: Long, time: Long): Single<List<Message>> {
        return imDb.getChatRoomDao().getUnreadMessagesForChatRoomUpToTime(id, time)
    }

    override fun changeOwner(owner: String) {
        imDb.close()
        imDb = Room.databaseBuilder(context.applicationContext, ImProviderDb::class.java, "im$owner").build()
        chatCounter.reinit()
    }

    override fun addMessage(message: Message) {
        val localMessage = imDb.getChatRoomDao().getMessageByExternalId(message.externalId)
        if (localMessage == null) {
            val id = imDb.getChatRoomDao().insertMessage(message)
            if (id != null || id == -1L) {
                message.id = id
                imDb.getChatRoomDao().updateChatTime(message.creationTime, message.chatId)
                observableMessagesAdded.onNext(listOf(message))
                getChatRoom(message.chatId)?.let {
                    observableChatRoomUpdated.onNext(listOf(it))
                }
            }
        }
    }

    override fun getMessageByExternalId(externalID : String) : Message? {
        return imDb.getChatRoomDao().getMessageByExternalId(externalID)
    }

    //must be of same conversation
    override fun addMessages(messages: List<Message>) {
        val messageIds = imDb.getChatRoomDao().insertMessages(messages)
        messageIds.forEachIndexed { index, id ->
            id?.let { messages[index].id = id }
        }
        val message = messages.sortedWith((compareBy(Message::creationTime))).last()
        imDb.getChatRoomDao().updateChatTime(message.creationTime, message.chatId)
        observableMessagesAdded.onNext(messages)
    }

    override fun updateMessages(messages: List<Message>): Int {
        val update = imDb.getChatRoomDao().updateMessages(messages)
        observableMessagesUpdated.onNext(messages)
        return update
    }

    override fun deleteMessages(messages: List<Message>) {
        imExecutorService.execute {
            imDb.getChatRoomDao().deleteMessages(messages)
            observableMessagesRemoved.onNext(messages)
        }
    }

    override fun markRead(messages: List<Message>) {
        if (messages.isNotEmpty()) {
            imExecutorService.execute {
                messages.forEach { it.readState = MessageReadState.READ }
                imDb.getChatRoomDao().markRead(messages.map { it.id })
                observableMessagesRead.onNext(messages)
            }
        }
    }

    override fun markReadChatRoom(chatRoomId: Long) {
        imExecutorService.execute {
            imDb.getChatRoomDao().markReadChatRoom(chatRoomId)
            val chatRooms = listOf(imDb.getChatRoomDao().getRoom(chatRoomId) ?: ChatRoom())
            observableChatRoomUpdated.onNext(chatRooms)
            observableChatRoomMessagesRead.onNext(chatRooms)

        }
    }

    override fun getNMessageWithOffsetForChat(id: Long, offset: Int, limit: Int): List<Message> {
        return imDb.getChatRoomDao().getNMessageWithOffsetForChat(id, offset, limit)
    }

    override fun getMessageById(id: Long): Maybe<Message> {
        return imDb.getChatRoomDao().getMessageById(id)
    }

    override fun getMessagesWithIds(messageIds: List<Long>) : Single<List<Message>> {
        return imDb.getChatRoomDao().getMessagesWithIds(messageIds)
    }

    override fun addRoomInfo(info: LocalRoomInfo) {
        serverRoomsMap[info.cleanedJid] = info
        observableServerRoomListChanged.onNext(serverRoomsMap.values.toList())
    }

    override fun getRoomInfoByJid(jid: String): LocalRoomInfo? {
        return serverRoomsMap[jid]
    }

    override fun removeRoomInfoWithJid(jid: String) {
        serverRoomsMap.remove(jid)
        observableServerRoomListChanged.onNext(serverRoomsMap.values.toList())
    }

    override fun updateRoomInfo(info: LocalRoomInfo) {
        removeRoomInfoWithJid(info.cleanedJid)
        serverRoomsMap[info.cleanedJid] = info
        observableServerRoomListChanged.onNext(serverRoomsMap.values.toList())
    }

    override fun replaceRoomInfos(infos: List<LocalRoomInfo>) {
        //rework this logic
        //not removing from local list
        //val filteredRooms = infos.filter { !localRoomJids.contains(ImpsUtils.removeDomainFromAddress(it.cleanedJid))}
        infos.forEach { info ->
            if (info.hasInfo || serverRoomsMap[info.cleanedJid]?.hasInfo != true) {
                serverRoomsMap[info.cleanedJid] = info
            }
        }
        //remove if not in new list
        serverRoomsMap.filter { severRoom -> infos.firstOrNull { it.cleanedJid == severRoom.key } == null }.forEach {
            serverRoomsMap.remove(it.key)
        }

        observableServerRoomListChanged.onNext(serverRoomsMap.values.toList())
    }


    override fun removeRoomInfosForAcc(accId: String) {
        serverRoomsMap.filter { it.value.accId == accId }.forEach {
            serverRoomsMap.remove(it.key)
        }
    }

    override fun roomNameInTheList(name: String): Boolean {
        return serverRoomsMap.values.toList().firstOrNull { it.name == name } != null
    }

    override fun getServerRoomList(query: String, withInfo: Boolean, hideLocalRooms: Boolean): MutableList<LocalRoomInfo> {
        //ntd remove rooms already in database
        return if (query.isEmpty()) {
            serverRoomsMap.values.toList().filter { !(withInfo && !it.hasInfo) && (!hideLocalRooms || !localRoomJids.contains(it.cleanedJid)) }.sortedBy {
                it
                        .name
                        .toLowerCase()
            }
                    .toMutableList()
        } else {
            serverRoomsMap.values.toList().filter {
                !(withInfo && !it.hasInfo)
                        && (!hideLocalRooms || !localRoomJids.contains(it.cleanedJid))
                        && (it.name.contains(query, true) || it.description.contains(query,
                        true))
            }
                    .toList().sortedBy { it.name.toLowerCase() }.toMutableList()
        }
    }



    override fun roomInTheServerList(jid: String): Boolean {
        return serverRoomsMap.containsKey(jid)
    }

    override fun getAllPrivateChatRooms(): Single<List<ChatRoom>> {
        return imDb.getChatRoomDao().getAllPrivateChatRooms()
    }

    override fun getFilteredMessagesWithOffsetForChat(id: Long, offset: Int, limit: Int, filter: String): List<Message> {
        return imDb.getChatRoomDao().getFilteredMessagesWithOffsetForChat(id, offset, limit, filter)
    }

    override fun destroy() {
        chatCounter.destroy()
        imDb.close()
    }
}

data class LocalRoomInfo(val name: String,
                         val accId: String,
                         val cleanedJid: String,
                         var description: String = "",
                         var anonymousMode: ChatRoomAnonymousMode = ChatRoomAnonymousMode.NON_ANONYMOUS,
                         var hasInfo: Boolean = false,
                         var creationTime: Long
) : IdentifiableDataItem {
    var topic: String = ""


    override fun getUniqueIdentifier(): String {
        return cleanedJid
    }


}
