package com.bria.common.controller.im.refactoring.roomdb.entities

import android.arch.persistence.room.*
import com.bria.common.controller.im.refactoring.db.entities.ImConversationData
import com.bria.common.controller.settings.core.upgrade.v1.types.SettingsStringEncoder_v1
import com.bria.common.uireusable.datatypes.ExtendedConversationData
import com.bria.common.util.TextBoundaries
import com.bria.common.util.im.ParticipantsSet

@Entity(tableName = "rooms")
data class ChatRoom(
        @PrimaryKey(autoGenerate = true)
        @ColumnInfo(name = "id") var id: Long = 0,
        @ColumnInfo(name = "type") var type: Long = ExtendedConversationData.CHAT_ROOM_TYPE.toLong(),
        @ColumnInfo(name = "chatKey") var chatKey: String = "",
        @ColumnInfo(name = "jid") var jid: String = "",
        @ColumnInfo(name = "name") var name: String = "",
        @ColumnInfo(name = "description") var description: String = "",
        @ColumnInfo(name = "owners") var owners: ParticipantsSet = ParticipantsSet(),
        @ColumnInfo(name = "members") var members: ParticipantsSet = ParticipantsSet(),
        @ColumnInfo(name = "invited") var invited: ParticipantsSet = ParticipantsSet(),
        @ColumnInfo(name = "password") var password: ChatRoomPassword = ChatRoomPassword(""),
        @ColumnInfo(name = "color") var color: Int = -1,
        @ColumnInfo(name = "accountId") var accountId: String = "",
        @ColumnInfo(name = "publicRoom") var publicRoom: Boolean = false,
        @ColumnInfo(name = "modTime") var modTime: Long = 0,
        @ColumnInfo(name = "creationTime") var creationTime: Long = 0,
        @ColumnInfo(name = "state") var state: ChatRoomState = ChatRoomState.ACTIVE,
        @ColumnInfo(name = "joinTime") var joinTime: Long = 0,
        @ColumnInfo(name = "topic") var topic: String = "",
        @ColumnInfo(name = "notificationsLevel") var notificationsLevel : ChatRoomNotificationLevel = ChatRoomNotificationLevel.ALL,
        @ColumnInfo(name = "isOpen") var open: Boolean = true,
        @ColumnInfo(name = "isModerated") var moderated : Boolean = false,
        @ColumnInfo(name = "isPersistent") var persistent : Boolean = true,
        @ColumnInfo(name = "roomCreator") var roomCreator : String = "",
        @ColumnInfo(name = "anonymousMode") var anonymousMode : ChatRoomAnonymousMode = ChatRoomAnonymousMode.NON_ANONYMOUS

) {
    @Ignore
    var lastViewedMessage = 0L

    fun isUserOwner(): Boolean {
        return owners.containsRemoteAddress(accountId)
    }

    fun getJoinTimeInSec() = joinTime / 1000

    fun isCollab() = this.type == ImConversationData.GROUPCHAT_COLLAB.toLong()

    fun isOffline() = this.state == ChatRoomState.REMOTELY_DELETED || this.state == ChatRoomState.OFFLINE

    companion object {
        fun getInitials(chatRoomName: String): String {
            val words = chatRoomName
                .toUpperCase()
                .trim()
                .split(" ")
                .filter { it.isNotBlank() }

            return when (words.size) {
                0 -> ""
                1 -> TextBoundaries.findGraphemesAndFlags(words.first()).first()
                else -> {
                    val firstWord = words.first();
                    val secondWord = words.drop(1).first();

                    val initial1 = TextBoundaries.findGraphemesAndFlags(firstWord).firstOrNull()
                    val initial2 = TextBoundaries.findGraphemesAndFlags(secondWord).firstOrNull()

                    return initial1 + initial2
                }
            }
        }
    }

    val initials
        get() = getInitials(name)
}

data class ChatRoomPassword(val plainPassword: String)

class ChatRoomTypeConverters {

    @TypeConverter
    fun toEncryptedPassword(plain: ChatRoomPassword): String {
        if (plain.plainPassword.isEmpty()) return ""
        return SettingsStringEncoder_v1.encrypt(plain.plainPassword ?: "") ?: ""
    }

    @TypeConverter
    fun toPlainPassword(encrypted: String): ChatRoomPassword {
        if (encrypted.isEmpty()) return ChatRoomPassword("")
        return ChatRoomPassword(SettingsStringEncoder_v1.decrypt(encrypted) ?: "")
    }

    @TypeConverter
    fun toParticipantsSet(participants: String): ParticipantsSet {
        return ParticipantsSet(participants)
    }

    @TypeConverter
    fun toParticipantsString(participantsSet: ParticipantsSet): String {
        return participantsSet.serializedParticipantKeys
    }

    @TypeConverter
    fun toChatRoomState(string: String): ChatRoomState = ChatRoomState.valueOf(string)

    @TypeConverter
    fun toStateString(state: ChatRoomState): String = state.name

    @TypeConverter
    fun toChatRoomNotificationsLevel(string: String): ChatRoomNotificationLevel = ChatRoomNotificationLevel.valueOf(string)

    @TypeConverter
    fun toChatRoomNotificationLevelString(level: ChatRoomNotificationLevel): String = level.name

    @TypeConverter
    fun toChatRoomAnonymousMode(string: String): ChatRoomAnonymousMode = ChatRoomAnonymousMode.valueOf(string)

    @TypeConverter
    fun toChatRoomAnonymousModeString(mode: ChatRoomAnonymousMode): String = mode.name

}

enum class ChatRoomState {
    ACTIVE, LEFT, REMOTELY_DELETED, OFFLINE
}

enum class ChatRoomNotificationLevel {
    ALL, MENTIONS, MUTE,
}

enum class ChatRoomAnonymousMode(val value: String , val intValue : Int) {
    NON_ANONYMOUS("nonanonymous", 0),
    SEMI_ANONYMOUS("semianonymous", 1),
    FULL_ANONYMOUS("fullanonymous", 2)


}


