package com.bria.common.controller.phone

import android.content.Context
import com.bria.common.controller.Controllers
import com.bria.common.controller.accounts.Account
import com.bria.common.controller.calllog.db.CallLogEntity
import com.bria.common.controller.calllog.db.CallType
import com.bria.common.controller.settings.EAccountSetting
import com.bria.common.controller.settings.ESetting
import com.bria.common.sdkwrapper.CallData
import com.bria.common.sdkwrapper.CallManager
import com.bria.common.util.Log
import com.bria.common.util.RecordingUtils
import com.bria.common.util.SipAddressUtils
import com.bria.common.util.Utils
import com.bria.common.util.sip.SipUri
import java.util.*

class CallLogHelper(private val context: Context, private val controllers: Controllers) {
    companion object {
        private const val TAG = "CallLogHelper"
    }

    fun generateCallLogEntity(callData: CallData): CallLogEntity? {
        if (callData.isSuppressLog) {
            Log.d(TAG, "Call Log suppressed")
            return null
        }

        val account = getAccount(callData)

        return CallLogEntity(
                number = getNumber(callData),
                callType = getCallType(callData),
                date = callData.callStartTimestamp,
                durationSeconds = getDurationSeconds(callData),
                provisioningUser = getProvisioningUser(),
                accountId = account?.id ?: -1,
                accountUsername = account?.getStr(EAccountSetting.UserName).orEmpty(),
                accountDomain = account?.getStr(EAccountSetting.Domain).orEmpty(),
                remoteDomain = getRemoteDomain(callData),
                remoteName = getRemoteDisplayName(callData),
                recordFileName = getRecordFileName(callData),
                forwardedTo = callData.forwardTo.orEmpty(),
                hostedCollab = callData.isHostedCollab,
                unread = true
        )
    }

    fun generateMissedCallLogEntity(account: Account, missedCallFromUri: String?): CallLogEntity {
        val fromSipUri = SipUri.parseSipContact(missedCallFromUri)

        return CallLogEntity(
                number = fromSipUri.userName,
                callType = CallType.MISSED,
                date = System.currentTimeMillis(),
                durationSeconds = 0,
                provisioningUser = getProvisioningUser(),
                accountId = account.id,
                accountUsername = account.getStr(EAccountSetting.UserName) ?: "",
                accountDomain = account.getStr(EAccountSetting.Domain) ?: "",
                remoteDomain = fromSipUri.domain,
                remoteName = fromSipUri.displayName,
                recordFileName = "",
                forwardedTo = "",
                hostedCollab = false,
                unread = true
        )
    }

    // generate 1000 records with the same numbers to show it as one group with 1000 sub-items
    fun generateCallLogEntity_test_1group1000(callData: CallData): List<CallLogEntity> {
        val list = mutableListOf<CallLogEntity>()
        repeat(1000) {i ->
            generateCallLogEntity(callData)?.let { list += it.apply { date -= i * 1000 } }
        }
        return list
    }

    // generate 1000 records with the unique numbers to show it as 1000 groups with 1 sub-item
    fun generateCallLogEntity_test_1000group1(callData: CallData): List<CallLogEntity> {
        val list = mutableListOf<CallLogEntity>()
        repeat(1000) {i ->
            generateCallLogEntity(callData)?.let {
                list += it.apply {
                    date -= i * 1000
                    if (i % 2 == 0) {
                        number = "${i}_$number"
                    }
                }
            }
        }
        return list
    }

    // generate 1000 records with the unique numbers to show it as 1000 groups with 1 sub-item, for each second entry the call type set as missed
    fun generateCallLogEntity_test_1000group1_missed(callData: CallData): List<CallLogEntity> {
        val list = mutableListOf<CallLogEntity>()
        repeat(1000) {i ->
            generateCallLogEntity(callData)?.let {
                list += it.apply {
                    date -= i * 1000
                    if (i % 2 == 0) {
                        number = "${i}_$number"
                        callType = CallType.MISSED
                    }
                }
            }
        }
        return list
    }

    private fun isVccsCall(callData: CallData) = controllers.collaboration.monitoredConference != null &&
            (callData.isHostedCollab || !callData.collabUrl.isNullOrEmpty())

    private fun isVoiceMail(callData: CallData) = callData.isVoiceMail(context) // todo: replace this deprecated method

    private fun getDurationSeconds(callData: CallData): Int {
        return if (callData.callAnswerTimestamp > 0 && callData.callDisconnectTimestamp > callData.callAnswerTimestamp) {
            (callData.callDisconnectTimestamp - callData.callAnswerTimestamp) / 1000
        } else {
            0
        }.toInt()
    }

    private fun getCallType(callData: CallData): CallType {
        if (callData.direction == CallData.DIRECTION_OUTGOING) {
            return when {
                isVccsCall(callData) -> CallType.CONFERENCE
                isVoiceMail(callData) -> CallType.VOICE_MAIL
                else -> CallType.OUTGOING
            }
        } else {
            val duration = getDurationSeconds(callData)
            var callType = CallType.INCOMING

            if (duration <= 0 && !callData.callRedirected) {
                callType = CallType.MISSED
            }

            when (callData.statusCode) {
                CallManager.STATUS_CODE_BUSY_HERE -> {
                    if (callData.callCancelled) {
                        callType = CallType.INCOMING
                    }
                }

                CallManager.STATUS_CODE_REQUEST_TERMINATED or
                CallManager.STATUS_CODE_MOVED_TEMPORARILY -> {
                    if (duration <= 0) {
                        callType = CallType.MISSED
                    }
                }
            }

            if (callData.isSuppressMissedCall) {
                callType = CallType.INCOMING
            }

            return callType
        }
    }

    private fun getRemoteExtension(callData: CallData): String {
        if (isVccsCall(callData)) {
            return callData.collabUrl ?: ""
        }

        if (controllers.settings.getBool(ESetting.FeatureGenband) && controllers.settings.getBool(ESetting.ShowUriDomain)) {
            return callData.remoteUserWithDomain
        }

        // If it is an outgoing call, use the string without dial plan settings applied
        // else use the callData to get the remote extension
        if (callData.direction == CallData.DIRECTION_OUTGOING) {
            return callData.remoteUserNoDialPlanApplied
        }

        if (callData.remoteUri.contains("@")) {
            // Remote uri contains username i.e. CLID enabled
            val tokenizer = StringTokenizer(callData.remoteUri, "@")
            tokenizer.nextToken()
            var remoteDomain = tokenizer.nextToken()

            if (remoteDomain.endsWith(">")) {
                remoteDomain = remoteDomain.substring(0, remoteDomain.length - 1)
            }

            val account = controllers.accounts.primaryAccount
            if (account != null && !account.getStr(EAccountSetting.Domain).equals(SipAddressUtils.removePort(remoteDomain), ignoreCase = true)) {
                if (!Utils.Brands.isPhytterBrand(context)) {
                    return SipAddressUtils.getAddress(callData.remoteUser, remoteDomain)
                }
            }
        }

        return callData.remoteUser
    }

    private fun getRemoteDisplayName(callData: CallData) =
            controllers.collaboration.monitoredConference?.details?.title ?: callData.remoteSipDisplayName.orEmpty()

    private fun getNumber(callData: CallData): String {
        if (isVccsCall(callData)) {
            return callData.collabUrl.orEmpty()
        }

        // If it is an outgoing call, use the string without dial plan settings applied
        val number = if (callData.direction == CallData.DIRECTION_OUTGOING) {
            callData.remoteUserNoDialPlanApplied.orEmpty()
        } else {
            callData.remoteUser.orEmpty()
        }

        return SipAddressUtils.removeDomain(number)
    }

    private fun getRemoteDomain(callData: CallData) =
            callData.remoteDomain?.takeIf { it.isNotEmpty() } ?: SipAddressUtils.getDomain(callData.remoteUri.orEmpty().replace(">", ""))

    private fun getAccount(callData: CallData): Account? = controllers.accounts.getAccount(callData.accountNickname)

    private fun getProvisioningUser(): String {
        return when {
            controllers.settings.getBool(ESetting.FeatureProvisioning) -> controllers.settings.getStr(ESetting.ProvisioningUsername)
            else -> ""
        }
    }

    private fun getRecordFileName(callData: CallData): String {
        if (controllers.phone.callRecordingEnabled()) {
            // auto-stop call recording (end of call)
            val callId = callData.callId
            var callRecordingStopped = false
            if (controllers.base.sipStackManager.callManager.isPartOfActiveCallRecording(callId)) {
                callRecordingStopped = controllers.base.sipStackManager.callManager.stopCallRecording(callId)
            }
            // always check if a temporary .wav file exists, matching this callId
            if (callRecordingStopped) {
                val lastId = controllers.callLog.api.getAll().blockingFirst(mutableListOf())
                        .maxBy { it.id }?.id?.toInt() ?: 0
                val recordFileName = RecordingUtils.getFileNameFromCallLogId(lastId + 1)
                if (RecordingUtils.renameTempFile(context, callId, recordFileName)) {
                    // successful rename (it means that we have recorded this call)
                    return recordFileName
                }
            }
        }
        return ""
    }
}
