package com.bria.common.modules

import android.annotation.SuppressLint
import android.app.Application
import android.content.ContentResolver
import android.os.Build
import android.support.v4.app.NotificationManagerCompat
import com.bria.common.ProcessLifecycleTracker
import com.bria.common.controller.ClientConfig
import com.bria.common.controller.ControllerObservable
import com.bria.common.controller.accounts.IAccountsCtrlActions
import com.bria.common.controller.accounts.IAccountsCtrlObserver
import com.bria.common.controller.commlog.ICommLogCtrlEvents
import com.bria.common.controller.commlog.ICommLogCtrlObserver
import com.bria.common.controller.contacts.buddy.IBuddyCtrlEvents
import com.bria.common.controller.contacts.buddy.IBuddyCtrlObserver
import com.bria.common.controller.contacts.ldap.LdapModule
import com.bria.common.controller.contacts.local.IContactsCtrlEvents
import com.bria.common.controller.coordinatedevents.CoordinatedEventAggregator
import com.bria.common.controller.coordinatedevents.ICoordinatedEventAggregatorListenerSide
import com.bria.common.controller.coordinatedevents.ICoordinatedEventAggregatorPublishSide
import com.bria.common.controller.im.IImCtrlEvents
import com.bria.common.controller.im.refactoring.DraftInstantMessages
import com.bria.common.controller.im.refactoring.adapter.ImListAdapterMentionsHelper
import com.bria.common.controller.im.refactoring.chatroom.ChatRoomApi
import com.bria.common.controller.im.refactoring.chatroom.ChatRoomMemberCache
import com.bria.common.controller.im.refactoring.roomdb.ChatRoomRepo
import com.bria.common.controller.license.ILicenseCtrlActions
import com.bria.common.controller.license.LicenseUtil
import com.bria.common.controller.phone.GoodCallQualityProvenObservable
import com.bria.common.controller.phone.IPhoneCtrlEvents
import com.bria.common.controller.presence.IPresenceCtrlEvents
import com.bria.common.controller.presence.PresenceDataProvider
import com.bria.common.controller.presence.PresenceStatuses
import com.bria.common.controller.presence.refactoring.EPresenceStatus
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.core.Settings
import com.bria.common.crashlytics.CrashlyticsImpl
import com.bria.common.crashlytics.CrashlyticsStub
import com.bria.common.crashlytics.ICrashlytics
import com.bria.common.modules.android.*
import com.bria.common.modules.store.GoodReviewDriver
import com.bria.common.network.ConnectivityObservable
import com.bria.common.network.NetworkModule
import com.bria.common.notification.NotificationsHandler
import com.bria.common.permission.PermissionChecker
import com.bria.common.teams.TeamsModule
import com.bria.common.uiframework.activities.Orion
import com.bria.common.uiframework.helpers.BriaShortcutManager
import com.bria.common.uireusable.dataprovider.ISimpleDataProvider
import com.bria.common.util.CpcCallingSchemesPatternFactory
import com.bria.common.util.CrashInDebug
import com.bria.common.util.IObservable
import com.bria.common.util.Utils
import com.bria.common.util.phone.SingleTouchToCallHelper
import java.util.*
import java.util.regex.Pattern

@SuppressLint("NewApi")
/*
* Why not Koin?
* Due to KoinContext.resolveInstance method being synchronized, creation of services is serialized.
* At this point in time this would work against us since we initialize on several threads during startup.
*
* Why not Dagger?
* Due to code generation, regular compile errors are obfuscated. Also, the build is prolonged.
*
* Use fun to express transient objects and val by lazy to define singletons.
* */
object BriaGraph {
    private val LOG_TAG = "BriaGraph"

    lateinit var application: Application

    fun initialize(application: Application) {
        this.application = application
    }

    private val stack: Deque<() -> Unit> = ArrayDeque()

    private fun <T> T.registerForShutdown(callback: (obj: T) -> Unit): T {

        val noArgCallback: () -> Unit = {
            callback(this)
        }

        stack.push(noArgCallback)

        return this
    }

    fun shutdown() {
        while (stack.isNotEmpty()) {
            val callback = stack.pop()
            try {
                callback()
            } catch (e: Throwable) {
                // shutting down, don't block the process
                CrashInDebug.with(LOG_TAG, e)
            }
        }
    }

    val settings: Settings by lazy {
        Settings.get(application)
    }

    val androidSoundInfo: AndroidSoundInfo by lazy {
        AndroidSoundInfo(application)
    }

    val screenOnOffReceiver: ScreenOnOffReceiver by lazy {
        ScreenOnOffReceiver()
    }

    val deviceInteractivityObservable: IObservable<IDeviceInteractivityChanged> by lazy {
        screenOnOffReceiver.deviceInteractivityObservable
    }

    val appLifetimeReceivers: AppLifetimeReceivers by lazy {
        AppLifetimeReceivers(application,
                BackgroundForegroundScreenOffReceiver(),
                screenOnOffReceiver)
    }

    val processLifecycleTracker: ProcessLifecycleTracker by lazy {
        ProcessLifecycleTracker(application)
    }

    val livingActivitiesTracker: LivingActivitiesTracker by lazy {
        LivingActivitiesTracker(application)
    }

    val networkModule: NetworkModule by lazy {
        NetworkModule(application, settings, ConnectivityObservable.getObservable())
                .registerForShutdown { it.destroy() }
    }

    val shortcutManager: BriaShortcutManager? by lazy {
        if (android.os.Build.VERSION.SDK_INT >= Build.VERSION_CODES.O && Settings.get(application).getBool(ESetting.FeatureShortcutManager)) {
            return@lazy BriaShortcutManager(application, callLogObservable, callLogActions, settings, contactCtrl.contactsApi).registerForShutdown { it.destroy() }
        } else return@lazy null
    }

    val orion: Orion by lazy {
        Orion(application)
                .registerForShutdown { it.destroy() }
    }

    val goodReviewDriver: GoodReviewDriver by lazy {
        GoodReviewDriver(
                settings,
                Utils.Build.getFullVersion(application),
                GoodCallQualityProvenObservable.getObservable(),
                ClientConfig.get().isDebugMode(),
                LicenseUtil.getAppBaseLicenseType(application))
                .registerForShutdown { it.shutdown() }
    }

    val cpcCallingSchemesPattern: Pattern by lazy {
        CpcCallingSchemesPatternFactory.compileSchemesPattern(application)
    }

    val notificationManager: NotificationManagerCompat by lazy {
        NotificationManagerCompat.from(application)
    }

    val notificationsHandler: NotificationsHandler by lazy {
        NotificationsHandler(application, ControllerObservable.getObservable(), notificationManager)
                .registerForShutdown { it.destroy() }
    }

    val presenceStatuses: PresenceStatuses by lazy {
        PresenceStatuses(settings, accountsController, accountsObservable)
                .registerForShutdown { it.shutdown() }
    }

    val presenceDataProvider: ISimpleDataProvider<EPresenceStatus> by lazy {
        PresenceDataProvider(presenceStatuses)
    }

    val accountsController: IAccountsCtrlActions by lazy {
        ControllerObservable.getObservable().blockingFirst().accountsCtrl.events
    }

    val accountsObservable: IObservable<IAccountsCtrlObserver> by lazy {
        ControllerObservable.getObservable().blockingFirst().accountsCtrl.observable
    }

    val licenseController: ILicenseCtrlActions by lazy {
        ControllerObservable.getObservable().blockingFirst().licenseCtrl.events
    }

    val presenceController: IPresenceCtrlEvents by lazy {
        ControllerObservable.getObservable().blockingFirst().presenceCtrl.events
    }

    val callLogActions: ICommLogCtrlEvents by lazy {
        ControllerObservable.getObservable().blockingFirst().localCommLogCtrl.events
    }

    val contactCtrl: IContactsCtrlEvents by lazy {
        ControllerObservable.getObservable().blockingFirst().contactsCtrl.events
    }

    val buddyCtrl: IBuddyCtrlEvents by lazy {
        ControllerObservable.getObservable().blockingFirst().buddyCtrl.events
    }

    val buddyObservable: IObservable<IBuddyCtrlObserver> by lazy {
        ControllerObservable.getObservable().blockingFirst().buddyCtrl.observable
    }

    val imCtrl: IImCtrlEvents by lazy {
        ControllerObservable.getObservable().blockingFirst().imCtrl.events
    }

    val callLogObservable: IObservable<ICommLogCtrlObserver> by lazy {
        if (settings.getBool(ESetting.BroadWorksEnabled) && settings.getBool(ESetting.FeatureBroadWorks)) {
            ControllerObservable.getObservable().blockingFirst().getBroadWorksCommLogCtrl().getObservable();
        } else {
            ControllerObservable.getObservable().blockingFirst().localCommLogCtrl.observable
        }
    }

    val phoneCtrl: IPhoneCtrlEvents by lazy {
        ControllerObservable.getObservable().blockingFirst().phoneCtrl.events
    }

    val crashlytics: ICrashlytics by lazy {
        // factory
        if (settings.getBool(ESetting.FeatureCrashlytics)) {
            if (!ClientConfig.get().isDebugMode) {
                return@lazy CrashlyticsImpl(application)
            }
        }

        return@lazy CrashlyticsStub()
    }

    val teams: TeamsModule by lazy {
        TeamsModule(accountsController)
    }

    val permissionChecker: PermissionChecker by lazy {
        PermissionChecker(application)
    }

    private val coordinatedEventAggregator: CoordinatedEventAggregator by lazy {
        CoordinatedEventAggregator()
    }

    val coordinatedEventAggregatorPublishSide: ICoordinatedEventAggregatorPublishSide
        get() = coordinatedEventAggregator

    val coordinatedEventAggregatorListenerSide: ICoordinatedEventAggregatorListenerSide
        get() = coordinatedEventAggregator

    val contentResolver: ContentResolver by lazy {
        application.contentResolver
    }
    val ldapModule: LdapModule? by lazy {
        LdapModule(application, networkModule)
                .registerForShutdown { it.destroy() }
    }

    val singleTouchToCallHelper: SingleTouchToCallHelper by lazy {
        SingleTouchToCallHelper(settings, accountsController)
    }

    val imListAdapterMentionsHelper: ImListAdapterMentionsHelper by lazy {
        ImListAdapterMentionsHelper(buddyCtrl, buddyObservable)
    }

    val draftInstantMessages: DraftInstantMessages by lazy {
        DraftInstantMessages()
    }

    val chatRoomRepo: ChatRoomRepo by lazy {
        imCtrl.imData.chatRoomRepo
    }

    val chatRoomApi: ChatRoomApi by lazy {
        imCtrl.chatApi.chatRoomApi
    }

    val chatRoomMemberCache: ChatRoomMemberCache by lazy {
        ChatRoomMemberCache()
    }


    // testing

    class A {
        var shutdownCallback: (() -> Unit)? = null

        fun shutdown() {
            shutdownCallback?.invoke()
        }
    }

    class BDependsOnA(val a: A) {
        var shutdownCallback: (() -> Unit)? = null

        fun shutdown() {
            shutdownCallback?.invoke()
        }
    }

    val a: A by lazy {
        A().registerForShutdown { it.shutdown() }
    }

    val bDependsOnA: BDependsOnA by lazy {
        BDependsOnA(a).registerForShutdown { it.shutdown() }
    }
}
