package com.bria.common.notification

import android.app.Notification
import android.app.NotificationChannel
import android.app.NotificationManager
import android.app.PendingIntent
import android.content.Context
import android.content.Intent
import android.support.v4.app.NotificationCompat
import android.support.v4.app.NotificationManagerCompat
import com.bria.common.R
import com.bria.common.controller.Controller
import com.bria.common.controller.commlog.CallLogObserverAdapter
import com.bria.common.controller.commlog.local.LocalCallLogController
import com.bria.common.controller.notifications.NotificationParams
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.core.Settings
import com.bria.common.modules.ModuleClassFinder
import com.bria.common.uiframework.coloring.Coloring
import com.bria.common.util.GlobalConstants
import com.bria.common.util.Log

class NotificationCallReminder(val mAppContext: Context, val mController: Controller, val mNotificationManager: NotificationManagerCompat) :
        AbstractNotification() {

    private val TAG = "NotificationCallReminder"

    private var mCallLogAdapterObserver: CallLogObserverAdapter

    init {
        mCallLogAdapterObserver = object : CallLogObserverAdapter() {
            override fun onCallReminder(name: String, date: String, content: String) {
                super.onCallReminder(name, date, content)
                updateCallReminderNotification(name, date, content)
            }
        }

        mController.localCommLogCtrl.observable.attachWeakObserver(mCallLogAdapterObserver)

        createNotificationChannel()
    }

    override fun createNotificationBuilder(notificationParams: NotificationParams): NotificationCompat.Builder {
        var builder = NotificationHelper.createNotificationBuilder(mAppContext, mController, notificationParams)
        val flags = Intent.FLAG_ACTIVITY_NEW_TASK or Intent.FLAG_ACTIVITY_CLEAR_TOP or Intent.FLAG_ACTIVITY_REORDER_TO_FRONT
        var notificationIntent = Intent(mAppContext, ModuleClassFinder.instance.mainActivityClass)
        notificationIntent.setAction(GlobalConstants.INTENT_ACTION_DECLINE_REMINDER)
        val dateName = notificationParams.data as String
        val arrayDateName = dateName.split("\\|\\|".toRegex()).dropLastWhile { it.isEmpty() }.toTypedArray()

        val date = arrayDateName[0]
        val name = arrayDateName[1]
        notificationIntent.putExtra(LocalCallLogController.KEY_CALL_REMINDER_NAME, name)
        notificationIntent.putExtra(LocalCallLogController.KEY_CALL_REMINDER_DATE, date)
        notificationIntent.setFlags(flags)
        val color = Coloring.get().decodeColor(Settings.get(mAppContext).getStr(ESetting.ColorBrandTint))
        builder.color = color
        builder.setColorized(true)
        val contentIntent = PendingIntent.getActivity(mAppContext, notificationParams.notificationID, notificationIntent,
                PendingIntent.FLAG_CANCEL_CURRENT)
        builder.setContentIntent(contentIntent)
        builder.setCategory(NotificationCompat.CATEGORY_REMINDER)
        return builder
    }

    override fun addActionButtons(builder: NotificationCompat.Builder?, notificationParams: NotificationParams) {
    }

    override fun createNotification(builder: NotificationCompat.Builder, notificationParams: NotificationParams) {
        val notification = builder.build()
        notification.flags = notification.flags or notificationParams.flags
        NotificationHelper.postNotification(notificationParams.notificationID, notification, mNotificationManager)
        NotificationHelper.addToActiveMap(notificationParams)
    }

    override fun createNotificationChannel() {
        val notificationManager = mAppContext.getSystemService(Context.NOTIFICATION_SERVICE) as NotificationManager
        if (android.os.Build.VERSION.SDK_INT >= android.os.Build.VERSION_CODES.O) {
            val remindersChannelName = mAppContext.getString(R.string.tNotificationChannelRemindersTitle)
            val remindersChannel = NotificationChannel(NotificationHelper.REMINDERS_CHANNEL_ID, remindersChannelName, NotificationManager.IMPORTANCE_HIGH)
            remindersChannel.enableVibration(true)
            remindersChannel.enableLights(true)
            notificationManager.createNotificationChannel(remindersChannel)
        }
    }

    override fun setupNotification() {
    }

    override fun destroy() {
        mController.localCommLogCtrl.observable.detachObserver(mCallLogAdapterObserver)
    }

    override fun handleAction(intent: Intent?) {
    }

    fun updateCallReminderNotification(name: String, date: String, content: String) {
        val notificationID = Math.abs(name.hashCode() * date.hashCode()) % 1023 // TODO what here?
        Log.d(TAG, "Call reminder notification update requested. Name = $name, date = $date, content = $content, id = $notificationID")

        val contentTitle = mAppContext.getString(R.string.tCallReminder)
        val soundVibration = true
        val notificationType = NotificationParams.ENotificationType.CallReminder
        val priority = Notification.PRIORITY_DEFAULT
        val whenTime = System.currentTimeMillis()

        val notificationParams = NotificationHelper.updateNotificationParams(notificationID, R.drawable.ic_snooze, null, contentTitle, content, null, -1,
                GlobalConstants.INTENT_ACTION_DECLINE_REMINDER, notificationType, null, soundVibration, priority, whenTime)
        notificationParams.data = "$date||$name"

        val builder = createNotificationBuilder(notificationParams)

        createNotification(builder, notificationParams)
    }
}
