package com.bria.common.notification

import android.app.NotificationChannel
import android.app.NotificationManager
import android.content.Context
import android.content.Intent
import android.support.v4.app.NotificationCompat
import android.support.v4.app.NotificationManagerCompat
import com.bria.common.R
import com.bria.common.controller.Controller
import com.bria.common.controller.im.refactoring.chatroom.LoggedInUserMentionDetector
import com.bria.common.controller.im.refactoring.roomdb.entities.Message
import com.bria.common.controller.notifications.NotificationParams
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.core.Settings
import com.bria.common.util.GlobalConstants
import com.bria.common.util.Log
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.schedulers.Schedulers
import java.util.concurrent.TimeUnit

class NotificationMention(val mAppContext: Context, val mController: Controller, val mNotificationManager: NotificationManagerCompat) : AbstractNotification() {
    private val mCompositeDisposable = CompositeDisposable()
    var mImData = mController.getImCtrl().getEvents().getImData()

    private val TAG = "NotificationMention"

    init {
        listenImDataObservables()
    }

    private fun listenImDataObservables() {
        mCompositeDisposable.clear()

        val dataChangedRoomDisposable = mImData.getChatRoomRepo().getDataChangedObservable()
                .sample(500, TimeUnit.MILLISECONDS, Schedulers.computation())
                .observeOn(Schedulers.io())
                .subscribeOn(Schedulers.io())
                .subscribe({ _ ->
                    if ((Settings.get(mAppContext).getBool(ESetting.FeaturePersistentChatRooms)
                                    && Settings.get(mAppContext).getBool(ESetting.PersistentChatRoomsEnabled))
                                    && Settings.get(mAppContext).getBool(ESetting.ImPresence)) {
                        val unreadMessages = mImData.chatRoomRepo.getUnreadMessagesAllChatRooms()
                        val unreadMentions : MutableList<Message> = mutableListOf()
                        for (im in unreadMessages.blockingGet()) {
                            mImData.chatRoomRepo.getChatRoom(im.chatId)?.let {
                                if (LoggedInUserMentionDetector.isLoggedInUserMentioned(im, it)) {
                                    unreadMentions.add(im)
                                }
                            }
                            updateUnreadMentions(unreadMentions)
                        }
                    }
                }, { throwable -> Log.e(TAG, "Throwable " + throwable.message) })
        mCompositeDisposable.add(dataChangedRoomDisposable)
    }

    private fun updateUnreadMentions(unreadMentions: MutableList<Message>) {
        for (unreadMention in unreadMentions) {
            val notificationID = NotificationHelper.generateMentionNotificationId(unreadMention.id)

            val contentTitle = mAppContext.getString(R.string.tNotificationMention)
            val soundVibration = true
            val notificationType = NotificationParams.ENotificationType.Mention
            val priority = NotificationManagerCompat.IMPORTANCE_DEFAULT
            val whenTime = unreadMention.creationTime

            val content = mAppContext.getString(R.string.tNotificationMentionContent) + mImData.chatRoomRepo.getChatRoom(unreadMention.chatId)?.name

            val notificationParams = NotificationHelper.updateNotificationParams(notificationID, R.drawable.ic_stat_notify_im, null, contentTitle, content,
                    null, -1, GlobalConstants.INTENT_ACTION_DECLINE_REMINDER, notificationType, null, soundVibration, priority, whenTime)

            val builder = createNotificationBuilder(notificationParams)

            createNotification(builder, notificationParams)
        }
    }

    override fun createNotificationBuilder(notificationParams: NotificationParams): NotificationCompat.Builder {
        return NotificationHelper.createNotificationBuilder(mAppContext, mController, notificationParams)
    }

    override fun addActionButtons(builder: NotificationCompat.Builder?, notificationParams: NotificationParams) {
    }

    override fun createNotification(builder: NotificationCompat.Builder, notificationParams: NotificationParams) {
        val notification = builder.build()
        notification.flags = notification.flags or notificationParams.flags
        NotificationHelper.postNotification(notificationParams.notificationID, notification, mNotificationManager)
        NotificationHelper.addToActiveMap(notificationParams)
    }

    override fun createNotificationChannel() {
        val notificationManager = mAppContext.getSystemService(Context.NOTIFICATION_SERVICE) as NotificationManager
        if (android.os.Build.VERSION.SDK_INT >= android.os.Build.VERSION_CODES.O) {
            val mentionsChannelName = mAppContext.getString(R.string.tNotificationChannelMentionsTitle)
            val mentionsChannel = NotificationChannel(NotificationHelper.MENTION_CHANNEL_ID, mentionsChannelName, NotificationManager.IMPORTANCE_HIGH)
            mentionsChannel.enableVibration(true)
            mentionsChannel.enableLights(true)
            notificationManager.createNotificationChannel(mentionsChannel)
        }
    }

    override fun setupNotification() {
    }

    override fun destroy() {
        mCompositeDisposable.dispose()
    }

    override fun handleAction(intent: Intent?) {
    }
}
