package com.bria.common.notification

import android.app.NotificationChannel
import android.app.NotificationManager
import android.app.PendingIntent
import android.app.Service
import android.content.Context
import android.content.Intent
import android.os.Build
import android.support.v4.app.NotificationCompat
import android.support.v4.app.NotificationManagerCompat
import com.bria.common.R
import com.bria.common.controller.IController
import com.bria.common.controller.accounts.*
import com.bria.common.controller.notifications.NotificationParams
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.branding.EAccountType
import com.bria.common.controller.settings.core.ISettingsObserver
import com.bria.common.controller.settings.core.Settings
import com.bria.common.modules.BriaGraph
import com.bria.common.modules.ModuleClassFinder
import com.bria.common.network.INetworkObserver
import com.bria.common.util.GlobalConstants
import com.bria.common.util.Log
import java.lang.ref.WeakReference

class NotificationStatus(val mAppContext: Context, val mNotificationManager: NotificationManagerCompat) : AbstractNotification(), IAccountsCtrlObserver {
    override fun onServiceConnectionStatusChanged(connected: Boolean) {

    }

    private val TAG = "NotificationStatus"
    private val mSettings: Settings = Settings.get(mAppContext)

    private var mSettingsObserver: ISettingsObserver
    private val SERVICE_NOTIFICATION_ID = "SERVICE_NOTIFICATION_ID".hashCode()
    private lateinit var mServiceInstance: WeakReference<Service>

    private var mNotificationsStatusEnabled: Boolean = false

    private lateinit var mController: IController
    private lateinit var mAccountsC: IAccountsCtrlActions

    init {
        mNotificationsStatusEnabled = mSettings.getBool(ESetting.NotificationsStatusEnabled)
        createNotificationChannel()

        mSettingsObserver = object : ISettingsObserver {
            override fun onBeforeOwnerChange(newOwner: String?) {}

            override fun onOwnerChanged() {}

            override fun onSettingsChanged(changedSettings: MutableSet<ESetting>) {
                if (changedSettings.contains(ESetting.NotificationsStatusEnabled)) {
                    mNotificationsStatusEnabled = mSettings.getBool(ESetting.NotificationsStatusEnabled)
                    if (mNotificationsStatusEnabled) {
                        updateAccountNotification()
                    } else {
                        val service = getService()
                        if (service != null && Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
                            service.stopForeground(true)
                        } else {
                            mNotificationManager.cancel(SERVICE_NOTIFICATION_ID)
                        }
                    }
                }
            }
        }

        BriaGraph.settings.attachWeakObserver(mSettingsObserver, setOf(ESetting.NotificationsRemoteDebugEnabled))


    }

    fun setController(controller: IController) {
        mController = controller
        mAccountsC = mController.accountsCtrl.events
        mController.accountsCtrl.observable.attachWeakObserver(this)
    }

    override fun createNotificationBuilder(notificationParams: NotificationParams): NotificationCompat.Builder {
        val builder = NotificationCompat.Builder(mAppContext, notificationParams.channelId)
        // BRIAN-5469. This type of notification, we should update existing one instead of creating new, in order to keep service's foreground status

        // Always goes through splash activity. It needed to properly handle permission requests delayed/skipped by the User on the first run.
        val notificationIntent = Intent(mAppContext, ModuleClassFinder.instance.mainActivityClass)
        notificationIntent.action = notificationParams.intentAction

        val contentIntent = PendingIntent.getActivity(mAppContext, 0, notificationIntent, PendingIntent.FLAG_CANCEL_CURRENT)

        builder.setContentIntent(contentIntent)
        builder.setSmallIcon(notificationParams.iconResId)
        builder.setTicker(notificationParams.tickerText)
        builder.setWhen(notificationParams.now)
        builder.setContentTitle(notificationParams.contentTitle)
        builder.setContentText(notificationParams.contentText)
        builder.setPriority(notificationParams.priority)

        val ongoingEventFlag = notificationParams.flags and (NotificationCompat.FLAG_NO_CLEAR or NotificationCompat.FLAG_ONGOING_EVENT)
        builder.setOngoing(ongoingEventFlag > 0)

        builder.setCategory(NotificationCompat.CATEGORY_STATUS)

        return builder
    }

    override fun addActionButtons(builder: NotificationCompat.Builder?, notificationParams: NotificationParams) {
    }

    override fun createNotification(builder: NotificationCompat.Builder, notificationParams: NotificationParams) {
        try {
            val notification = builder.build()
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O) {
                val service = getService()
                service?.startForeground(notificationParams.notificationID, notification)
            } else {
                NotificationHelper.postNotification(notificationParams.notificationID, notification, mNotificationManager)
            }
            NotificationHelper.addToActiveMap(notificationParams)
        } catch (e: Exception) {
//            Log.fail(TAG, "Notification build exception, catching OS bug, where " + "mAppContext: " + mAppContext
//                    + ", NotificationParams.action: " + NotificationHelper.mNotificationParams.intentAction + ", contentIntent: "
//                    + (contentIntent?.toString() ?: "") + ", iconResId: " + NotificationHelper.mNotificationParams.iconResId
//                    + ", tickerText: " + NotificationHelper.mNotificationParams.tickerText + ", now: " + NotificationHelper.mNotificationParams.now + ", contentTitle: "
//                    + NotificationHelper.mNotificationParams.contentTitle + ", contentText: " + NotificationHelper.mNotificationParams.contentText + ", stacktrace: " + e)
        }
    }

    override fun createNotificationChannel() {
        val notificationManager = mAppContext.getSystemService(Context.NOTIFICATION_SERVICE) as NotificationManager
        if (android.os.Build.VERSION.SDK_INT >= android.os.Build.VERSION_CODES.O) {
            val connectionChannelName = mAppContext.getString(R.string.tNotificationChannelConnectivityTitle)
            val connectionChannelDescription = mAppContext.getString(R.string.tNotificationChannelConnectivityDescription)
            val connectionNotificationChannel = NotificationChannel(NotificationHelper.CONNECTIVITY_CHANNEL_ID, connectionChannelName, NotificationManager.IMPORTANCE_LOW)
            connectionNotificationChannel.description = connectionChannelDescription
            connectionNotificationChannel.setShowBadge(false)
            notificationManager.createNotificationChannel(connectionNotificationChannel)
        }
    }

    override fun setupNotification() {
        updateAccountNotification()
    }

    override fun destroy() {
        val service = getService()
        service?.stopForeground(true)
        mController.accountsCtrl.observable.detachObserver(this)
    }

    override fun handleAction(intent: Intent?) {
    }

    private fun getService(): Service? {
        return mServiceInstance.get()
    }

    fun setService(service: WeakReference<Service>) {
        mServiceInstance = service
    }

    // FIXME unnecessarily complex method
    private fun updateAccountNotification() {
        if (!mNotificationsStatusEnabled) {
            return
        }

        Log.d(TAG, "Account registration status changed")

        val title = mAppContext.getString(R.string.tStatusBarNotification)
        var text: String? = null
        var icon = 0
        var intentAction: String? = null
        var priority = NotificationManagerCompat.IMPORTANCE_MIN

        // Retrieve account information for both types of accounts (SIP and XMPP)
        val acct = mAccountsC.getPrimaryAccount()
        val numberOfAccounts = mAccountsC.getAccounts().size
        // Count both SIP and XMPP accounts
        val numberOfActiveAccounts = mAccountsC.getActiveAccounts(EAccountType.Sip).size + mAccountsC.getActiveAccounts(EAccountType.Xmpp).size
        val numberOfActivePushAccounts = mAccountsC.getAccountsWithActivePush(EAccountType.Sip).size
        val numberofFailedPushAccounts = mAccountsC.getPushRegistrationFailedAccounts(EAccountType.Sip).size
        val areAccountsNotRegistered = numberOfActiveAccounts < numberOfAccounts

        var xmppAccountFailed = false
        if (mAccountsC.getActiveAccounts(EAccountType.Xmpp).size == 0) {
            for (xmppAccount in mAccountsC.getAccounts(AccountsFilter.XMPP)) {
                if (xmppAccount.getAccountStatus() == EAccountStatus.RegistrationFailed) {
                    xmppAccountFailed = true
                    break
                }
            }
        }

        if (numberOfActiveAccounts == 0 && (acct == null || (acct.getAccountStatus() == EAccountStatus.Disabled
                        || acct.getAccountStatus() == EAccountStatus.Deregistered || acct.getAccountStatus() == EAccountStatus.Unregistered
                        || acct.getAccountStatus() == EAccountStatus.RegistrationFailed || acct.getAccountStatus() == EAccountStatus.TryingToRegister))) {

            // Google Cloud Messaging - Push Feature
            if (mSettings.getBool(ESetting.FeaturePush) &&
                    (mSettings.getBool(ESetting.PushRegistered) || mSettings.getBool(ESetting.PushAccountsManagedByClient))
                    && (numberOfActivePushAccounts > 0 || numberofFailedPushAccounts > 0)) {
                if (numberOfActivePushAccounts == 1) {
                    icon = R.drawable.ic_stat_notify_push
                    if (numberofFailedPushAccounts > 0) {
                        text = String.format(mAppContext.getString(R.string.tMultiplePushRegisteredAndFailed), numberOfActivePushAccounts,
                                numberofFailedPushAccounts)
                        intentAction = GlobalConstants.INTENT_ACTION_SHOW_ACCOUNTS
                    } else {
                        text = mAppContext.getString(R.string.tSingleLoggedInPush)
                    }
                } else {
                    icon = R.drawable.ic_stat_notify_push
                    if (numberofFailedPushAccounts > 0) {
                        if (numberOfActivePushAccounts > 0) {
                            text = String.format(mAppContext.getString(R.string.tMultiplePushRegisteredAndFailed),
                                    numberOfActivePushAccounts, numberofFailedPushAccounts)
                        } else {
                            if (numberofFailedPushAccounts == 1) {
                                text = mAppContext.getString(R.string.tSinglePushFailed)
                            } else {
                                text = String.format(mAppContext.getString(R.string.tMultiplePushFailed), numberofFailedPushAccounts)
                            }
                        }
                        intentAction = GlobalConstants.INTENT_ACTION_SHOW_ACCOUNTS
                    } else {
                        text = String.format(mAppContext.getString(R.string.tMultipleLoggedInPush), numberOfActivePushAccounts)
                    }
                }
            } else {
                icon = R.drawable.ic_stat_notify_notregistered
                intentAction = GlobalConstants.INTENT_NOTIFICATION_UNREGISTERED_ACCOUNT

                if (acct != null && acct.accountStatus == EAccountStatus.RegistrationFailed || xmppAccountFailed) {
                    text = mAppContext.getString(R.string.tAccountRegistrationFailed)
                    if (mSettings.getBool(ESetting.FeatureWifiOnlyRegistration)) {

                        val eNetworkType = BriaGraph.networkModule.connectionType

                        if (eNetworkType != INetworkObserver.ENetworkType.WIFI) {
                            text = mAppContext.getString(R.string.tCheckYourWiFi)
                        }
                    }
                    priority = NotificationManagerCompat.IMPORTANCE_DEFAULT
                } else {
                    text = mAppContext.getString(R.string.tLoggedOut)
                }
            }
        } else {
            // When registration fails by automatic means (i.e. receive
            // 408) display partial registration if there are saved
            // but unregistered accounts
            var wasAcctManuallyDisabled = true
            if (acct != null && (acct.accountStatus == EAccountStatus.Unregistered || acct.accountStatus == EAccountStatus.Deregistered || acct
                            .accountStatus == EAccountStatus.RegistrationFailed)) {
                wasAcctManuallyDisabled = false
            }

            if (!wasAcctManuallyDisabled && areAccountsNotRegistered) {
                icon = R.drawable.ic_stat_notify_registered
                text = mAppContext.getString(R.string.tPartialLoggedIn)
                intentAction = GlobalConstants.INTENT_ACTION_SHOW_ACCOUNTS
            } else {
                if (numberOfActiveAccounts == 1) {
                    if (numberOfActivePushAccounts > 0) {
                        if (numberofFailedPushAccounts > 0) {
                            // there is 1 active, 1 or more push registered and 1 or more push failed accounts
                            text = String.format(mAppContext.getString(R.string.tMultipleLoggedInWithPushAndFailed),
                                    numberOfActiveAccounts, numberOfActivePushAccounts, numberofFailedPushAccounts)
                            intentAction = GlobalConstants.INTENT_ACTION_SHOW_ACCOUNTS
                        } else {
                            // there is 1 active and 1 or more push registered accounts
                            text = String.format(mAppContext.getString(R.string.tMultipleLoggedInWithPush), numberOfActiveAccounts, numberOfActivePushAccounts)
                        }
                        icon = R.drawable.ic_stat_notify_push
                    } else {
                        if (numberofFailedPushAccounts > 0) {
                            // there is 1 active and 1 or more push failed accounts
                            text = String.format(mAppContext.getString(R.string.tMultipleLoggedInWithPushFailed),
                                    numberOfActiveAccounts, numberofFailedPushAccounts)
                            intentAction = GlobalConstants.INTENT_ACTION_SHOW_ACCOUNTS
                        } else {
                            // there is 1 active account
                            text = mAppContext.getString(R.string.tSingleLoggedIn)
                        }
                        icon = R.drawable.ic_stat_notify_registered
                    }
                } else if (numberOfActiveAccounts > 1) {
                    if (numberOfActivePushAccounts > 0) {
                        if (numberofFailedPushAccounts > 0) {
                            // there is 1 active, 1 or more push registered and 1 or more push failed accounts
                            text = String.format(mAppContext.getString(R.string.tMultipleLoggedInWithPushAndFailed),
                                    numberOfActiveAccounts, numberOfActivePushAccounts, numberofFailedPushAccounts)
                            intentAction = GlobalConstants.INTENT_ACTION_SHOW_ACCOUNTS
                        } else {
                            // there is more than 1 active, and 1 or more push registered accounts
                            text = String.format(mAppContext.getString(R.string.tMultipleLoggedInWithPush), numberOfActiveAccounts, numberOfActivePushAccounts)
                        }
                        icon = R.drawable.ic_stat_notify_push
                    } else {
                        // there is more than 1 active account
                        icon = R.drawable.ic_stat_notify_registered
                        if (numberofFailedPushAccounts > 0) {
                            // there is 1 active and 1 or more push failed accounts
                            text = String.format(mAppContext.getString(R.string.tMultipleLoggedInWithPushFailed),
                                    numberOfActiveAccounts, numberofFailedPushAccounts)
                            intentAction = GlobalConstants.INTENT_ACTION_SHOW_ACCOUNTS
                        } else {
                            text = String.format(mAppContext.getString(R.string.tMultipleLoggedIn), numberOfActiveAccounts)
                        }
                    }
                }
            }
        }

        val whenTime = System.currentTimeMillis()

        val notificationParams = NotificationHelper.updateNotificationParams(SERVICE_NOTIFICATION_ID, icon, null, title, text, null, -1, intentAction,
                NotificationParams.ENotificationType.Ongoing, null, false, priority, whenTime)
        val builder = createNotificationBuilder(notificationParams)
        createNotification(builder, notificationParams)
    }

    fun createTemporaryStatus() {
        val title = mAppContext.getString(R.string.tStatusBarNotification)
        val text = mAppContext.getString(R.string.tStartingApplication)
        val icon = R.drawable.ic_stat_notify_notregistered
        val priority = NotificationManagerCompat.IMPORTANCE_MIN
        val intentAction = GlobalConstants.INTENT_NOTIFICATION_STATUS_TEMPORARY

        val whenTime = System.currentTimeMillis()

        val notificationParams = NotificationHelper.updateNotificationParams(SERVICE_NOTIFICATION_ID, icon, null, title, text, null, -1, intentAction,
                NotificationParams.ENotificationType.Ongoing, null, false, priority, whenTime)
        val builder = createNotificationBuilder(notificationParams)
        createNotification(builder, notificationParams)
    }

    override fun onPrimaryAccountChanged(primaryAccount: Account?) {
        updateAccountNotification()
    }

    override fun onAccountsChanged(changes: AccountsChangeInfo?) {
        updateAccountNotification()
    }

    override fun onAccountVMNumberChanged(voiceMail: VoiceMail?) {
    }

    override fun onAccountStatusChanged(account: Account, isPushTunnel: Boolean) {
        if (account.accountStatus != EAccountStatus.TryingToRegister) {
            updateAccountNotification()
        }
    }
}
