package com.bria.common.uiframework.helpers

import android.annotation.TargetApi
import android.content.Context
import android.content.Intent
import android.content.pm.ShortcutInfo
import android.content.pm.ShortcutManager
import android.graphics.Bitmap
import android.graphics.drawable.BitmapDrawable
import android.graphics.drawable.Icon
import android.net.Uri
import android.os.Build
import android.support.annotation.RequiresApi
import com.bria.common.R
import com.bria.common.controller.Controllers
import com.bria.common.controller.calllog.compareNumber
import com.bria.common.controller.calllog.db.CallLogEntity
import com.bria.common.controller.calllog.getAssociatedContact
import com.bria.common.controller.calllog.getContactDisplayName
import com.bria.common.controller.calllog.getNumberForContacts
import com.bria.common.controller.commlog.CallLogDataObject
import com.bria.common.controller.commlog.CallLogObserverAdapter
import com.bria.common.controller.commlog.ICommLogCtrlEvents
import com.bria.common.controller.commlog.ICommLogCtrlObserver
import com.bria.common.controller.contacts.local.Contact
import com.bria.common.controller.contacts.local.ContactsApi
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.core.Settings
import com.bria.common.modules.ModuleClassFinder
import com.bria.common.modules.glide.GlideApp
import com.bria.common.util.*
import io.reactivex.Maybe
import io.reactivex.disposables.Disposable
import io.reactivex.rxkotlin.subscribeBy
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.PublishSubject

private const val TAG = "BriaShortcutManager"

private const val NUMBER_OF_SHOWN_CALL_LOGS = 3

@RequiresApi(Build.VERSION_CODES.N_MR1)
@TargetApi(Build.VERSION_CODES.M)
class BriaShortcutManager(
        private val context: Context,
        @Deprecated("will be replaced by new CallLogApi")
        private val callLogCtrlObserver: IObservable<ICommLogCtrlObserver>,
        @Deprecated("will be replaced by new CallLogApi")
        private val callLogController: ICommLogCtrlEvents,
        private val settings: Settings,
        private val contactsApi: ContactsApi) : IBriaShortcutManager {

    private val shortcutManager = context.getSystemService(ShortcutManager::class.java)
    private var composeNewMessageShortcut: ShortcutInfo

    private var publishSubject: PublishSubject<Any> = PublishSubject.create<Any>()
    private var disposable: Disposable? = null

    @Deprecated("will be replaced by new CallLogApi")
    private val observer = object: CallLogObserverAdapter() {
        override fun onCallLogCursorRefreshed() {
            publishSubject.onNext(Any())
        }
    }

    init {
        val composeNewMessageIntent = Intent(context, ModuleClassFinder.instance.mainActivityClass).apply { action = GlobalConstants.FAST_SHORTCUT_ACTION_MESSAGE }

        composeNewMessageShortcut =
                ShortcutInfo.Builder(context, "composeIM")
                        .setShortLabel(context.getString(R.string.tComposeNewIM))
                        .setLongLabel(context.getString(R.string.tComposeNewIM))
                        .setIcon(Icon.createWithResource(context, R.drawable.icon_message_active))
                        .setIntent(composeNewMessageIntent)
                        .build()

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.O && settings.getBool(ESetting.FeatureShortcutManager)) {
            if (settings.getBool(ESetting.CallLogV2)) {
                disposable = callLogController.api.getAll()
                        .subscribeBy(
                                onNext = this::updateShortcutsNew,
                                onError = { CrashInDebug.with(TAG, it) }
                        )
            } else {
                Log.d(TAG, "Call Log v2 disabled, old API is using")
                callLogCtrlObserver.attachWeakObserver(observer)

                disposable = publishSubject
                        .subscribeOn(Schedulers.io())
                        .map { Controllers.get().callLog.allCallLogs.take(NUMBER_OF_SHOWN_CALL_LOGS) }
                        .observeOn(Schedulers.io()) // move the delivery of new events (see usage of onNext above) to IO thread
                        .subscribeBy(
                                onNext = this::updateShortcutsOld,
                                onError = { CrashInDebug.with(TAG, it) }
                        )
            }
        }
    }

    override fun update() {
        val composeNewMessageIntent = Intent(context, ModuleClassFinder.instance.mainActivityClass).apply { action = GlobalConstants.FAST_SHORTCUT_ACTION_MESSAGE }

        composeNewMessageShortcut =
                ShortcutInfo.Builder(context, "composeIM")
                        .setShortLabel(context.getString(R.string.tComposeNewIM))
                        .setLongLabel(context.getString(R.string.tComposeNewIM))
                        .setIcon(Icon.createWithResource(context, R.drawable.icon_message_active))
                        .setIntent(composeNewMessageIntent)
                        .build()
        callLogController.api.getAll()
                .subscribeBy(
                        onNext = this::updateShortcutsNew,
                        onError = { CrashInDebug.with(TAG, it) }
                )
    }

    override fun destroy() {
        callLogCtrlObserver.detachObserver(observer)
        disposable?.dispose()
    }

    private fun updateShortcutsNew(allCallLogs: List<CallLogEntity>) {
        val fewGroupedCallLogs = toGroupedCallLogs(allCallLogs, NUMBER_OF_SHOWN_CALL_LOGS)
        shortcutManager.removeAllDynamicShortcuts()

        shortcutManager.addDynamicShortcuts(listOf(composeNewMessageShortcut))

        shortcutManager.addDynamicShortcuts(
                fewGroupedCallLogs.map {
                    val displayName = it.getContactDisplayName()
                    val icon = Icon.createWithBitmap(getIconFromContact(it.getAssociatedContact().blockingGet()))
                    val callIntent = getCallIntent(it.getNumberForContacts())
                    return@map ShortcutInfo.Builder(context, it.id.toString())
                            .setShortLabel(displayName)
                            .setLongLabel(displayName)
                            .setIcon(icon)
                            .setIntent(callIntent)
                            .build()
                }
        )
    }

    private fun toGroupedCallLogs(allCallLogs: List<CallLogEntity>, n: Int): List<CallLogEntity> {
        val fewGroupedCallLogs = mutableListOf<CallLogEntity>()
        var lastGroupEntity: CallLogEntity? = null
        for (entity in allCallLogs) {
            if (lastGroupEntity == null || !lastGroupEntity.compareNumber(entity)) {
                lastGroupEntity = entity
                fewGroupedCallLogs.add(lastGroupEntity)
            }

            if (fewGroupedCallLogs.size == n) break
        }

        return fewGroupedCallLogs
    }

    private fun getIconFromContact(contact: Contact?): Bitmap {
        val defaultPhoto = Maybe.fromCallable { (AndroidUtils.getDrawable(context, R.drawable.default_avatar) as BitmapDrawable).bitmap }
        contact ?: return defaultPhoto.blockingGet()
        return contact.getPhotoAsMaybe(context)
                .map { GlideApp.with(context).asBitmap().load(it).circleCrop().into(100, 100).get() }
                .switchIfEmpty(defaultPhoto)
                .blockingGet()
    }

    private fun getCallIntent(number: String): Intent {
        return Intent(context, ModuleClassFinder.instance.mainActivityClass).apply {
            action = GlobalConstants.FAST_SHORTCUT_ACTION_CALL
            data = Uri.fromParts("sip", number, null)
        }
    }

    @Deprecated("will be replaced by new CallLogApi")
    private fun updateShortcutsOld(fewLastCallLogs: List<CallLogDataObject>) {
        shortcutManager.removeAllDynamicShortcuts()

        shortcutManager.addDynamicShortcuts(listOf(composeNewMessageShortcut))

        shortcutManager.addDynamicShortcuts(
                fewLastCallLogs.map {
                    val contact = getContact(it)
                    val labelString = if (it.remoteName.isNullOrEmpty()) it.number else it.remoteName
                    val callIntent = getCallIntent(it.number)
                    val icon = Icon.createWithBitmap(getIconFromContact(contact))
                    return@map ShortcutInfo.Builder(context, it.id.toString())
                            .setShortLabel(labelString)
                            .setLongLabel(labelString)
                            .setIcon(icon)
                            .setIntent(callIntent)
                            .build()
                }
        )
    }

    @Deprecated("will be replaced by new CallLogApi")
    private fun getContact(callLogDataObject: CallLogDataObject): Contact? {
        val name = callLogDataObject.name
        return try {
            contactsApi.contact.byId(callLogDataObject.contactId).withAllData().asObservable()
                    .onErrorResumeNext(contactsApi.contact.byNumber(getNumberWithoutDomain(callLogDataObject)).andDisplayName(name).withAllData().asObservable())
                    .onErrorResumeNext(contactsApi.contact.byNumber(getSanitizedNumberWithDomain(callLogDataObject)).andDisplayName(name).withAllData().asObservable())
                    .onErrorResumeNext(contactsApi.contact.byNumber(getNumberWithDomain(callLogDataObject)).andDisplayName(name).withAllData().asObservable())
                    .onErrorResumeNext(contactsApi.contact.byNumber(getSanitizedNumberWithoutDomain(callLogDataObject)).andDisplayName(name).withAllData().asObservable())
                    .blockingFirst()
        } catch (exception: Exception) {
            null
        }
    }

    @Deprecated("will be replaced by new CallLogApi")
    private fun getNumberWithoutDomain(callLogDataObject: CallLogDataObject): String {
        return callLogController.getNumberWithoutDomain(callLogDataObject)
    }

    @Deprecated("will be replaced by new CallLogApi")
    private fun getSanitizedNumberWithDomain(callLogDataObject: CallLogDataObject): String {
        return callLogController.getSanitizedNumberWithDomain(callLogDataObject)
    }

    @Deprecated("will be replaced by new CallLogApi")
    private fun getNumberWithDomain(callLogDataObject: CallLogDataObject): String {
        return callLogController.getNumberWithDomain(callLogDataObject)
    }

    @Deprecated("will be replaced by new CallLogApi")
    private fun getSanitizedNumberWithoutDomain(callLogDataObject: CallLogDataObject): String {
        return callLogController.getSanitizedNumberWithoutDomain(callLogDataObject)
    }
}
