package com.bria.common.uireusable.dataprovider

import android.annotation.SuppressLint
import android.content.ContentResolver
import android.content.Context
import android.database.Cursor
import android.database.MatrixCursor
import android.net.Uri
import android.provider.ContactsContract
import android.support.annotation.Size
import android.text.TextUtils
import com.bria.common.controller.ClientConfig
import com.bria.common.controller.Controllers
import com.bria.common.controller.buddy.BuddyAvailability
import com.bria.common.controller.contacts.ContactsSortBy
import com.bria.common.controller.contacts.buddy.Buddy
import com.bria.common.controller.contacts.buddy.IBuddyCtrlEvents
import com.bria.common.controller.contacts.buddy.IBuddyCtrlObserver
import com.bria.common.controller.contacts.buddy.SipBuddy
import com.bria.common.controller.contacts.local.ContactsApi
import com.bria.common.controller.contacts.local.data.ContactsDB
import com.bria.common.controller.contacts.local.provider.ContactQueries
import com.bria.common.controller.settings.ESetting
import com.bria.common.uireusable.datatypes.IPersonListItem
import com.bria.common.util.IObservable
import com.bria.common.util.Log
import com.bria.common.util.t9.ContactDataItem
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.schedulers.Schedulers
import java.util.*
import java.util.concurrent.TimeUnit


class ContactDataProviderV2
@SuppressLint("RxLeakedSubscription", "RxSubscribeOnError") // todo: avoid using it
constructor(
        context: Context,
        contactChangedObservable: Observable<Uri>,
        private val contactsApi: ContactsApi,
        iBuddyObservable: IObservable<IBuddyCtrlObserver>
) : AbstractFilterableCursorDataProvider<IPersonListItem>(context), IContactActionDataProvider<IPersonListItem>, IBuddyCtrlObserver {

    private var iContactDataProviderListener: IExtendedFilterableDataProvider? = null

    init {
        // todo: care of disposable
        val disposable = contactChangedObservable
                .subscribeOn(Schedulers.io())
                .throttleFirst(1, TimeUnit.SECONDS, Schedulers.io())
                .observeOn(Schedulers.io())
                .subscribe { _ ->
                    refilter()
                }
        iBuddyObservable.attachWeakObserver(this)
    }

    companion object {

        private val TAG = "ContactDataProviderV2"

        private val SOFTPHONE_MIMETYPE = "" // using an empty string for a slight optimization

        private val CONTACT_ROW_PROJECTION = arrayOf(ContactsContract.Contacts.DISPLAY_NAME, ContactsContract.Contacts.PHOTO_URI, ContactsContract.Contacts.PHOTO_ID)
        private val CONTACT_ROW_SELECTION = String.format("%s = ?", ContactsContract.Contacts._ID)

        internal fun getContactRow(contactId: String, contentResolver: ContentResolver?): Cursor? {
            return contentResolver?.query(
                    ContactsContract.Contacts.CONTENT_URI,
                    CONTACT_ROW_PROJECTION,
                    CONTACT_ROW_SELECTION,
                    arrayOf(contactId), null)
        }

        internal// the desired sort order for the returned Cursor. In honeycomb and later, the primary sort key allows for localization.
        // in earlier versions use the display name as the sort key
        val sortOrderQuery: String
            get() = if (ContactsSortBy.fromRawIntValue(Controllers.get().settings.getInt(ESetting.ContactSortOrder)) === ContactsSortBy.LastName)
                ContactsContract.Contacts.SORT_KEY_ALTERNATIVE
            else
                ContactsContract.Contacts.SORT_KEY_PRIMARY
    }

    override fun initializeLoadingData() {
        initialize()
    }

    override fun constructDataObject(data: Cursor, position: Int): IPersonListItem {
        data.moveToPosition(position)
        return ContactDataItem.createBuilder()
                .setId(data.getString(ContactQueries.FilterResult.ID))
                .setFirstName(data.getString(ContactQueries.FilterResult.FIRST_NAME))
                .setLastName(data.getString(ContactQueries.FilterResult.LAST_NAME))
                .setQueryData(data.getString(ContactQueries.FilterResult.QUERY_DATA))
                .setQueryDataType(data.getString(ContactQueries.FilterResult.QUERY_DATA_TYPE))
                .setImageUri(data.getString(ContactQueries.FilterResult.PHOTO_URI))
                .setImageId(data.getInt(ContactQueries.FilterResult.PHOTO_ID))
                .build()
    }

    @SuppressLint("Recycle") // we will close the Cursor later
    override fun provideFullDataSet(contentResolver: ContentResolver): Cursor {
        val projection = ContactQueries.AllContacts.PROJECTION
        val selection = if (Controllers.get().settings.getBool(ESetting.ShowAllContacts))
            ContactQueries.AllContacts.SELECTION_NO_GROUP
        else
            ContactQueries.AllContacts.SELECTION
        val sortOrder = ContactDataProvider.getSortOrderQuery()

        var result = MatrixCursor(ContactQueries.FilterResult.PROJECTION, 1)

        try {
            contentResolver.query(ContactQueries.AllContacts.CONTENT_URI, projection, selection, null, sortOrder)!!.use { cursor ->
                if (cursor.moveToFirst()) {
                    result = MatrixCursor(ContactQueries.FilterResult.PROJECTION, cursor.count)

                    do {
                        val firstLastName = getNames(cursor.getString(ContactQueries.AllContacts.DISPLAY_NAME))

                        // Add in this order: id, first name, last name, query data, query data type (for UI), photo URI, photo ID
                        result.newRow()
                                .add(cursor.getInt(ContactQueries.AllContacts.ID))
                                .add(firstLastName[0])
                                .add(firstLastName[1])
                                .add("")
                                .add("")
                                .add(cursor.getString(ContactQueries.AllContacts.PHOTO_THUMBNAIL_DATA))
                                .add(cursor.getInt(ContactQueries.AllContacts.PHOTO_ID))
                    } while (cursor.moveToNext())
                }
            }
        } catch (ignored: Exception) {
            // it can crash if contacts permission not granted yet
            Log.w(TAG, "Contacts exception non fatal " + ignored.message)
        }

        return result
    }

    @SuppressLint("UseSparseArrays")
    override fun provideFilteredDataSet(contentResolver: ContentResolver, query: String): Cursor {
        val selection = if (Controllers.get().settings.getBool(ESetting.ShowAllContacts))
            ContactQueries.Filter.SELECTION_NO_GROUP
        else
            ContactQueries.Filter.SELECTION
        val selectionArgs = arrayOf("$query%", "$query%", "$query%", "$query%")
        val sortOrder = ContactDataProvider.getSortOrderQuery()

        val softphones: Map<String, List<String>>
        if (context == null) {
            softphones = HashMap()
        } else {
            softphones = ContactsDB.get(context!!).getIdSoftphonePairForQuery(query)
        }

        val visitedMap = HashMap<String, Boolean>()

        for ((key) in softphones) {
            // None of these softphones have been passed through, marking them all as false
            visitedMap[key] = false
        }

        var result: MatrixCursor? = null

        try {
            contentResolver.query(ContactQueries.Filter.URI, ContactQueries.Filter.PROJECTION, selection, selectionArgs, sortOrder)!!.use { cursor ->
                if (cursor != null && cursor.moveToFirst()) {
                    result = MatrixCursor(ContactQueries.FilterResult.PROJECTION, cursor.count)

                    val seen = HashSet<String>()

                    do {
                        val contactId = cursor.getString(ContactQueries.Filter.ID)
                        val mimeType = cursor.getString(ContactQueries.Filter.MIMETYPE)
                        var data: String? = null

                        // Eliminate all rows that are matched only by name, if we have matches by other criteria (i.e. phone number, email...)
                        if (!(ContactsContract.CommonDataKinds.StructuredName.CONTENT_ITEM_TYPE == mimeType && seen.contains(contactId))) {
                            seen.add(contactId)
                            data = cursor.getString(ContactQueries.Filter.DATA)
                        }

                        val phones = softphones[contactId]
                        if (phones != null && !visitedMap[contactId]!!) {
                            var i = 0
                            val size = phones.size
                            while (i < size) {
                                val softphone = phones[i]
                                seen.add(contactId)
                                visitedMap[contactId] = true
                                fillRow(cursor, result!!, contactId, SOFTPHONE_MIMETYPE, softphone)
                                i++
                            }
                        }

                        if (data != null) {
                            fillRow(cursor, result!!, contactId, mimeType, data)
                        }
                    } while (cursor.moveToNext())
                }
            }
        } catch (e: Exception) {
            Log.e(TAG, "Contacts exception", e)
        }

        if (result == null) {
            result = MatrixCursor(ContactQueries.FilterResult.PROJECTION, visitedMap.entries.size)
        }

        // Add all the contacts that have only softphones (won't be picked up by the upper cursor)
        for ((contactId, value) in visitedMap) {
            // If a softphone ID has not been visited at all (possible when a contact only has softphones)
            if (!value) {
                try {
                    ContactDataProvider.getContactRow(contactId, mContentResolver)!!.use { contactRow ->
                        if (contactRow != null && contactRow.moveToFirst()) {
                            val displayName = contactRow.getString(0)
                            val softphone = softphones[contactId]?.get(0)
                            val photoUri = contactRow.getString(1)
                            val photoId = contactRow.getInt(2)

                            fillRow(result!!, contactId, SOFTPHONE_MIMETYPE, softphone.toString(), displayName, photoUri, photoId)
                        }
                    }
                } catch (e: Exception) {
                    Log.e(TAG, "Contacts exception", e)
                }

            }
        }

        return result as MatrixCursor
    }

    private fun fillRow(cursor: Cursor, result: MatrixCursor, id: String, mimeType: String, data: String) {
        val photoUri = if (TextUtils.isEmpty(cursor.getString(ContactQueries.Filter.PHOTO_URI))) "" else cursor.getString(ContactQueries.Filter.PHOTO_URI)
        fillRow(result, id, mimeType, data, cursor.getString(ContactQueries.Filter.DISPLAY_NAME), photoUri,
                cursor.getInt(ContactQueries.Filter.PHOTO_ID))
    }

    private fun fillRow(result: MatrixCursor, id: String, mimeType: String, data: String, displayName: String, photoUri: String,
                        photoId: Int) {
        val firstLastName = getNames(displayName)

        // Add in this order: id, first name, last name, query data, query data type (for UI), photo URI, photo ID
        result.newRow()
                .add(id)
                .add(firstLastName[0])
                .add(firstLastName[1])
                .add(data)
                .add(convertMimetypeForUi(mimeType))
                .add(photoUri)
                .add(photoId)
    }

    @Size(2)
    private fun getNames(displayName: String): Array<String> {
        var displayName = displayName
        displayName = displayName.trim { it <= ' ' }
        val spaceIndex = displayName.indexOf(" ")
        var firstName = displayName
        var secondName = ""

        if (spaceIndex > 0) {
            firstName = displayName.substring(0, spaceIndex)
            // Because we have trimmed the string before, we can safely call substring on spaceIndex + 1  here.
            secondName = displayName.substring(spaceIndex + 1)
        }

        return arrayOf(firstName, secondName)
    }

    // TODO: Send int IDs to the UI instead, and resolve in the adapter, for translations to work.
    private fun convertMimetypeForUi(rawMimetype: String): String {
        when (rawMimetype) {
            SOFTPHONE_MIMETYPE -> return "Softphone"
            ContactsContract.CommonDataKinds.StructuredName.CONTENT_ITEM_TYPE -> return ""
            ContactsContract.CommonDataKinds.Phone.CONTENT_ITEM_TYPE -> return "Phone"
            ContactsContract.CommonDataKinds.Email.CONTENT_ITEM_TYPE -> return "Email"
            ContactsContract.CommonDataKinds.SipAddress.CONTENT_ITEM_TYPE -> return "Sip Address"
            else -> {
                if (ClientConfig.get().isDebugMode) {
                    throw RuntimeException("Unknown mimetype: $rawMimetype")
                }
                return "Unknown"
            }
        }
    }

    @SuppressLint("RxLeakedSubscription", "CheckResult")
    override fun removeActiveItem(position: Int?) {

        val contactId = getItemAt(position!!)!!.uniqueIdentifier
        contactsApi.removeContact(contactId)
                .subscribeOn(Schedulers.io())
                .observeOn(AndroidSchedulers.mainThread())
                .subscribe({ }, { throwable -> })

    }

    override fun attachWeakListenerV2(iExtendedFilterableDataProvider: IExtendedFilterableDataProvider) {
        attachWeakListener(iExtendedFilterableDataProvider)
        this.iContactDataProviderListener = iExtendedFilterableDataProvider
    }

    override fun onBuddyListUpdated() {
    }

    override fun onBuddyPresenceChanged(buddy: Buddy?) {
        if (buddy!!.buddyType == Buddy.EBuddyType.SIP) {
            val contact = buddy as SipBuddy
            if (mDisplayedData!!.moveToFirst()) {
                do {
                    val cursorContactId = mDisplayedData!!.getString(ContactQueries.Filter.ID)
                    if (cursorContactId == contact.contactId) {
                        iContactDataProviderListener!!.updatedItem(mDisplayedData!!.position)
                        return
                    }

                } while (mDisplayedData!!.moveToNext())
            }
        }


    }

    override fun onBuddyFilterChanged(aBuddyFilterType: IBuddyCtrlEvents.EBuddyFilterType?) {
    }

    override fun filterDataProvider(query: CharSequence) {
        filterData(query)
    }

    override fun filterDataProvider(subFilter: BuddyAvailability) {
    }

}
