package com.bria.common.util

import java.text.BreakIterator

object TextBoundaries {

    /*
    * Character boundary analysis allows users to interact with characters as they expect to, for example, when moving the cursor through a
    * text string. Character boundary analysis provides correct navigation through character strings, regardless of how the character is
    * stored. The boundaries returned may be those of supplementary characters, combining character sequences, or ligature clusters. For
    * example, an accented character might be stored as a base character and a diacritical mark. What users consider to be a character
    * can differ between languages.
    *
    * https://developer.android.com/reference/java/text/BreakIterator.html#character
    *
    * https://en.wikipedia.org/wiki/Unicode_equivalence#Combining_and_precomposed_characters
    *
    * http://www.unicode.org/glossary/#grapheme
    *
    * Additionally, this implementation handles flags.
    * */
    fun findGraphemesAndFlags(s: String): List<String> {
        val graphemes = mutableListOf<String>()

        val breakIterator = BreakIterator.getCharacterInstance()
        breakIterator.setText(s)

        var start = breakIterator.first()
        var end = breakIterator.next()
        while (end != BreakIterator.DONE) {
            graphemes.add(s.substring(start, end))

            start = end
            end = breakIterator.next()
        }

        if (graphemes.size < 2) {
            return graphemes
        }

        val result = mutableListOf<String>()

        var previousRegionalIndicatorSymbol: String? = null
        for (grapheme in graphemes) {
            if (isRegionalIndicatorSymbol(grapheme)) {
                // part of the flag

                if (previousRegionalIndicatorSymbol == null) {
                    // wait for next
                    previousRegionalIndicatorSymbol = grapheme
                } else {
                    // flag
                    val flag = previousRegionalIndicatorSymbol + grapheme
                    result.add(flag)
                    previousRegionalIndicatorSymbol = null
                }
            } else {
                // not part of the flag
                if (previousRegionalIndicatorSymbol == null) {
                    result.add(grapheme)
                } else {
                    result.add(previousRegionalIndicatorSymbol)
                    result.add(grapheme)
                    previousRegionalIndicatorSymbol = null
                }
            }
        }

        return result
    }

    fun isRegionalIndicatorSymbol(s: String): Boolean {
        if (s.length < 2) {
            return false
        }

        val toTest = s.substring(0, 2)
        val REGIONAL_INDICATOR_SYMBOL_LETTER_A = "\uD83C\uDDE6";
        val REGIONAL_INDICATOR_SYMBOL_LETTER_Z = "\uD83C\uDDFF";

        val result = toTest >= REGIONAL_INDICATOR_SYMBOL_LETTER_A && toTest <= REGIONAL_INDICATOR_SYMBOL_LETTER_Z

        return result
    }
}
