package com.bria.voip.ui.main.calllog.refactoring

import android.Manifest
import android.content.BroadcastReceiver
import android.content.Context
import android.content.Intent
import android.content.IntentFilter
import android.content.pm.PackageManager
import android.os.Bundle
import android.util.SparseArray
import com.bria.common.R
import com.bria.common.controller.calllog.CallLogApi
import com.bria.common.controller.calllog.db.CallLogEntity
import com.bria.common.controller.calllog.db.CallType
import com.bria.common.controller.calllog.getDefaultDisplayName
import com.bria.common.controller.contacts.buddy.Buddy
import com.bria.common.controller.contacts.buddy.IBuddyCtrlEvents
import com.bria.common.controller.contacts.buddy.IBuddyCtrlObserver
import com.bria.common.controller.contacts.local.ContactsApi
import com.bria.common.controller.im.ImpsUtils
import com.bria.common.uiframework.presenters.AbstractPresenter
import com.bria.common.uiframework.presenters.IPresenterEventTypeEnum
import com.bria.common.uireusable.dataprovider.ISimpleDataProvider
import com.bria.common.util.Log
import com.bria.common.util.RecordingUtils
import com.bria.common.util.SipAddressUtils
import io.reactivex.Completable
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.disposables.Disposable
import io.reactivex.rxkotlin.plusAssign
import io.reactivex.rxkotlin.subscribeBy
import io.reactivex.schedulers.Schedulers
import java.util.concurrent.CopyOnWriteArrayList
import java.util.concurrent.TimeUnit
import kotlin.math.abs
import kotlin.properties.Delegates
import kotlin.reflect.KProperty

enum class CallLogFilterType(val nameResId: Int) {
    ALL(R.string.tCallLogFilterAll),
    MISSED(R.string.tCallLogFilterMissed),
    CONFERENCE(R.string.tCallLogFilterConference)
}

abstract class CallLogAbstractPresenter : AbstractPresenter(), ISimpleDataProvider<CallLogItemModel> {
    companion object {
        private const val TAG = "CallLogAbstractPresenter"
    }

    enum class Events : IPresenterEventTypeEnum {
        DATA_SET_CHANGED,
        DATA_SET_UPDATED,

        SHOW_MESSAGE_LONG,

        SHOW_DELETE_CALL_RECORD_CONFIRMATION,

        SHOW_PLAY_RECORD_VIEW,
        SHOW_SHARE_RECORD_VIEW,

        UPDATE_SELECTED_TAB,

        // define new events here
        ;
        override fun getType(): IPresenterEventTypeEnum = this
    }

    var filterType: CallLogFilterType by Delegates.observable(CallLogFilterType.ALL, onChange = ::onFilterTypePropertyChanged)

    var isInTabletMode = false

    open val isTabsVisible = true

    protected lateinit var mCallLogApi: CallLogApi
    protected lateinit var mContactsApi: ContactsApi

    protected val mGroupedList: MutableList<CallLogItemModel> = CopyOnWriteArrayList<CallLogItemModel>()

    private val mDisposables = CompositeDisposable()

    private var mGetAllDisposable: Disposable? = null

    private var mRestartUpdatingContactInfoDisposable: Disposable? = null
    private val mDoUpdatingContactInfoDisposables = CompositeDisposable()
    private val mUpdateGroupDisposables = SparseArrayDisposable()

    private var mFirstVisiblePosition = 0
    private var mLastVisiblePosition = 10
    private var mOrientation = resources.configuration.orientation

    // Time listener is required to keep relative date time updated, see CallLogGroupItem#getCallInfo
    private val mTimeTickReceiver = object : BroadcastReceiver() {
        override fun onReceive(context: Context, intent: Intent) {
            firePresenterEvent(Events.DATA_SET_CHANGED)
        }
    }

    private val mBuddyObserver = object : IBuddyCtrlObserver {
        override fun onBuddyFilterChanged(buddyFilterType: IBuddyCtrlEvents.EBuddyFilterType) {

        }

        override fun onBuddyListUpdated() {
            onContactsChanged()
        }

        override fun onBuddyPresenceChanged(buddy: Buddy) {
            findGroupItemBy(ImpsUtils.getBuddyKey(buddy))?.let { firePresenterEvent(Events.DATA_SET_CHANGED) }
        }
    }

    // Life-cycle

    override fun onCreate() {
        super.onCreate()

        mCallLogApi = mControllers.callLog.api
        mContactsApi = mControllers.contacts.contactsApi

        if (isTabsVisible) {
            mDisposables += mCallLogApi.getCountOfUnreadMissedCalls()
                    .subscribeBy(
                            onNext = this::onCountOfUnreadMissedCallsChanged,
                            onError = { Log.e(TAG, "CallLogApi.getCountOfUnreadMissedCalls error", it) },
                            onComplete = { Log.w(TAG, "CallLogApi.getCountOfUnreadMissedCalls complete") }
                    )
        }

        restartGetAll()

        mDisposables += mContactsApi.observable
                .subscribeOn(Schedulers.io())
                .throttleFirst(1, TimeUnit.SECONDS, Schedulers.io())
                .observeOn(Schedulers.io())
                .subscribeBy(
                        onNext = { onContactsChanged() },
                        onError = { Log.e(TAG, "ContactsApi.getObservable error", it) },
                        onComplete = { Log.w(TAG, "ContactsApi.getObservable complete") }
                )
    }

    private fun restartGetAll() {
        mGetAllDisposable?.dispose()
        mGetAllDisposable = mCallLogApi.getAll()
                // already working on IO thread, since we do not switch threads here observation also will be on IO thread
                .subscribeBy(
                        onNext = this::onCallLogDataChanged,
                        onError = { Log.e(TAG, "CallLogApi.getAll error", it) },
                        onComplete = { Log.w(TAG, "CallLogApi.getAll complete") }
                )
    }

    override fun onDestroy() {
        super.onDestroy()
        mGetAllDisposable?.dispose()
        mDisposables.dispose()
        mDoUpdatingContactInfoDisposables.dispose()
        mUpdateGroupDisposables.dispose()
    }

    override fun onSubscribe() {
        super.onSubscribe()
        firePresenterEvent(Events.DATA_SET_CHANGED)
        firePresenterEvent(Events.UPDATE_SELECTED_TAB)
        context.applicationContext.registerReceiver(mTimeTickReceiver, IntentFilter(Intent.ACTION_TIME_TICK))
        mObservables.buddy.attachWeakObserver(mBuddyObserver)
    }

    override fun onUnsubscribe() {
        super.onUnsubscribe()
        context.applicationContext.unregisterReceiver(mTimeTickReceiver)
        mObservables.buddy.detachObserver(mBuddyObserver)
        stopUpdatingContactInfo()
    }

    // Data Provider

    override fun getItemsCount() = mGroupedList.size

    override fun getItemAt(position: Int) = mGroupedList.getOrNull(position)

    override fun clean() { }

    // = PUBLIC section

    abstract fun getTitle(): String

    abstract fun onScreenNewConfig(bundle: Bundle?)

    // Tab settings
    fun isConferenceTabVisible() = isTabsVisible && mControllers.collaboration.isCollabAvailable

    // Item actions

    fun onItemDeleteCallRecordButtonClicked(position: Int) =
            // Explicit call of method with three arguments to avoid interpret passed int value as delay
            firePresenterEvent(Events.SHOW_DELETE_CALL_RECORD_CONFIRMATION, position, 0)

    fun onItemSaveCallRecordButtonClicked(position: Int) {
        // todo: implement me, what we should todo here?
    }

    fun onItemShareCallRecordButtonClicked(position: Int) {
        val contentItem = getContentItemAt(position) ?: return

        val recordFileUri = RecordingUtils.getFileUri(context, contentItem.recordFileName) ?: return

        val message = getString(R.string.tCallRecordingShareText, findGroupItemBy(contentItem)?.displayName.orEmpty())

        firePresenterEvent(Events.SHOW_SHARE_RECORD_VIEW, Pair(recordFileUri, message))
    }

    fun onItemPlayCallRecordButtonClicked(position: Int) {
        if (mControllers.phone.callCount != 0) {
            firePresenterEvent(Events.SHOW_MESSAGE_LONG, getString(R.string.tCannotPlayRecordingDuringCall))
            Log.w(TAG, "Call in progress, cannot play recording during a call.")
            return
        }

        val item = getContentItemAt(position) ?: return

        val recordFileUri = RecordingUtils.getFileUri(context, item.recordFileName) ?: return

        firePresenterEvent(Events.SHOW_PLAY_RECORD_VIEW, recordFileUri)
    }

    fun onDeleteCallRecordConfirmed(position: Int) {
        val item = getContentItemAt(position) ?: return
        // Remove file from storage
        if (!RecordingUtils.deleteFile(context, item.recordFileName)) {
            firePresenterEvent(CallLogAbstractPresenter.Events.SHOW_MESSAGE_LONG, getString(R.string.tFailedToDeleteRecording))
            return
        }

        // Reset recordPath value in DB
        // todo: avoid blocking call here
        val callLogEntity = mCallLogApi.get(item.id.toLong()).blockingGet()
        callLogEntity?.let {
            it.recordFileName = ""
            mDisposables += mCallLogApi.put(it).subscribeBy(onError = { Log.w(TAG, "UPDATE entity in DB error", it) })
        }
    }

    // = PROTECTED section

    protected abstract fun buildGroupedList(newEntityList: List<CallLogEntity>): List<CallLogItemModel>

    protected fun buildGroup(callLogEntity: CallLogEntity) : CallLogGroupItem {
        return CallLogGroupItem(
                id = "g${callLogEntity.id}",
                number = SipAddressUtils.removeDomain(callLogEntity.number), // domain can still be exist in DB (old call log entries)
                displayName = callLogEntity.getDefaultDisplayName(),
                date = callLogEntity.date,
                // Below uses for options only
                accountId = callLogEntity.accountId,
                accountDomain = callLogEntity.accountDomain,
                remoteDomain = callLogEntity.remoteDomain,
                hostedCollab = callLogEntity.hostedCollab
        ).apply {
            callTypes += callLogEntity.callType
            items += buildContent(callLogEntity)
        }
    }

    protected fun buildContent(callLogEntity: CallLogEntity) : CallLogContentItem {
        return CallLogContentItem(
                id = callLogEntity.id.toString(),
                remoteName = callLogEntity.remoteName,
                callType = callLogEntity.callType,
                date = callLogEntity.date,
                durationSeconds = callLogEntity.durationSeconds,
                recordFileName = callLogEntity.recordFileName,
                forwardedTo = callLogEntity.forwardedTo,
                isCallForwardedInfoVisible = callLogEntity.forwardedTo.isNotEmpty(),
                isCallDateColored = callLogEntity.callType == CallType.MISSED,
                isCallDurationVisible = callLogEntity.callType != CallType.MISSED && callLogEntity.durationSeconds != -1,
                isRecordInfoVisible = callLogEntity.recordFileName.isNotEmpty(),
                isSaveButtonVisible = false,
                hostedCollab = callLogEntity.hostedCollab,
                externalId = callLogEntity.externalId
        )
    }

    @Synchronized
    protected fun onCallLogDataChanged(newEntityList: List<CallLogEntity>) {
        stopUpdatingContactInfo()

        val filteredEntityList =
                // Null means any call type is acceptable, so if null then just use original list, else apply filter
                mapFilterType()?.let { filter -> newEntityList.filter { it.callType == filter } } ?: newEntityList
        val newGroupedList = buildGroupedList(filteredEntityList)
        updateGroupsDisplayName(newGroupedList)

        // switch to main thread
        post {
            mGroupedList.clear()
            mGroupedList += newGroupedList
            firePresenterEvent(Events.DATA_SET_CHANGED)
            restartUpdatingContactInfo()
        }
    }

    // Helper methods

    protected fun findGroupItemBy(buddyKey: String) =
            mGroupedList.find { it.isGroup && (it as CallLogGroupItem).buddyKey == buddyKey } as? CallLogGroupItem

    protected fun findGroupItemBy(contentItem: CallLogContentItem) =
            mGroupedList.find { it.isGroup && (it as CallLogGroupItem).items.contains(contentItem) } as? CallLogGroupItem

    protected fun getGroupItemAt(position: Int) = getItemAt(position) as? CallLogGroupItem

    protected fun getContentItemAt(position: Int) = getItemAt(position) as? CallLogContentItem

    protected fun isGroup(position: Int): Boolean = getItemAt(position)?.isGroup == true

    protected fun getExpandedPosition() = mGroupedList.indexOfFirst { it.isGroup && (it as CallLogGroupItem).isExpanded }

    protected fun getSelectedPosition() = mGroupedList.indexOfFirst { it.isGroup && (it as CallLogGroupItem).isSelected }

    // = PRIVATE section

    // Try to keep using remote name if it exist
    private fun updateGroupsDisplayName(groupedList: List<CallLogItemModel>) {
        groupedList.forEach { (it as? CallLogGroupItem)?.updateDisplayName() }
    }

    private fun onFilterTypePropertyChanged(property: KProperty<*>, oldValue: CallLogFilterType, newValue: CallLogFilterType) {
        Log.d(TAG, "FilterType property changed: $oldValue -> $newValue")
        stopUpdatingContactInfo()
        restartGetAll()
    }

    private fun mapFilterType() = when (filterType) {
        CallLogFilterType.ALL -> null // Null means any call type is acceptable
        CallLogFilterType.MISSED -> CallType.MISSED
        CallLogFilterType.CONFERENCE -> CallType.CONFERENCE
    }

    private fun onCountOfUnreadMissedCallsChanged(count: Int) {
        if (count > 0) {
            filterType = CallLogFilterType.MISSED
            firePresenterEvent(Events.UPDATE_SELECTED_TAB)
        }
    }

    // Update info with contacts data

    private fun onContactsChanged() {
        stopUpdatingContactInfo()
        mGroupedList.mapNotNull { it as? CallLogGroupItem }.forEach { it.isContactInfoDeprecated = true }
        restartUpdatingContactInfo()
    }

    override fun onRequestPermissionsResult(requestCode: Int, permissions: Array<String>, grantResults: IntArray) {
        super.onRequestPermissionsResult(requestCode, permissions, grantResults)

        if (grantResults.firstOrNull() == PackageManager.PERMISSION_GRANTED) {
            when (permissions.firstOrNull()) {
                Manifest.permission.READ_CONTACTS -> restartUpdatingContactInfo()
            }
        }
    }

    @Synchronized
    private fun restartUpdatingContactInfo(delay: Long = 0L) {
        val f: (Any) -> Unit = {
            stopUpdatingContactInfo()

            val startUpdatingPosition = mFirstVisiblePosition
            val updatingItemCount = mLastVisiblePosition - mFirstVisiblePosition + 1

            startUpdatingContactInfo(startUpdatingPosition, updatingItemCount, true)
        }

        mRestartUpdatingContactInfoDisposable?.dispose()

        if (delay == 0L) {
            f.invoke(0L)
        } else {
            mRestartUpdatingContactInfoDisposable = Observable.timer(delay, TimeUnit.MILLISECONDS, Schedulers.computation())
                    .observeOn(Schedulers.io())
                    .subscribeBy(onNext = f)
        }
    }

    private fun stopUpdatingContactInfo() {
        mRestartUpdatingContactInfoDisposable?.dispose()
        mDoUpdatingContactInfoDisposables.clear()
        mUpdateGroupDisposables.clear()
    }

    private fun startUpdatingContactInfo(startPosition: Int, count: Int, isScrollDown: Boolean) {
        mDoUpdatingContactInfoDisposables += doUpdatingContactInfo(startPosition, count, isScrollDown)
                .subscribeOn(Schedulers.io())
                .subscribeBy(onError = { Log.d(TAG, "Update data with contacts error", it) })
    }

    private fun doUpdatingContactInfo(startPosition: Int, count: Int, isScrollDown: Boolean): Completable {
        return Completable.create {
            val startTime = System.currentTimeMillis()

            val endPosition = startPosition + if (isScrollDown) (count - 1) else -(count - 1)

            debug("callLogTest: update start, range: $startPosition-$endPosition, count: $count, isScrollDown: $isScrollDown)")

            if (isScrollDown) {
                val iterator = mGroupedList.listIterator(startPosition)
                while (iterator.hasNext()) {
                    val index = iterator.nextIndex()
                    val group = iterator.next()

                    if (index > endPosition || it.isDisposed) {
                        break
                    }

                    updateGroup(group, index)
                }
            } else {
                val iterator = mGroupedList.listIterator(startPosition + 1)
                while (iterator.hasPrevious()) {
                    val index = iterator.previousIndex()
                    val group = iterator.previous()

                    if (index < endPosition || it.isDisposed) {
                        break
                    }

                    updateGroup(group, index)
                }
            }

            debug("callLogTest: update end, range: $startPosition-$endPosition, count: $count, time: ${System.currentTimeMillis() - startTime}")

            it.onComplete()
        }
    }

    private fun updateGroup(group: CallLogItemModel, index: Int) {
        if (!group.isGroup || group !is CallLogGroupItem) {
            return
        }
        if (group.isConference() || group.isVoiceMail()) {
            return
        }
        debug("callLogTest: index $index, start")
        val disposable= group.getAssociatedContact()
                .subscribeOn(Schedulers.io())
                .observeOn(AndroidSchedulers.mainThread())
                .subscribeBy(
                        onSuccess = { contact ->
                            debug("callLogTest: index $index, success")
                            group.contactId = contact.id
                            group.buddyKey = contact.buddyKey ?: group.buddyKey
                            group.displayName = contact.displayName ?: group.displayName
                            group.contactPhoto = contact.getPhotoAsMaybe(context).subscribeOn(Schedulers.io()).blockingGet()
                            group.numberType = contact.allPhoneNumbers.find { group.compareNumber(it.number) }?.getSubTypeString(context) ?: group.numberType
                            group.externalId = contact.buddyKey.orEmpty()
                            group.isContactInfoDeprecated = false

                            // Explicit call of method with three arguments to avoid interpret passed index as delay
                            firePresenterEvent(Events.DATA_SET_UPDATED, index, 0)
                        },
                        onError = {
                            debug("callLogTest: index $index, error: $it", true)
                        },
                        onComplete = {
                            debug("callLogTest: index $index, complete")
                            // Contact not found
                            if (group.contactId.isEmpty() && !group.isContactInfoDeprecated) {
                                // Not found at all, ignore
                            } else {
                                // Contact was exist before, but now looks like removed or its number changed. Update with original data
                                val callLogEntity = mCallLogApi.get(group.getEntityId()).blockingGet()
                                group.contactId = ""
                                group.buddyKey = null
                                group.updateDisplayName(callLogEntity?.getDefaultDisplayName())
                                group.contactPhoto = null
                                group.numberType = ""
                                group.externalId = ""

                                // Explicit call of method with three arguments to avoid interpret passed index as delay
                                firePresenterEvent(Events.DATA_SET_UPDATED, index, 0)
                            }
                        }
                )
        mUpdateGroupDisposables.add(index, disposable)
    }

    fun onScrolling(newFirstVisiblePosition: Int, newLastVisiblePosition: Int, newVisibleItemCount: Int, dy: Int) {
        if (dy == 0) { // true when orientation changed or list expanded/collapsed
            if (mOrientation != resources.configuration.orientation) {
                mFirstVisiblePosition = newFirstVisiblePosition
                mLastVisiblePosition = newLastVisiblePosition
                mOrientation = resources.configuration.orientation
                debug("callLogTest: orientation changed: $mOrientation", true)
                restartUpdatingContactInfo()
            } else if (mLastVisiblePosition != newLastVisiblePosition) {
                mFirstVisiblePosition = newFirstVisiblePosition
                mLastVisiblePosition = newLastVisiblePosition
                restartUpdatingContactInfo(400)
            }
            return
        }

        val isScrollDown = dy > 0
        val startUpdatingPosition: Int
        val updatingItemCount: Int
        if (isScrollDown) {
            val diffLast = abs(mLastVisiblePosition - newLastVisiblePosition)
            if (diffLast == 0) {
                return
            }

            startUpdatingPosition = mLastVisiblePosition + 1
            updatingItemCount = diffLast
        } else {
            val diffFirst = abs(mFirstVisiblePosition - newFirstVisiblePosition)
            if (diffFirst == 0) {
                return
            }

            startUpdatingPosition = newFirstVisiblePosition
            updatingItemCount = diffFirst
        }

        mFirstVisiblePosition = newFirstVisiblePosition
        mLastVisiblePosition = newLastVisiblePosition

        startUpdatingContactInfo(startUpdatingPosition, updatingItemCount, isScrollDown)
    }

    fun onScrollStopped(newFirstVisiblePosition: Int, newLastVisiblePosition: Int, newVisibleItemCount: Int) {
        debug("callLogTest: pool size ${mUpdateGroupDisposables.size()}")

        mFirstVisiblePosition = newFirstVisiblePosition
        mLastVisiblePosition = newLastVisiblePosition
    }

    fun onViewRecycled(position: Int) {
        if (position >= 0) {
            debug("callLogTest: index $position, remove")
            mUpdateGroupDisposables.remove(position)
        }
    }

}

private class SparseArrayDisposable : Disposable {
    private val mSparseArray = SparseArray<Disposable>()
    @Volatile private var mDisposed = false

    override fun isDisposed() = mDisposed

    override fun dispose() {
        if (mDisposed) {
            return
        }
        synchronized(mSparseArray) {
            if (mDisposed) {
                return
            }
            clear()
            mDisposed = true
        }
    }

    fun add(key: Int, disposable: Disposable) {
        if (!mDisposed) {
            synchronized(mSparseArray) {
                if (!mDisposed) {
                    mSparseArray.get(key)?.dispose()
                    mSparseArray.put(key, disposable)
                    return
                }
            }
        }
        disposable.dispose()
    }

    fun remove(key: Int) {
        if (mDisposed) {
            return
        }
        synchronized(mSparseArray) {
            if (mDisposed) {
                return
            }
            mSparseArray.get(key)?.dispose()
            mSparseArray.remove(key)
        }
    }

    fun clear() {
        if (mDisposed) {
            return
        }
        synchronized(mSparseArray) {
            if (mDisposed) {
                return
            }
            for (i in 0 until mSparseArray.size()) {
                mSparseArray.valueAt(i)?.dispose()
            }
            mSparseArray.clear()
        }
    }

    fun size(): Int {
        if (mDisposed) {
            return 0
        }
        synchronized(mSparseArray) {
            return if (mDisposed) 0 else mSparseArray.size()
        }
    }
}
