package com.bria.voip.ui.main.calllog.refactoring

import android.app.Dialog
import android.content.Intent
import android.net.Uri
import android.os.Bundle
import android.os.Parcelable
import android.support.annotation.CallSuper
import android.support.annotation.MainThread
import android.support.design.widget.TabLayout
import android.support.v7.app.AlertDialog
import android.support.v7.widget.LinearLayoutManager
import android.support.v7.widget.RecyclerView
import android.widget.TextView
import com.bria.common.uiframework.annotations.Inject
import com.bria.common.uiframework.annotations.Layout
import com.bria.common.uiframework.coloring.Coloring
import com.bria.common.uiframework.presenters.IPresenterEvent
import com.bria.common.uiframework.screens.AbstractScreen
import com.bria.common.util.AndroidUtils
import com.bria.common.util.Log
import com.bria.voip.R
import com.bria.voip.ui.main.calllog.refactoring.DialogId.DELETE_CALL_RECORD_CONFIRM_DIALOG_ID
import com.bria.voip.ui.shared.helpers.setVisible

@Layout(R.layout.calllog_list_screen)
abstract class CallLogAbstractScreen<Presenter : CallLogAbstractPresenter> : AbstractScreen<Presenter>() {
    companion object {
        private const val TAG = "CallLogAbstractScreen"

        internal const val KEY_ITEM_POSITION = "KEY_ITEM_POSITION"

        internal const val KEY_LINEAR_LAYOUT_MANAGER_STATE = "KEY_LINEAR_LAYOUT_MANAGER_STATE"
    }

    @Inject(id = R.id.calllog_list_screen_tabs)
    protected lateinit var mTabLayout: TabLayout

    @Inject(R.id.calllog_list_screen_recycler_view)
    protected lateinit var mRecyclerView: RecyclerView

    @Inject(R.id.calllog_list_screen_empty_view)
    protected lateinit var mEmptyView: TextView

    protected lateinit var mAdapter: CallLogListAdapter
    protected lateinit var mLinearLayoutManager: LinearLayoutManager

    @CallSuper
    override fun onPresenterEvent(event: IPresenterEvent<*, *>) {
        when (event.type) {
            CallLogAbstractPresenter.Events.DATA_SET_CHANGED -> handleDataSetChanged()
            CallLogAbstractPresenter.Events.DATA_SET_UPDATED -> mAdapter.notifyDataChanged(event.data as Int, Unit) // Unit is a mark to disable animation

            CallLogAbstractPresenter.Events.SHOW_MESSAGE_LONG -> toastLong(event.data as String)
            CallLogAbstractPresenter.Events.SHOW_DELETE_CALL_RECORD_CONFIRMATION -> showDeleteCallRecordConfirmationDialog(event.data as Int)
            CallLogAbstractPresenter.Events.SHOW_PLAY_RECORD_VIEW -> showPlayRecordActivity(event.data as Uri)
            CallLogAbstractPresenter.Events.SHOW_SHARE_RECORD_VIEW -> (event.data as Pair<Uri, String>).let { showShareRecordActivity(it.first, it.second) }

            CallLogAbstractPresenter.Events.UPDATE_SELECTED_TAB -> updateSelectedTab()
        }
    }

    @MainThread
    override fun onCreate(bundle: Bundle?) {
        super.onCreate(bundle)
        presenter.isInTabletMode = isInTabletMode
        mAdapter = CallLogListAdapter(mRecyclerView, presenter)
        mLinearLayoutManager = LinearLayoutManager(activity)
        mRecyclerView.adapter = mAdapter
        mRecyclerView.layoutManager = mLinearLayoutManager

        setupTabs()

        doRestoreState(bundle)
    }

    @MainThread
    override fun onStart(bundle: Bundle?) {
        super.onStart(bundle)
        presenter.subscribe(this)
        mRecyclerView.addOnScrollListener(mOnScrollListener)
        mRecyclerView.setRecyclerListener { presenter.onViewRecycled(it.adapterPosition) }
    }

    @MainThread
    override fun onNewConfig(bundle: Bundle?) {
        super.onNewConfig(bundle)
        presenter.onScreenNewConfig(bundle)
    }

    @MainThread
    override fun onStop(finishing: Boolean) {
        super.onStop(finishing)
        presenter.unsubscribe()
        mRecyclerView.removeOnScrollListener(mOnScrollListener)
    }

    @MainThread
    override fun onDestroy(finishing: Boolean) {
        super.onDestroy(finishing)
    }

    @MainThread
    override fun onSaveState(stateBundle: Bundle) {
        stateBundle.putParcelable(KEY_LINEAR_LAYOUT_MANAGER_STATE, mLinearLayoutManager.onSaveInstanceState())
        super.onSaveState(stateBundle)
    }

    @MainThread
    override fun onRestoreState(stateBundle: Bundle) {
        super.onRestoreState(stateBundle)
        doRestoreState(stateBundle)
    }

    @CallSuper
    protected open fun doRestoreState(stateBundle: Bundle?) {
        mLinearLayoutManager.onRestoreInstanceState(stateBundle?.getParcelable(KEY_LINEAR_LAYOUT_MANAGER_STATE))
    }

    private fun saveLayoutState() {
        save(presenter.filterType.name, mLinearLayoutManager.onSaveInstanceState())
    }

    private fun restoreLayoutState() {
        restore(presenter.filterType.name)?.let {
            save(presenter.filterType.name, null) // reset saved state
            mLinearLayoutManager.onRestoreInstanceState(it as Parcelable)
        }
    }

    override fun getTitle() = presenter.getTitle()

    private fun handleDataSetChanged() {
        mAdapter.notifyDataChanged()

        restoreLayoutState()

        mEmptyView.setVisible(presenter.itemsCount == 0)
        mRecyclerView.setVisible(presenter.itemsCount != 0)

        invalidateMenu()
    }

    private fun showDeleteCallRecordConfirmationDialog(position: Int) {
        showDialog(DELETE_CALL_RECORD_CONFIRM_DIALOG_ID, Bundle(1).apply { putInt(KEY_ITEM_POSITION, position) })
    }

    private fun showPlayRecordActivity(recordFileUri: Uri) {
        var intent = Intent(Intent.ACTION_VIEW)
        intent.setDataAndType(recordFileUri, "audio/*")
        intent.addFlags(Intent.FLAG_GRANT_READ_URI_PERMISSION)
        intent = Intent.createChooser(intent, getString(R.string.tIntentChooserTitle))
        if (!AndroidUtils.canHandleIntent(intent, activity)) {
            Log.w(TAG, "Can't handle intent $intent")
            return
        }
        activity.startActivity(intent)
    }

    private fun showShareRecordActivity(recordFileUri: Uri, message: String) {
        var intent = Intent(Intent.ACTION_SEND)
        intent.addFlags(Intent.FLAG_GRANT_READ_URI_PERMISSION)
        intent.type = "audio/*"
        intent.putExtra(Intent.EXTRA_STREAM, recordFileUri)
        intent.putExtra(Intent.EXTRA_SUBJECT, getString(R.string.tCallRecordingShareSubject))
        intent.putExtra(Intent.EXTRA_TEXT, message)
        intent = Intent.createChooser(intent, getString(R.string.tCallRecordingShareChooserTitle))
        if (!AndroidUtils.canHandleIntent(intent, activity)) {
            Log.w(TAG, "Can't handle intent $intent")
            return
        }
        activity.startActivity(intent)
    }

    override fun createDialog(which: Int, data: Bundle?): Dialog? {
        when(which) {
            DELETE_CALL_RECORD_CONFIRM_DIALOG_ID ->
                return AlertDialog.Builder(activity)
                        .setMessage(R.string.tAreYouSureToDeleteCallRecording)
                        .setPositiveButton(R.string.tYes) { _, _ -> presenter.onDeleteCallRecordConfirmed(data?.getInt(KEY_ITEM_POSITION) ?: -1) }
                        .setNegativeButton(R.string.tNo, null)
                        .setCancelable(true)
                        .create()
        }

        return super.createDialog(which, data)
    }

    private fun setupTabs() {
        if (!presenter.isTabsVisible) {
            mTabLayout.setVisible(false)
            return
        }

        mTabLayout.removeAllTabs()
        mTabLayout.clearOnTabSelectedListeners()

        mTabLayout.addTab(mTabLayout.newTab().setText(CallLogFilterType.ALL.nameResId).setTag(CallLogFilterType.ALL), false)
        mTabLayout.addTab(mTabLayout.newTab().setText(CallLogFilterType.MISSED.nameResId).setTag(CallLogFilterType.MISSED), false)

        if (presenter.isConferenceTabVisible()) {
            mTabLayout.addTab(mTabLayout.newTab().setText(CallLogFilterType.CONFERENCE.nameResId).setTag(CallLogFilterType.CONFERENCE), false)
        }

        Coloring.get().getCustomizer(TabLayout::class.java).apply { setTarget(mTabLayout) }.applyChanges()

        updateSelectedTab() // set initial selection

        mTabLayout.addOnTabSelectedListener(object : TabLayout.OnTabSelectedListener {
            override fun onTabSelected(tab: TabLayout.Tab) {
                tab.tag?.let { tag ->
                    mRecyclerView.stopScroll()
                    saveLayoutState()
                    presenter.filterType = tag as CallLogFilterType
                }
            }

            override fun onTabUnselected(tab: TabLayout.Tab) {}
            override fun onTabReselected(tab: TabLayout.Tab) {}
        })
    }

    private fun updateSelectedTab() {
        if (presenter.isTabsVisible) {
            repeat(mTabLayout.tabCount) { index ->
                mTabLayout.getTabAt(index)?.let { tab ->
                    if (tab.tag == presenter.filterType) {
                        tab.select()
                    }
                }
            }
        }
    }

    private val mOnScrollListener = object : RecyclerView.OnScrollListener() {
        override fun onScrollStateChanged(recyclerView: RecyclerView, newState: Int) {
            super.onScrollStateChanged(recyclerView, newState)
            when (newState) {
                RecyclerView.SCROLL_STATE_IDLE -> {
                    presenter.onScrollStopped(
                            mLinearLayoutManager.findFirstVisibleItemPosition(),
                            mLinearLayoutManager.findLastVisibleItemPosition(),
                            mLinearLayoutManager.childCount
                    )
                }
            }
        }

        override fun onScrolled(recyclerView: RecyclerView, dx: Int, dy: Int) {
            super.onScrolled(recyclerView, dx, dy)
            presenter.onScrolling(
                    mLinearLayoutManager.findFirstVisibleItemPosition(),
                    mLinearLayoutManager.findLastVisibleItemPosition(),
                    mLinearLayoutManager.childCount,
                    dy
            )
        }
    }
}
