package com.bria.voip.ui.main.calllog.refactoring

import android.content.Context
import android.provider.ContactsContract
import android.text.format.DateUtils
import com.bria.common.controller.Controllers
import com.bria.common.controller.calllog.CallLogUtils
import com.bria.common.controller.calllog.db.CallType
import com.bria.common.controller.contacts.local.Contact
import com.bria.common.controller.contacts.local.data.PhoneNumber
import com.bria.common.controller.settings.ESetting
import com.bria.common.util.*
import com.bria.voip.R
import com.bria.voip.ui.main.calllog.refactoring.CallLogGroupItem.Companion.TAG
import io.reactivex.Maybe

// Group item

fun CallLogGroupItem.compareNumber(other: CallLogGroupItem) = compareNumber(other.getNumberWithDomain())

fun CallLogGroupItem.compareNumber(otherNumber: String) = CallLogUtils.compareNumbers(getNumberWithDomain(), otherNumber)

fun CallLogGroupItem.isAnonymous() = Validator.isAnonymousAddress(getNumberWithDomain())

fun CallLogGroupItem.isVoiceMail() = callTypes.first() == CallType.VOICE_MAIL

fun CallLogGroupItem.isConference() = callTypes.first() == CallType.CONFERENCE

fun CallLogGroupItem.getAssociatedContact(): Maybe<Contact> {
    val fetchContactByNumber = Controllers.get().contacts.contactsApi.contact.byNumber(getNumberWithDomain())
            .withPhones()
            .withSoftPhones()
            .asMaybe()
            .doOnSubscribe { Log.inDebug(TAG, "callLogTest: fetch contact by number ${getNumberWithDomain()}") }

    val getContactById = Controllers.get().contacts.contactsApi.contact.byId(contactId)
            .withPhones()
            .withSoftPhones()
            .asMaybe()
            .doOnSubscribe { Log.inDebug(TAG, "callLogTest: get contact by ID $contactId") }

    // When contacts info is deprecated we should try to fetch contact by number only,
    // since contact ID can be point to deprecated data (contact can be removed or its number can be changed)
    return if (isContactInfoDeprecated) fetchContactByNumber else getContactById.switchIfEmpty(Maybe.defer { fetchContactByNumber })
}

// Useful for tests
//fun CallLogGroupItem.getAssociatedContact(): Maybe<Contact> {
//    return Maybe.fromCallable {
//        //Thread.sleep(1000)
//        Contact(Contact.Type.LOCAL, id).apply { displayName = "display name $id" }
//    }
//}

fun CallLogGroupItem.getNumberWithDomain() = SipAddressUtils.getAddress(number, getDomain())

fun CallLogGroupItem.getDomain() = SipAddressUtils.removePort(remoteDomain.takeIf { it.isNotEmpty() } ?: accountDomain)

fun CallLogGroupItem.getNumberForContacts() =
        if (Controllers.get().settings.getBool(ESetting.RemoveFullyQualifiedNameFromContacts) || Validator.isNumeric(number)) {
            number
        } else {
            getNumberWithDomain()
        }

fun CallLogGroupItem.getPhoneTypeForContacts() =
        if (Controllers.get().settings.getBool(ESetting.HaveSoftphoneType) && !Validator.isNumeric(getNumberForContacts())) {
            PhoneNumber.SOFTPHONE_CODE
        } else {
            ContactsContract.CommonDataKinds.Phone.TYPE_MOBILE
        }

fun CallLogGroupItem.isLocalContactExists() =
        contactId.isNotEmpty() &&
        contactId != GlobalConstants.CALL_LOG_GENBAND_FRIEND_ID &&
        contactId != GlobalConstants.CALL_LOG_LDAP_CONTACT &&
        contactId != GlobalConstants.CALL_LOG_XMPP_BUDDY_ID &&
        contactId != GlobalConstants.CALL_LOG_VCCS_CONFERENCE

// Group Entity ID its ID of first call log entity in group
fun CallLogGroupItem.getEntityId() = items.last().id.toLong() // Not empty list expected here, so let it crash otherwise

fun CallLogGroupItem.getCallInfo() = if (isExpanded) getNumberInfo() else getNumberTypeAndDate()

fun CallLogGroupItem.isCallInfoVisible() = !isExpanded || (displayName != getNumberForContacts() && !isConference())

fun CallLogGroupItem.isCallTypeIconsVisible() = !isExpanded

fun CallLogGroupItem.isRecordIndicatorVisible() = items.find { it.isRecordInfoVisible } != null

fun CallLogGroupItem.updateDisplayName(defaultDisplayName: String? = null) {
    displayName = getRemoteName() ?: defaultDisplayName ?: displayName
}

fun CallLogGroupItem.getRemoteName() = items.find { it.remoteName.isNotBlank() }?.remoteName

private fun CallLogGroupItem.getNumberInfo() = "$numberType ${getNumberForContacts()}".trim()

private fun CallLogGroupItem.getNumberTypeAndDate(): String {
    val relativeTime = DateUtils.getRelativeTimeSpanString(date, System.currentTimeMillis(),
            DateUtils.MINUTE_IN_MILLIS, DateUtils.FORMAT_ABBREV_RELATIVE).toString()
    val typeAndDate = if (numberType.isNotEmpty()) "$numberType, $relativeTime" else relativeTime
    return if (items.size > callTypes.size) "(${items.size}) $typeAndDate" else typeAndDate
}

// Content item

fun CallLogContentItem.getCallTypeIconResId() = when (callType) {
    CallType.INCOMING -> com.bria.voip.R.drawable.call_incoming
    CallType.OUTGOING -> com.bria.voip.R.drawable.call_outgoing
    CallType.MISSED -> com.bria.voip.R.drawable.call_missed
    CallType.VOICE_MAIL -> com.bria.voip.R.drawable.call_outgoing
    CallType.CONFERENCE -> com.bria.voip.R.drawable.call_outgoing
}

fun CallLogContentItem.getCallInfo() = when (callType) {
    CallType.INCOMING -> R.string.comm_log_incoming_call
    CallType.OUTGOING -> R.string.comm_log_outgoing_call
    CallType.MISSED -> R.string.comm_log_missed_call
    CallType.VOICE_MAIL -> R.string.comm_log_outgoing_call
    CallType.CONFERENCE -> if (hostedCollab) R.string.comm_log_vccs_hosted_call else R.string.comm_log_vccs_call
}

fun CallLogContentItem.getCallDate(context: Context) = DateUtils.formatDateTime(context, date,
        DateUtils.FORMAT_SHOW_TIME or DateUtils.FORMAT_SHOW_DATE or DateUtils.FORMAT_SHOW_YEAR or DateUtils.FORMAT_SHOW_WEEKDAY)!!
        .capitalize()

fun CallLogContentItem.getCallDuration(context: Context): String {
    val minutes = durationSeconds / 60
    val seconds = durationSeconds % 60
    val minutesString = context.getString(if (minutes <= 1) R.string.logMinuteSingle else R.string.logMinutePlural, minutes)
    val secondsString = context.getString(if (seconds <= 1) R.string.logSecondSingle else R.string.logSecondPlural, seconds)
    return if (minutes > 0) "$minutesString $secondsString" else secondsString
}

fun CallLogContentItem.getCallForwardedInfo(context: Context) = "${context.getString(R.string.tCallLogForwardedTo)} $forwardedTo"

fun CallLogContentItem.getRecordInfo(context: Context) = formatRecordDuration(context).let { if (it.isEmpty()) recordFileName else "$recordFileName ($it)" }

private fun CallLogContentItem.formatRecordDuration(context: Context): String {
    return if (recordFileName.isNotEmpty()) {
        try {
            val durationMillis = RecordingUtils.getDurationMillis(context, recordFileName)
            val durationSeconds = durationMillis / 1000
            val minutes = durationSeconds / 60
            val seconds = durationSeconds % 60
            "$minutes:$seconds"
        } catch (e: Exception) {
            ""
        }
    } else {
        ""
    }
}

fun CallLogGroupItem.getPresenceResId() = buddyKey?.let { Controllers.get().buddy.getBuddy(it)?.presence?.status?.iconResourceId } ?: 0
