package com.bria.voip.ui.main.calllog.refactoring

import android.net.Uri
import android.os.Bundle
import com.bria.common.R
import com.bria.common.analytics.BIAnalytics
import com.bria.common.controller.ClientConfig
import com.bria.common.controller.accounts.Account
import com.bria.common.controller.accounts.DialPlanHelper
import com.bria.common.controller.calllog.compareNumber
import com.bria.common.controller.calllog.db.CallLogEntity
import com.bria.common.controller.calllog.db.CallType
import com.bria.common.controller.collaboration.rx.impl.ParticipantManager
import com.bria.common.controller.commlog.local.LocalCallLogController
import com.bria.common.controller.contacts.local.Contact
import com.bria.common.controller.contacts.local.data.PhoneNumber
import com.bria.common.controller.im.ImpsUtils
import com.bria.common.controller.im.refactoring.BuddyKeyUtils
import com.bria.common.controller.im.refactoring.db.entities.ImConversationData
import com.bria.common.controller.license.features.EFeature
import com.bria.common.controller.phone.IPhoneCtrlEvents
import com.bria.common.controller.settings.EAccountSetting
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.branding.EAccountType
import com.bria.common.controller.settings.branding.EGuiVisibility
import com.bria.common.controller.settings.gui.EGuiElement
import com.bria.common.sdkwrapper.CallData
import com.bria.common.uiframework.presenters.IPresenterEventTypeEnum
import com.bria.common.util.*
import com.bria.voip.service.BriaVoipService
import com.bria.voip.ui.main.contacts.ContactEditPresenter
import com.bria.voip.ui.main.contacts.PhoneNumberActionsScreen
import io.reactivex.Completable
import io.reactivex.Observable
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.rxkotlin.plusAssign
import io.reactivex.rxkotlin.subscribeBy
import io.reactivex.schedulers.Schedulers
import java.util.*
import java.util.concurrent.TimeUnit

open class CallLogListPresenter : CallLogAbstractPresenter() {
    companion object {
        private const val TAG = "CallLogListPresenter"
        private const val MAX_CALL_TYPE_ICONS = 4
    }

    enum class Events : IPresenterEventTypeEnum {
        DATA_SET_COLLAPSED,
        DATA_SET_EXPANDED,
        SHOW_CALL_REMINDER,
        SHOW_CONFERENCE_CALL_CONFIRMATION,
        SHOW_HOSTED_CONFERENCE_CALL_CONFIRMATION,
        SHOW_WEB_VIEW,
        SHOW_PHONE_NUMBER_ACTION_PICKER,
        SHOW_CONTACT_PICKER,
        SHOW_CONTACT_EDIT,
        SHOW_CONTACT_DETAILS,
        SHOW_BUDDY_DETAILS,
        SHOW_CONVERSATION,
        SHOW_CALL_LOG_DETAILS,
        // define new events here
        ;
        override fun getType(): IPresenterEventTypeEnum = this
    }

    private enum class CallAction {
        AudioCall,
        PrefixCall,
        VideoCall
    }

    private val mCompositeDisposable = CompositeDisposable()

    // Life-cycle

    override fun onDestroy() {
        super.onDestroy()
        mCompositeDisposable.dispose()
    }

    override fun onSubscribe() {
        super.onSubscribe()
        if (isInTabletMode) {
            mCompositeDisposable += Observable.timer(100, TimeUnit.MILLISECONDS, Schedulers.computation())
                    .subscribeBy(
                            onNext = { setSelectedGroup(getSelectedPosition().takeUnless { pos -> pos == -1 } ?: 0) },
                            onError = { Log.e(TAG, "Update/set selection on subscription error", it) }
                    )
        }
    }

    // = PUBLIC section

    override fun getTitle() = getString(R.string.tCallLog)

    override fun onScreenNewConfig(bundle: Bundle?) = handleCallReminder(bundle)

    // Menu settings

    fun isMenuDeleteAllEnabled() = itemsCount != 0

    fun isMenuDeleteMissedEnabled() = itemsCount != 0

    fun isMenuDeleteConferenceEnabled() = itemsCount != 0 && mControllers.collaboration.isCollabAvailable

    fun isMenuDeleteEnabled() = itemsCount != 0

    fun isMenuCallEnabled(position: Int) = getGroupItemAt(position)?.isAnonymous() == false

    open fun isMenuAddContactEnabled(position: Int): Boolean {
        val group = getGroupItemAt(position) ?: return false
        return !group.isAnonymous() &&
                !group.isConference() &&
                !group.isVoiceMail() &&
                !group.getNumberForContacts().isEmpty() &&
                !group.isLocalContactExists()
    }

    fun isMenuViewContactEnabled(position: Int) = getGroupItemAt(position)?.isLocalContactExists() == true

    open fun isMenuUpdateContactEnabled(position: Int) = isMenuAddContactEnabled(position) && mContactsApi.count().onErrorReturnItem(0).blockingGet() > 0

    fun isMenuSendSmsEnabled(position: Int): Boolean {
        val group = getGroupItemAt(position) ?: return false

        if (group.isAnonymous()) {
            return false
        }

        val account = mControllers.accounts.getAccount(group.accountId) ?: return false

        val featureImps = mControllers.license.checkFeature(EFeature.IMPS)
        val sms = mControllers.settings.getBool(ESetting.Sms) && account.getBool(EAccountSetting.IsSMS)
        val listOfSmsApiAccounts = mControllers.accounts.getAccounts { it.type == EAccountType.SmsApi && it.getBool(EAccountSetting.IsSMS) }
        val smsApi = mControllers.settings.getBool(ESetting.FeatureSMSApi) && listOfSmsApiAccounts.size > 0

        return featureImps && (sms || smsApi)
    }

    open fun isMenuSendImEnabled(position: Int): Boolean {
        val group = getGroupItemAt(position) ?: return false

        if (group.isAnonymous()) {
            return false
        }

        if (!group.isLocalContactExists()) {
            return false
        }

        val account = mControllers.accounts.getAccount(group.accountId) ?: return false

        val featureImps = mControllers.license.checkFeature(EFeature.IMPS)
        val im = mControllers.settings.getBool(ESetting.ImPresence) &&
                (account.type == EAccountType.Xmpp || account.getBool(EAccountSetting.IsIMPresence))
        val voiceMail = group.isVoiceMail() && group.number == SipAddressUtils.removeDomain(account.getStr(EAccountSetting.VMNumber))

        return featureImps && im && !voiceMail
    }

    open fun isMenuVideoCallEnabled(position: Int): Boolean {
        if (!isMenuCallEnabled(position)) {
            return false
        }

        val group = getGroupItemAt(position) ?: return false
        val account = mControllers.accounts.getAccount(group.accountId) ?: mControllers.accounts.primaryAccount ?: return false
        val voiceMail = group.isVoiceMail() && group.number == SipAddressUtils.removeDomain(account.getStr(EAccountSetting.VMNumber))

        return mControllers.accounts.isVideoEnabled(account) && !account.getBool(EAccountSetting.AlwaysOfferVideo) && !voiceMail
    }

    open fun isMenuTransferCallEnabled(position: Int): Boolean {
        if (ClientConfig.get().getGuiVisibility(EGuiElement.CallControlTransfer) == EGuiVisibility.Hidden) {
            return false
        }

        val group = getGroupItemAt(position) ?: return false

        if (mControllers.collaboration.hasConferenceCall() || group.isConference()) {
            return false
        }

        if (group.isAnonymous()) {
            return false
        }

        if (mControllers.phone.callCount != 1) {
            return false
        }

        if (mControllers.phone.phoneState != IPhoneCtrlEvents.EPhoneState.eInCall) {
            return false
        }

        return true
    }

    fun isMenuPrefixCallEnabled(position: Int) = mControllers.phone.isPrefixCallingEnabled && isMenuCallEnabled(position)

    fun isMenuConferenceCallEnabled(position: Int) = mControllers.collaboration.isCollabAvailable && !isMenuHostedConferenceCallEnabled(position)

    fun isMenuHostedConferenceCallEnabled(position: Int) = mControllers.collaboration.isCollabAvailable && (getGroupItemAt(position)?.hostedCollab ?: false)

    fun isMenuOnlyConferenceRelatedActionsAllowed(position: Int) = getGroupItemAt(position)?.isConference() == true

    // Menu actions

    fun onMenuDeleteAll() {
        mCompositeDisposable += mCallLogApi.deleteAll().subscribeBy(onError = { Log.w(TAG, "DELETE ALL entities from DB error", it) })
    }

    fun onMenuDeleteMissed() {
        mCompositeDisposable += mCallLogApi.delete(CallType.MISSED).subscribeBy(onError = { Log.w(TAG, "DELETE MISSED entities from DB error", it) })
    }

    fun onMenuDeleteConference() {
        mCompositeDisposable += mCallLogApi.delete(CallType.CONFERENCE).subscribeBy(onError = { Log.w(TAG, "DELETE CONFERENCE entities from DB error", it) })
    }

    fun getContextMenuTitle(position: Int) : String {
        return getGroupItemAt(position)?.displayName ?: ""
    }

    fun onMenuDelete(position: Int) {
        val group = getGroupItemAt(position) ?: return
        val entities = group.items.map { item ->
            CallLogEntity(
                    id = item.id.toLong(),
                    callType = item.callType, // required for BW
                    externalId = item.externalId) // required for BW
        }
        val recordFileNames = group.items.asSequence().filter { it.isRecordInfoVisible }.map { it.recordFileName }.toList()
        mCompositeDisposable += mCallLogApi.delete(*entities.toTypedArray()).subscribeBy(onError = { Log.w(TAG, "DELETE entities from DB error", it) })
        mCompositeDisposable += Completable.fromAction { recordFileNames.forEach { RecordingUtils.deleteFile(context, it) } }
                .subscribeOn(Schedulers.io()).subscribeBy(onError = { Log.w(TAG, "DELETE recording files from storage error", it) })
    }

    fun onMenuCall(position: Int) {
        callNumber(getGroupItemAt(position), CallAction.AudioCall)
        BIAnalytics.get().reportUIClick(BIAnalytics.DIAL_FROM_CALL_LOG_LISTING)
    }

    fun onMenuAddContact(position: Int) {
        val group = getGroupItemAt(position) ?: return

        val phoneNumber = group.getNumberForContacts()
        val phoneType = group.getPhoneTypeForContacts()
        val displayName = group.getRemoteName() ?: ""

        val bundle = Bundle(5)
        bundle.putSerializable(ContactEditPresenter.KEY_SCREEN_ACTION, ContactEditPresenter.ACTION_ADD)
        bundle.putSerializable(GlobalConstants.KEY_CONTACT_TYPE, GlobalConstants.EContactsFilterType.CONTACTS)
        bundle.putString(ContactEditPresenter.KEY_CONTACT_PHONE_NUMBER, phoneNumber)
        bundle.putInt(ContactEditPresenter.KEY_PHONE_TYPE, phoneType)
        bundle.putString(ContactEditPresenter.KEY_DISPLAY_NAME, displayName)

        firePresenterEvent(Events.SHOW_CONTACT_EDIT, bundle)
    }

    open fun onMenuViewContact(position: Int) {
        val group = getGroupItemAt(position) ?: return

        val contactId = group.contactId
        if (contactId == GlobalConstants.CALL_LOG_XMPP_BUDDY_ID) {
            val xmppBuddy = mControllers.buddy.getXmppBuddyByPhoneNumber(group.number) ?: return
            val bundle = Bundle(1)
            bundle.putString(GlobalConstants.KEY_BUDDY_ID, ImpsUtils.getBuddyKey(xmppBuddy))

            firePresenterEvent(Events.SHOW_BUDDY_DETAILS, bundle)
        } else {
            val bundle = Bundle(2)
            bundle.putSerializable(GlobalConstants.KEY_CONTACT_TYPE, GlobalConstants.EContactsFilterType.CONTACTS)
            bundle.putString(GlobalConstants.KEY_CONTACT_ID, contactId)

            firePresenterEvent(Events.SHOW_CONTACT_DETAILS, bundle)
        }
    }

    fun onMenuUpdateContact(position: Int) =
            // Explicit call of method with three arguments to avoid interpret passed int value as delay
            firePresenterEvent(Events.SHOW_CONTACT_PICKER, position, 0)

    fun onMenuTransferCall(position: Int) {
        val group = getGroupItemAt(position) ?: return

        if (mControllers.phone.callCount == 1) {
            val call = mControllers.phone.callListCopy.first()
            if (call.isTransferPossible) {
                val accountNickname = mControllers.accounts.getAccount(group.accountId)?.getStr(EAccountSetting.Nickname) ?: ""
                mControllers.phone.transfer(call.callId, group.getNumberWithDomain(), accountNickname, false)
            } else {
                firePresenterEvent(CallLogAbstractPresenter.Events.SHOW_MESSAGE_LONG, getString(R.string.tPhoneTabTransferNotPossible))
            }
        }
    }

    fun onMenuVideoCall(position: Int) = callNumber(getGroupItemAt(position), CallAction.VideoCall)

    fun onMenuSendSms(position: Int) {
        val group = getGroupItemAt(position) ?: return

        var account: Account? = mControllers.accounts.primaryAccount

        if (group.isLocalContactExists()) {
            account = mControllers.accounts.getAccount(group.accountId)
            if (account?.getBool(EAccountSetting.IsSMS) != true) {
                account = mControllers.accounts.primaryAccount
            }
        }

        val listOfSmsApiAccounts = mControllers.accounts.getAccounts { it.isEnabled && it.type == EAccountType.SmsApi && it.getBool(EAccountSetting.IsSMS) }
        if (mControllers.settings.getBool(ESetting.FeatureSMSApi) && listOfSmsApiAccounts.isNotEmpty()) {
            account = listOfSmsApiAccounts.first()
        }

        if (account == null || !account.isRegistered || !account.getBool(EAccountSetting.IsSMS)) {
            firePresenterEvent(CallLogAbstractPresenter.Events.SHOW_MESSAGE_LONG, getString(R.string.tNoBuddiesToastAccountsSMS))
            return
        }

        val accountId = account.getStr(EAccountSetting.Nickname)
        var number = group.getNumberForContacts()
        var conversationType = ImConversationData.SMS

        if (account.type == EAccountType.SmsApi) {
            number = DialPlanHelper.applyDialPlan(number, account)
            conversationType = ImConversationData.SMS_API
        }

        number = ImpsUtils.getSanitizedSMSNumber(
                PhoneNumber(PhoneNumber.MainType.PHONE_NUMBER, group.getPhoneTypeForContacts(), number),
                account)

        val numbers = ArrayList<String>().apply { add(number) }

        val bundle = Bundle(3)
        bundle.putString(GlobalConstants.KEY_CHAT_ACCOUNT, accountId)
        bundle.putInt(GlobalConstants.KEY_CONVERSATION_TYPE, conversationType)
        bundle.putStringArrayList(GlobalConstants.KEY_CHAT_SMS_NUMBERS, numbers)

        firePresenterEvent(Events.SHOW_CONVERSATION, bundle)
    }

    fun onMenuSendIm(position: Int) {
        val group = getGroupItemAt(position) ?: return

        if (!group.isLocalContactExists()) {
            return
        }

        var account: Account? = mControllers.accounts.getAccount(group.accountId)

        // If buddy's corresponding account is deleted or unregistered try to get other registered accounts
        // with enabled IM on same domain (Bria iPhone default behavior)
        // If there are no appropriate accounts error message will be shown (account will remain null or unregistered)
        if (account == null || !account.isRegistered) {
            val accountsForDomain = mControllers.accounts.getActiveImAccountsForDomain(group.accountDomain)
            if (accountsForDomain.isNotEmpty()) {
                account = accountsForDomain.first()
            }
        }

        if (account == null || !account.isRegistered || !account.getBool(EAccountSetting.IsIMPresence)) {
            firePresenterEvent(CallLogAbstractPresenter.Events.SHOW_MESSAGE_LONG, getString(R.string.tNoBuddiesToastAccountsIM))
            return
        }

        var remoteAddress = group.getNumberForContacts()
        if (!remoteAddress.contains("@")) {
            remoteAddress = "$remoteAddress@${account.getStr(EAccountSetting.Domain)}"
        }

        val buddyKey = BuddyKeyUtils.getNewBuddyKey(account.type, SipAddressUtils.getAddress(account), remoteAddress)
        val buddyKeyList = ArrayList<String>().apply { add(buddyKey) }
        val conversationType = if (account.type == EAccountType.Xmpp) ImConversationData.IM_XMPP else ImConversationData.IM_SIP

        val bundle = Bundle(2)
        bundle.putStringArrayList(GlobalConstants.KEY_BUDDY_IDS, buddyKeyList)
        bundle.putSerializable(GlobalConstants.KEY_CONVERSATION_TYPE, conversationType)

        firePresenterEvent(Events.SHOW_CONVERSATION, bundle)
    }

    fun onMenuPrefixCall(position: Int) = callNumber(getGroupItemAt(position), CallAction.PrefixCall)

    fun onMenuConferenceCall(position: Int) {
        val group = getGroupItemAt(position) ?: return

        val vccsLink = group.number

        if (mControllers.collaboration.hasConferenceCall() &&  mControllers.collaboration.equalsVccsLink(vccsLink)) {
            BriaVoipService.get()!!.showCallUi(GlobalConstants.COLLABORATION_CALL_ORIGIN)
        } else {
            firePresenterEvent(Events.SHOW_WEB_VIEW, vccsLink)
        }
    }

    fun onMenuHostedConferenceCall(position: Int) {
        if (ParticipantManager.isCurrentUserModerator()) {
            BriaVoipService.get()!!.showCallUi(GlobalConstants.COLLABORATION_CALL_ORIGIN)
        } else {
            mControllers.collaboration.startCollaboration(null)
        }
    }

    fun onHostedConferenceCallConfirmed(position: Int) = mControllers.collaboration.startCollaboration(null)

    fun onConferenceCallConfirmed(position: Int) = mControllers.collaboration.startCollaboration(Uri.parse(getGroupItemAt(position)?.number))

    fun onContactForUpdateSelected(contactId: String, position: Int) {
        val group = getGroupItemAt(position) ?: return

        val phoneNumber = group.getNumberForContacts()
        val phoneType = group.getPhoneTypeForContacts()

        val bundle = Bundle(4)
        bundle.putSerializable(ContactEditPresenter.KEY_SCREEN_ACTION, ContactEditPresenter.ACTION_EDIT)
        bundle.putString(GlobalConstants.KEY_CONTACT_ID, contactId)
        bundle.putString(ContactEditPresenter.KEY_CONTACT_PHONE_NUMBER, phoneNumber)
        bundle.putInt(ContactEditPresenter.KEY_PHONE_TYPE, phoneType)

        firePresenterEvent(Events.SHOW_CONTACT_EDIT, bundle)
    }

    // Item actions

    fun onItemClicked(position: Int) {
        if (isGroup(position)) {
            if (isInTabletMode) {
                setSelectedGroup(position)
            } else {
                toggleExpandedGroup(position)
            }
        }
    }

    fun onItemCallButtonClicked(position: Int) {
        val group = getGroupItemAt(position) ?: return

        if (group.isConference()) {
            if (group.hostedCollab) {
                if (mControllers.collaboration.hasConferenceCall()) {
                    // Explicit call of method with three arguments to avoid interpret passed int value as delay
                    firePresenterEvent(Events.SHOW_HOSTED_CONFERENCE_CALL_CONFIRMATION, position, 0)
                } else {
                    mControllers.collaboration.startCollaboration(null)
                }
            } else {
                if (mControllers.collaboration.hasConferenceCall()) {
                    // Explicit call of method with three arguments to avoid interpret passed int value as delay
                    firePresenterEvent(Events.SHOW_CONFERENCE_CALL_CONFIRMATION, position, 0)
                } else {
                    firePresenterEvent(Events.SHOW_WEB_VIEW, group.number)
                }
            }
        } else {
            if (mControllers.settings.getBool(ESetting.SingleTouchtoCall) &&
                    !mControllers.phone.isCallbackEnabled &&
                    !isMenuVideoCallEnabled(position) &&
                    !isMenuPrefixCallEnabled(position) &&
                    !isMenuTransferCallEnabled(position)) {
                callNumber(group, CallAction.AudioCall)
            } else {
                val account = mControllers.accounts.primaryAccount
                if (account == null || !account.accountStatus.isRegistered) {
                    firePresenterEvent(CallLogAbstractPresenter.Events.SHOW_MESSAGE_LONG, getString(R.string.tNoActiveAccount))
                    return
                }

                val contact: Contact? = group.getAssociatedContact().subscribeOn(Schedulers.io()).blockingGet()
                val phoneNumber =
                        contact?.allPhoneNumbers?.find { group.compareNumber(it.number) } ?:
                        PhoneNumber(PhoneNumber.MainType.PHONE_NUMBER, group.getPhoneTypeForContacts(), group.getNumberForContacts())

                val bundle = Bundle(4)
                bundle.putInt(PhoneNumberActionsScreen.KEY_SELECTED_ACCOUNT, account.id)
                bundle.putSerializable(PhoneNumberActionsScreen.KEY_CONTACT_TYPE, contact?.type)
                bundle.putString(PhoneNumberActionsScreen.KEY_CONTACT_NAME, contact?.displayName.orEmpty())
                bundle.putParcelable(PhoneNumberActionsScreen.KEY_PHONE_NUMBER, phoneNumber)

                firePresenterEvent(Events.SHOW_PHONE_NUMBER_ACTION_PICKER, bundle)
            }
        }
    }

    fun onItemContactButtonClicked(position: Int) {
        if (isMenuViewContactEnabled(position) || isMenuAddContactEnabled(position)) {
            val group = getGroupItemAt(position) ?: return
            if (group.contactId.isNotEmpty()) {
                onMenuViewContact(position)
            } else {
                onMenuAddContact(position)
            }
        }
    }

    // = PRIVATE section

    private fun handleCallReminder(reminderBundle: Bundle?) {
        val nameKey = LocalCallLogController.KEY_CALL_REMINDER_NAME
        val dateKey = LocalCallLogController.KEY_CALL_REMINDER_DATE
        if (reminderBundle != null && reminderBundle.containsKey(nameKey) && reminderBundle.containsKey(dateKey)) {
            val message = getString(R.string.tCallReminderDisplayText, reminderBundle.getString(nameKey), reminderBundle.getString(dateKey))
            firePresenterEvent(Events.SHOW_CALL_REMINDER, message)
        }
    }

    // Grouped List management

    override fun buildGroupedList(newEntityList: List<CallLogEntity>): List<CallLogItemModel> {
        val newGroupedList = mutableListOf<CallLogItemModel>()
        var groupEntity: CallLogEntity? = null
        var group: CallLogGroupItem? = null

        // Build groups
        for (entity in newEntityList) {
            if (group != null && groupEntity != null && groupEntity.compareNumber(entity)) {
                // Fill the current group with content items
                group.items += buildContent(entity)

                if (group.callTypes.size < MAX_CALL_TYPE_ICONS) {
                    group.callTypes += entity.callType
                }
            } else {
                // Create a new group
                groupEntity = entity
                group = buildGroup(groupEntity)
                newGroupedList += group
            }
        }

        // Update expanded group
        val oldExpandedPosition = getExpandedPosition()
        val oldExpandedGroup = getGroupItemAt(oldExpandedPosition)
        if (oldExpandedGroup != null && newGroupedList.isNotEmpty()) {
            val newExpandedPosition = newGroupedList
                    .indexOfFirst { it.isGroup && (it as CallLogGroupItem).getEntityId() == oldExpandedGroup.getEntityId() }
            if (newExpandedPosition != -1) {
                // Found. Update group and expanded position by new value
                val newExpandedGroup = newGroupedList[newExpandedPosition] as CallLogGroupItem
                newExpandedGroup.isExpanded = true
                newGroupedList.addAll(newExpandedPosition + 1, newExpandedGroup.items)
            }
        }

        // Update selected group
        val oldSelectedPosition = getSelectedPosition()
        val oldSelectedGroup = getGroupItemAt(oldSelectedPosition)
        if (oldSelectedGroup != null && newGroupedList.isNotEmpty()) {
            val newSelectedGroup = newGroupedList
                    .find { it.isGroup && (it as CallLogGroupItem).getEntityId() == oldSelectedGroup.getEntityId() }
            (newSelectedGroup as? CallLogGroupItem)?.let { it.isSelected = true }
        }

        return newGroupedList
    }

    private fun toggleExpandedGroup(position: Int) {
        var currentPosition = position
        val expandedPosition = getExpandedPosition()
        val expandedGroup = getItemAt(expandedPosition)
        synchronized(mGroupedList) {
            if (expandedGroup is CallLogGroupItem) {
                // Expanded. Collapse it first
                expandedGroup.isExpanded = false
                // Explicit call of method with three arguments to avoid interpret passed int value as delay
                firePresenterEvent(CallLogAbstractPresenter.Events.DATA_SET_UPDATED, expandedPosition, 0) // isExpanded changed

                mGroupedList.removeAll(expandedGroup.items)
                firePresenterEvent(Events.DATA_SET_COLLAPSED, Pair(expandedPosition + 1, expandedGroup.items.size))

                // Take care of current position after list was collapsed
                if (currentPosition > expandedPosition) {
                    currentPosition -= expandedGroup.items.size
                }
            }

            val currentGroup = getItemAt(currentPosition)
            if (expandedPosition == currentPosition) {
                // Already collapsed by code above, do nothing
            } else if (currentGroup is CallLogGroupItem && !currentGroup.isExpanded) {
                // Collapsed. Expand it now
                currentGroup.isExpanded = true
                // Explicit call of method with three arguments to avoid interpret passed int value as delay
                firePresenterEvent(CallLogAbstractPresenter.Events.DATA_SET_UPDATED, currentPosition, 0) // isExpanded changed

                mGroupedList.addAll(currentPosition + 1, currentGroup.items)
                firePresenterEvent(Events.DATA_SET_EXPANDED, Pair(currentPosition + 1, currentGroup.items.size))
            }
        }
    }

    private fun setSelectedGroup(position: Int) {
        val oldSelectedPosition = getSelectedPosition()
        if (oldSelectedPosition != position) {
            // Deselect old selected item if found
            val oldSelectedGroup = getGroupItemAt(oldSelectedPosition)
            oldSelectedGroup?.let {
                it.isSelected = false
                // Explicit call of method with three arguments to avoid interpret passed index as delay
                firePresenterEvent(CallLogAbstractPresenter.Events.DATA_SET_UPDATED, oldSelectedPosition, 0)
            }

            // Select new item
            val group = getGroupItemAt(position) ?: return

            group.isSelected = true
            // Explicit call of method with three arguments to avoid interpret passed index as delay
            firePresenterEvent(CallLogAbstractPresenter.Events.DATA_SET_UPDATED, position, 0)
        }

        goToDetails(position)
    }

    private fun goToDetails(position: Int) {
        val group = getGroupItemAt(position) ?: return
        val groupId = group.getEntityId()
        val bundle = Bundle(1).apply { putLong(CallLogDetailsPresenter.KEY_CALL_LOG_DETAILS_GROUP_ID, groupId) }
        firePresenterEvent(Events.SHOW_CALL_LOG_DETAILS, bundle)
    }

    // Helper methods

    private fun callNumber(group: CallLogGroupItem?, callAction: CallAction) =
            group?.let { callNumber(it.getNumberForContacts(), it.displayName, it.accountId, callAction) }

    private fun callNumber(number: String, displayName : String, accountId: Int, callAction: CallAction) {
        if (!Validator.isValidUserName(number) && !Validator.isValidPhoneNumber(number)) {
            firePresenterEvent(CallLogAbstractPresenter.Events.SHOW_MESSAGE_LONG, getString(R.string.tDialNumberIncorrect1))
            return
        }

        val accountNickname = mControllers.accounts.getAccount(accountId)?.getStr(EAccountSetting.Nickname) ?: ""

        when (callAction) {
            CallAction.AudioCall -> mControllers.phone.call(number, accountNickname, displayName, CallData.ECallType.Generic)
            CallAction.PrefixCall -> mControllers.phone.prefixCall(number, accountNickname, displayName)
            CallAction.VideoCall -> mControllers.phone.callVideo(number, accountNickname, displayName)
        }.also { success ->
            if (!success) {
                mControllers.phone.lastError?.let { firePresenterEvent(CallLogAbstractPresenter.Events.SHOW_MESSAGE_LONG, it.description) }
            }
        }
    }
}
