package com.bria.voip.ui.main.calllog.refactoring

import android.Manifest
import android.app.Dialog
import android.content.Intent
import android.content.pm.PackageManager
import android.net.Uri
import android.os.Bundle
import android.support.annotation.MainThread
import android.support.v7.app.AlertDialog
import android.support.v7.widget.RecyclerView
import android.view.ContextMenu
import android.view.MenuItem
import android.view.View
import com.bria.common.permission.PermissionRequestCode
import com.bria.common.uiframework.annotations.Layout
import com.bria.common.uiframework.annotations.Menu
import com.bria.common.uiframework.dialogs.ScreenHolderDialog
import com.bria.common.uiframework.presenters.IPresenterEvent
import com.bria.common.uiframework.screens.IScreenEnum
import com.bria.common.util.AndroidUtils
import com.bria.common.util.Log
import com.bria.voip.R
import com.bria.voip.ui.main.calllog.refactoring.DialogId.BUDDY_DISPLAY_SCREEN_HOLDER_DIALOG_ID
import com.bria.voip.ui.main.calllog.refactoring.DialogId.CALL_REMINDER_DIALOG_ID
import com.bria.voip.ui.main.calllog.refactoring.DialogId.CONFERENCE_CALL_CONFIRM_DIALOG_ID
import com.bria.voip.ui.main.calllog.refactoring.DialogId.CONTACT_DISPLAY_SCREEN_HOLDER_DIALOG_ID
import com.bria.voip.ui.main.calllog.refactoring.DialogId.CONTACT_EDIT_SCREEN_HOLDER_DIALOG_ID
import com.bria.voip.ui.main.calllog.refactoring.DialogId.HOSTED_CONFERENCE_CALL_CONFIRM_DIALOG_ID
import com.bria.voip.ui.main.calllog.refactoring.DialogId.PHONE_NUMBER_ACTION_PICKER_ID
import com.bria.voip.ui.main.misc.EScreenList
import com.bria.voip.ui.shared.helpers.TabletListSelectionHelper
import com.bria.voip.ui.shared.pickers.ContactPickerScreen

@Layout(R.layout.calllog_list_screen)
@Menu(R.menu.call_log_list_menu)
open class CallLogListScreen<Presenter : CallLogListPresenter> : CallLogAbstractScreen<Presenter>() {
    companion object {
        // Screen Tag
        private const val TAG = "CallLogListScreen"

        // Context menu IDs
        private const val CONTEXT_MENU_DELETE = 1
        private const val CONTEXT_MENU_CALL = 2
        private const val CONTEXT_MENU_ADD_CONTACT = 3
        private const val CONTEXT_MENU_VIEW_CONTACT = 4
        private const val CONTEXT_MENU_UPDATE_CONTACT = 5
        private const val CONTEXT_MENU_SEND_SMS = 6
        private const val CONTEXT_MENU_SEND_IM = 7
        private const val CONTEXT_MENU_VIDEO_CALL = 8
        private const val CONTEXT_MENU_TRANSFER_CALL = 9
        private const val CONTEXT_MENU_PREFIX_CALL = 10
        private const val CONTEXT_MENU_CONFERENCE_CALL = 11
        private const val CONTEXT_MENU_HOSTED_CONFERENCE_CALL = 12

        // Bundle Keys
        private const val KEY_CALL_REMINDER_MESSAGE = "KEY_CALL_REMINDER_MESSAGE"

        private const val KEY_LAST_LONG_CLICKED_ITEM_POS = "KEY_LAST_LONG_CLICKED_ITEM_POS"
        private const val KEY_LAST_CLICKED_ITEM_CONTACT_POS = "KEY_LAST_CLICKED_ITEM_CONTACT_POS"
    }

    private val mTabletListSelectionHelper = TabletListSelectionHelper()
    private var mLastLongClickedItemPosition: Int = RecyclerView.NO_POSITION
    private var mLastClickedItemContactPosition: Int = RecyclerView.NO_POSITION

    override fun onPresenterEvent(event: IPresenterEvent<*, *>) {
        super.onPresenterEvent(event)

        when (event.type) {
            CallLogListPresenter.Events.DATA_SET_EXPANDED -> (event.data as? Pair<Int, Int>)?.let { dataSetExpanded(it.first, it.second) }
            CallLogListPresenter.Events.DATA_SET_COLLAPSED -> (event.data as? Pair<Int, Int>)?.let { dataSetCollapsed(it.first, it.second) }

            CallLogListPresenter.Events.SHOW_CALL_REMINDER -> showCallReminderDialog(event.data as String)
            CallLogListPresenter.Events.SHOW_CONFERENCE_CALL_CONFIRMATION ->  showConferenceCallConfirmationDialog(event.data as Int)
            CallLogListPresenter.Events.SHOW_HOSTED_CONFERENCE_CALL_CONFIRMATION ->  showHostedConferenceCallConfirmationDialog(event.data as Int)
            CallLogListPresenter.Events.SHOW_WEB_VIEW -> showWebViewActivity(event.data as String)
            CallLogListPresenter.Events.SHOW_PHONE_NUMBER_ACTION_PICKER -> showPhoneNumberActionPicker(event.data as Bundle)
            CallLogListPresenter.Events.SHOW_CONTACT_PICKER ->  showContactPickerScreen(event.data as Int)
            CallLogListPresenter.Events.SHOW_CONTACT_EDIT -> showContactEditScreen(event.data as Bundle)
            CallLogListPresenter.Events.SHOW_CONTACT_DETAILS -> showContactDetailsScreen(event.data as Bundle)
            CallLogListPresenter.Events.SHOW_BUDDY_DETAILS -> showBuddyDetailsScreen(event.data as Bundle)
            CallLogListPresenter.Events.SHOW_CONVERSATION -> showConversationScreen(event.data as Bundle)
            CallLogListPresenter.Events.SHOW_CALL_LOG_DETAILS -> showCallLogDetailsScreen(event.data as Bundle)
        }
    }

    override fun getPresenterClass(): Class<out Presenter> = CallLogListPresenter::class.java as Class<out Presenter>

    @MainThread
    override fun onCreate(bundle: Bundle?) {
        super.onCreate(bundle)
        mAdapter.onItemClickListener = { _, position -> presenter.onItemClicked(position) }
        mAdapter.onCreateContextMenuListener = mOnCreateContextMenuListener
        mAdapter.onItemCallButtonClickListener = { _, position -> presenter.onItemCallButtonClicked(position) }
        mAdapter.onItemContactButtonClickListener = { _, position ->
            mLastClickedItemContactPosition = position
            presenter.onItemContactButtonClicked(position)
        }
        mAdapter.onItemDeleteRecordButtonClickListener = { _, position -> presenter.onItemDeleteCallRecordButtonClicked(position) }
        mAdapter.onItemSaveRecordButtonClickListener = { _, position -> presenter.onItemSaveCallRecordButtonClicked(position) }
        mAdapter.onItemShareRecordButtonClickListener = { _, position -> presenter.onItemShareCallRecordButtonClicked(position) }
        mAdapter.onItemPlayRecordButtonClickListener = { _, position -> presenter.onItemPlayCallRecordButtonClicked(position) }

        mTabletListSelectionHelper.checkForFlags(bundle)
    }

    override fun onSaveState(stateBundle: Bundle) {
        stateBundle.putInt(KEY_LAST_LONG_CLICKED_ITEM_POS, mLastLongClickedItemPosition)
        stateBundle.putInt(KEY_LAST_CLICKED_ITEM_CONTACT_POS, mLastClickedItemContactPosition)
        super.onSaveState(stateBundle)
    }

    override fun doRestoreState(stateBundle: Bundle?) {
        super.doRestoreState(stateBundle)
        mLastLongClickedItemPosition = stateBundle?.getInt(KEY_LAST_LONG_CLICKED_ITEM_POS, RecyclerView.NO_POSITION) ?: RecyclerView.NO_POSITION
        mLastClickedItemContactPosition = stateBundle?.getInt(KEY_LAST_CLICKED_ITEM_CONTACT_POS, RecyclerView.NO_POSITION) ?: RecyclerView.NO_POSITION
    }

    override fun updateMenuItems(menu: android.view.Menu, updateKey: String?) {
        super.updateMenuItems(menu, updateKey)
        if (presenter.isMenuDeleteAllEnabled().not()) menu.removeItem(R.id.call_log_list_delete_all)
        if (presenter.isMenuDeleteMissedEnabled().not()) menu.removeItem(R.id.call_log_list_delete_missed)
        if (presenter.isMenuDeleteConferenceEnabled().not()) menu.removeItem(R.id.call_log_list_delete_conference)
    }

    @MainThread
    override fun onMenuItemClick(menuItem: MenuItem): Boolean {
        return when (menuItem.itemId) {
            R.id.call_log_list_delete_all -> { presenter.onMenuDeleteAll(); true }
            R.id.call_log_list_delete_missed -> { presenter.onMenuDeleteMissed(); true }
            R.id.call_log_list_delete_conference -> { presenter.onMenuDeleteConference(); true }
            else -> false
        } || super.onMenuItemClick(menuItem)
    }

    @MainThread
    override fun onNewMessage(message: Bundle, sender: IScreenEnum) {
        super.onNewMessage(message, sender)
        if (sender == EScreenList.CONTACT_PICKER) {
            val position = message.getInt(KEY_ITEM_POSITION, -1)
            val contactId = message.getStringArrayList(ContactPickerScreen.SELECTED_CONTACT_IDS)?.get(0) ?: ""
            if (position != -1 && contactId.isNotEmpty()) {
                presenter.onContactForUpdateSelected(contactId, position)
            }
        }
    }

    override fun createDialog(which : Int, data : Bundle?) : Dialog? {
        when(which) {
            CALL_REMINDER_DIALOG_ID ->
                return AlertDialog.Builder(activity)
                        .setTitle(R.string.tCallReminder)
                        .setMessage(data?.getString(KEY_CALL_REMINDER_MESSAGE))
                        .setNeutralButton(R.string.tOk, null)
                        .setCancelable(true)
                        .create()
//            CALL_IN_PROGRESS_DIALOG_ID ->
//                return AlertDialog.Builder(activity)
//                        .setMessage(R.string.tCallInProgressWait)
//                        .setPositiveButton(R.string.tOk, null)
//                        .setCancelable(true)
//                        .create()
            CONFERENCE_CALL_CONFIRM_DIALOG_ID ->
                return AlertDialog.Builder(activity)
                        .setTitle(R.string.tCollab_AnotherConferenceTitle)
                        .setMessage(R.string.tCollab_AnotherConferenceText)
                        .setPositiveButton(R.string.tCollab_AnotherConferencePositive) { _, _ -> presenter.onConferenceCallConfirmed(data?.getInt(KEY_ITEM_POSITION) ?: -1) }
                        .setNegativeButton(R.string.tCollab_AnotherConferenceNegative, null)
                        .setCancelable(true)
                        .create()
            HOSTED_CONFERENCE_CALL_CONFIRM_DIALOG_ID ->
                return AlertDialog.Builder(activity)
                        .setTitle(R.string.tCollab_AnotherConferenceTitle)
                        .setMessage(R.string.tCollab_AnotherConferenceText)
                        .setPositiveButton(R.string.tCollab_AnotherConferencePositive) { _, _ -> presenter.onHostedConferenceCallConfirmed(data?.getInt(KEY_ITEM_POSITION) ?: -1) }
                        .setNegativeButton(R.string.tCollab_AnotherConferenceNegative, null)
                        .setCancelable(true)
                        .create()
            CONTACT_DISPLAY_SCREEN_HOLDER_DIALOG_ID ->
                return createScreenHolderDialog(EScreenList.CONTACT_DISPLAY, ScreenHolderDialog.DIALOG_FULL_SCREEN, data)
            BUDDY_DISPLAY_SCREEN_HOLDER_DIALOG_ID ->
                return createScreenHolderDialog(EScreenList.BUDDY_DISPLAY, ScreenHolderDialog.DIALOG_FULL_SCREEN, data)
            CONTACT_EDIT_SCREEN_HOLDER_DIALOG_ID ->
                return createScreenHolderDialog(EScreenList.CONTACT_EDIT, ScreenHolderDialog.DIALOG_MAX_SIZE_NO_MARGIN, data)
            PHONE_NUMBER_ACTION_PICKER_ID ->
                return ScreenHolderDialog
                        .builder(activity)
                        .screen(EScreenList.PHONE_ACTION_SELECT)
                        .bottomSheet()
                        .bundle(data)
                        .build()
        }
        return super.createDialog(which, data)
    }

    private fun createScreenHolderDialog(screen: EScreenList, @ScreenHolderDialog.DialogStyle style: Int, bundle: Bundle?): Dialog {
        return ScreenHolderDialog
                .builder(activity)
                .screen(screen)
                .style(style)
                .showTitle(false)
                .bundle(bundle)
                .build()
    }

    private fun showCallReminderDialog(message : String) {
        showDialog(CALL_REMINDER_DIALOG_ID, Bundle(1).apply { putString(KEY_CALL_REMINDER_MESSAGE, message) })
    }

    private fun showConferenceCallConfirmationDialog(position: Int) {
        showDialog(CONFERENCE_CALL_CONFIRM_DIALOG_ID, Bundle(1).apply { putInt(KEY_ITEM_POSITION, position) })
    }

    private fun showHostedConferenceCallConfirmationDialog(position: Int) {
        showDialog(HOSTED_CONFERENCE_CALL_CONFIRM_DIALOG_ID, Bundle(1).apply { putInt(KEY_ITEM_POSITION, position) })
    }

    private fun showWebViewActivity(link : String) {
        val webIntent = Intent(Intent.ACTION_VIEW, Uri.parse(link))
        if (!AndroidUtils.canHandleIntent(webIntent, activity)) {
            Log.w(TAG, "Can't handle intent $webIntent")
            return
        }
        activity.startActivity(webIntent)
    }

    private fun showPhoneNumberActionPicker(bundle: Bundle) {
        showDialog(PHONE_NUMBER_ACTION_PICKER_ID, bundle)
    }

    private fun showContactPickerScreen(position : Int) {
        val bundle = Bundle(2).apply {
            putInt(KEY_ITEM_POSITION, position);
            putInt(ContactPickerScreen.MAX_NUMBER_OF_CHOICES, 1)
        }
        showScreenForResult(EScreenList.CONTACT_PICKER, ScreenHolderDialog.DIALOG_MAX_SIZE_NO_MARGIN, bundle)
    }

    private fun showContactEditScreen(bundle: Bundle) {
        if (checkOrRequestContactsPermission(PermissionRequestCode.CP_PERMISSION_WRITE_CONTACTS_FROM_CALLLOG_LIST)) {
            goTo(EScreenList.CONTACT_EDIT, CONTACT_EDIT_SCREEN_HOLDER_DIALOG_ID, bundle)
        }
    }

    private fun showContactDetailsScreen(bundle: Bundle) {
        goTo(EScreenList.CONTACT_DISPLAY, CONTACT_DISPLAY_SCREEN_HOLDER_DIALOG_ID, bundle)
    }

    private fun showBuddyDetailsScreen(bundle: Bundle) {
        goTo(EScreenList.BUDDY_DISPLAY, BUDDY_DISPLAY_SCREEN_HOLDER_DIALOG_ID, bundle)
    }

    private fun showConversationScreen(bundle: Bundle) {
        mCoordinator.flow(bundle).goTo(EScreenList.CONVERSATION)
    }

    private fun showCallLogDetailsScreen(bundle: Bundle) {
        mCoordinator.flow(bundle).goTo(EScreenList.CALL_LOG_DETAILS_V2)
    }

    private fun goTo(screen: EScreenList, which: Int, bundle: Bundle) {
        if (isInTabletMode) {
            showDialog(which, bundle)
        } else {
            mCoordinator.flow(bundle).goTo(screen)
        }
    }

    private fun dataSetExpanded(position: Int, count: Int) {
        mAdapter.notifyDataInserted(position, count)
        val lastVisiblePosition = mLinearLayoutManager.findLastCompletelyVisibleItemPosition()
        if (position > lastVisiblePosition) {
            mRecyclerView.scrollToPosition(position)
        }
    }

    private fun dataSetCollapsed(position: Int, count: Int) {
        mAdapter.notifyDataRemoved(position, count)
    }

    private fun checkOrRequestContactsPermission(permissionCode : Int) : Boolean {
        if (checkPermission(Manifest.permission.WRITE_CONTACTS)) {
            return true
        }
        if (isPermissionNeverAskChecked(Manifest.permission.WRITE_CONTACTS)) {
            toastLong(getString(R.string.tContactsNoPermission))
            return false
        }
        requestPermission(Manifest.permission.WRITE_CONTACTS, permissionCode, getString(R.string.tPermissionContacts))
        return false
    }

    @MainThread
    override fun onRequestPermissionsResult(requestCode: Int, permissions: Array<String>, grantResults: IntArray) {
        super.onRequestPermissionsResult(requestCode, permissions, grantResults)

        if (mLastLongClickedItemPosition == RecyclerView.NO_POSITION && mLastClickedItemContactPosition == RecyclerView.NO_POSITION) {
            return
        }

        if (grantResults.firstOrNull() == PackageManager.PERMISSION_GRANTED) {
            when (requestCode) {
                PermissionRequestCode.CP_PERMISSION_ADD_CONTACT_FROM_CALLLOG_LIST -> presenter.onMenuAddContact(mLastLongClickedItemPosition)
                PermissionRequestCode.CP_PERMISSION_VIEW_CONTACT_FROM_CALLLOG_LIST -> presenter.onMenuViewContact(mLastLongClickedItemPosition)
                PermissionRequestCode.CP_PERMISSION_UPDATE_CONTACT_FROM_CALLLOG_LIST -> showContactPickerScreen(mLastLongClickedItemPosition)
                PermissionRequestCode.CP_PERMISSION_WRITE_CONTACTS_FROM_CALLLOG_LIST -> presenter.onMenuAddContact(mLastClickedItemContactPosition)
            }
        }
    }

    private val mOnCreateContextMenuListener = { menu: ContextMenu, _: View, _: ContextMenu.ContextMenuInfo?, position: Int ->

        mLastLongClickedItemPosition = position

        menu.setHeaderTitle(presenter.getContextMenuTitle(position))

        if (presenter.isMenuOnlyConferenceRelatedActionsAllowed(position)) {
            if (presenter.isMenuConferenceCallEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_CONFERENCE_CALL, ContextMenu.NONE, R.string.comm_log_vccs_call /*todo number*/)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuHostedConferenceCallEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_HOSTED_CONFERENCE_CALL, ContextMenu.NONE, R.string.tCollab_CreateNewConferenceAgain)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuDeleteEnabled()) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_DELETE, ContextMenu.NONE, R.string.tDelete)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
        } else {
            if (presenter.isMenuDeleteEnabled()) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_DELETE, ContextMenu.NONE, R.string.tDelete)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuCallEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_CALL, ContextMenu.NONE, R.string.callLog /*todo number*/)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuAddContactEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_ADD_CONTACT, ContextMenu.NONE, R.string.addLogToContacts)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuViewContactEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_VIEW_CONTACT, ContextMenu.NONE, R.string.viewLogContactDetails)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuUpdateContactEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_UPDATE_CONTACT, ContextMenu.NONE, R.string.updateExistingLogContactDetails)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuSendSmsEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_SEND_SMS, ContextMenu.NONE, R.string.tSendSms)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuSendImEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_SEND_IM, ContextMenu.NONE, R.string.tBuddySendIM)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuVideoCallEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_VIDEO_CALL, ContextMenu.NONE, R.string.videoCallLog /*todo number*/)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuTransferCallEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_TRANSFER_CALL, ContextMenu.NONE, R.string.callTransferLog /*todo number*/)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
            if (presenter.isMenuPrefixCallEnabled(position)) {
                menu.add(ContextMenu.NONE, CONTEXT_MENU_PREFIX_CALL, ContextMenu.NONE, R.string.tPrefixCall /*todo number*/)
                        .setOnMenuItemClickListener(mOnContextMenuItemClickListener)
            }
        }
    }

    private val mOnContextMenuItemClickListener = MenuItem.OnMenuItemClickListener { menuItem ->
        var handled = true
        when (menuItem.itemId) {
            CONTEXT_MENU_DELETE -> presenter.onMenuDelete(mLastLongClickedItemPosition)
            CONTEXT_MENU_CALL -> presenter.onMenuCall(mLastLongClickedItemPosition)
            CONTEXT_MENU_ADD_CONTACT -> {
                if (checkOrRequestContactsPermission(PermissionRequestCode.CP_PERMISSION_ADD_CONTACT_FROM_CALLLOG_LIST)) {
                    presenter.onMenuAddContact(mLastLongClickedItemPosition)
                }
            }
            CONTEXT_MENU_VIEW_CONTACT -> {
                if (checkOrRequestContactsPermission(PermissionRequestCode.CP_PERMISSION_VIEW_CONTACT_FROM_CALLLOG_LIST)) {
                    presenter.onMenuViewContact(mLastLongClickedItemPosition)
                }
            }
            CONTEXT_MENU_UPDATE_CONTACT -> {
                if (checkOrRequestContactsPermission(PermissionRequestCode.CP_PERMISSION_UPDATE_CONTACT_FROM_CALLLOG_LIST)) {
                    presenter.onMenuUpdateContact(mLastLongClickedItemPosition)
                }
            }
            CONTEXT_MENU_SEND_SMS -> presenter.onMenuSendSms(mLastLongClickedItemPosition)
            CONTEXT_MENU_SEND_IM -> presenter.onMenuSendIm(mLastLongClickedItemPosition)
            CONTEXT_MENU_VIDEO_CALL -> presenter.onMenuVideoCall(mLastLongClickedItemPosition)
            CONTEXT_MENU_TRANSFER_CALL -> presenter.onMenuTransferCall(mLastLongClickedItemPosition)
            CONTEXT_MENU_PREFIX_CALL -> presenter.onMenuPrefixCall(mLastLongClickedItemPosition)
            CONTEXT_MENU_CONFERENCE_CALL -> presenter.onMenuConferenceCall(mLastLongClickedItemPosition)
            CONTEXT_MENU_HOSTED_CONFERENCE_CALL -> presenter.onMenuHostedConferenceCall(mLastLongClickedItemPosition)

            else -> handled = false
        }
        return@OnMenuItemClickListener handled
    }
}
