package com.bria.voip.ui.main.im.chatroom.add

import android.annotation.SuppressLint
import android.os.Bundle
import com.bria.common.controller.Controllers
import com.bria.common.controller.accounts.Account
import com.bria.common.controller.accounts.AccountsFilter
import com.bria.common.controller.im.ImpsUtils
import com.bria.common.controller.im.refactoring.BuddyKeyUtils
import com.bria.common.controller.im.refactoring.chatroom.ChatRoomFieldValidator
import com.bria.common.controller.im.refactoring.chatroom.getAvatarColor
import com.bria.common.controller.im.refactoring.roomdb.entities.ChatRoom
import com.bria.common.controller.settings.EAccountSetting
import com.bria.common.controller.settings.branding.EAccountType
import com.bria.common.uiframework.presenters.AbstractPresenter
import com.bria.common.uiframework.presenters.IPresenterEventTypeEnum
import com.bria.common.util.*
import com.bria.common.util.im.ParticipantsSet
import com.bria.voip.R
import io.reactivex.Observable
import io.reactivex.Single
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.BehaviorSubject
import java.util.concurrent.TimeUnit

class ChatRoomConfPresenter : AbstractPresenter() {

    companion object {
        const val TAG = "ChatRoomConfPresenter"
    }

    private var chatRoomApi = Controllers.get().im.chatApi.chatRoomApi

    private var chatRoomRepo = Controllers.get().im.imData.chatRoomRepo

    private var chatRoom = ChatRoom(publicRoom = false)

    var disposable = CompositeDisposable()



    enum class Events : IPresenterEventTypeEnum {
        INVALIDATE_MENU,
        PROPERTY_VALIDATION_FAILED,
        ROOM_NAME_DUPLICATED,
        UPDATE_SCREEN,
        NO_ACTIVE_ACCOUNT,
        UPDATE_AVATAR_COLOR,
        DISMISSED,
        SHOW_DISCARD_CHANGES_DIALOG,
        UPDATE_INITIALS,
        CHAT_ROOM_SAVED,
        TOAST;

        override fun getType(): IPresenterEventTypeEnum = this
    }

    init {
        //for Stretto we will consider single XMPP account case
    }

    var title: String = context.getString(R.string.tCreatChatRoom)

    fun handleAccept() {
        if (mIsSavingSubject.value == true) {
            firePresenterEvent(Events.TOAST, getString(R.string.tStillWorking))
            return
        }
        chatRoom.name = chatRoom.name.trim().trimStart()
        val account = Controllers.get().accounts.getAccounts(AccountsFilter.XMPP).firstOrNull { it.isRegistered }
        if (account == null) {
            firePresenterEvent(Events.NO_ACTIVE_ACCOUNT)
            return
        }

        if (chatRoom.accountId.isEmpty()) {
            chatRoom.owners = ParticipantsSet(getSelfInfoKey(account = account))
            chatRoom.accountId = ImpsUtils.removePortFromDomain(account.getUserAtDomain())
        }
        mIsSavingSubject.onNext(true)
        disposable.add(validateRoom().subscribeOn(Schedulers.io()).subscribe ({valid ->
            if (valid) {
                dataValidCreateChatRoom()
            } else {
                mIsSavingSubject.onNext(false)
            }
        },{}))
    }

    private fun dataValidCreateChatRoom() {
        val time = System.currentTimeMillis()
        chatRoom.creationTime = time
        chatRoom.joinTime = time
        chatRoom.modTime = time
        if (chatRoom.publicRoom) {
            chatRoom.invited.clear()
        }

        // lock it down, so user input does not break filter
        val chatRoomName = chatRoom.name

        val addOrUpdateChatRoomObservable = chatRoomRepo.getOnRoomAddedObservable()
                .mergeWith(chatRoomRepo.getOnRoomUpdatedObservable())
                .flatMapIterable { it }
                .filter { it.name == chatRoomName }
                .map { succeedWith<ChatRoom, SaveError>(it) }

        val timeoutObservable = Observable
                // save operation can sometimes take a while
                .timer(30, TimeUnit.SECONDS, Schedulers.computation())
                .map { failWith<ChatRoom, SaveError>(SaveError.Timeout) }

        val observableWithTimeout = addOrUpdateChatRoomObservable.ambWith(timeoutObservable)

        disposable.add(
                observableWithTimeout
                        .subscribeOn(Schedulers.computation())

                        // safe as either timeout will occur or we'll get ChatRoom
                        .firstOrError()

                        // in any case hide progress bar
                        .doAfterTerminate { mIsSavingSubject.onNext(false) }

                        .subscribe({
                            when (it) {
                                is Result.Success<ChatRoom, SaveError> -> {
                                    val bundle = Bundle(1)
                                    bundle.putLong(GlobalConstants.KEY_CHAT_ROOM_ID, it.value.id)
                                    cleanup()
                                    firePresenterEvent(Events.CHAT_ROOM_SAVED, bundle)
                                }
                                is Result.Failure<ChatRoom, SaveError> -> {
                                    when (it.value) {
                                        SaveError.Timeout -> {
                                            Log.d(TAG, "Saving of chat room timed out.")
                                            firePresenterEvent(Events.TOAST, getString(R.string.tOperationHasTimedOut))
                                        }
                                    }
                                }
                            }
                        }, {
                            CrashInDebug.with(TAG, it)
                        })
        )

        chatRoomApi.createRoom(chatRoom = chatRoom)
    }

    @SuppressLint("CheckResult")
    private fun validateRoom() : Single<Boolean>{
        Log.d(TAG, "ValidateRoom 1")
        return Single.fromCallable {
            val validationResult = ChatRoomFieldValidator.validate(chatRoom, context)
            if (!validationResult.isValid) {
                Log.d(TAG, "ValidateRoom 2 not valid property")
                firePresenterEvent(Events.PROPERTY_VALIDATION_FAILED, validationResult)
                false
            } else {
                Log.d(TAG, "ValidateRoom 3 checking local room list")
                chatRoomRepo.getAllChatRooms().blockingGet().firstOrNull { it.name == chatRoom.name && it.chatKey != chatRoom.chatKey }?.let {
                    firePresenterEvent(Events.ROOM_NAME_DUPLICATED)
                    return@fromCallable false
                }
                Log.d(TAG, "ValidateRoom 4 get Room list")
                chatRoomApi.updateRoomList(false)
                chatRoomRepo.getOnServerRoomListChangedObservable().first(emptyList()).blockingGet()
                Log.d(TAG, "ValidateRoom 45 Room list Received")
                val found = chatRoomRepo.roomNameInTheList(chatRoom.name)
                        if (found)  firePresenterEvent(Events.ROOM_NAME_DUPLICATED)
                !found

            }
        }
    }

    fun invitedMembersUpdated(buddyList: List<String>) {
        chatRoom.invited.clear()
        chatRoom.invited = ParticipantsSet(BuddyKeyUtils.getStringFromBuddyKeys(buddyList.toSet()))
    }

    fun getMembersText(): String {
        val members = ParticipantsSet(chatRoom.members)
        members.addAll(chatRoom.invited)
        return members.getShortParticipantList(context)
    }

    fun getInvitedMembers(): String {
        return chatRoom.invited.serializedParticipantKeys
    }

    fun getNameText(): String = chatRoom.name

    fun getOwnersText(): String = chatRoom.owners.participantsString

    fun getTopicText(): String = chatRoom.topic

    fun updateTopic(topic: String) {
        chatRoom.topic = topic
    }

    val activeMembers: Set<String>
        get() = (chatRoom.members + chatRoom.owners).asSequence().map { it.remoteAddress }.toSet()


    val tentativeMembers: Set<String>
        get() = chatRoom.invited.asSequence().map { it.remoteAddress }.toSet()

    fun updateName(name: String) {
        chatRoom.name = name
        firePresenterEvent(Events.UPDATE_INITIALS, getInitials())
        firePresenterEvent(Events.UPDATE_AVATAR_COLOR)
    }

    fun getInitials(): String {
        val initials = chatRoom.initials
        return if (initials.isEmpty()) "#" else initials
    }

    fun getAvatarColor() = chatRoom.getAvatarColor()

    private fun getSelfInfoKey(account: Account): String {
        return BuddyKeyUtils.getNewBuddyKey(EAccountType.Xmpp, account.getUserAtDomain(), account.getUserAtDomain())
    }

    fun handleDismissRoomBackKey() {
        firePresenterEvent(Events.DISMISSED)
    }

    fun handleDismissRoomUpKey() {
        firePresenterEvent(Events.DISMISSED)
    }

    fun updateDescription(description: String) {
        chatRoom.description = description
    }

    fun setActionNew() {
        Controllers.get().accounts.getAccounts(AccountsFilter.XMPP).asSequence().filter { it.isRegistered }.firstOrNull()?.let {
            chatRoom.owners = ParticipantsSet(getSelfInfoKey(account = it))
            chatRoom.accountId = ImpsUtils.removePortFromDomain(it.getUserAtDomain())
        }
        firePresenterEvent(Events.UPDATE_SCREEN)
    }

    fun isPublicRoom() = chatRoom.publicRoom
    fun getDescriptionText() = chatRoom.description

    fun cleanup() {
        chatRoom = ChatRoom(publicRoom = false)
        mIsSavingSubject.onNext(false)
    }

    fun forceDismissChatRoom() {
        cleanup()
        firePresenterEvent(Events.DISMISSED)
    }

    fun setIsPublic(public: Boolean) {
        chatRoom.publicRoom = public
    }

    enum class SaveError {
        Timeout
    }

    private val mIsSavingSubject = BehaviorSubject.createDefault(false)

    val isSavingObservable
        get() = mIsSavingSubject
}

fun Account.getUserAtDomain(): String = ImpsUtils.removePortFromDomain("${this.accountSettings[EAccountSetting.UserName]}@${this.accountSettings.get
(EAccountSetting.Domain)}")
