package com.bria.voip.ui.main.im.chatroom.add

import android.app.Dialog
import android.os.Bundle
import android.support.annotation.MainThread
import android.support.design.widget.TextInputLayout
import android.support.v7.app.AlertDialog
import android.text.Editable
import android.text.TextWatcher
import android.view.Gravity
import android.view.Menu
import android.view.MenuItem
import android.view.View
import android.widget.*
import com.bria.common.controller.Controllers
import com.bria.common.controller.buddy.BuddyAvailability
import com.bria.common.controller.contacts.buddy.Buddy
import com.bria.common.controller.im.refactoring.roomdb.entities.ChatRoom
import com.bria.common.controller.settings.ESetting
import com.bria.common.uiframework.annotations.Clickable
import com.bria.common.uiframework.annotations.Inject
import com.bria.common.uiframework.annotations.Layout
import com.bria.common.uiframework.coloring.Coloring
import com.bria.common.uiframework.dialogs.ScreenHolderDialog
import com.bria.common.uiframework.presenters.IPresenterEvent
import com.bria.common.uiframework.screens.AbstractScreen
import com.bria.common.uiframework.screens.IScreenEnum
import com.bria.common.util.CrashInDebug
import com.bria.common.util.Log
import com.bria.common.util.im.ParticipantsSet
import com.bria.common.validation.ValidationResult
import com.bria.voip.R
import com.bria.voip.ui.main.misc.EScreenList
import com.bria.voip.ui.shared.helpers.setVisible
import com.bria.voip.ui.shared.pickers.ChatRoomMembersPickerScreen
import com.bria.voip.ui.shared.pickers.generic.AbstractPickerScreen
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.rxkotlin.addTo

@Layout(R.layout.chat_room_conf_screen)
@com.bria.common.uiframework.annotations.Menu(R.menu.edit_menu)
class ChatRoomConfScreen<Presenter : ChatRoomConfPresenter> : AbstractScreen<Presenter>() {

    companion object {
        private const val TAG = "ChatRoomConfScreen"

        private const val STATE_SAVED = "STATE_SAVED"

        private const val TYPE_KEY = "TYPE"
        private const val NAME_KEY = "NAME"
        private const val NAME_CONTAINER_ERROR = "NAME_CONTAINER_ERROR"
        private const val TOPIC_CONTAINER_ERROR = "TOPIC_CONTAINER_ERROR"
        private const val DESCRIPTION_CONTAINER_ERROR = "DESCRIPTION_CONTAINER_ERROR"
        private const val MEMBERS_KEY = "MEMBERS"
        private const val OWNERS_KEY = "OWNERS"
        private const val TOPIC_KEY = "TOPIC"


        private const val DISCARD_CHANGES_DIALOG = 2
    }

    @Inject(id = R.id.chat_room_conf_progress_bar, fore = ESetting.ColorBrandTint)
    private lateinit var progressBar: ProgressBar

    @Inject(R.id.chat_room_conf_type)
    private lateinit var isPublicSwitch: Switch

    @Inject(R.id.chat_room_conf_type_desc)
    private lateinit var typeDescription: TextView

    @Inject(R.id.chat_room_conf_initials)
    private lateinit var initials: TextView

    @Inject(R.id.chat_room_conf_name_container)
    private lateinit var nameContainer: TextInputLayout

    @Inject(R.id.chat_room_conf_name)
    private lateinit var name: EditText

    @Inject(R.id.chat_room_conf_owners_container)
    private lateinit var ownersContainer: TextInputLayout

    @Inject(R.id.chat_room_conf_owners)
    private lateinit var owners: EditText

    @Inject(R.id.chat_room_conf_owners_add)
    @Clickable
    private lateinit var addOwners: ImageButton

    @Inject(R.id.chat_room_conf_members_add)
    @Clickable
    private lateinit var addMembers: ImageButton

    @Inject(R.id.chat_room_conf_members_container)
    private lateinit var membersContainer: TextInputLayout

    @Inject(R.id.chat_room_conf_members)
    private lateinit var members: EditText

    @Inject(R.id.chat_room_conf_description_container)
    private lateinit var descriptionContainer: TextInputLayout

    @Inject(R.id.chat_room_conf_description)
    private lateinit var description: EditText

    @Inject(R.id.chat_room_conf_passcode_container)
    private lateinit var passwordContainer: TextInputLayout

    @Inject(R.id.chat_room_conf_passcode)
    private lateinit var password: EditText

    @Inject(R.id.chat_room_conf_topic_container)
    private lateinit var topicContainer: TextInputLayout

    @Inject(R.id.chat_room_topic)
    private lateinit var topic: EditText

    override fun onCreate(bundle: Bundle?) {
        super.onCreate(bundle)
        if (bundle?.containsKey(STATE_SAVED) != true) {
            // start with clean slate
            presenter.cleanup()
        }

        presenter.subscribe(this)
        membersContainer.isEnabled = false
        presenter.setActionNew()
        addWatchers()
        bundle?.let { bundle ->
            if (bundle.containsKey(TYPE_KEY)) {
                isPublicSwitch.isChecked = bundle.getBoolean(TYPE_KEY)
            }
            if (bundle.containsKey(NAME_KEY)) {
                name.setText(bundle.getString(NAME_KEY, ""))
            }

            if (bundle.containsKey(NAME_CONTAINER_ERROR)) {
                nameContainer.error = bundle.getString(NAME_CONTAINER_ERROR, "")
            }

            if (bundle.containsKey(TOPIC_CONTAINER_ERROR)) {
                topicContainer.error = bundle.getString(TOPIC_CONTAINER_ERROR, "")
            }

            if (bundle.containsKey(DESCRIPTION_CONTAINER_ERROR)) {
                topicContainer.error = bundle.getString(DESCRIPTION_CONTAINER_ERROR, "")
            }

            if (bundle.containsKey(MEMBERS_KEY)) {
                val membersString = bundle.getString(MEMBERS_KEY)
                val membersSet = ParticipantsSet(membersString)
                presenter.invitedMembersUpdated(membersSet.map { it.key })
                members.setText(presenter.getMembersText())
            }
            if (bundle.containsKey(OWNERS_KEY)) {
                val ownersText = bundle.getString(OWNERS_KEY)
                owners.setText(ownersText)
            }
            if (bundle.containsKey(TOPIC_KEY)) {
                val topicText = bundle.getString(TOPIC_KEY)
                topic.setText(topicText)
            }
        }

        // calc progress bar visibility
        presenter.isSavingObservable
                .observeOn(AndroidSchedulers.mainThread())
                .subscribe({
                    progressBar.setVisible(it)
                }, {
                    CrashInDebug.with(TAG, it)
                })
                .addTo(mStartStopDisposables)
    }

    override fun onDestroy(finishing: Boolean) {
        super.onDestroy(finishing)
        presenter.unsubscribe()
        if (finishing) presenter.cleanup()
    }

    override fun onRestoreState(stateBundle: Bundle) {
        super.onRestoreState(stateBundle)
    }

    override fun onSaveState(stateBundle: Bundle) {
        stateBundle.putBoolean(STATE_SAVED, true)

        stateBundle.putBoolean(TYPE_KEY, presenter.isPublicRoom())
        stateBundle.putString(NAME_KEY, presenter.getNameText())
        stateBundle.putString(NAME_CONTAINER_ERROR, nameContainer.error?.toString())
        stateBundle.putString(TOPIC_CONTAINER_ERROR, topicContainer.error?.toString())
        stateBundle.putString(DESCRIPTION_CONTAINER_ERROR, descriptionContainer.error?.toString())
        stateBundle.putString(MEMBERS_KEY, presenter.getInvitedMembers())
        stateBundle.putString(OWNERS_KEY, owners.text.toString())
        stateBundle.putString(TOPIC_KEY, topic.text.toString())

        super.onSaveState(stateBundle)
    }

    private fun updateType() {
        if (isPublicSwitch.isChecked) {
            addMembers.visibility = View.GONE
            membersContainer.visibility = View.GONE
            typeDescription.text = getString(R.string.tPublicRoomDescription)
            members.setText(presenter.getMembersText())
        } else {
            typeDescription.text = getString(R.string.tPrivateRoomDescription)
            addMembers.visibility = View.VISIBLE
            membersContainer.visibility = View.VISIBLE
        }
    }

    override fun onMenuItemClick(menuItem: MenuItem?): Boolean {
        when (menuItem?.itemId) {
            R.id.mi_edit_save -> {
                presenter.handleAccept()
            }
            R.id.mi_edit_discard -> {
                coordinator.flow().goBack()
            }
        }
        return super.onMenuItemClick(menuItem)
    }

    override fun updateMenuItems(menu: Menu, updateKey: String?) {
        super.updateMenuItems(menu, updateKey)
        if (isInsideDialog) {
            menu.removeItem(R.id.mi_edit_discard)
        }
    }

    private fun removeWatchers() {
        Log.d(TAG, "removeWatchers")
        name.removeTextChangedListener(nameWatcher)
        description.removeTextChangedListener(descriptionWatcher)
        topic.removeTextChangedListener(topicWatcher)
        isPublicSwitch.setOnClickListener(null)
    }

    private fun addWatchers() {
        Log.d(TAG, "addWatcher")
        name.addTextChangedListener(nameWatcher)
        description.addTextChangedListener(descriptionWatcher)
        topic.addTextChangedListener(topicWatcher)
        isPublicSwitch.setOnCheckedChangeListener { _, isChecked ->
            presenter.setIsPublic(isChecked)
            updateType()
        }
    }

    private fun getSimpleTextWatcher(f: (String) -> Unit) = object : TextWatcher {
        override fun afterTextChanged(s: Editable) {
            f(s.toString())
        }

        override fun beforeTextChanged(s: CharSequence?, start: Int, count: Int, after: Int) {}
        override fun onTextChanged(s: CharSequence?, start: Int, before: Int, count: Int) {}
    }

    private fun updateName(name: String) {
        nameContainer.error = ""
        presenter.updateName(name)
    }

    private fun updateDescription(desc: String) {
        descriptionContainer.error = ""
        presenter.updateDescription(desc)
    }

    private fun updateTopic(topic: String) {
        topicContainer.error = ""
        presenter.updateTopic(topic)
    }

    private val nameWatcher = getSimpleTextWatcher(::updateName)
    private val descriptionWatcher = getSimpleTextWatcher(::updateDescription)
    private val topicWatcher = getSimpleTextWatcher(::updateTopic)

    override fun onClick(v: View) {
        when (v.id) {
            R.id.chat_room_conf_members_add -> {
                val bundle = Bundle()
                bundle.putStringArrayList(ChatRoomMembersPickerScreen.FIXED_MEMBERS, ArrayList(presenter.activeMembers))

                val initiallySelectedIds = Controllers.get().buddy
                        .getListOfBuddiesSorted(Buddy.EBuddyType.XMPP, BuddyAvailability.All)
                        .filter { presenter.tentativeMembers.contains(it.imAddress) }
                        .map { it.uniqueIdentifier }
                        .toTypedArray()

                bundle.putStringArray(AbstractPickerScreen.INITIALLY_SELECTED_IDS, initiallySelectedIds)

                showScreenForResult(EScreenList.CHAT_ROOMS_MEMBERS_PICKER_SCREEN, bundle)
            }
        }
        super.onClick(v)

    }

    @MainThread
    override fun onNewMessage(message: Bundle, sender: IScreenEnum) {
        super.onNewMessage(message, sender)

        when (sender) {
            EScreenList.CHAT_ROOMS_MEMBERS_PICKER_SCREEN -> {
                val buddyList = message.getStringArrayList(ChatRoomMembersPickerScreen.CHOSEN_BUDDY_IDS)
                presenter.invitedMembersUpdated(buddyList)
                members.setText(presenter.getMembersText())
            }
        }
    }

    override fun onPresenterEvent(event: IPresenterEvent<*, *>) {
        val type = event.type as ChatRoomConfPresenter.Events
        Log.d(TAG, "OnPresenterEvent ${type.name}")
        when (type) {
            ChatRoomConfPresenter.Events.INVALIDATE_MENU -> TODO()
            ChatRoomConfPresenter.Events.UPDATE_INITIALS -> {
                initials.text = event.data as String
            }
            ChatRoomConfPresenter.Events.UPDATE_AVATAR_COLOR -> {
                initials.setBackgroundColor(presenter.getAvatarColor())
            }
            ChatRoomConfPresenter.Events.PROPERTY_VALIDATION_FAILED -> {
                val validationResult = event.data as ValidationResult<ChatRoom>
                for (error in validationResult.errors) {
                    when (error.property) {
                        ChatRoom::name -> nameContainer.error = error.errorMessage
                        ChatRoom::description -> descriptionContainer.error = error.errorMessage
                        ChatRoom::topic -> topicContainer.error = error.errorMessage
                    }
                }
            }
            ChatRoomConfPresenter.Events.ROOM_NAME_DUPLICATED -> {
                nameContainer.error = getString(R.string.tChatRoomNameDuplicated)
            }

            ChatRoomConfPresenter.Events.UPDATE_SCREEN -> {
                updateScreen()
            }
            ChatRoomConfPresenter.Events.CHAT_ROOM_SAVED -> {
                if (isInsideDialog) {
                    dismissScreenHolderDialog()
                }
                coordinator.flow(event.data as Bundle).goTo(EScreenList.CONVERSATION)
            }
            ChatRoomConfPresenter.Events.NO_ACTIVE_ACCOUNT -> toastShort(R.string.tAddingXmppBuddyWithoutActiveAccount)
            ChatRoomConfPresenter.Events.DISMISSED -> dismissScreen()
            ChatRoomConfPresenter.Events.SHOW_DISCARD_CHANGES_DIALOG -> showDialog(DISCARD_CHANGES_DIALOG)
            ChatRoomConfPresenter.Events.TOAST -> {
                hideKeyboard() // otherwise snack is behind
                showSnack(event.data as String)
            }
        }
    }

    private fun updateScreen() {
        removeWatchers()
        isPublicSwitch.isChecked = presenter.isPublicRoom()
        name.setText(presenter.getNameText())
        owners.setText(presenter.getOwnersText())
        members.setText(presenter.getMembersText())
        description.setText(presenter.getDescriptionText())
        topic.setText(presenter.getTopicText())
        initials.text = presenter.getInitials()
        initials.setBackgroundColor(presenter.getAvatarColor())
        updateType()
        addWatchers()
    }

    override fun createDialogForResult(whichScreen: IScreenEnum, data: Bundle): ScreenHolderDialog? {
        if (whichScreen === EScreenList.CHAT_ROOMS_MEMBERS_PICKER_SCREEN) {
            if (!isInTabletMode) {
                return ScreenHolderDialog
                        .builder(activity)
                        .screen(whichScreen)
                        .style(ScreenHolderDialog.DIALOG_MAX_SIZE_NO_MARGIN)
                        .bundle(data)
                        .build()
            } else {
                val color = Coloring.get().decodeColor(presenter.toolbarColor)
                return ScreenHolderDialog
                        .builder(activity)
                        .screen(whichScreen)
                        .anchorView(toolbar!!.findViewById(R.id.conversation_list_new))
                        .gravity(Gravity.BOTTOM)
                        .arrowColor(color)
                        .backgroundColor(color)
                        .bundle(data)
                        .build()
            }
        }

        return super.createDialogForResult(whichScreen, data)
    }

    override fun createDialog(which: Int, data: Bundle?): Dialog? {
        when (which) {
            DISCARD_CHANGES_DIALOG ->
                return AlertDialog.Builder(activity)
                        .setMessage(getString(R.string.tContactEditDiscardMessage))
                        .setCancelable(false)
                        .setPositiveButton(getString(R.string.tContactEditDiscardShort)) { _, _ -> presenter.forceDismissChatRoom() }
                        .setNegativeButton(getString(R.string.tContactEditKeepEditing)) { dialog, _ -> dialog.dismiss() }
                        .create()
        }
        return super.createDialog(which, data)
    }

    @Suppress("UNCHECKED_CAST")
    override fun getPresenterClass(): Class<Presenter> = ChatRoomConfPresenter::class.java as Class<Presenter>

    override fun getTitle(): String? = presenter.title

    @MainThread
    override fun onBackPressed(willGoToParent: Boolean): Boolean {
            presenter.handleDismissRoomBackKey()
        return true
    }

    @MainThread
    override fun onUpNavigationClicked(v: View) {
        presenter.handleDismissRoomUpKey()
    }

    private fun dismissScreen() {
        if (isInsideDialog) {
            dismissScreenHolderDialog()
        } else {
            coordinator.flow().goTo(EScreenList.CONVERSATION_LIST)
        }
    }

}


fun EditText.afterChange(cb: (String) -> Unit) {
    this.addTextChangedListener(object : TextWatcher {
        override fun afterTextChanged(s: Editable?) {
            cb(s.toString())
        }

        override fun beforeTextChanged(s: CharSequence?, start: Int, count: Int, after: Int) {}
        override fun onTextChanged(s: CharSequence?, start: Int, before: Int, count: Int) {}
    })
}
