package com.bria.voip.ui.main.im.chatroom.details

import android.os.Bundle
import android.text.format.DateUtils
import com.bria.common.R
import com.bria.common.controller.Controllers
import com.bria.common.controller.coordinatedevents.CoordinatedEventType
import com.bria.common.controller.im.refactoring.BuddyKeyUtils
import com.bria.common.controller.im.refactoring.roomdb.entities.ChatRoom
import com.bria.common.controller.im.refactoring.roomdb.entities.ChatRoomNotificationLevel
import com.bria.common.controller.im.refactoring.roomdb.entities.ChatRoomState
import com.bria.common.modules.BriaGraph
import com.bria.common.uiframework.presenters.AbstractPresenter
import com.bria.common.uiframework.presenters.IPresenterEventTypeEnum
import com.bria.common.util.CrashInDebug
import com.bria.common.util.GlobalConstants
import com.bria.common.util.im.ParticipantsSet
import io.reactivex.disposables.Disposable
import java.util.*

class ChatRoomDetailsPresenter : AbstractPresenter() {

    private val TAG = "ChatRoomDetailsPresenter"

    enum class Events : IPresenterEventTypeEnum {
        UPDATE_SCREEN, SHOW_INFO, GO_UP;
        override fun getType(): IPresenterEventTypeEnum = this
    }


    // dependencies
    private val chatRoomApi
        get() = Controllers.get().im.chatApi.chatRoomApi

    private val chatRoomRepo
        get() = Controllers.get().im.imData.chatRoomRepo

    private val coordinatedEventAggregator
        get() = BriaGraph.coordinatedEventAggregatorPublishSide
    //


    var chatRoom : ChatRoom? = null

    var creatorName : String = ""

    private var roomUpdatedDisposable : Disposable? = null

    fun setChatId(id: Long) {
        offloadLightWork {
            Controllers.get().im.imData.chatRoomRepo.getChatRoom(id)?.let {
                chatRoom = it
                creatorName = it.owners.firstParticipant?.displayName ?: ""
                firePresenterEvent(Events.UPDATE_SCREEN)
            }

        }
    }

    fun getName() = chatRoom?.name ?: ""

    fun getDescription() = chatRoom?.description ?: ""

    fun getTopic() = chatRoom?.topic ?: ""

    fun getCreationInfo() : String {
        chatRoom?.let {
            if (it.isUserOwner()) {
            return context.getString(R.string.tChatRoomCreatedInfo, DateUtils.getRelativeTimeSpanString(it.creationTime,
                        System.currentTimeMillis(), DateUtils.DAY_IN_MILLIS, DateUtils.FORMAT_SHOW_DATE), creatorName)
            } else {
                return context.getString(R.string.tChatRoomJoinedInfo, DateUtils.getRelativeTimeSpanString(it.joinTime,
                        System.currentTimeMillis(), DateUtils.DAY_IN_MILLIS, DateUtils.FORMAT_SHOW_DATE))
            }
        }
      return ""

    }

    fun getId() = chatRoom?.id ?: -1

    fun leaveRoom() {
        offloadLightWork {

            chatRoom?.apply {
                if (Controllers.get().im.chatApi.isServiceAvailable(this.accountId)) {
                    firePresenterEvent(Events.GO_UP, getChatRoomData())
                    Controllers.get().im.chatApi.chatRoomApi.deleteRoom(this)
                } else {
                    firePresenterEvent(Events.SHOW_INFO, getString(R.string.tChatRoomServiceNotReady))
                }
            }

        }
    }

    val activeMembers: Set<String>
        get() = (chatRoom!!.members + chatRoom!!.owners).asSequence().map { it.key }.toSet()


    val tentativeMembers: Set<String>
        get() = chatRoom!!.invited.asSequence().map { it.remoteAddress }.toSet()

    fun getChatRoomData(): Bundle {
        val bundle = Bundle(1)
        bundle.putLong(GlobalConstants.KEY_CHAT_ROOM_ID, chatRoom?.id ?: 0)
        return bundle
    }

    fun inviteMembers(buddyList: ArrayList<String>) {
        chatRoom!!.invited.clear()
        chatRoom!!.invited = ParticipantsSet(BuddyKeyUtils.getStringFromBuddyKeys(buddyList.toSet()))
        chatRoomApi.inviteMembers(chatRoom!!)

        trackSuccessOfInvite(chatRoom!!, buddyList);
    }

    private fun trackSuccessOfInvite(chatRoom: ChatRoom, newBuddyKeys: List<String>) {
        offloadLightWork {
            Thread.sleep(2 * 1000)

            chatRoomApi.requestMembersList(chatRoom)

            Thread.sleep(2 * 1000)

            val updatedChatRoom = chatRoomRepo.getChatRoom(chatRoom.id)
            if (updatedChatRoom == null) {
                CrashInDebug.with(TAG, "Chat room not found.")
                return@offloadLightWork
            }

            val allMembers = updatedChatRoom.members.setOfBuddyKeys.asIterable() + updatedChatRoom.owners.setOfBuddyKeys
            val success = allMembers.containsAll(newBuddyKeys)

            if (!success) {
                val bundle = Bundle()
                bundle.putString(GlobalConstants.SHOW_TOAST_KEY, String.format(getString(R.string.tChatRoomFailedToAddMembersToRoom), chatRoom.name))
                // using global toast as user might have navigated away
                coordinatedEventAggregator.sendCoordinatedEvent(CoordinatedEventType.SHOW_TOAST_EVENT, bundle)
            }
        }
    }

    fun getNotificationValue(): String {
        return when(chatRoom?.notificationsLevel ?: ChatRoomNotificationLevel.ALL) {
            ChatRoomNotificationLevel.ALL -> context.getString(R.string.tAllNewMessages)
            ChatRoomNotificationLevel.MENTIONS -> context.getString(R.string.tMentions)
            ChatRoomNotificationLevel.MUTE -> context.getString(R.string.tMute)
        }
    }

    fun isPublic(): Boolean = chatRoom?.publicRoom ?: false

    fun isUserOwner(): Boolean = chatRoom?.isUserOwner() ?: false
    fun cleanup() {
        chatRoom = null
        creatorName =""
    }

    fun isLeft(): Boolean {
        return chatRoom?.state == ChatRoomState.LEFT
    }

    override fun onCreate() {
        super.onCreate()

        roomUpdatedDisposable?.dispose()
        roomUpdatedDisposable = chatRoomRepo.getOnRoomUpdatedObservable()
            .flatMapIterable { list -> list }
            .filter { it.id == chatRoom?.id }
            .subscribe({
                chatRoom = it
                firePresenterEvent(Events.UPDATE_SCREEN)
            }, {
                CrashInDebug.with(TAG, it)
            })
    }

    override fun onDestroy() {
        roomUpdatedDisposable?.dispose()

        super.onDestroy()
    }
}
