package com.bria.voip.ui.main.im.chatroom.edit

import android.os.Bundle
import com.bria.common.controller.Controllers
import com.bria.common.controller.accounts.AccountsFilter
import com.bria.common.controller.im.ImpsUtils
import com.bria.common.controller.im.refactoring.chatroom.ChatRoomFieldValidator
import com.bria.common.controller.im.refactoring.roomdb.entities.ChatRoom
import com.bria.common.modules.BriaGraph
import com.bria.common.uiframework.presenters.AbstractPresenter
import com.bria.common.uiframework.presenters.IPresenterEventTypeEnum
import com.bria.common.util.CrashInDebug
import com.bria.common.util.GlobalConstants
import com.bria.common.util.Log
import com.bria.voip.ui.main.im.chatroom.add.getUserAtDomain
import io.reactivex.Observable
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.disposables.Disposable
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.BehaviorSubject

class ChatRoomEditPresenter : AbstractPresenter() {

    companion object {
        const val TAG = "ChatRoomEditPresenter"
    }

    // dependencies
    private val accountsCtrl
        get() = BriaGraph.accountsController

    private var chatRoomApi = Controllers.get().im.chatApi.chatRoomApi

    private var chatRoomRepo = Controllers.get().im.imData.chatRoomRepo



    private var chatRoom = ChatRoom(publicRoom = false)

    private val mNameErrorSubject = BehaviorSubject.create<String>()
    private val mDescriptionErrorSubject = BehaviorSubject.create<String>()
    private val mTopicErrorSubject = BehaviorSubject.create<String>()

    private var originalChatRoom: ChatRoom = chatRoom

    private var disposables = CompositeDisposable()
    private var waitForSavedDisposable : Disposable? = null

    enum class Events : IPresenterEventTypeEnum {
        ROOM_NAME_DUPLICATED,
        UPDATE_SCREEN,
        NO_ACTIVE_ACCOUNT,
        SHOW_DISCARD_CHANGES_DIALOG,
        DISMISSED,
        CHAT_ROOM_SAVED;

        override fun getType(): IPresenterEventTypeEnum = this
    }

    val nameErrorObservable: Observable<String>
        get() = mNameErrorSubject

    val descriptionErrorObservable: Observable<String>
        get() = mDescriptionErrorSubject

    val topicErrorObservable: Observable<String>
        get() = mTopicErrorSubject


    init {
        disposables.add(chatRoomRepo.getOnRoomDeletedObservable().observeOn(Schedulers.io()).subscribe({ list ->
            if (list.firstOrNull { it.chatKey == chatRoom.chatKey } != null) {
                firePresenterEvent(Events.DISMISSED)
            }
        },{
            Log.e(TAG, "error: ${it.message}")
        }))
    }
    fun handleSave() {
        chatRoom.name = chatRoom.name.trim()

        val account = accountsCtrl
            .getAccounts(AccountsFilter.XMPP)
            .firstOrNull { it.isRegistered }

        if (account == null) {
            firePresenterEvent(Events.NO_ACTIVE_ACCOUNT)
            return
        }

        if (chatRoom.accountId.isEmpty()) {
            chatRoom.accountId = ImpsUtils.removePortFromDomain(account.getUserAtDomain())
        }

        val validationResult = ChatRoomFieldValidator.validate(chatRoom, context)
        val propertiesUserCanEdit = setOf(ChatRoom::name, ChatRoom::topic, ChatRoom::description)
        val errors = validationResult.errors.filter { propertiesUserCanEdit.contains(it.property) }

        for (error in errors) {
            when (error.property) {
                ChatRoom::name -> mNameErrorSubject.onNext(error.errorMessage)
                ChatRoom::topic -> mTopicErrorSubject.onNext(error.errorMessage)
                ChatRoom::description -> mDescriptionErrorSubject.onNext(error.errorMessage)
            }
        }

        if (errors.any()) {
            return
        }

        if(chatRoomRepo.roomNameInTheList(chatRoom.name) && chatRoom.name != originalChatRoom.name){
            firePresenterEvent(Events.ROOM_NAME_DUPLICATED)
            return
        }

        //for Stretto we will consider single XMPP account case

        waitForSavedDisposable?.dispose()
        waitForSavedDisposable = chatRoomRepo.getOnRoomAddedObservable()
            .mergeWith(chatRoomRepo.getOnRoomUpdatedObservable())
            .subscribe({ list ->
                list
                    .firstOrNull { it.name == chatRoom.name }
                    ?.let {
                        val bundle = Bundle(1)
                        bundle.putLong(GlobalConstants.KEY_CHAT_ROOM_ID, it.id)
                        firePresenterEvent(Events.CHAT_ROOM_SAVED, bundle)
                    }
            }, {
                CrashInDebug.with(TAG, it)
            })

        chatRoomApi.updateRoomDetails(chatRoom)
    }


    fun getNameText(): String = chatRoom.name

    fun updateName(name: String) {
        chatRoom.name = name

        // clear error
        mNameErrorSubject.onNext("")
    }


    fun getDescriptionText() = chatRoom.description

    fun updateDescription(description: String) {
        chatRoom.description = description

        // clear error
        mDescriptionErrorSubject.onNext("")
    }


    fun getTopicText(): String = chatRoom.topic

    fun updateTopic(topic: String) {
        chatRoom.topic = topic

        // clear error
        mTopicErrorSubject.onNext("")
    }



    fun isDataChangeEnabled() = chatRoom.isUserOwner()

    fun handleDismissRoomBackKey() {
        if (chatRoom != originalChatRoom) {
            firePresenterEvent(Events.SHOW_DISCARD_CHANGES_DIALOG)
        } else {
            firePresenterEvent(Events.DISMISSED, getChatRoomData())
        }
    }

    fun handleDismissRoomUpKey() {
        if (chatRoom != originalChatRoom) {
            firePresenterEvent(Events.SHOW_DISCARD_CHANGES_DIALOG)
        } else {
            firePresenterEvent(Events.DISMISSED, getChatRoomData())
        }
    }

    fun setChatID(id: Long) {
        // clean slate
        chatRoom = ChatRoom(publicRoom = false)
        mNameErrorSubject.onNext("")
        mDescriptionErrorSubject.onNext("")
        mTopicErrorSubject.onNext("")

        offloadLightWork {
            chatRoomRepo.getChatRoom(id)?.let { chatRoom ->
                this.chatRoom = chatRoom
                this.originalChatRoom = chatRoom.copy()
                firePresenterEvent(Events.UPDATE_SCREEN)
            } ?: firePresenterEvent(Events.DISMISSED)
        }
    }

    private fun getChatRoomData(): Bundle {
        val bundle = Bundle(1)
        bundle.putLong(GlobalConstants.KEY_CHAT_ROOM_ID, chatRoom.id)

        return bundle
    }

    fun forceDismissChatRoom() {
        firePresenterEvent(Events.DISMISSED, getChatRoomData())
    }

    override fun onUnsubscribe() {
        super.onUnsubscribe()

        disposables.clear()
        waitForSavedDisposable?.dispose()
    }
}
