package com.bria.voip.ui.main.im.chatroom.edit

import android.app.Dialog
import android.os.Bundle
import android.support.annotation.MainThread
import android.support.design.widget.TextInputLayout
import android.support.v7.app.AlertDialog
import android.text.Editable
import android.text.TextWatcher
import android.view.Menu
import android.view.MenuItem
import android.view.View
import android.widget.EditText
import com.bria.common.uiframework.annotations.Inject
import com.bria.common.uiframework.annotations.Layout
import com.bria.common.uiframework.presenters.IPresenterEvent
import com.bria.common.uiframework.screens.AbstractScreen
import com.bria.common.util.CrashInDebug
import com.bria.common.util.GlobalConstants
import com.bria.common.util.Log
import com.bria.voip.R
import io.reactivex.Observable
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.rxkotlin.addTo

@Layout(R.layout.chat_room_edit_screen)
@com.bria.common.uiframework.annotations.Menu(R.menu.edit_menu)
class ChatRoomEditScreen<Presenter : ChatRoomEditPresenter> : AbstractScreen<Presenter>() {

    companion object {
        private const val TAG = "ChatRoomEditScreen"

        private const val DISCARD_CHANGES_DIALOG = 2
    }

    @Inject(R.id.chat_room_edit_name_container)
    private lateinit var nameContainer: TextInputLayout

    @Inject(R.id.chat_room_edit_name)
    private lateinit var name: EditText
    
    @Inject(R.id.chat_room_edit_description_container)
    private lateinit var descriptionContainer: TextInputLayout

    @Inject(R.id.chat_room_edit_description)
    private lateinit var description: EditText

    @Inject(R.id.chat_room_edit_topic_container)
    private lateinit var topicContainer: TextInputLayout

    @Inject(R.id.chat_room_edit_topic)
    private lateinit var topic: EditText

    override fun onStart(bundle: Bundle?) {
        super.onStart(bundle)

        updateScreen()

        presenter.subscribe(this)
        addWatchers()
        bundle?.let { bundle ->
                        bundle.getLong(GlobalConstants.KEY_CHAT_ROOM_ID, -1L).let { id ->
                            if (id != -1L) {
                                presenter.setChatID(id)
                            }
                    }
        }

        startObservingErrors(presenter.nameErrorObservable, nameContainer)
        startObservingErrors(presenter.topicErrorObservable, topicContainer)
        startObservingErrors(presenter.descriptionErrorObservable, descriptionContainer)
    }

    private fun startObservingErrors(observable: Observable<String>, textInputLayout: TextInputLayout) {
        observable
            .observeOn(AndroidSchedulers.mainThread())
            .subscribe({
                textInputLayout.error = it
            }, {
                CrashInDebug.with(TAG, it)
            })
            .addTo(mStartStopDisposables)
    }

    override fun onStop(finishing: Boolean) {
        presenter.unsubscribe()

        super.onStop(finishing)
    }

    override fun onMenuItemClick(menuItem: MenuItem?): Boolean {
        when (menuItem?.itemId) {
            R.id.mi_edit_save -> {
                presenter.handleSave()
            }
        }

        return super.onMenuItemClick(menuItem)
    }

    override fun updateMenuItems(menu: Menu, updateKey: String?) {
        menu.removeItem(R.id.mi_edit_discard)
        super.updateMenuItems(menu, updateKey)
    }


    private fun removeWatchers() {
        Log.d(TAG, "removeWatchers")
        name.removeTextChangedListener(nameWatcher)
        description.removeTextChangedListener(descriptionWatcher)
        topic.removeTextChangedListener(topicWatcher)
    }

    private fun addWatchers() {
        Log.d(TAG, "addWatcher")
        name.addTextChangedListener(nameWatcher)
        description.addTextChangedListener(descriptionWatcher)
        topic.addTextChangedListener(topicWatcher)
    }

    private fun getSimpleTextWatcher(f: (String) -> Unit) = object : TextWatcher {
        override fun afterTextChanged(s: Editable) {
            f(s.toString())
        }

        override fun beforeTextChanged(s: CharSequence?, start: Int, count: Int, after: Int) {}
        override fun onTextChanged(s: CharSequence?, start: Int, before: Int, count: Int) {}
    }

    private fun updateName(name: String) {
        presenter.updateName(name)
    }

    private fun updateDescription(desc: String) {
        presenter.updateDescription(desc)
    }

    private fun updateTopic(topic: String) {
        presenter.updateTopic(topic)
    }

    private val nameWatcher = getSimpleTextWatcher(::updateName)
    private val descriptionWatcher = getSimpleTextWatcher(::updateDescription)
    private val topicWatcher = getSimpleTextWatcher(::updateTopic)

    override fun onPresenterEvent(event: IPresenterEvent<*, *>) {
        val type = event.type as ChatRoomEditPresenter.Events
        Log.d(TAG, "OnPresenterEvent ${type.name}")
        when (type) {
            ChatRoomEditPresenter.Events.ROOM_NAME_DUPLICATED -> {
                nameContainer.error = getString(R.string.tChatRoomNameDuplicated)
            }
            ChatRoomEditPresenter.Events.UPDATE_SCREEN -> {
                updateScreen()
            }
            ChatRoomEditPresenter.Events.CHAT_ROOM_SAVED -> {
                dismissScreen(event.data)
            }
            ChatRoomEditPresenter.Events.NO_ACTIVE_ACCOUNT -> toastShort(R.string.tAddingXmppBuddyWithoutActiveAccount)
            ChatRoomEditPresenter.Events.DISMISSED -> {
                dismissScreen(event.data)
            }
            ChatRoomEditPresenter.Events.SHOW_DISCARD_CHANGES_DIALOG -> showDialog(DISCARD_CHANGES_DIALOG)
        }
    }

    private fun dismissScreen(data : Any?) {
        if (isInsideDialog) {
            dismissScreenHolderDialog()
        } else {
            if (data != null) {
                coordinator.flow(data as Bundle).goUp()
            } else {
                coordinator.flow().goUp()
            }
        }
    }

    private fun updateScreen() {
        removeWatchers()

        nameContainer.isEnabled = presenter.isDataChangeEnabled()
        descriptionContainer.isEnabled = presenter.isDataChangeEnabled()

        name.setText(presenter.getNameText())
        description.setText(presenter.getDescriptionText())
        topic.setText(presenter.getTopicText())

        addWatchers()
    }

    override fun createDialog(which: Int, data: Bundle?): Dialog? {
        when (which) {
            DISCARD_CHANGES_DIALOG ->
                return AlertDialog.Builder(activity)
                        .setMessage(getString(R.string.tContactEditDiscardMessage))
                        .setCancelable(false)
                        .setPositiveButton(getString(R.string.tContactEditDiscardShort)) { dialog, id -> presenter.forceDismissChatRoom() }
                        .setNegativeButton(getString(R.string.tContactEditKeepEditing)) { dialog, id -> dialog.dismiss() }
                        .create()
        }

        return super.createDialog(which, data)
    }

    @Suppress("UNCHECKED_CAST")
    override fun getPresenterClass(): Class<Presenter> = ChatRoomEditPresenter::class.java as Class<Presenter>

    override fun getTitle(): String? = getString(R.string.tEditChatRoom)

    @MainThread
    override fun onBackPressed(willGoToParent: Boolean): Boolean {
        presenter.handleDismissRoomBackKey()
        return true
    }

    @MainThread
    override fun onUpNavigationClicked(v: View) {
        presenter.handleDismissRoomUpKey()
    }

}
