package com.bria.voip.ui.main.im.chatroom.join

import android.os.Bundle
import android.os.Parcelable
import android.support.v7.widget.LinearLayoutManager
import android.support.v7.widget.RecyclerView
import android.view.MenuItem
import android.view.View
import android.widget.TextView
import com.bria.common.uiframework.annotations.Inject
import com.bria.common.uiframework.annotations.Layout
import com.bria.common.uiframework.lists.recycler.AbstractMultiSelectRecyclerAdapter
import com.bria.common.uiframework.lists.recycler.OnRecyclerItemClickListener
import com.bria.common.uiframework.presenters.IPresenterEvent
import com.bria.common.uiframework.screens.AbstractScreen
import com.bria.common.uireusable.CustomSearchView
import com.bria.common.util.GlobalConstants
import com.bria.common.util.Log
import com.bria.voip.R
import com.bria.voip.ui.main.misc.EScreenList


@Layout(R.layout.join_room_screen)
@com.bria.common.uiframework.annotations.Menu(R.menu.join_room_menu)
class JoinRoomScreen : AbstractScreen<JoinRoomPresenter>(), OnRecyclerItemClickListener {

    private companion object {
        private const val TAG = "JoinRoomScreen"

        private const val KEY_LINEAR_LAYOUT_STATE = "LINEAR_LAYOUT_STATE"
        private const val ADAPTER_STATE = "ADAPTER_STATE"
        private val SEARCH_STATE = "JOIN_ROOM_SCREEN_SEARCH_STATE"

        public const val JOINED_ROOMS = "JOINED_ROOMS"
    }

    @Inject(id = com.bria.voip.R.id.join_room_screen_search_view, col = false)
    private lateinit var mSearchView: CustomSearchView

    @Inject(R.id.join_room_screen_recycler_view)
    private lateinit var mList: RecyclerView

    @Inject(R.id.join_room_screen_info)
    private lateinit var mInfo: TextView

    private lateinit var mLinearLayoutManager: LinearLayoutManager

    private lateinit var mAdapter: JoinRoomAdapter

    override fun getPresenterClass(): Class<out JoinRoomPresenter> = JoinRoomPresenter::class.java


    override fun onPresenterEvent(event: IPresenterEvent<*, *>) {
        val type = event.type as JoinRoomPresenter.Events
        Log.d(TAG, "OnPresenterEvent ${type.name}")
        when (type) {
            JoinRoomPresenter.Events.DATA_LOADED -> {
                if (event.data != null) {
                    if (event.data as Boolean) {
                        mList.visibility = View.GONE
                        mInfo.visibility = View.VISIBLE
                        mInfo.text = getString(if (presenter.activeFilter.isEmpty()) R.string.tEmptyRoomList else R.string.tNoResults)
                    } else {
                        Log.d(TAG, "DATA CHANGED!")
                        mList.visibility = View.VISIBLE
                        mInfo.visibility = View.GONE
                        mAdapter.notifyDataChanged()
                    }
                }
            }
            JoinRoomPresenter.Events.DATA_LOADING -> {
                mList.visibility = View.GONE
                mInfo.visibility = View.VISIBLE
                mInfo.text = getString(R.string.tLoading)
            }
            JoinRoomPresenter.Events.GO_UP -> mCoordinator.flow().goUp()
            JoinRoomPresenter.Events.ERROR_JOINING_THE_ROOM -> {
                toastLong(R.string.tErrorJoiningChatRoom)
                mCoordinator.flow().goUp()
            }
            JoinRoomPresenter.Events.GO_TO -> {
                val bundle = Bundle(1)
                bundle.putLong(GlobalConstants.KEY_CHAT_ROOM_ID, event.data as Long)
                mCoordinator.flow().withBundle(bundle).goTo(EScreenList.CONVERSATION)
            }
        }
    }

    override fun onRestoreState(stateBundle: Bundle) {
        super.onRestoreState(stateBundle)
        //restoreState(stateBundle)

    }

    private val mJoinRoomAdapterCallback = object : JoinRoomAdapter.JoinRoomAdapterCallback {
        override fun getRoomInfoFor(accId: String, chatKey: String) {
            presenter.getRoomInfoFor(accId,  chatKey)
        }
    }

    override fun onCreate(bundle: Bundle?) {
        super.onCreate(bundle)

        val firstCreate = bundle == null || !bundle.containsKey(ADAPTER_STATE)
        if (firstCreate) {
            // not rotating
            presenter.initialize()
        }

        presenter.subscribe(this)
        mLinearLayoutManager = LinearLayoutManager(activity)
        mList.layoutManager = mLinearLayoutManager
        mAdapter = JoinRoomAdapter(
                presenter.settings,
                mList,
                mJoinRoomAdapterCallback,
                presenter.dataProvider,
                presenter.adapterState)
        mAdapter.selectionMode = AbstractMultiSelectRecyclerAdapter.SelectionMode.Multi
        mAdapter.setOnItemClickListener(this)
        mList.adapter = mAdapter
        restoreState(bundle)
        mSearchView.setEventHandler(object : CustomSearchView.EventHandler {
            override fun searchFor(searchString: String?) {
                presenter.filterBy(searchString)
            }

            override fun onClosed() {}
            override fun onOpened() {}
            override fun onClear() {
                presenter.filterBy("")
            }
        })

        mSearchView.setDelayedSearchInterval(0)

        //presenter.filterBy("")

        //presenter.updateList()
    }

    override fun onStart(bundle: Bundle?) {
        super.onStart(bundle)
        presenter.updateList()
    }

    fun updateVisibilities() {
        if (presenter.dataProvider.itemsCount > 0) {
            mList.visibility = View.VISIBLE
            mInfo.visibility = View.GONE
        } else {
            mList.visibility = View.GONE
            mInfo.visibility = View.VISIBLE
        }
    }


    fun restoreState(stateBundle: Bundle?) {
        if (stateBundle != null) {
            val searchState = stateBundle.getBundle(SEARCH_STATE)
            if (searchState != null) {
                mSearchView.restoreState(searchState)
            }

            val adapterState = stateBundle.getBundle(ADAPTER_STATE)
            if (adapterState != null) run {
                mAdapter.restoreState(adapterState)
            }

            val linearLayoutState = stateBundle.getParcelable<Parcelable>(KEY_LINEAR_LAYOUT_STATE);
            mLinearLayoutManager.onRestoreInstanceState(linearLayoutState)
        }
    }

    override fun onDestroy(finishing: Boolean) {
        presenter.unsubscribe()
        super.onDestroy(finishing)
    }

//    override fun onStop(finishing: Boolean) {
//        if (finishing) {
//            mSearchView.close()
//            mSearchView.setEventHandler(null)
//        }
//
//        super.onStop(finishing)
//    }

    override fun onSaveState(stateBundle: Bundle) {
        super.onSaveState(stateBundle)
        stateBundle.putBundle(ADAPTER_STATE, mAdapter.saveState())
        stateBundle.putParcelable(KEY_LINEAR_LAYOUT_STATE, mLinearLayoutManager.onSaveInstanceState())
        stateBundle.putBundle(SEARCH_STATE, mSearchView.saveState())
    }

    override fun onResume() {
        super.onResume()
        updateVisibilities()
        //mAdapter.notifyDataChanged()
    }

    override fun onPause(finishing: Boolean) {
        super.onPause(finishing)
    }

    override fun onMenuItemClick(menuItem: MenuItem?): Boolean {
        when (menuItem?.itemId) {
            R.id.join_room_mi_search -> {
                mSearchView.open(true)
                return true
            }
            R.id.join_room_mi_add -> {
                if (mAdapter.selectedIds.isNotEmpty()) {
                    val roomsBundle = Bundle()
                    roomsBundle.putStringArrayList(JOINED_ROOMS, ArrayList(mAdapter.selectedIds))
                    publishResult(roomsBundle)
                    if (isInsideDialog) {
                        dismissScreenHolderDialog()
                    }
                    //presenter.joinSelectedRooms(mAdapter.selectedIds.toList())
                } else {
                    toastLong(getString(R.string.tNoRoomSelected))
                }
            }
        }
        return super.onMenuItemClick(menuItem)

    }

    override fun getTitle(): String? {
        return presenter.title
    }

    override fun onItemClick(item: View?, index: Int) {

    }
}
