package com.bria.voip.ui.main.im.chatroom.members

import android.content.Context
import android.support.annotation.DrawableRes
import android.support.v7.widget.RecyclerView
import android.view.View
import android.widget.Button
import android.widget.ImageView
import android.widget.TextView
import com.bria.common.R
import com.bria.common.controller.contacts.ContactNamesArrangement
import com.bria.common.controller.contacts.ContactsSortBy
import com.bria.common.controller.contacts.buddy.XmppBuddy
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.core.ISettingsObserver
import com.bria.common.controller.settings.core.Settings
import com.bria.common.ui.tintWithColorRes
import com.bria.common.ui.util.ContactsSearchQueryColorizer
import com.bria.common.ui.util.ExpandableAreaController
import com.bria.common.uiframework.coloring.Coloring
import com.bria.common.uiframework.lists.recycler.AbstractMultiSelectRecyclerAdapter
import com.bria.common.uiframework.lists.recycler.AbstractRecyclerAdapter
import com.bria.common.uireusable.dataprovider.FilterQueryMatcher
import com.bria.common.uireusable.dataprovider.IDataChangeListener
import com.bria.common.uireusable.dataprovider.IFilterableDataProvider
import com.bria.common.util.ThreadExecutors
import com.bria.voip.ui.main.contacts.ContactUiHelper
import com.bria.voip.ui.shared.helpers.setVisible

class ChatRoomMembersAdapter(
        private val settings: Settings,
        recyclerView: RecyclerView,
        private val queriesHandler: QueriesHandler,
        private val actionsHandler: ActionsHandler,
        private val chatRoomMembersDataProvider: ChatRoomMembersDataProvider,
        private val state: State)
    : AbstractMultiSelectRecyclerAdapter<XmppBuddy, ChatRoomMembersAdapter.ViewHolder>(
        recyclerView,
        chatRoomMembersDataProvider,
        R.layout.chat_room_members_list_item) {

    // tint attribute in Button / TextView supported from api level 23
    private fun getTintedDrawable(context: Context, @DrawableRes id: Int) =
            context.getDrawable(id)!!.tintWithColorRes(context, R.color.primary_accent)

    private val callDrawable = getTintedDrawable(recyclerView.context, R.drawable.btn_call_small)
    private val messageDrawable = getTintedDrawable(recyclerView.context, R.drawable.vccs_message)
    private val removeDrawable = getTintedDrawable(recyclerView.context, R.drawable.vccs_delite)

    private var searchQueryColorizer = ContactsSearchQueryColorizer(Coloring.get().decodeColor(settings.getStr(ESetting.ColorBrandTint)))

    private val mSettingsObserver = object : ISettingsObserver {
        override fun onBeforeOwnerChange(newOwner: String?) {}

        override fun onSettingsChanged(changedSettings: MutableSet<ESetting>?) {
            if (changedSettings?.contains(ESetting.ColorBrandTint) == true) {
                searchQueryColorizer = ContactsSearchQueryColorizer(Coloring.get().decodeColor(settings.getStr(ESetting.ColorBrandTint)))
            }

            notifyDataChanged()
        }

        override fun onOwnerChanged() {}
    }

    private val mDataChangeListener = object : IDataChangeListener {
        override fun itemAdded(position: Int) {
            ThreadExecutors.scheduleForegroundWork {
                notifyDataChanged()
            }
        }

        override fun itemRemoved(position: Int) {
            ThreadExecutors.scheduleForegroundWork {
                notifyDataChanged()
            }
        }

        override fun itemChanged(position: Int) {
            ThreadExecutors.scheduleForegroundWork {
                notifyDataChanged(position)
            }
        }

        override fun dataChanged() {
            ThreadExecutors.scheduleForegroundWork {
                notifyDataChanged()
            }
        }
    }

    private val dataProviderListener = object : IFilterableDataProvider.DataProviderListener {
        override fun onDataLoaded(isEmpty: Boolean) {
            ThreadExecutors.scheduleForegroundWork {
                notifyDataChanged()
            }
        }

        override fun onDataLoading() {}
    }

    init {
        settings.attachWeakObserver(mSettingsObserver, setOf(ESetting.ContactSortOrder, ESetting.ContactDisplayOrder, ESetting.ColorBrandTint))

        chatRoomMembersDataProvider.attachWeakListener(dataProviderListener)

        chatRoomMembersDataProvider.observable.attachWeakObserver(mDataChangeListener)
    }

    interface QueriesHandler {
        fun amIOwner(): Boolean
    }

    interface ActionsHandler {
        fun onBuddyClicked(buddy: XmppBuddy)
        fun onCallClicked(buddy: XmppBuddy)
        fun onMessageClicked(buddy: XmppBuddy)
        fun onRemoveClicked(buddy: XmppBuddy)
        fun onResendInvitationClicked(buddy: XmppBuddy)
    }

    inner class ViewHolder internal constructor(itemView: View)
        : AbstractRecyclerAdapter.AbstractViewHolder(itemView) {

        private val buddyContainer: View = itemView.findViewById(R.id.chat_room_members_item_buddy_container)

        private val avatar: ImageView = itemView.findViewById(R.id.chat_room_members_item_avatar)
        private val presence: ImageView = itemView.findViewById(R.id.chat_room_members_item_presence_image)

        private val name1: TextView = itemView.findViewById(R.id.chat_room_members_item_name1)
        private val nameSpacer: TextView = itemView.findViewById(R.id.chat_room_members_item_name_spacer)
        private val name2: TextView = itemView.findViewById(R.id.chat_room_members_item_name2)

        private val firstName: TextView
        private val lastName: TextView

        private val moreArrow: View = itemView.findViewById(R.id.chat_room_members_item_more)
        private val moreContainer: View = itemView.findViewById(R.id.chat_room_members_item_more_container)

        private val expandableAreaController = ExpandableAreaController(moreArrow, moreContainer, ::onExpandChanged)

        private val btnCall: Button = itemView.findViewById(R.id.chat_room_members_item_action_call)
        private val btnMessage: Button = itemView.findViewById(R.id.chat_room_members_item_action_message)
        private val btnRemove: Button = itemView.findViewById(R.id.chat_room_members_item_action_remove)
        private val btnResendInvitation: Button = itemView.findViewById(R.id.chat_room_members_item_action_resend_invitation)

        private var buddy: XmppBuddy? = null

        init {
            val displayFirstNameFirst = ContactNamesArrangement.fromRawIntValue(settings.getInt(ESetting.ContactDisplayOrder)) == ContactNamesArrangement.FirstNameThenLastName

            firstName = if (displayFirstNameFirst) name1 else name2
            lastName = if (!displayFirstNameFirst) name1 else name2

            btnCall.setCompoundDrawablesRelativeWithIntrinsicBounds(callDrawable, null, null, null)
            btnMessage.setCompoundDrawablesRelativeWithIntrinsicBounds(messageDrawable, null, null, null)
            btnRemove.setCompoundDrawablesRelativeWithIntrinsicBounds(removeDrawable, null, null, null)

            buddyContainer.setOnClickListener {
                if (buddy != null) {
                    actionsHandler.onBuddyClicked(buddy!!)
                }
            }

            btnCall.setOnClickListener {
                if (buddy != null) {
                    actionsHandler.onCallClicked(buddy!!)
                }
            }

            btnMessage.setOnClickListener {
                if (buddy != null) {
                    actionsHandler.onMessageClicked(buddy!!)
                }
            }
            btnRemove.setOnClickListener {
                if (buddy != null) {
                    actionsHandler.onRemoveClicked(buddy!!)
                }
            }

            btnResendInvitation.setOnClickListener {
                if (buddy != null) {
                    actionsHandler.onResendInvitationClicked(buddy!!)
                }
            }
        }

        fun bindToModel(buddy: XmppBuddy) {
            this.buddy = buddy

            if (buddy.avatarIcon != null) {
                avatar.setImageBitmap(buddy.avatarIcon)
            } else {
                avatar.setImageDrawable(avatar.context.getDrawable(com.bria.voip.R.drawable.default_avatar))
            }

            presence.setImageDrawable(buddy.getPresenceIcon(presence.context))

            firstName.text = buddy.firstName
            lastName.text = buddy.lastName

            // if first info is empty, there will be a margin issue and the thing will look weird
            if (name1.text.isNullOrEmpty()) {
                name1.visibility = View.GONE
                nameSpacer.visibility = View.GONE
            } else {
                name1.visibility = View.VISIBLE
                nameSpacer.visibility = View.VISIBLE
            }

            ContactUiHelper.signifySortOrderWithTypeface(firstName, lastName, ContactsSortBy.fromRawIntValue(settings.getInt(ESetting.ContactSortOrder)))

            val filterQueryMatcher = FilterQueryMatcher(chatRoomMembersDataProvider.searchQuery)
            searchQueryColorizer.colorizeContactQuerySearch(name1, filterQueryMatcher.queryTokens)
            searchQueryColorizer.colorizeContactQuerySearch(name2, filterQueryMatcher.queryTokens)


            expandableAreaController.setExpanded(state.expandedItems.contains(buddy))
            btnCall.visibility = if (buddy.vCard?.phoneList?.isNotEmpty() == true) View.VISIBLE else View.GONE
            //btnRemove.visibility = if (queriesHandler.amIOwner()) View.VISIBLE else View.GONE
            btnRemove.visibility = View.GONE

            btnResendInvitation.setVisible(!state.isPublicRoom && state.isOwner)
        }

        fun onExpandChanged(isExpanded: Boolean) {
            if (buddy == null) {
                return
            }

            if (isExpanded) {
                state.expandedItems.add(buddy!!)
            } else {
                state.expandedItems.remove(buddy!!)
            }
        }
    }

    override fun provideNewViewHolder(inflatedView: View, viewType: Int): ViewHolder {
        return ViewHolder(inflatedView)
    }

    override fun onNewViewHolderReady(viewHolder: ViewHolder, viewType: Int) {}

    override fun onUpdateNeeded(viewHolder: ViewHolder, position: Int) {
        val buddy = mDataProvider.getItemAt(position) as XmppBuddy?
        if (buddy == null) {
            return
        }

        viewHolder.bindToModel(buddy)
    }

    class State {
        val expandedItems: MutableSet<XmppBuddy> = mutableSetOf()
        var isPublicRoom: Boolean = false
        var isOwner: Boolean = false
    }
}

