package com.bria.voip.ui.main.im.chatroom.members

import android.os.Bundle
import com.bria.common.R
import com.bria.common.controller.contacts.buddy.Buddy
import com.bria.common.controller.contacts.buddy.XmppBuddy
import com.bria.common.controller.contacts.local.data.PhoneNumber
import com.bria.common.controller.im.refactoring.BuddyKeyUtils
import com.bria.common.controller.im.refactoring.chatroom.ChatRoomCacheState
import com.bria.common.controller.im.refactoring.chatroom.ChatRoomMembers
import com.bria.common.controller.im.refactoring.db.entities.ImConversationData
import com.bria.common.controller.im.refactoring.roomdb.entities.ChatRoom
import com.bria.common.controller.settings.EAccountSetting
import com.bria.common.controller.settings.branding.EAccountType
import com.bria.common.modules.BriaGraph
import com.bria.common.sdkwrapper.CallData
import com.bria.common.uiframework.presenters.AbstractPresenter
import com.bria.common.uiframework.presenters.IPresenterEventTypeEnum
import com.bria.common.uireusable.dataprovider.IFilterableDataProvider
import com.bria.common.util.CrashInDebug
import com.bria.common.util.GlobalConstants
import com.bria.common.util.Log
import com.bria.voip.BriaSipGraph
import com.bria.voip.ui.main.contacts.ChooseNumberOfXmppBuddyPresenter
import io.reactivex.disposables.CompositeDisposable
import io.reactivex.disposables.Disposable
import io.reactivex.rxkotlin.addTo
import java.util.*

class ChatRoomMembersPresenter : AbstractPresenter() {

    companion object {
        private const val TAG = "ChatRoomMembersPresenter"
    }

    // dependencies
    val settings
        get() = BriaGraph.settings

    val accountsCtrl
        get() = BriaGraph.accountsController

    private val phoneCtrl
        get() = BriaGraph.phoneCtrl

    private val singleTouchToCallHelper
        get() = BriaGraph.singleTouchToCallHelper

    private val buddyCtrl
        get() = BriaGraph.buddyCtrl

    private val chatRoomApi
        get() = BriaGraph.chatRoomApi

    private val chatRoomRepo
        get() = BriaGraph.chatRoomRepo

    private val chatRoomMemberCache
        get() = BriaGraph.chatRoomMemberCache


    private var updateRoomStateDisposable: Disposable? = null

    enum class Events : IPresenterEventTypeEnum {
        SHOW_NUMBER_CHOOSER_FOR_CALL,
        GO_TO_IM_CONVERSATION,
        SHOW_MESSAGE_LONG,
        DATA_LOADED,
        DATA_LOADING,
        GO_UP,
        SHOW_SNACK,
        SHOW_PHONE_ITEM_ACTIONS_DIALOG;

        override fun getType(): IPresenterEventTypeEnum = this
    }

    private val dataProviderListener = object : IFilterableDataProvider.DataProviderListener {
        override fun onDataLoaded(isEmpty: Boolean) {
            firePresenterEvent(Events.DATA_LOADED, isEmpty)
        }

        override fun onDataLoading() {
            // reserving loading event for trip to server
        }
    }


    var chatRoomId: Long = 0
    var chatRoom: ChatRoom? = null
    var adapterState = ChatRoomMembersAdapter.State()

    val dataProvider = BriaSipGraph.chatRoomMembersDataProvider()

    val compositeDisposable = CompositeDisposable()

    init {
        dataProvider.attachWeakListener(dataProviderListener)

        chatRoomMemberCache.observable
            .filter { it.chatRoomId == chatRoomId }
            .subscribe({
                setBuddiesToDataProvider(it.chatRoomMembers)
            }, {
                CrashInDebug.with(TAG, it)
            })
            .addTo(compositeDisposable)

        chatRoomApi.onErrorEventObservable
            .filter { it.chatKey == chatRoom?.chatKey }
            .subscribe({
                firePresenterEvent(Events.SHOW_SNACK, getString(R.string.tSomethingWrongTryAgain))

                val isEmpty = true
                firePresenterEvent(Events.DATA_LOADED, isEmpty)
            }, {

            })
            .addTo(compositeDisposable)
    }

    fun initForRoom(chatRoomId: Long) {
        this.chatRoomId = chatRoomId

        adapterState = ChatRoomMembersAdapter.State()


        firePresenterEvent(Events.DATA_LOADING)

        offloadLightWork {
            chatRoom = chatRoomRepo.getChatRoom(chatRoomId)
            if (chatRoom == null) {
                CrashInDebug.with(TAG, "Chat room not found for id: $chatRoomId")
                firePresenterEvent(Events.GO_UP)

                return@offloadLightWork
            }

            chatRoom?.let { room ->
                adapterState.isPublicRoom = room.publicRoom
                adapterState.isOwner = room.isUserOwner()

                // refresh member list
                if (!chatRoomApi.requestMembersList(room)) {
                    // if user rushed to room, chat object will not be ready, and requestMembersList call returns false

                    val isEmpty = true
                    firePresenterEvent(Events.DATA_LOADED, isEmpty)

                    firePresenterEvent(Events.SHOW_SNACK, getString(R.string.tSomethingWrongTryAgain))
                }
            }

            val optional = chatRoomMemberCache.getAllMembers(chatRoomId)
            if (optional.hasValue) {
                setBuddiesToDataProvider(optional.value)
            }
        }

        initChatRoomStateWatcher()
    }

    private fun initChatRoomStateWatcher() {
        updateRoomStateDisposable = chatRoomApi.getOnChatRoomStateObservable()
            .filter { chatRoom?.chatKey == it.key && it.state == ChatRoomCacheState.REMOTELY_LEFT }
            .subscribe({
                firePresenterEvent(Events.SHOW_SNACK, getString(com.bria.common.R.string.tUserLeftChatRoom))
                firePresenterEvent(Events.GO_UP)
            }, {
                CrashInDebug.with(TAG, it)
            })
    }

    fun sendMessageTo(buddy: Buddy) {
        val account = mControllers.accounts.getAccount(buddy.accountNickname)
        if (account != null && account.isRegistered && (account.type == EAccountType.Xmpp || account.getBool(EAccountSetting.IsIMPresence))) {
            val bundle = Bundle()
            val buddyKeys = ArrayList<String>()
            buddyKeys.add(BuddyKeyUtils.getNewBuddyKey(buddy))
            bundle.putStringArrayList(GlobalConstants.KEY_BUDDY_IDS, buddyKeys)
            bundle.putInt(GlobalConstants.KEY_CONVERSATION_TYPE, ImConversationData.IM_XMPP)
            firePresenterEvent(Events.GO_TO_IM_CONVERSATION, bundle)
        } else {
            firePresenterEvent(Events.SHOW_MESSAGE_LONG, getString(com.bria.common.R.string.tNoAccountActive))
        }
    }

    fun removeFromChatRoom(buddy: Buddy) {
        // chatRoom?.let { imCtrl.chatApi.chatRoomApi.removeMember(it, Participant(buddy.uniqueIdentifier)) }
    }

    fun call(buddy: XmppBuddy) {
        val account = accountsCtrl.primaryAccount
        if (account == null) {
            firePresenterEvent(Events.SHOW_SNACK, getString(com.bria.common.R.string.tNoActiveAccount))
            return
        }

        val numbers = ChooseNumberOfXmppBuddyPresenter.getNumbers(buddy)
        if (numbers.isEmpty()) {
            Log.e(TAG, "callXmppBuddy - unexpected case: xmpp buddy has no phone numbers")
        } else if (numbers.size == 1) {
            callXmppNumber(buddy, numbers[0])
        } else if (numbers.size > 1) {
            firePresenterEvent(Events.SHOW_NUMBER_CHOOSER_FOR_CALL, buddy)
        }
    }

    data class PhoneActionsDialogParams(
            val contactName: String?,
            val phoneNumber: PhoneNumber)

    fun callXmppNumber(xmppChatParticipantKey: String?, number: String?) {
        if (xmppChatParticipantKey == null) {
            return
        }

        val buddy = buddyCtrl.getBuddyByNewKey(xmppChatParticipantKey) as? XmppBuddy
        if (buddy == null) {
            CrashInDebug.with(TAG, "buddy not found for xmppChatParticipantKey: $xmppChatParticipantKey")
            return
        }

        callXmppNumber(buddy, number)
    }

    private fun callXmppNumber(buddy: Buddy, number: String?) {
        val account = mControllers.accounts.primaryAccount
        if (account == null) {
            firePresenterEvent(Events.SHOW_SNACK, getString(com.bria.common.R.string.tNoActiveAccount))
            return
        }

        if (number.isNullOrBlank()) {
            return
        }

        if (singleTouchToCallHelper.shouldCallOnSingleTouch()) {
            val nickname = account.getStr(EAccountSetting.Nickname)
            phoneCtrl.call(number, nickname, buddy.displayName, CallData.ECallType.Generic)
        } else {
            val phoneNumber = PhoneNumber(PhoneNumber.MainType.PHONE_NUMBER, PhoneNumber.SOFTPHONE_CODE, number!!)
            val phoneActionsDialogParams = PhoneActionsDialogParams(buddy.displayName, phoneNumber)
            firePresenterEvent(Events.SHOW_PHONE_ITEM_ACTIONS_DIALOG, phoneActionsDialogParams)
        }
    }

    fun filterBy(searchString: String?) {
        dataProvider.filter(searchString)
    }

    override fun onUnsubscribe() {
        updateRoomStateDisposable?.dispose()

        super.onUnsubscribe()
    }

    fun resendInvitation(buddy: XmppBuddy) {
        chatRoomApi.reInviteToPrivateRoom(chatRoomId, buddy)
        firePresenterEvent(Events.SHOW_SNACK, getString(com.bria.common.R.string.tChatRoomInvitationSent))
    }

    override fun onDestroy() {
        compositeDisposable.clear()

        super.onDestroy()
    }

    private fun setBuddiesToDataProvider(chatRoomMembers: ChatRoomMembers) {
        val buddies  = chatRoomMembers.owners
            .plus(chatRoomMembers.members)

            // BRACE-12968 duplicate member appears
            // could it be that owner is repeated in members?
            .distinct()

            // remove self
            .filter { BuddyKeyUtils.getImAdressFromNewBuddyKey(it) != chatRoom?.accountId }

            .map { buddyCtrl.getBuddyByNewKey(it) as? XmppBuddy }
            .filterNotNull()


        dataProvider.setBuddies(buddies)
    }
}
