package com.bria.voip.ui.main.im.chatroom.mentions

import android.content.Context
import android.database.DataSetObservable
import android.database.DataSetObserver
import android.view.LayoutInflater
import android.view.View
import android.view.ViewGroup
import android.widget.*
import com.bria.common.controller.contacts.ContactsSortBy
import com.bria.common.controller.contacts.buddy.*
import com.bria.common.controller.im.refactoring.chatroom.ChatRoomApi
import com.bria.common.controller.im.refactoring.roomdb.ChatRoomRepo
import com.bria.common.controller.im.refactoring.roomdb.entities.ChatRoom
import com.bria.common.controller.settings.ESetting
import com.bria.common.controller.settings.core.Settings
import com.bria.common.uireusable.dataprovider.FilterQueryMatcher
import com.bria.common.util.*
import com.bria.voip.R
import io.reactivex.android.schedulers.AndroidSchedulers
import io.reactivex.disposables.Disposable
import io.reactivex.rxkotlin.withLatestFrom
import io.reactivex.schedulers.Schedulers
import io.reactivex.subjects.BehaviorSubject

class ChatRoomMentionsAdapter(
        private val context: Context,
        private val settings: Settings,
        private val buddyController: IBuddyCtrlEvents,
        buddyObservable: IObservable<IBuddyCtrlObserver>,
        chatRoomRepo: ChatRoomRepo,
        private val chatRoomApi: ChatRoomApi
) : ListAdapter, Filterable {

    private val chatRoomSubject = BehaviorSubject.createDefault(Optional.missing<ChatRoom>())
    private val xmppBuddiesSubject = BehaviorSubject.createDefault(buddyController.allBuddies
        .filter { it.buddyType == Buddy.EBuddyType.XMPP }
        .map { it as XmppBuddy })

    private var subscription: Disposable?

    // hold a reference to observers
    private var mBuddyCtrlObserver: IBuddyCtrlObserver = object : IBuddyCtrlObserver {
        override fun onBuddyListUpdated() {
            xmppBuddiesSubject.onNext(buddyController.allBuddies
                .filter { it.buddyType == Buddy.EBuddyType.XMPP }
                .map { it as XmppBuddy }
            )
        }

        override fun onBuddyPresenceChanged(buddy: Buddy) {
            val isDisplayed = mData.any { it.buddy == buddy }
            if (isDisplayed) {
                notifyDataSetChanged()
            }
        }

        override fun onBuddyFilterChanged(aBuddyFilterType: IBuddyCtrlEvents.EBuddyFilterType?) {}
    }


    init {
        buddyObservable.attachWeakObserver(mBuddyCtrlObserver)

        // when room is updated in the DB - e.g. member list is changed
        val roomUpdatedObservable = chatRoomRepo.getOnRoomUpdatedObservable()
            .flatMapIterable { it }
            .filter { chatRoomSubject.value?.hasValue == true && chatRoomSubject.value?.value?.id == it.id }

        // or we are looking at the new chat room
        val chatRoomObservable = chatRoomSubject
            .filter { it.hasValue }
            .map { it.value }
            .distinctUntilChanged()

        val chatRoomMemberKeysObservable = roomUpdatedObservable
            .mergeWith(chatRoomObservable)
            .observeOn(Schedulers.computation())
            .map { it.members.setOfBuddyKeys + it.owners.setOfBuddyKeys }
            .distinctUntilChanged()

        val chatRoomMembersObservable = chatRoomMemberKeysObservable
            .observeOn(Schedulers.computation())

            // project XmppBuddy objects
            .withLatestFrom(xmppBuddiesSubject.observeOn(Schedulers.computation())) { buddyKeys, buddies ->
                buddies.filter { buddy ->
                    buddyKeys.contains(buddy.uniqueIdentifier)
                }
            }

        subscription = chatRoomMembersObservable
            .observeOn(AndroidSchedulers.mainThread())
            .subscribe({
                mMembers = it

                // if refilter in case popup is open
                mData = calcDisplayedData(mLastFilter?.toString())
                notifyDataSetChanged()


            }, {
                CrashInDebug.with(TAG, it)
            })
    }

    private val predefinedItems by lazy {
        ChatRoomMentionAtItem.getKnownItems(context)
            .sortedBy { it.atText }
            .map { ChatRoomMentionItem(null, it) }
    }


    fun setChatRoom(chatRoom: Optional<ChatRoom>) {
        chatRoomSubject.onNext(chatRoom)

        if (chatRoom.hasValue && chatRoomSubject.hasValue() && chatRoomSubject.value?.hasValue == true && chatRoomSubject.value?.value?.id != chatRoom.value.id) {
            chatRoomApi.requestMembersList(chatRoom.value)
        }
    }

    fun close() {
        subscription?.dispose()
    }

    override fun getView(position: Int, convertView: View?, parent: ViewGroup): View {
        val view = convertView ?: LayoutInflater.from(context).inflate(R.layout.chat_room_mention_item, parent, false)

        val item = getItem(position) as ChatRoomMentionItem

        bind(view, item)

        return view
    }

    private fun bind(view: View, item: ChatRoomMentionItem) {
        val buddyContainer = view.findViewById<View>(R.id.buddy_container)
        val atContainer = view.findViewById<View>(R.id.at_container)

        if (item.buddy != null) {
            buddyContainer.visibility = View.VISIBLE
            atContainer.visibility = View.GONE

            bindBuddy(view, item.buddy)
        } else {
            buddyContainer.visibility = View.GONE
            atContainer.visibility = View.VISIBLE

            bindAtItem(view, item.atItem!!)
        }
    }

    private fun bindBuddy(view: View, buddy: XmppBuddy) {
        val avatar = view.findViewById<ImageView>(R.id.avatar)
        val displayName = view.findViewById<TextView>(R.id.display_name)
        val presence = view.findViewById<ImageView>(R.id.presence_image)

        if (buddy.avatarIcon != null) {
            avatar.setImageBitmap(buddy.avatarIcon)
        } else {
            avatar.setImageDrawable(AndroidUtils.getDrawable(context, R.drawable.default_avatar))
        }

        displayName.text = buddy.displayName
        presence.setImageDrawable(buddy.getPresenceIcon(context))
    }

    private fun bindAtItem(view: View, atItem: ChatRoomMentionAtItem) {
        val description = view.findViewById<TextView>(R.id.notify_whom_text)
        description.setText(atItem.description)

        val text = view.findViewById<TextView>(R.id.at_text)
        text.text = atItem.atText
    }

    private var mMembers = listOf<XmppBuddy>()

    private var mLastFilter: CharSequence? = null
    private var mData = listOf<ChatRoomMentionItem>()

    override fun getCount(): Int {
        return mData.size
    }

    override fun isEmpty(): Boolean {
        return mData.isEmpty()
    }

    override fun getItem(position: Int): Any {
        return mData.get(position)
    }

    companion object {
        private const val TAG = "ChatRoomMentionsAdapter"

        private fun ChatRoomMentionItem.matches(atSearchString: CharSequence): Boolean {
            if (this.buddy != null) {
                return buddy.matches(atSearchString.trimStart('@'));
            } else {
                return this.atItem!!.matches(atSearchString);
            }
        }

        private fun XmppBuddy.matches(searchStringWithoutAt: CharSequence): Boolean {
            val match = this.firstName?.startsWith(searchStringWithoutAt, true) == true ||
                    this.lastName?.startsWith(searchStringWithoutAt, true) == true ||
                    this.displayName?.startsWith(searchStringWithoutAt, true) == true ||
                    this.imAddress?.startsWith(searchStringWithoutAt, true) == true

            if (match) {
                return true
            }

            if(searchStringWithoutAt.contains(' ')){
                // do not match parts i.e. we don't want to match Jon Conor if query was 'Sarah Conor'
                return false
            }

            val filterQueryMatcher = FilterQueryMatcher(searchStringWithoutAt.toString())
            if (filterQueryMatcher.isMatch(displayName)) {
                return true
            }

            return false
        }

        private fun ChatRoomMentionAtItem.matches(prefixIncludesAt: CharSequence): Boolean {
            return this.atText.startsWith(prefixIncludesAt, true) ||
                    this.description.startsWith(prefixIncludesAt.trim('@'), true)
        }

        fun filter(list: List<ChatRoomMentionItem>, atSearchString: CharSequence?): List<ChatRoomMentionItem> {
            if (atSearchString?.startsWith("@") == true) {
                return list.filter { it.matches(atSearchString) }
            } else {
                return listOf()
            }
        }
    }

    inner class SimpleFilter : Filter() {

        override fun performFiltering(atSearchString: CharSequence?): Filter.FilterResults {
            mLastFilter = atSearchString

            val results = Filter.FilterResults()

            val list = calcDisplayedData(atSearchString?.toString())

            results.values = list
            results.count = list.size

            return results
        }

        override fun publishResults(constraint: CharSequence?, results: Filter.FilterResults) {
            mData = results.values as List<ChatRoomMentionItem>

            notifyDataSetChanged()
        }
    }

    private fun calcDisplayedData(filterString: String?): List<ChatRoomMentionItem> {
        val memberItems = mMembers
            .sortedWith(BuddyComparator(ContactsSortBy.fromRawIntValue(settings.getInt(ESetting.ContactSortOrder))))
            .map { ChatRoomMentionItem(it, null) }


        val list = filter(memberItems + predefinedItems, filterString)

        return list
    }

    // mandatory implementations, no use case specific logic

    private val mDataSetObservable = DataSetObservable()

    override fun registerDataSetObserver(observer: DataSetObserver?) {
        mDataSetObservable.registerObserver(observer)
    }

    override fun unregisterDataSetObserver(observer: DataSetObserver?) {
        mDataSetObservable.unregisterObserver(observer)
    }

    private val mFilter = SimpleFilter()

    override fun getFilter(): Filter {
        return mFilter
    }


    override fun getItemId(position: Int): Long {
        return position.toLong()
    }

    override fun hasStableIds(): Boolean {
        return false
    }

    override fun areAllItemsEnabled(): Boolean {
        return true
    }

    override fun isEnabled(position: Int): Boolean {
        return true
    }

    private fun notifyDataSetChanged() {
        ThreadExecutors.scheduleForegroundWork {
            mDataSetObservable.notifyChanged()
        }
    }

    private fun notifyDataSetInvalidated() {
        ThreadExecutors.scheduleForegroundWork {
            mDataSetObservable.notifyInvalidated()
        }
    }

    override fun getViewTypeCount(): Int {
        return 1
    }

    override fun getItemViewType(position: Int): Int {
        return 0
    }
}
