package com.bria.voip.ui.main.im.chatroom.notifications

import android.os.Bundle
import android.support.annotation.MainThread
import android.view.Gravity
import android.view.Menu
import android.view.MenuItem
import android.view.View
import android.widget.FrameLayout
import android.widget.Switch
import android.widget.TextView
import com.bria.common.controller.im.refactoring.roomdb.entities.ChatRoomNotificationLevel
import com.bria.common.uiframework.annotations.Clickable
import com.bria.common.uiframework.annotations.Inject
import com.bria.common.uiframework.annotations.Layout
import com.bria.common.uiframework.presenters.IPresenterEvent
import com.bria.common.uiframework.screens.AbstractScreen
import com.bria.common.util.GlobalConstants
import com.bria.common.util.Log
import com.bria.voip.R

@Layout(R.layout.chat_room_notification_settings)
@com.bria.common.uiframework.annotations.Menu(R.menu.edit_menu)
class ChatRoomNotificationsSettingsScreen<Presenter : ChatRoomNotificationsSettingsPresenter> : AbstractScreen<Presenter>() {

    companion object {
        private const val TAG = "ChatRoomNotificationsSettingsScreen"
        const val NOTIFICATION_ACTIVE_LEVEL_ID = "NOTIFICATION_ACTIVE_LEVEL_ID"
    }

    @Inject(R.id.chat_room_notification_settings_mute)
    private lateinit var muteSwitch: Switch


    @Inject(R.id.chat_room_notification_setting_level_label)
    private lateinit var levelLabel: TextView


    @Inject(R.id.chat_room_notification_setting_level_value)
    private lateinit var levelValue: TextView

    @Clickable
    @Inject(R.id.chat_room_notification_setting_level_container)
    private lateinit var levelContainer: FrameLayout

    @Suppress("UNCHECKED_CAST")
    override fun getPresenterClass(): Class<Presenter> = ChatRoomNotificationsSettingsPresenter::class.java as Class<Presenter>

    override fun onStart(bundle: Bundle?) {
        super.onStart(bundle)
        presenter.subscribe(this)

        bundle?.getString(NOTIFICATION_ACTIVE_LEVEL_ID)?.let {
            presenter.activeLevel = ChatRoomNotificationLevel.valueOf(it)
        }

        bundle?.getLong(GlobalConstants.KEY_CHAT_ROOM_ID, -1)?.let { id ->
            presenter.setChatId(id)
        }
    }

    override fun onSaveState(stateBundle: Bundle) {
        stateBundle.putLong(GlobalConstants.KEY_CHAT_ROOM_ID, presenter.getId())
        stateBundle.putString(NOTIFICATION_ACTIVE_LEVEL_ID, presenter.activeLevel?.name ?: ChatRoomNotificationLevel.ALL.name)
        super.onSaveState(stateBundle)
    }
    override fun onStop(finishing: Boolean) {
        super.onStop(finishing)
        presenter.activeLevel = null
        presenter.unsubscribe()
    }

    override fun onDestroy(finishing: Boolean) {
        super.onDestroy(finishing)
        if (finishing) destroyPresenter()
    }

    override fun onClick(v: View) {
        when (v.id) {
            R.id.chat_room_notification_setting_level_container -> {
                showPopupMenu(R.menu.chat_room_notifications_level, levelLabel, Gravity.CENTER_HORIZONTAL or Gravity.BOTTOM)
            }
        }
        super.onClick(v)
    }

    override fun updateMenuItems(menu: Menu, updateKey: String?) {
        super.updateMenuItems(menu, updateKey)

        menu.removeItem(R.id.mi_edit_discard)
    }

    @MainThread
    override fun onMenuItemClick(menuItem: MenuItem): Boolean {
        return when(menuItem.itemId) {
            R.id.chat_room_notifications_all -> {
                presenter.activeLevel = ChatRoomNotificationLevel.ALL
                true
            }
            R.id.chat_room_notifications_mentions -> {
                presenter.activeLevel = ChatRoomNotificationLevel.MENTIONS
                true
            }
            R.id.mi_edit_save -> {
                presenter.actionSave()
                true
            }
            else -> super.onMenuItemClick(menuItem)
        }
    }

    override fun onPresenterEvent(event: IPresenterEvent<*, *>) {
        val type = event.type as ChatRoomNotificationsSettingsPresenter.Events
        Log.d(TAG, "OnPresenterEvent ${type.name}")
        when (type) {
            ChatRoomNotificationsSettingsPresenter.Events.UPDATE_SCREEN -> updateScreen()
            ChatRoomNotificationsSettingsPresenter.Events.GO_UP -> {
                val bundle = Bundle(1)
                bundle.putLong(GlobalConstants.KEY_CHAT_ROOM_ID, presenter.getId())
                coordinator.flow(bundle).goBack()
            }
        }
    }

    private fun updateScreen() {
        muteSwitch.setOnCheckedChangeListener(null)
        when (presenter.activeLevel) {
            ChatRoomNotificationLevel.ALL, null -> {
                muteSwitch.isChecked = false
                levelValue.text = getString(R.string.tAllNewMessages)
                levelValue.visibility = View.VISIBLE
                levelLabel.visibility = View.VISIBLE
            }
            ChatRoomNotificationLevel.MENTIONS -> {
                muteSwitch.isChecked = false
                levelValue.text = getString(R.string.tMentions)
                levelValue.visibility = View.VISIBLE
                levelLabel.visibility = View.VISIBLE
            }
            ChatRoomNotificationLevel.MUTE -> {
                muteSwitch.isChecked = true
                levelValue.visibility = View.GONE
                levelLabel.visibility = View.GONE
            }
        }
        muteSwitch.setOnCheckedChangeListener { _, isChecked ->
            presenter.setMute(isChecked)
        }
    }


    override fun getTitle() = getString(R.string.tNotifications)
}
