package com.bria.voip.ui.shared.pickers

import android.os.Bundle
import android.support.annotation.MainThread
import android.support.design.widget.Snackbar
import android.view.View
import android.widget.AdapterView
import com.bria.common.controller.buddy.BuddyAvailability
import com.bria.common.controller.contacts.buddy.Buddy
import com.bria.common.customelements.internal.SpinnerUtils
import com.bria.common.uiframework.annotations.Layout
import com.bria.common.uiframework.annotations.Menu
import com.bria.common.uiframework.lists.recycler.AbstractMultiSelectRecyclerAdapter
import com.bria.common.uiframework.presenters.IPresenterEvent
import com.bria.common.uireusable.adapters.BuddyListAdapter
import com.bria.voip.R
import com.bria.voip.ui.shared.pickers.generic.AbstractPickerScreen
import com.seraphim.chips.Chip

@Layout(R.layout.picker_screen)
@Menu(R.menu.picker_screen_menu)
class ChatRoomMembersPickerScreen : AbstractPickerScreen<ChatRoomMembersPickerPresenter, Buddy>() {

    companion object {
        val FIXED_MEMBERS = "fixed_members"

        val CHOSEN_BUDDY_IDS = "chosen"
        val CHOSEN_BUDDY_NAMES = "chosen_names"
        val CHOSEN_BUDDY_ADDRESSES = "chosen_addresses"
    }

    private val SPINNER_ALL_BUDDIES = 0
    private val SPINNER_ONLINE_BUDDIES = 1
    private val SPINNER_STATE = "buddy_picker_spinner_state"

    @MainThread
    override fun onCreate(bundle: Bundle?) {
        super.onCreate(bundle)

        SpinnerUtils.populateSpinner(mSpinner, getString(R.string.tDisplayAllBuddies), getString(R.string.tDisplayOnlyOnlineBuddies))
        SpinnerUtils.setSpinnerListener(mSpinner, object : SpinnerUtils.OnItemSelectedAdapter() {
            override fun onItemSelected(parent: AdapterView<*>, view: View, position: Int, id: Long) {
                mRecyclerView.isLayoutFrozen = true
                if (position == SPINNER_ALL_BUDDIES) {
                    presenter.setOnlineStatusFilter(BuddyAvailability.All)
                } else if (position == SPINNER_ONLINE_BUDDIES) {
                    presenter.setOnlineStatusFilter(BuddyAvailability.Online)
                }
            }
        })

        if (bundle != null) {
            val fixedMembers = bundle.getStringArrayList(ChatRoomMembersPickerScreen.FIXED_MEMBERS) ?: arrayListOf()

            presenter.setMembers(fixedMembers.toSet())

            val spinnerPosition = bundle.getInt(SPINNER_STATE, -1)
            if (spinnerPosition != -1) {
                mSpinner.setSelection(spinnerPosition)
            }
        }

        mSpinner.visibility = View.VISIBLE
    }

    @MainThread
    override fun onSaveState(stateBundle: Bundle) {
        super.onSaveState(stateBundle)
        stateBundle.putInt(SPINNER_STATE, mSpinner.selectedItemPosition)
    }

    override fun getPresenterClass(): Class<out ChatRoomMembersPickerPresenter> {
        return ChatRoomMembersPickerPresenter::class.java
    }

    override fun getTitle(): String? {
        return activity.getString(R.string.tAddMembers)
    }

    override fun getMaximumNumberOfSelections(bundle: Bundle?): Int {
        return 10000
    }

    override fun createAdapter(): AbstractMultiSelectRecyclerAdapter<Buddy, *> {
        return BuddyListAdapter(mRecyclerView, presenter.dataProvider, R.layout.buddy_list_item_new)
    }

    override fun getNoDataString(): Int {
        return R.string.tBuddiesEmpty
    }

    override fun getNoMatchesFoundString(): Int {
        return R.string.tNoMatchesFound
    }

    override fun onDataProcessed(data: Any?) {
        if (shouldPublishResult() && data is MutableList<*>) {
            val ids = ArrayList<String>()
            val names = ArrayList<String>()
            val addresses = ArrayList<String>()
            for (buddy in data) {
                if (buddy is Buddy) {
                    ids.add(buddy.uniqueIdentifier)
                    names.add(buddy.displayName)
                    addresses.add(buddy.imAddress)
                }
            }

            val bundle = Bundle()
            bundle.putStringArrayList(CHOSEN_BUDDY_IDS, ids)
            bundle.putStringArrayList(CHOSEN_BUDDY_NAMES, names)
            bundle.putStringArrayList(CHOSEN_BUDDY_ADDRESSES, addresses)

            mBundle.putAll(bundle)
            publishResult(mBundle)
        }

        dismissScreenHolderDialog()
    }

    override fun onChipAdded(chip: Chip<Buddy>) {
        super.onChipAdded(chip)
    }

    override fun onChipDeleted(chip: Chip<Buddy>) {
        super.onChipDeleted(chip)
//        if (mChipsView.chips.isEmpty()) {
//            presenter.stopPerAccountFiltering()
//        }
    }

    @MainThread
    override fun onItemClick(item: View, index: Int) {
        if (!presenter.isNetworkAvailable()) {
            Snackbar.make(layout, R.string.tNoInternetConnection, Snackbar.LENGTH_SHORT).show()
            return
        }

        if (presenter.isDataLoading) {
            Snackbar.make(layout, R.string.tBuddiesLoading, Snackbar.LENGTH_SHORT).show()
            return
        }

        super.onItemClick(item, index)
    }

    override fun onPresenterEvent(event: IPresenterEvent<*, *>) {
        if (event.type === ChatRoomMembersPickerPresenter.ChatRoomMembersPickerPresenterEvents.NETWORK_CONNECTION_LOST) {
            mChipsView.setChipsListener(null) // stop observing while clearing data to avoid spamming of adapter by onChipDeleted
            mChipsView.clearAllChips()
            mChipsView.setChipsListener(this) // continue observing

            mAdapter.clearSelections()
        } else if (event.type === ChatRoomMembersPickerPresenter.ChatRoomMembersPickerPresenterEvents.PRESENCE_CHANGED) {
            mAdapter.notifyItemChanged((event.data as Long).toInt())
        } else {
            super.onPresenterEvent(event)
        }
    }
}
